#include <string>
#include <iostream>
using namespace std;
#include "foci_c.h"
extern "C" {
#include <caml/mlvalues.h>
#include <caml/callback.h>
#include <caml/alloc.h>
#include <caml/memory.h>
}

value_wrapper::value_wrapper(value v) {
    wrapped = v;
    caml_register_global_root(&wrapped);
}

value_wrapper::value_wrapper(const value_wrapper &vw) {
    wrapped = vw.wrapped;
    caml_register_global_root(&wrapped);
}

value_wrapper::~value_wrapper() {
    caml_remove_global_root(&wrapped);
}

// --- helpers --

/*value alloc_list_value2 ( value v1, value v2 ) {
    CAMLparam2(v1, v2);
    CAMLlocal2(r, r2);

    r = caml_alloc (2,0);
    Store_field(r, 0, v2);
    Store_field(r, 1, Val_int(0));

    r2 = caml_alloc (2, 0);
    Store_field(r2, 0, v1);
    Store_field(r2, 1, r);

    CAMLreturn(r2);
}

value mk_tuple2 ( value v1, value v2 ) {
    CAMLparam2(v1, v2);
    CAMLlocal1(v);
    v = caml_alloc_tuple(2);
    Store_field(v,0,v1);
    Store_field(v,1,v2);
    CAMLreturn(v);
}
*/

value alloc_list_value2( value v1, value v2 ) {
    static value * cl = NULL;
    if (cl == NULL) cl = caml_named_value("list2");
    return (caml_callback2(*cl,v1,v2));
}

value mk_tuple2( value v1, value v2 ) {
    static value * cl = NULL;
    if (cl == NULL) cl = caml_named_value("tuple2");
    return (caml_callback2(*cl,v1,v2));
}


// ----------  to_str ---

string foci_expr_to_str(foci_expr e) {
    static value * cl = NULL;
    if (cl == NULL) cl = caml_named_value("eToString");
    return string(String_val(caml_callback(*cl, e.val()))); 
}

string foci_pred_to_str(foci_pred p) {
    static value * cl = NULL;
    if (cl == NULL) cl = caml_named_value("pToString");
    return string(String_val(caml_callback(*cl, p.val()))); 
}

foci_pred foci_pred_elim_ite ( foci_pred p ) {
    static value * cl = NULL;
    if (cl == NULL) cl = caml_named_value("pElimIte");
    return foci_pred ( caml_callback(*cl, p.val()) ); 
}
// -------- 

foci_expr foci_mk_const ( int n ) {
    static value * cl = NULL;
    if (cl == NULL) cl = caml_named_value("eConstant");
    return foci_expr( caml_callback(*cl, Val_int(n)) ); 
}

foci_expr foci_mk_var ( const string& s ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("eVariable");
    value v;
    v = caml_copy_string(s.c_str());
    return foci_expr( caml_callback(*cl, v) );
}

foci_pred foci_mk_atom_pred ( const string& s ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pAtom");
    value v;
    v = caml_copy_string(s.c_str());
    return foci_pred( caml_callback(*cl, v) );
}

foci_expr foci_mk_ite ( foci_pred c, foci_expr t, foci_expr f ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("eIte");
    value v;
    v = caml_alloc_tuple(3);
    Store_field(v,0,c.val());
    Store_field(v,1,t.val());
    Store_field(v,2,f.val());
    return foci_expr( caml_callback(*cl, v) );
}

foci_expr foci_mk_sum ( foci_expr e1, foci_expr e2 ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("eSum");
    value v;
    v = alloc_list_value2(e1.val(), e2.val());
    return foci_expr( caml_callback(*cl, v) );
}

foci_expr foci_mk_sub ( foci_expr e1, foci_expr e2 ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("eSub");
    value v;
    v = alloc_list_value2(e1.val(), e2.val());
    return foci_expr( caml_callback(*cl, v) );
}

foci_expr foci_mk_mul ( foci_expr e1, foci_expr e2 ) {
    assert(0);
}

foci_pred foci_mk_true() {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pTrue");
    return foci_pred( caml_callback(*cl, Val_unit ) );
}

foci_pred foci_mk_false() {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pFalse");
    return foci_pred( caml_callback(*cl, Val_unit ) );
}

foci_pred foci_mk_and ( foci_pred e1, foci_pred e2 ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pAnd");
    value v;
    v = alloc_list_value2(e1.val(), e2.val());
    return foci_pred( caml_callback(*cl, v) );
}

foci_pred foci_mk_or ( foci_pred e1, foci_pred e2 ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pOr");
    value v;
    v = alloc_list_value2(e1.val(), e2.val());
    return foci_pred( caml_callback(*cl, v) );
}

/*foci_pred foci_mk_not ( foci_pred p ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pNot");
    return foci_pred( caml_callback(*cl, p) );
}*/

foci_pred foci_mk_not ( foci_pred p ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pNot");
    return foci_pred(caml_callback(*cl,p.val()));
}

foci_pred foci_mk_eq ( foci_expr e1, foci_expr e2 ) {
 //cerr << foci_expr_to_str(e1) << "\n";
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pEquality");
    value v;
    v = mk_tuple2(e1.val(),e2.val());
    foci_pred r = foci_pred( caml_callback(*cl,v));
 /////cerr << foci_pred_to_str(r) << "\n";
 return r;

}

foci_pred foci_mk_implies ( foci_pred p1, foci_pred p2 ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pImplies");
    value v;
    v = mk_tuple2(p1.val(),p2.val());
    return foci_pred( caml_callback(*cl, v) );
}


foci_pred foci_mk_leq ( foci_expr e1, foci_expr e2 ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pLeq");
    value v;
    v = mk_tuple2(e1.val(),e2.val());
    return foci_pred( caml_callback(*cl, v) );
}


foci_pred foci_mk_lt ( foci_expr e1, foci_expr e2 ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pLt");
    value v;
    v = mk_tuple2(e1.val(),e2.val());
    return foci_pred( caml_callback(*cl, v) );
}


foci_pred foci_mk_geq ( foci_expr e1, foci_expr e2 ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pGeq");
    value v;
    v = mk_tuple2(e1.val(),e2.val());
    return foci_pred( caml_callback(*cl, v) );
}

foci_pred foci_mk_gt ( foci_expr e1, foci_expr e2 ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pGt");
    value v;
    v = mk_tuple2(e1.val(),e2.val());
    return foci_pred( caml_callback(*cl, v) );
}

bool foci_interpolate ( foci_pred A, foci_pred B, foci_pred& res ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("interpolate");
    //value v, res_v;
    value res_v;
    
    //v = mk_tuple2(A.val(),B.val());
    res_v = caml_callback2_exn(*cl, A.val(), B.val() ) ;
    if ( Is_exception_result(res_v) ) {
        return false;
    }
    else {
        res = res_v;
        return true;
    }
}

foci_expr foci_shift_expr_to_zero ( foci_expr e ) {
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("eShift");
    return foci_expr( caml_callback(*cl, e.val()) );
}

foci_pred foci_shift_pred_to_zero ( foci_pred p ) {
   // cerr << "begin pred_shift_to_zero: pred is " << foci_pred_to_str(p) <<"\n";
    static value * cl = NULL;
    if ( cl == NULL ) cl = caml_named_value("pShift");
    //return foci_pred( caml_callback(*cl, p.val()) );
    foci_pred p_ =  foci_pred( caml_callback(*cl, p.val()) );
    //cerr << "end pred_shift_to_zero: pred is " << foci_pred_to_str(p_) <<"\n";
    return p_;
}
