let eConstant (n: int) = FociDag.eConstant( FociDag.Constant.Int n )
(* let eVariable ( s : string) = FociDag.eVariable( FociDag.Symbol.string s ) *)
let pTrue () = FociDag.pTrue
let pFalse () = FociDag.pFalse

let list2 a b  = [a;b]
let tuple2 a b = (a,b)

let eSub ( e1, e2 ) =
    FociDag.eSum [
        e1;
        FociDag.eCoeff ( FociDag.Constant.Int(-1), e2)
    ]


let pLt (e1, e2) =
    FociDag.pAnd [
        FociDag.pLeq (e1, e2);
        FociDag.pNot (FociDag.pEquality (e1, e2))
    ]
 
let pGeq (e1, e2) =
    FociDag.pNot (pLt (e1, e2))

let pGt (e1, e2) =
    FociDag.pNot (FociDag.pLeq (e1, e2))

let _ = Callback.register "list2" list2
let _ = Callback.register "tuple2" tuple2

let _ = Callback.register "eConstant" eConstant
let _ = Callback.register "eVariable" FociDag.eVariable
let _ = Callback.register "eApplication" FociDag.eApplication
let _ = Callback.register "eSum" FociDag.eSum
let _ = Callback.register "eCoeff" FociDag.eCoeff
let _ = Callback.register "eIte" FociDag.eIte

let _ = Callback.register "pTrue" pTrue
let _ = Callback.register "pFalse" pFalse
let _ = Callback.register "pAnd" FociDag.pAnd
let _ = Callback.register "pOr" FociDag.pOr
let _ = Callback.register "pNot" FociDag.pNot
let _ = Callback.register "pImplies" FociDag.pImplies
let _ = Callback.register "pEquality" FociDag.pEquality
let _ = Callback.register "pLeq" FociDag.pLeq
let _ = Callback.register "pAtom" FociDag.pAtom
let _ = Callback.register "pForall" FociDag.pForall


let _ = Callback.register "pLt" pLt
let _ = Callback.register "pGeq" pGeq
let _ = Callback.register "pGt" pGt
let _ = Callback.register "eSub" eSub


let simplify_list l =
    let rec simplify_list' todo result =
        match todo with
        | []       -> List.rev result
        | h::todo' ->
            if List.mem h result then
                simplify_list' todo' result
            else
                simplify_list' todo' (h::result)
    in
    simplify_list' l []

let lToString' b l p s =
    match l with
    | [] -> failwith "lToString' of empty"
    | [h] -> p b h
    | h::t ->
        p b h;
        List.iter (fun e -> Buffer.add_string b s; p b e) t

let rec eToString' b e =
    let add_string = Buffer.add_string b in
    let rec eToString'' e =
        match e with
        | (FociDag.Constant c, _)         ->
                add_string (FociDag.Constant.toString c)
        | (FociDag.Variable v, _)         ->
                add_string v
        | (FociDag.Application (v, l), _) ->
                failwith "Application not supported"
        | (FociDag.Sum l, _)              ->
                lToString' b l eToString' " + "
        | (FociDag.Coeff (c, e), _)       ->
                add_string (FociDag.Constant.toString c);
                add_string "*";
                eToString'' e
        | (FociDag.Ite (p, e1, e2), _)    ->
                add_string "ite(";
                pToString' b p;
                add_string ",";
                eToString'' e1;
                add_string ",";
                eToString'' e2
    in
    if false then
        (add_string "{";
        eToString'' e;
        add_string ":";
        add_string (string_of_int (snd e));
        add_string "}"
        )
    else
        eToString'' e

and pToString' b p =
    let add_string = Buffer.add_string b in
    let rec pToString'' p =
        match p with
        | (FociDag.True, _)              ->
            add_string "true"
        | (FociDag.False, _)             ->
            add_string "false"
        | (FociDag.And l, _)             ->
            let l = List.filter (fun x -> x <> FociDag.pTrue) l in
            let l = simplify_list l in
            (match l with
            | []  -> pToString'' FociDag.pTrue
            | [h] -> pToString'' h
            | _   ->
                add_string "(";
                lToString' b l pToString' " & ";
                add_string ")"
            )
        | (FociDag.Or l, _)              ->
            let l = List.filter (fun x -> x <> FociDag.pFalse) l in
            let l = simplify_list l in
            (match l with
            | []  -> pToString'' FociDag.pFalse
            | [h] -> pToString'' h
            | _   ->
                add_string "(";
                lToString' b l pToString' " | ";
                add_string ")"
            )
        | (FociDag.Not p, _)             ->
            add_string "~";
            pToString'' p
        | (FociDag.Implies (p1, p2), _)  ->
            add_string "(";
            pToString'' p1;
            add_string " -> ";
            pToString'' p2;
            add_string ")"
        | (FociDag.Equality (e1, e2), _) ->
            eToString' b e1;
            add_string " == ";
            eToString' b e2
        | (FociDag.Leq (e1, e2), _)      ->
            eToString' b e1;
            add_string " <= ";
            eToString' b e2
        | (FociDag.Atom s, _)            ->
            add_string s
        | (FociDag.Forall (l, p), _)     ->
            add_string "forall ";
            lToString' b l Buffer.add_string ",";
            add_string ": ";
            pToString'' p
    in
    if false then
        (add_string "[";
        pToString'' p;
        add_string ":";
        add_string (string_of_int (snd p));
        add_string "]"
        )
    else
        pToString'' p

let eToString e =
  let b = Buffer.create 16 in
  eToString' b e;
  Buffer.contents b

let pToString p =
  let b = Buffer.create 16 in
  pToString' b p;
  Buffer.contents b

let pSimplify p =
    let identity e = e in
    let savings = ref 0 in
    let saved n = savings := !savings + n in
    let mergeAnds l =
        let rec mergeAnds' l r =
            match l with
            | []   -> r
            | h::t ->
                (match h with
                | (FociDag.And l', _) -> mergeAnds' t (l' @ r)
                | _                   -> mergeAnds' t (h::r)
                )
        in
        mergeAnds' l []
    in
    let mergeOrs l =
        let rec mergeOrs' l r =
            match l with
            | []   -> r
            | h::t ->
                (match h with
                | (FociDag.Or l', _) -> mergeOrs' t (l' @ r)
                | _                  -> mergeOrs' t (h::r)
                )
        in
        mergeOrs' l []
    in
    let mergeAnds l = l in
    let mergeOrs l = l in
    
    let pSimplify' p =
        match p with
        | (FociDag.And l, _)             ->
            let l = mergeAnds l in
            let n = List.length l in
            let l = List.filter (fun x -> x <> FociDag.pTrue) l in
            let l = simplify_list l in
            if List.mem FociDag.pFalse l then
                (saved n; FociDag.pFalse)
            else
                (match l with
                | []  -> saved n; FociDag.pTrue
                | [h] -> saved n; h
                | _   -> saved (n - (List.length l)); FociDag.pAnd l
                )
        | (FociDag.Or l, _)              ->
            let l = mergeOrs l in
            let n = List.length l in
            let l = List.filter (fun x -> x <> FociDag.pFalse) l in
            let l = simplify_list l in
            if List.mem FociDag.pTrue l then
                (saved n; FociDag.pTrue)
            else
                (match l with
                | []  -> saved n; FociDag.pFalse
                | [h] -> saved n; h
                | _   -> saved (n - (List.length l)); FociDag.pOr l
                )
        | (FociDag.Not (FociDag.Not p', _), _) ->
            saved 2; p'
        | _ -> p
        (*
        | (FociDag.True, _)              -> check FociDag.pTrue
        | (FociDag.False, _)             -> check FociDag.pFalse
        | (FociDag.Not p', _)            -> check (FociDag.pNot p')
        | (FociDag.Implies (p1, p2), _)  -> check (FociDag.pImplies (p1, p2))
        | (FociDag.Equality (e1, e2), _) -> check (FociDag.pEquality (e1, e2))
        | (FociDag.Leq (e1, e2), _)      -> check (FociDag.pLeq (e1, e2))
        | (FociDag.Atom s, _)            -> check (FociDag.pAtom s)
        | (FociDag.Forall (l, p'), _)    -> check (FociDag.pForall (l, p'))
        *)
    in
    let p' = FociDag.Predicate.map pSimplify' identity p in
    Printf.fprintf stderr "Savings: %d\n" !savings;
    p'

(*let pSimplify p n =
    p
*)

let pElimIteSimplify p =
    pSimplify (FociDag.Predicate.elim_ite p)  

let _ = Callback.register "pElimIte" (pElimIteSimplify : FociDag.predicate -> FociDag.predicate )

let interpolate' a b = 
    (*let a = FociDag.Predicate.elim_ite a in
    let b = FociDag.Predicate.elim_ite b in
    *)
    flush stderr;
    try 
        (*let a = pSimplify a "a" in
        let b = pSimplify b "b" in
        *)
        (*Printf.fprintf stderr "Interpolating...\n";
        Printf.fprintf stderr " A is %s\n\n" (pToString a);
        Printf.fprintf stderr " B is %s\n\n" (pToString b);
        flush stderr;
        *)
        let r = Foci.interpolate (a,b) in
        Printf.fprintf stderr "UNSAT: Interpolant is %s\n" (pToString r);
        (* let r = pSimplify r "r" in*)
        flush stderr;
        r
    with Foci.SAT (p, l) ->
        Printf.fprintf stderr "SAT:\n";
        (*Printf.fprintf stderr "  p = [%s]\n"
            (String.concat " , "
                (List.map FociDag.Predicate.toString p));
        Printf.fprintf stderr "  l = [\n";
        List.iter
            (fun (c, n) ->
                Printf.fprintf stderr "    %s = %s\n"
                    (FociDag.Concrete.toString c)
                    (Num.string_of_num n)
            )
            l;
        Printf.fprintf stderr "  ]\n";
        *)
        flush stderr;
        
        raise (Foci.SAT (p, l))
    |    e ->
        Printf.fprintf stderr "Mean Exception: %s\n" (Printexc.to_string e);
        flush stderr;
        assert false;
        raise e


(* let _ = Callback.register "interpolate" Foci.interpolate *)
let _ = Callback.register "interpolate" interpolate'
let _ = Callback.register "fst" fst
let _ = Callback.register "snd" snd




let rec range (i: int) (j: int): int list =
    if i < j then
        []
    else
        i::(range (i+1) j)

let isAlpha (ch: char): bool =
  (ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <='z')

let isDigit (ch: char): bool =
  (ch >= '0' && ch <= '9')

let parseStateVariable (s: string): string * int =
    try
        let u = String.rindex s '_' in
        let l = String.length s in
        for i = 0 to l-1 do
          let ch = s.[i] in
          (*Printf.fprintf stderr "parseStateVariable: %s\n" s; 
          flush stderr;
          *)
          if i < u then assert (isAlpha ch || isDigit ch || ch == '_' || ch == '.');
          if i == u then assert (ch == '_');
          if i > u then assert(isDigit ch);
        done;
        (String.sub s 0 u, int_of_string (String.sub s (u+1) (l-u-1)))
    with Not_found ->
        assert false

let eShift e =
    match e with
    | (FociDag.Variable name, _) ->
        let (pre, post) = parseStateVariable(name) in
        FociDag.eVariable (Printf.sprintf "%s_%d" pre (0))
    | _                                             ->
        e

(* let x p =
  match p with
  | (And [(Not x, _); (Not y, _)], _) when rem("[A-Z]*", x) -> Not (Or [x, y])
  | (And [(Not x, _); (Not y, _)], _)  -> Not (Or [x, y])
*)

let pShift p =
    let pAtomShift p =
        match p with 
        | (FociDag.Atom name, _) ->
            let (pre, post) = parseStateVariable(name) in
            FociDag.pAtom (Printf.sprintf "%s_%d" pre (0))
        | _ -> p
    in
    FociDag.Predicate.map pAtomShift eShift p

let _ = Callback.register "eShift" eShift
let _ = Callback.register "pShift" pShift

(*
let _ = Callback.register "eToString" FociDag.Expression.toString
let _ = Callback.register "pToString" FociDag.Predicate.toString
*)

let _ = Callback.register "eToString" eToString
let _ = Callback.register "pToString" pToString

(* let ePrintToStr e = 
    let buf = Buffer.create 0 in
    FociDag.Expression.print (Format.formatter_of_buffer buf) e;
    output_string stderr "toString: ";
    output_string stderr (FociDag.Expression.toString e);
    output_string stderr "\n";
    output_string stderr "print   : ";
    output_string stderr (Buffer.contents buf);
    output_string stderr "\n";
    flush stderr;
    Buffer.contents buf

let pPrintToStr p = 
    let buf = Buffer.create 0 in
    FociDag.Predicate.print (Format.formatter_of_buffer buf) p;
    output_string stderr "toString: ";
    output_string stderr (FociDag.Predicate.toString p);
    output_string stderr "\n";
    output_string stderr "print   : ";
    output_string stderr (Buffer.contents buf);
    output_string stderr "\n";
    flush stderr;
    Buffer.contents buf

let _ = Callback.register "pPrintToStr" pPrintToStr
let _ = Callback.register "ePrintToStr" ePrintToStr
*)

(*let _ = Callback.register "pPrint" (FociDag.Predicate.print Format.std_formatter)*)

(*
let _ = Gc.create_alarm ( fun () -> Printf.fprintf stderr "GC invoked\n"; flush (stderr) ) 
let _ = Gc.create_alarm ( fun () -> Gc.print_stat stderr ; flush (stderr) ) 
*)

(*
open FociDag
let x = eVariable "x_12"
let y = eVariable "yyy_2"
let three = eConstant (FociDag.Constant.Int 3)
let four = eConstant (FociDag.Constant.Int 4)
let p = pAnd [ pLeq (x, three); pLeq(y, four) ]
let p' = pShift p
let () = Printf.printf "p  is (%s)\n" (Predicate.toString p)
let () = Printf.printf "p' is (%s)\n" (Predicate.toString p')
*)

(*
open FociDag
let const (n: int) = eConstant( Constant.Int n )
let p2e p = eIte (p, const 1, const 0)
let pEqPP p1 p2 = pEquality (p2e p1, p2e p2)
let pEqPP' p1 p2 = pOr [pAnd [p1; p2]; pAnd [pNot p1; pNot p2]]

let a = pAtom "a"
let x = eVariable "x"
let y = eVariable "y"
let z = eVariable "z"
let a = Predicate.elim_ite (pLeq (z, eIte (a, x, y)))
let b = pFalse
let i = interpolate' a b
let _ = Printf.fprintf stderr "XXX: %s\n" (Predicate.toString i)
*)
