extern "C" {
#include <caml/mlvalues.h>
}

class value_wrapper {
    public:

    value_wrapper(value v);
    value_wrapper(const value_wrapper &vw);
    ~value_wrapper();

    const value_wrapper &operator=(const value_wrapper &vw) {
        wrapped = vw.wrapped;
        return *this;
    }

    const value val() const { return wrapped; }
    value val() { return wrapped; }
    void set_val(value val) { wrapped = val; }

    private:
    value wrapped;
};

struct foci_pred : public value_wrapper {
    foci_pred() : value_wrapper(Val_unit) {}
    foci_pred(value v) : value_wrapper(v) {}
    foci_pred(const foci_pred &p) : value_wrapper(p.val()) {} 
    const foci_pred &operator=(const foci_pred &p) { set_val(p.val()); return (*this); }
};

struct foci_expr : public value_wrapper {
    foci_expr() : value_wrapper(Val_unit) {}
    foci_expr(value v) : value_wrapper(v) {}
    foci_expr(const foci_expr &p) : value_wrapper(p.val()) {}
    const foci_expr &operator=(const foci_expr &p) { set_val(p.val()); return (*this); }
};


string foci_expr_to_str(foci_expr e) ;
string foci_pred_to_str(foci_pred p) ;

foci_pred foci_pred_elim_ite ( foci_pred p );

foci_expr foci_mk_const ( int n ) ;
foci_expr foci_mk_var ( const string& s ) ;
foci_expr foci_mk_ite ( foci_pred c, foci_expr t, foci_expr f ) ;
foci_pred foci_mk_atom_pred ( const string& s ) ;

foci_pred foci_mk_true();
foci_pred foci_mk_false();

foci_pred foci_mk_and ( foci_pred p1, foci_pred p2 ) ;
foci_pred foci_mk_or ( foci_pred p1, foci_pred p2 ) ;
foci_pred foci_mk_implies ( foci_pred p1, foci_pred p2 ) ;
foci_pred foci_mk_not ( foci_pred p ) ;
foci_pred foci_mk_eq ( foci_expr e1, foci_expr e2 ) ;
foci_pred foci_mk_diseq ( foci_expr e1, foci_expr e2 ) ;
foci_pred foci_mk_leq ( foci_expr e1, foci_expr e2 ) ;
foci_pred foci_mk_lt ( foci_expr e1, foci_expr e2 ) ;
foci_pred foci_mk_geq ( foci_expr e1, foci_expr e2 ) ;
foci_pred foci_mk_gt ( foci_expr e1, foci_expr e2 ) ;

foci_expr foci_mk_sum ( foci_expr e1, foci_expr e2 ) ;
foci_expr foci_mk_sub ( foci_expr e1, foci_expr e2 ) ;
foci_expr foci_mk_mul ( foci_expr e1, foci_expr e2 ) ;

bool foci_interpolate ( foci_pred A, foci_pred B, foci_pred& res ) ;

foci_expr foci_shift_expr_to_zero ( foci_expr e ) ;
foci_pred foci_shift_pred_to_zero ( foci_pred e ) ;


