#ifndef ACT_H
#define ACT_H
#include"stmts.h"
#include"graph.h"
namespace META = ttl::meta;
namespace IMPL = META::impl;

class Comm_act;
typedef std::list<Comm_act> Comm_act_list;
typedef std::list<Const> Const_list;

template <typename V>
struct variant_equal_visitor {
    template <typename T>
        void operator()(const T &val1) { r = (val1 == VAR::get<T>(v2)); }

    variant_equal_visitor(const V &_v1, const V &_v2) :
        v2(_v2), r(_v1.which() == _v2.which()) 
        {
            Util::msg(5,"v1 is "+_v1.to_str() + " v2 is "+_v2.to_str() + "_v1.which= "+ int2str(_v1.which())+ "_v2.which = "+ int2str(_v2.which()) + "\n");
            if( !r ) return;
            VAR::apply_visitor(*this, _v1);
            Util::msg(5,"r is %d\n",r);
        }
    bool get() const { return r; }

    private:
    const V &v2;
    bool    r;
};


template<class Act>
string act_to_str(const Act& act);
template<class Act>
Edge_type act_get_type(const Act& act);
template <typename V>
bool act_equals(const V &v1, const V &v2) ;


template<class T>
struct comm_act_cons_vis {

    comm_act_cons_vis(Comm_act& _act, const T& t) : act(_act)
    { VAR::apply_visitor(*this, t); }

    void operator() (const Send& s) { act = s; }
    void operator() (const Recv& r) { act = r; }
    template <typename C>
        void operator()(const C &val1) { assert(0); }

    private:
    Comm_act &act;
};

struct comm_act_is_concrete{
    bool ret;

    comm_act_is_concrete(const Comm_act& _act) 
    { VAR::apply_visitor(*this, _act); }

    void operator() (const Send& s) { ret = s.is_concrete(); }
    void operator() (const Recv& r) { ret = r.is_concrete(); }
    template <typename C>
        void operator()(const C &val1) { assert(0); }

};

struct comm_act_get_vr{
    Varref vr;

    comm_act_get_vr(const Comm_act& act) 
    { VAR::apply_visitor(*this,act); }

    void operator() (const Send& s) { vr = s.get_vr(); }
    void operator() (const Recv& r) { vr = r.get_vr(); }
    template <typename C>
        void operator()(const C &val1) { assert(0); }

};

struct comm_act_get_const{
    Const c;

    comm_act_get_const(const Comm_act& act) 
    { VAR::apply_visitor(*this,act); }

    void operator() (const Send& s) { c = s.get_const_arg(); }
    void operator() (const Recv& r) { c = r.get_const_arg(); }
    template <typename C>
        void operator()(const C &val1) { assert(0); }

};

/*
template<class T>
struct comm_act_concretize {

    comm_act_concretize(const T& _act, const Const_list& _clist, list<T>& _conc_acts )  
       :  clist(_clist), conc_acts( _conc_acts )
    { VAR::apply_visitor(*this, _act); }

    void operator() (const Send& s) { 
        const Varref& vr = s.get_vr();
        const_foreach(it, clist) {
            Send_arg sa(*it); 
            Send_args sas(1,sa);
            Send s(vr,sas);
            conc_acts.push_back( T(s) );
        }
    }

    void operator() (const Recv& r) {  
        const Varref& vr = r.get_vr();
        const_foreach(it, clist) {
            Recv_arg ra(*it); 
            Recv_args ras(1,ra);
            Recv r(vr,ras);
            conc_acts.push_back( T(r) );
        }
    }

    template <typename C>
        void operator()(const C &val1) { assert(0); }

    private:
    const Const_list &clist;
    list<T> &conc_acts;
};
*/

class Comm_act : public VAR::variant<Send, Recv> {
    typedef VAR::variant<Send, Recv> Super;
    public: 
    Comm_act() {}
    Comm_act (const Super& s) : Super(s) {}
    Comm_act (const Send& s) : Super(s) {}
    Comm_act (const Recv& r) : Super(r) {}
    template<class T>
        Comm_act (const T& t) { comm_act_cons_vis<T> vis(*this,t); }


    /*void concretize ( const Const_list& clist, Comm_act_list& conc_acts) const {
        comm_act_concretize<Comm_act> vis(*this, clist, conc_acts);
    }*/

    void concretize( int a, Const_list& l ) const {}

    string to_str() const { return ::act_to_str(*this); }
    Edge_type get_type() const { return ::act_get_type(*this); }
    bool operator == (const Comm_act& cact) const { 
        return act_equals<Comm_act>(*this, cact);
    }

    bool is_concrete() const { comm_act_is_concrete vis (*this); return vis.ret; }
    Varref get_vr() const { comm_act_get_vr vis (*this); return vis.vr; }
    Const get_const_arg() const { comm_act_get_const vis(*this); return vis.c; }
};


//struct Action_vis : VAR::static_visitor<> {
struct Action_vis  {
    enum mode {TO_STR, GET_TYPE};
    mode m;

    string s;
    Edge_type etype;

    Action_vis(mode _m) : m(_m) {}

    void operator() (const Assign& a) {
        switch(m) {
            case TO_STR:
                s = a.to_str();
                break;
            case GET_TYPE:
                etype = OTHER;
                break;
            default: assert(0);
        }
    }

    void operator() (const DAssign& a) {
        switch(m) {
            case TO_STR:
                s = a.to_str();
                break;
            case GET_TYPE:
                etype = OTHER;
                break;
            default: assert(0);
        }
    }

    void operator() (const Send& send) {
        switch(m) {
            case TO_STR:
                s = send.to_str();
                break;
            case GET_TYPE:
                etype = SEND;
                break;
            default: assert(0);
        }
    }


    void operator() (const Recv& r) {
        switch(m) {
            case TO_STR:
                s = r.to_str();
                break;
            case GET_TYPE:
                etype = RECV;
                break;
            default: assert(0);
        }
    }

    void operator() (const Pause& p) {
        switch(m) {
            case TO_STR:
                s = p.to_str();
                break;
            case GET_TYPE:
                etype = OTHER;
                break;
            default: assert(0);
        }
    }


    template<typename T> void operator() (const T& t) {printf("Not handled\n"); assert(0); }
};


template<class Action>
string act_to_str(const Action& a) {
    Action_vis va(Action_vis::TO_STR);
    VAR::apply_visitor(va,a);
    return va.s;
}

//template string act_to_str<Comm_act>(const Comm_act&);

template<class Action>
Edge_type act_get_type(const Action& a) {
    Action_vis va (Action_vis::GET_TYPE);
    VAR::apply_visitor(va, a);
    return va.etype;
}


template <typename V>
bool act_equals(const V &v1, const V &v2) {
    return variant_equal_visitor<V>(v1, v2).get();
}

class _Action_;

struct Action_rw_vis {
    RW& rw;

    Action_rw_vis ( const _Action_& act, RW& _rw) : rw(_rw)
        { VAR::apply_visitor(*this,act); }

    void operator() (const Assign& a) { a.get_rw(rw); }
    void operator() (const DAssign& a) { a.get_rw(rw); }
    void operator() (const Send& send) { send.get_rw(rw); }
    void operator() (const Recv& r) { r.get_rw(rw); }
    void operator() (const Pause& p) { }

    template<typename T> void operator() (const T& t) {printf("Not handled\n"); assert(0); }
};

template<class O>
struct Action_fix_const_vis {
    const O& o;
    Pid pid;

    Action_fix_const_vis ( _Action_& act, const O& _o, Pid _pid) : o(_o), pid(_pid) 
        { VAR::apply_visitor(*this,act); }

    void operator() (Assign& a) { a.fix_const_types(o,pid); }
    void operator() (DAssign& a) { a.fix_const_types(o,pid); }
    void operator() (Send& send) { assert(0); }
    void operator() (Recv& r) { assert(0); }
    void operator() (Pause& p) { }

    template<typename T> void operator() (T& t) {printf("Not handled\n"); assert(0); }
};


template<class T>
struct act_cons_vis {

    act_cons_vis(_Action_& _act, const T& t) : act(_act)
    { 

        VAR::apply_visitor(*this, t); 
    }

    void operator() (const Assign& a) { act = a; }
    void operator() (const DAssign& a) { act = a; }
    void operator() (const Send& s) { act = s; }
    void operator() (const Recv& r) { act = r; }
    void operator() (const Pause& p) { act = p; }
    template <typename C>
        void operator()(const C &c) { assert(0); }

    private:
    _Action_ &act;
};

class _Action_ : public VAR::variant<Assign, DAssign, Send, Recv, Pause> {
    typedef VAR::variant<Assign, DAssign, Send,  Recv, Pause> Super;
    public: 
    typedef _Action_ Lab;
    _Action_ () {}
    _Action_ (const Assign& a) : Super(a) {} 
    _Action_ (const DAssign& a) : Super(a) {} 
    _Action_ (const Send& s) : Super(s) { }
    _Action_ (const Recv& r) : Super(r) { }
    _Action_ (const Pause& p) : Super(p) { }
//    _Action_ (const Composite& c)  { assert(0); } //FIX: this makes Prop_act and CFA_action different
    //then will have to write a constructor from CFA to PropFA
    //_Action_ (const Super& s) : Super(s) {}
    template<class T>
        _Action_ (const T& t) { act_cons_vis<T> vis(*this,t); }


    string to_str() const { 
        if( is_eps() )  return "eps";
        else return ::act_to_str(*this); 
    }
    Edge_type get_type() const { return ::act_get_type(*this); }

    bool operator == (const _Action_& act) const { 
        bool ret = act_equals<_Action_>(*this, act);
        Util::msg(6,"_Action_::< - to_str() = "+this->to_str() + " , act.to_str() = " + act.to_str() + ", ret ="+ int2str(ret) +"\n");
        return ret;
    }

    bool matches (const _Action_& act) const { 
        return (*this == act);
    }

    bool operator < (const _Action_& act) const {
        bool retval = (to_str() < act.to_str());
        Util::msg(6,"_Action_::< - to_str() = "+to_str() + " , act.to_str() = " + act.to_str() + ", retval ="+ int2str(retval) +"\n");
        return retval;
    }

    bool operator != (const _Action_& act) const { assert(0); return false;}
    bool operator > (const _Action_& act) const { assert(0); return false;}
    bool operator <= (const _Action_& act) const { assert(0); return false;}
    bool operator >= (const _Action_& act) const { assert(0); return false;}

    static _Action_ get_eps() { return _Action_(); }
    bool is_eps() const { return Super::is_singular(); }
    void get_rw(RW& rw) const { Action_rw_vis(*this,rw); }
    template<class O>
    void fix_const_types(const O& oracle, Pid pid) { Action_fix_const_vis<O> vis(*this,oracle,pid); }
};



template<class N, class Lab_type>
class Lab_edge : public Edge<N> {
    protected:
        Lab_type s;
        typedef Edge<N> Super;
    public:
        typedef typename Lab_type::Lab Act;
        Lab_edge()   { s = Lab_type(); }
        Lab_edge(const Lab_type& _s)     { s = _s; }
        Lab_edge(const N& n1, const Lab_type& _s, const N& n2)     { s = _s; Super::set_ends(n1,n2); }
        //virtual ~Lab_edge() {}
        void set_lab(const Lab_type& _s) { s = _s; }
        void set_act(const Act& act) { s = act; }

        Edge_type get_type() const { return s.get_type(); }
        bool is_comm_type() const { return (get_type() == SEND || get_type() == RECV);}
        const Lab_type& get_lab() const { return s;}
        const Lab_type& get_act() const { return s; } //override if s is not an action variant, e.g. CFG_stmt
        Lab_type& get_act() { return s; }
        const Lab_type& get_comm_act() const { CHK(is_comm_type()); return s; } //same as above

        Lab_edge get_epsilon() const { return Lab_edge(); }
        bool is_epsilon() const { return (s == Lab_type()); }

        /*Var_dcl eval_chan_name(const Sys_state& s, int pid) const { Note: don't make Lab_edge dep on Sys_state
          }*/

        //equivalence
        bool operator == (const Lab_edge& le) const { return get_lab() == le.get_lab(); }
        //some kind of preorder
        bool matches(const Lab_edge& le) const { 
            //return act_equals<Lab_type>(get_lab(), le.get_lab());
            return get_lab().matches( le.get_lab() );
        }

        bool operator < (const Lab_edge& le) const { return get_lab() < le.get_lab(); }

        bool operator != (const Lab_edge& le) const { assert(0); return false; }
        bool operator <= (const Lab_edge& le) const { assert(0); return false; }
        bool operator >= (const Lab_edge& le) const { assert(0); return false; }
        bool operator > (const Lab_edge& le) const { assert(0); return false; }

        //virtual abstract functions of Super
        string to_str() { return string( s.to_str() +" ->");}
        string to_str_const() const { return string( s.to_str_const() +" ->");}
        bool is_eps () const { return s.is_eps(); }

};

/*
template<typename G, typename Act>
void get_graph_comm_sigma(const G& g, set<Act>& sigma) {
    const typename G::Edges& edges = g.get_edges();
    typename_const_foreach(it, edges) {
        if( it->is_comm_type() ) {
            sigma.insert( Act(it->get_comm_act()) );
        }
    }
}*/

Assign get_assign_from_str(const string& str);
Send get_send_from_str(const string& str);
Recv get_recv_from_str(const string& str);

template<class Act>
Act get_act_from_str(const string& str) {
    if( str.find('=') != string::npos ) { //ASSIGN
        return get_assign_from_str(str);
    }
    else if(str.find('!') != string::npos) { //SEND
        return get_send_from_str(str);

    }
    else if(str.find('?') != string::npos) { //RECV
        return get_recv_from_str(str);
    }
    else assert(0);
}

#endif
