#ifndef AG_CHECKER_H
#define AG_CHECKER_H

#include"system.h"
#include"teacher.h"
#include"learner.h"


enum AG_mode {NC,CIRC};


template<class Learner, class Teacher, class Sys_CE>
class AG_checker {

    public:
        typedef typename Teacher::System System;
        typedef typename Teacher::Sys_explorer Sys_explorer;
        typedef typename Teacher::CE Learner_CE;
        typedef typename Teacher::Sig Sig;
        typedef typename Teacher::Prop Prop;
        typedef typename Teacher::_DFA DFA;

    protected:

        typedef set<Sig> Sigma;
        typedef list<Sig> Sig_list;

        const System& sys;
        const Prop& spec;
        AG_mode m;


    public:


        AG_checker(const System& _sys, const Prop& _spec, AG_mode _m ) 
            : sys(_sys), spec(_spec), m(_m) { }


        bool run(Sys_CE& ce) {
            if(sys.num_cfg_components() == 1) {
                Util::msg(1,"running sys expl for last component\n");
            int VL = Util::VERBOSITY_LEVEL; 
            Util::VERBOSITY_LEVEL = 3;
                bool res = Sys_explorer(sys).check_prop(spec, ce);
            Util::VERBOSITY_LEVEL = VL;
                return res;
            } else CHK (sys.num_cfg_components() > 1);

            if(m == NC) return run_nc(ce);
            else if(m == CIRC) return run_circ(ce);
            else CHK (0);
            return false;
        }

    protected:


        //learner sigma for AG rule: curr_sys || A \sat P, rest_sys \sat A

        void compute_learner_sigma_nc ( System& curr_sys, System& rest_sys, const Prop& curr_spec, 
                                     Sigma& learner_sigma) {
            RW curr, rest;
            RW p_rw;
            rest = rest_sys.get_glb_rw();
            curr = curr_sys.get_glb_rw();

            //Util::msg(1,"lsigma: rest is "+rest.to_str());
            //Util::msg(1,"lsigma: curr is "+curr.to_str());
            RW _env_rw = curr_sys.get_env_rw();
            //Util::msg(1,"lsigma: env rw is "+_env_rw.to_str());
            curr.add(curr_sys.get_env_rw());

            fa_get_glb_rw(curr_spec,p_rw);

            set<Var_dcl> curr_r = curr.get_r();
            set<Var_dcl> curr_w = curr.get_w();

            set<Var_dcl> p_r = p_rw.get_r();
            set<Var_dcl> p_w = p_rw.get_w();

            set<Var_dcl> l_basis = curr_r;
            l_basis.insert(curr_w.begin(),curr_w.end());
            l_basis.insert(p_r.begin(),p_r.end());
            l_basis.insert(p_w.begin(),p_w.end());

            RW l_basis_rw (l_basis, l_basis);
            curr_sys.get_concrete_comm_sigma(l_basis_rw,learner_sigma);
            


/*          
            set<Var_dcl> rest_r = rest.get_r();
            set<Var_dcl> rest_w = rest.get_w();

            list<Var_dcl> toErase;
            //prune stuff from rest_r so that rest_r matches curr_w 
            foreach(it, rest_r) {
                if( 
                   //curr_w.find(it->first) == curr_w.end()  
                   !curr.has_write(*it) &&
                   !p_rw.has_read_or_write(*it)
                  ) 
                { toErase.push_back(*it); }
            }
            foreach(it, toErase) { rest_r.erase(*it); }
            toErase.clear();
            //prune stuff from rest_w so that rest_w matches curr_r 
            foreach(it, rest_w) {
                //if( curr_r.find(it->first) == curr_r.end() ) 
                if ( !curr.has_read(*it) &&
                     !p_rw.has_read_or_write(*it)
                   )
                { toErase.push_back(*it); }
            }

            foreach(it, toErase) { rest_w.erase(*it); }
            // --- added NMT
            rest.add(curr);
            rest.add(p_rw);

            //now get CFA_actions
            Util::msg(1,"lsigma: finally rest is "+rest.to_str());
            CHK ( !rest.empty() );
            rest_sys.get_concrete_comm_sigma(rest, learner_sigma);
            */

            CHK ( !learner_sigma.empty() );
            //Util::msg(1,"learner sigma is "+set_to_str(learner_sigma) );
        }
        /*
            S_env || M_i || S_a \sat P

            S_env has all glb vars which M_j to left of M_i may read/write
            S_a has all glb vars which M_j to right of M_i may read/write

            glb vars = glb_write_vars + glb read vars
            to find fixpoint, one needs only glb write vars

            glb_write_vars = glb state vars + glb logic vars
            
            fix_env will have all glb_w_vars 
            fix_a will have all glb_w_vars 
            M_i will have only glb + loc state vars



        */

        bool run_nc(Sys_CE& ce) {

            set<size_t> indices;

            //make curr_sys with first component
            int N = sys.num_cfg_components();
            indices.insert(0);
            const RW& env_rw = sys.get_env_rw();
            System curr_sys(sys, indices, env_rw);

            //make rest_sys with n-1 components
            RW rest_env_rw = curr_sys.get_glb_rw();
            CHK( !rest_env_rw.empty() );
            rest_env_rw.add(env_rw); 
            indices.clear();
            for(size_t i=1; i < sys.num_cfg_components(); ++i) indices.insert(i);
            System rest_sys(sys, indices, rest_env_rw); //should project env_rw on its alphabet?

            //obtain learner_sigma 
            Sigma learner_sigma;
            compute_learner_sigma_nc (curr_sys, rest_sys, spec, learner_sigma);
            //Util::msg(1,"learner sigma is "+set_to_str(learner_sigma) +"\n");


            //should make local vars of AG_checker for subst(NMT)
            Teacher mat(curr_sys,spec,learner_sigma);
            Learner learner(&mat);

            //top-level loop
            bool retval = false;
            while(1) {
                //generate candidate for R1: M_1 || A \subseteq P
                while(1) {
                    Learner_CE ce;
                    Util::msg(1,"------------------------------");
                    Util::message(1,"N%d: Asking cand query ...\n",N);      
                    learner.get_new_candidate().to_str_const();
                    bool res = mat.ask_cand_q(learner.get_new_candidate(),ce);
                    if(res) break;      
                    Util::message(1,"N%d: learning from ce=%s ...\n",N,ce.to_str().c_str());      
                    learner.learn_ce(ce);
                }
                Util::message(1,"N%d: finished learning assumptions ...assmp is\n",N);
                Util::msg(1,learner.get_prev_candidate().to_str_const());
                Util::message(1,"N%d: now checking rest system ...\n",N);

                //check rest against A : M_2 \subseteq A
                //use AG recursively
                Prop rest_spec(true);
                rest_spec.construct_from(learner.get_prev_candidate()); 
                rest_spec.negate();
                Sys_CE rest_ce;
                bool rest_res = AG_checker(rest_sys,rest_spec,m).run(rest_ce);
                //if model checking succeeds
                Util::message(1,"N%d: over checking rest system ..\n",N); 
                if(rest_res) { retval = true; break; }
                //otherwise check if ce is correct or not, ie M || ce falsifies P
                Util::message(1,"N%d: Obtained CE: checking if [%s] is spurious/actual\n",N,rest_ce.to_str().c_str());
                Learner_CE rest_ce_abs;
                Sys_CE all_ce;
                bool curr_res = mat.check_M_ce_satisfy_spec(rest_ce, rest_ce_abs, all_ce);

                //if CE model checking fails then return false else add CE to assumption and repeat
                if(!curr_res) {
                    Util::message(1,"N%d: CE is actual..returning to earlier level..\n",N); 
                    ce =  all_ce; retval = false; break;
                } else { 
                    //CE proj_ce; //rest_ce.project(mat.get_sigma(),proj_ce); //proj_ce.EliminateEpsilons();
                    Util::message(1,"N%d: CE is spurious..learning..\n",N); 
                    learner.learn_ce(rest_ce_abs);
                } 
            }
            return retval;
        }


        /*void compute_l_sigma_vec( const System& sys, vector<Sigma>& l_sigma_vec) {
            vector<RW_map> sys_rw_vec;
            size_t N = sys.num_cfg_components();
            size_t i,j;

            // determine l_sigma_vec 

            for(i=0; i < N; ++i) {
                RW_map rw_i;
                //TODO sys.get_cfg_rw(i,rw_i);
                sys_rw_vec.push_back( rw_i );
            }

            for(i=0; i < N ; ++i) {
                RW_map l_rw_i;
                //possible to make this rw_i computation incremental
                // get_l_rw_i:
                for( j=0; j < N; ++j) {
                    if ( j == i ) continue;
                    l_rw_i.add( sys_rw_vec[j] );
                }//end get_l_rw_i

                // TODO : prune stuff not in sys_rw_vec[i] or prop: match n prune 

                Sigma sigma_i;
                sys.get_concrete_comm_sigma(l_rw_i, sigma_i);
                l_sigma_vec.push_back(sigma_i);
            }

        }
        */

        void compute_learner_sigma_circ ( System& curr_sys, System& rest_sys, const Prop& curr_spec, 
                                     Sigma& learner_sigma) {
            //earlier learner_sigma = rest_sigma \cap ( matches(curr_sigma) \cup prop_sigma)
            // now learner_sigma = global_acts(rest_sigma) \cap matches(curr_sigma), 
            //read/writes rest_sys can do
            RW curr, rest; //make these RW
            RW p_rw;
            rest = rest_sys.get_glb_rw();
            curr = curr_sys.get_glb_rw();
            curr.add(curr_sys.get_env_rw());

            /*
            rest.add(curr);
            RW to_add(rest.get_w(), rest.get_r());
            rest.add(to_add);
            to_add = RW(curr.get_w(),curr.get_r());
            rest.add(to_add);
            rest_sys.get_concrete_comm_sigma(rest, learner_sigma);
            */


            fa_get_glb_rw(curr_spec,p_rw);

            set<Var_dcl> curr_r = curr.get_r();
            set<Var_dcl> curr_w = curr.get_w();

            set<Var_dcl> p_r = p_rw.get_r();
            set<Var_dcl> p_w = p_rw.get_w();

            set<Var_dcl> l_basis = curr_r;
            l_basis.insert(curr_w.begin(),curr_w.end());
            l_basis.insert(p_r.begin(),p_r.end());
            l_basis.insert(p_w.begin(),p_w.end());

            RW l_basis_rw (l_basis, l_basis);
            curr_sys.get_concrete_comm_sigma(l_basis_rw,learner_sigma);
            CHK ( !learner_sigma.empty() );
            //Util::msg(1,"learner sigma is "+set_to_str(learner_sigma) );
        }

        bool run_circ(Sys_CE& ce) ;



};


//template<class System, class Pclass Sys_CE>
//bool check_ag(const System& sys, const PropFA& spec, Sys_CE& ce) ;

#endif
