#include<string>
#include<iostream>
#include"ast.h"

namespace Ast {

/* constructor function for each of the node types */
Num mk_number (int n) {return n;}

Symbol mk_symbol (string s) { return s;}

/* Decl types */
Var_dcl mk_simplevardcl(Symbol s) { return Var_dcl(SimpleVarDcl(s));}
/*Var_dcl mk_bitvardcl(Symbol s, Num n) { return Var_dcl(BvVarDcl(s,n));}*/
Var_dcl mk_arrayvardcl(Symbol s, Num n) { return Var_dcl(ArrayVarDcl(s,n));}

//Varspec
/*typedef TUP::tuple<Var_dcl, Expr*> Varspec;
typedef list<Varspec> Varspec_list;
*/
Varspec mk_varspec (Var_dcl v, Expr* e) { return Varspec(v,e); }

// Var_decl
/*typedef TUP::tuple<Ty, Varspec> Var_decl;
typedef list<Var_decl> Var_decl_list;
typedef Var_decl_list Var_decls;
*/

Var_decl_list declare_vars(Ty t, Varspec_list lv) {
  Var_decl_list ld;
  for(Varspec_list::iterator it = lv.begin(); it != lv.end(); ++it) {
    ld.push_back(Var_decl(t,*it));
  }
  return ld;
}






/* elementary types */
Ty mk_ty_bit () { return BitTy;}
Ty mk_ty_bool () {return BoolTy;}
Ty mk_ty_byte () { return ByteTy;}
Ty mk_ty_int () {return IntTy;}
Ty mk_ty_ev() {return EvTy;}
Ty mk_ty_chan() {return ChanTy;}
//Ty mk_ty_mtype () {return MtypeTy;}
Ty mk_ty_pid () {return PidTy;}
Ty mk_ty_short () {return ShortTy;}
Ty mk_ty_unsigned () {return UnsignedTy;}
Ty mk_ty_named (Symbol s) {return NamedTy(s);}
Ty mk_ty_bv (Num w) {return BvTy(w);}
Ty mk_ty_array (bTy t, Num s) {return (ArrayTy (t, s)); }



/* expressions */


Expr* mk_expr_binop (Expr* l, op o, Expr* r) { return ExprManager::get_expr(l, o, r);}
Expr* mk_expr_ternop (op o, Expr* e1, Expr* e2, Expr* e3) { return ExprManager::get_expr(o, e1, e2, e3);}
Expr* mk_expr_unop (uop u, Expr* e) { return ExprManager::get_expr(u, e);}
Expr* mk_expr_varref (Varref v) { return ExprManager::get_expr(v);}
Expr* mk_expr_const (Const c) { return ExprManager::get_expr(c);}
Expr* mk_expr_timeout () { return ExprManager::get_expr(Timeout);}
Expr* mk_expr_chan_init (Ty_list t) { return ExprManager::get_expr(t);}

/* operators */

op mk_op_lor () { return LorOp; }
op mk_op_land () { return LandOp; }
op mk_op_or () { return OrOp; }
op mk_op_xor () { return XorOp; }
op mk_op_and () { return AndOp; }
op mk_op_equal () { return EqualOp; }
op mk_op_not_equal () { return NotEqualOp; }
op mk_op_less () { return LessOp; }
op mk_op_greater () { return GreaterOp; }
op mk_op_less_equal () { return LessEqualOp; }
op mk_op_greater_equal () { return GreaterEqualOp; }
op mk_op_lshift () { return LshiftOp; }
op mk_op_rshift () { return RshiftOp; }
op mk_op_plus () { return PlusOp; }
op mk_op_minus () { return MinusOp; }
op mk_op_times () { return TimesOp; }
op mk_op_div () { return DivOp; }
op mk_op_mod () { return ModOp; }
op mk_op_ite () { return IteOp; }
uop mk_uop_plus () { return PlusUOp; }
uop mk_uop_minus () { return MinusUOp; }
uop mk_uop_neg () { return NegUOp; }
uop mk_uop_not () { return NotUOp; }
/*
uop mk_uop_len () { return LenUOp; }
uop mk_uop_full () { return FullUOp; }
uop mk_uop_empty () { return EmptyUOp; }
uop mk_uop_nfull () { return NFullUOp; }
uop mk_uop_nempty () { return NEmptyUOp; }
*/



/* Varref */
Varref mk_varref_symbol (Symbol s) { return Varref(s);}
Varref mk_varref_array (Symbol s, Expr* e) { return Varref(ArrayRef(s, e));}
Varref mk_varref_bv (Symbol s, Expr* e) { return Varref(BvRef(s, e));}


/* Const*/

Const mk_const_int (Num n) { return Const(n); }
Const mk_const_bool (bool b)  { return Const(b); }

/* Send_arg, Recv_arg */
/*typedef Expr_list Send_args;
typedef VAR::variant<Const,Varref> Recv_arg;
typedef list<Recv_arg> Recv_args;
*/
Recv_arg mk_recv_arg_varref (Varref v) { return Recv_arg(v);}
Recv_arg mk_recv_arg_const (Const c) { return Recv_arg(c);}

/* sequence, stmts */
/*class Opts;
class Sequence;
typedef Varref Xr;
typedef Varref Xs;
typedef Opts* If;
typedef Opts* Do;
enum Stmt_enum {Else, Break};
typedef TUP::tuple<Varref,Expr*> Assign;
typedef Varref Inc;
typedef Varref Dec;
typedef Symbol Goto;
typedef Symbol Label;
typedef Expr* Assert;
typedef Sequence* Atomic;
typedef TUP::tuple<Varref,Send_args> Send;
typedef TUP::tuple<Varref,Recv_args> Recv;
typedef VAR::variant<Xr,Xs,Expr*,If,Do,Stmt_enum,Assign,Inc,Dec,Goto,Label,Assert,Atomic,Send,Recv> SStmt;
*/

/* Step */
/*typedef VAR::variant<Sequence*,SStmt> Stmt;
typedef VAR::variant<Var_decls, Stmt> Step;

// Opt
struct Sequence : list<Step> {} ;
typedef Sequence Opt;
struct Opts : list<Opt> {} ;
*/

//SStmt mk_stmt_xr (Varref v) {  return SStmt(Xr(v));}
//SStmt mk_stmt_xs (Varref v) { return SStmt(Xs(v));}
SStmt mk_stmt_expr (Expr* e) { return SStmt(e);}
SStmt mk_stmt_if (Opts o) { return SStmt(If(new Opts(o)));}
SStmt mk_stmt_do (Opts o) { return SStmt(Do(new Opts(o)));}
SStmt mk_stmt_else () { return SStmt(ELSE_STMT);}
SStmt mk_stmt_pause () { return SStmt(PAUSE_STMT);}
SStmt mk_stmt_break () { return SStmt(Break);}
SStmt mk_stmt_assign (Varref& v, Expr* e) { return SStmt(Assign(v, e));}
SStmt mk_stmt_dassign (Varref& v, Expr* e) { return SStmt(DAssign(v, e));}
SStmt mk_stmt_inc (Varref& v) { 
    Expr* e = ExprManager::get_expr(
                mk_expr_varref(v),mk_op_plus(),mk_expr_const(1) 
              );
    SStmt s(Assign(v,e)); return s;
}

SStmt mk_stmt_dec (Varref& v) { 
    Expr* e = ExprManager::get_expr(
                mk_expr_varref(v),mk_op_minus(),mk_expr_const(1) 
              );
    SStmt s(Assign(v,e)); return s;
}

SStmt mk_stmt_goto (Symbol s) { return SStmt(Goto(s));}
SStmt mk_stmt_label (Symbol s) { return SStmt(Label(s));}
SStmt mk_stmt_assert (Expr* e) { return SStmt(Assert (e));}
//SStmt mk_stmt_atomic (Sequence q) { return SStmt(Atomic(new Sequence(q)));}
SStmt mk_stmt_send (Varref v, Send_args a) { return SStmt((Send (v, a)));}
SStmt mk_stmt_recv (Varref v, Recv_args a) { return SStmt((Recv (v, a)));}


Step mk_step_var_decl (Var_decls v) { return Step(v);}
Step mk_step_stmt (SStmt s) { 
    //Sequence* seq = new Sequence(); 
    //seq->push_back(Step(s)); 
    return Step(Stmt(s));
}


Opt mk_opt (Sequence q) { return Opt(q);}


//Type_decl, Proc_decl, Init_decl
/*typedef TUP::tuple<Symbol,Expr_list> Run_step;
typedef list<Run_step> Run_step_list;
typedef Run_step_list Init_decl;
Run_step mk_run_step (Symbol s, Expr_list l) {return  Run_step(s, l);}
Init_decl mk_init_decl (Run_step_list l) { return l; }
*/

/*typedef TUP::tuple<Symbol, Symbol_list> Type_decl; // type name, enum values
typedef TUP::tuple<int, Symbol, Var_decls, Sequence> Proc_decl;
*/
Var_decl mk_var_decl (Symbol s, Ty t, Expr* e) { return Var_decl(t, Varspec(Var_dcl(s),e));}
Type_decl mk_type_decl (Symbol_list l, Symbol s) { 
    CHK( s!= Symbol());
    return Type_decl(s, l);
}

Proc_decl mk_proc_decl (int n, Symbol s, Var_decls& v, Body& b) { return Proc_decl (n, s, v, b); }
Body mk_body (SStmts& logic, SStmts& init, Sequence& seq) { return Body(logic, init, seq); }

// Decl
/*typedef VAR::variant<Var_decl, Type_decl, Proc_decl, Init_decl>  Decl;
typedef list<Decl> Decls;
*/
Decls mk_decl_var (Var_decls v) { 
  Decls d;
  //d.push_back(v);
  for(Var_decls::iterator it= v.begin(); it!= v.end(); it++) { d.push_back(Decl(*it));  }    
  return d;
  //return Decl(v);
}
Decl mk_decl_type (Type_decl t) { return Decl(t);}
Decl mk_decl_proc (Proc_decl p) { return Decl(p);}
//Decl mk_decl_init (Init_decl i) { return Decl(i);}
}
