#ifndef AST_H
#define AST_H

#include<string>
#include<iostream>
#include"common.h"
#include"expr.h"
#include"stmts.h"


namespace Ast {

typedef list<Expr*> Expr_list;

/* constructor function for each of the node types */
Num mk_number (int n);

typedef list<Symbol> Symbol_list;
Symbol mk_symbol (string s);


/* Decl types */
Var_dcl mk_simplevardcl(Symbol s) ;
/*Var_dcl mk_bitvardcl(Symbol s, Num n) ; */
Var_dcl mk_arrayvardcl(Symbol s, Num n) ;

//Varspec
typedef TUP::tuple<Var_dcl, Expr*> Varspec;
typedef list<Varspec> Varspec_list;
Varspec mk_varspec (Var_dcl v, Expr* e = NULL) ;

// Var_decl
typedef TUP::tuple<Ty, Varspec> Var_decl;
typedef list<Var_decl> Var_decl_list;
typedef Var_decl_list Var_decls;

Var_decl_list declare_vars(Ty t, Varspec_list lv) ;






/* elementary types */
Ty mk_ty_bit () ;
Ty mk_ty_bool () ;
Ty mk_ty_byte () ;
Ty mk_ty_int () ;
Ty mk_ty_ev() ;
Ty mk_ty_chan() ;
Ty mk_ty_mtype () ;
Ty mk_ty_pid () ;
Ty mk_ty_short () ;
Ty mk_ty_unsigned () ;
Ty mk_ty_named (Symbol s) ;
Ty mk_ty_bv (Num w) ;
Ty mk_ty_array (bTy t, Num s) ;



/* expressions */


Expr* mk_expr_binop (Expr* l, op o, Expr* r) ;
Expr* mk_expr_ternop (op o, Expr* l, Expr* m, Expr* r) ;
Expr* mk_expr_unop (uop u, Expr* e) ;
Expr* mk_expr_varref (Varref v) ;
Expr* mk_expr_const (Const c) ;
Expr* mk_expr_timeout () ;
Expr* mk_expr_chan_init (Ty_list t) ;

/* operators */

op mk_op_lor () ;
op mk_op_land () ;
op mk_op_or () ;
op mk_op_xor () ;
op mk_op_and () ;
op mk_op_equal () ;
op mk_op_not_equal () ;
op mk_op_less () ;
op mk_op_greater () ;
op mk_op_less_equal () ;
op mk_op_greater_equal () ;
op mk_op_lshift () ;
op mk_op_rshift () ;
op mk_op_plus () ;
op mk_op_minus () ;
op mk_op_times () ;
op mk_op_div () ;
op mk_op_mod () ;
op mk_op_ite () ;
uop mk_uop_plus () ;
uop mk_uop_minus () ;
uop mk_uop_neg () ;
uop mk_uop_not () ;
uop mk_uop_len () ;
uop mk_uop_full () ;
uop mk_uop_empty () ;
uop mk_uop_nfull () ;
uop mk_uop_nempty () ;




/* Varref */
Varref mk_varref_symbol (Symbol s) ;
Varref mk_varref_array (Symbol s, Expr* e) ;
Varref mk_varref_bv (Symbol s, Expr* e) ;


/* Const*/

Const mk_const_int (Num n) ;
Const mk_const_bool (bool b)  ;

/* Send_arg, Recv_arg */
//class Send_arg : Expr*
//typedef vector<Send_arg> Send_args;
//class Recv_arg : public VAR::variant<Const,Varref> 
//typedef vector<Recv_arg> Recv_args;
Recv_arg mk_recv_arg_varref (Varref v) ;
Recv_arg mk_recv_arg_const (Const c) ;


/* sequence, stmts */
class Opts;
class Sequence;
struct Xr  { Varref v;  Xr() {} Xr(const Varref& _v) : v(_v){} };
struct Xs  { Varref v; Xs() {} Xs(const Varref& _v) : v(_v){} };
struct Inc  { Varref v; Inc() {} Inc(const Varref& _v) : v(_v){} };
struct Dec { Varref v;  Dec() {} Dec(const Varref& _v) : v(_v){} };
struct If : TUP::tuple<const Opts*> { If(const Opts* opts_ptr) : TUP::tuple<const Opts*>(opts_ptr){} };
struct Do : TUP::tuple<const Opts*> { Do(const Opts* opts_ptr) : TUP::tuple<const Opts*>(opts_ptr){} };
enum Stmt_enum {PAUSE_STMT, ELSE_STMT, Break, Continue};
//typedef TUP::tuple<Varref,Expr*> Assign;
struct Goto : Symbol { Goto(const Symbol& s) : Symbol(s){} };
struct Label : Symbol { Label(const Symbol& s) : Symbol(s) {} };
struct Assert : TUP::tuple<Expr*> { Assert(Expr* e) : TUP::tuple<Expr*>(e){} };
typedef Sequence* Atomic;
//typedef TUP::tuple<Varref,Send_args> Send;
//typedef TUP::tuple<Varref,Recv_args> Recv;
typedef VAR::variant<Expr*,If,Do,Stmt_enum,Assign,DAssign,Inc,Dec,Goto,Label,Assert,Send,Recv> SStmt;
typedef list<SStmt> SStmts;


/* Step */
typedef VAR::variant<Sequence*,SStmt> Stmt;
typedef VAR::variant<Var_decls, Stmt> Step;

// Opt
struct Sequence : list<Step> {} ;
typedef Sequence Opt;
struct Opts : list<Opt> {} ;

//SStmt mk_stmt_xr (Varref v) ;
//SStmt mk_stmt_xs (Varref v) ;
SStmt mk_stmt_expr (Expr* e) ;
SStmt mk_stmt_if (Opts o) ;
SStmt mk_stmt_do (Opts o) ;
SStmt mk_stmt_else () ;
SStmt mk_stmt_break () ;
SStmt mk_stmt_pause () ;
SStmt mk_stmt_assign (Varref& v, Expr* e) ;
SStmt mk_stmt_dassign (Varref& v, Expr* e) ;
SStmt mk_stmt_inc (Varref& v) ;
SStmt mk_stmt_dec (Varref& v) ;
SStmt mk_stmt_goto (Symbol s) ;
SStmt mk_stmt_label (Symbol s) ;
SStmt mk_stmt_assert (Expr* e) ;
//SStmt mk_stmt_atomic (Sequence q) ;
SStmt mk_stmt_send (Varref v, Send_args a) ;
SStmt mk_stmt_recv (Varref v, Recv_args a) ;


Step mk_step_var_decl (Var_decls v) ;
Step mk_step_stmt (SStmt s) ;


Opt mk_opt (Sequence q) ;


//Type_decl, Proc_decl, Init_decl
/*typedef TUP::tuple<Symbol,Expr_list> Run_step;
typedef list<Run_step> Run_step_list;
typedef Run_step_list Init_decl;
Run_step mk_run_step (Symbol s, Expr_list l) ;
Init_decl mk_init_decl (Run_step_list l = Run_step_list()) ;
*/

typedef TUP::tuple<SStmts, SStmts, Sequence> Body; /* logic, init, sequence */
typedef TUP::tuple<Symbol, Symbol_list> Type_decl; // type name, enum values
typedef TUP::tuple<int, Symbol, Var_decls, Body> Proc_decl;

Var_decl mk_var_decl (Symbol s, Ty t, Expr* e) ;
Type_decl mk_type_decl (Symbol_list l, Symbol s= Symbol()) ;
Proc_decl mk_proc_decl (int n, Symbol s, Var_decls& v, Body& b) ;
Body mk_body (SStmts&, SStmts&, Sequence&);

// Decl
typedef VAR::variant<Var_decl, Type_decl, Proc_decl>  Decl;
typedef list<Decl> Decls;
Decls mk_decl_var (Var_decls v) ;
Decl mk_decl_type (Type_decl t) ;
Decl mk_decl_proc (Proc_decl p) ;
//Decl mk_decl_init (Init_decl i) ;


}

using namespace Ast;
#endif
