#ifndef _BVALS_H
#define _BVALS_H

#include"common.h"

typedef string Symbol;
typedef int Num;
class Const;

string to_str_num(Num n);

struct Bv_const {
    uint v;
    uint size;

    Bv_const (uint _v, uint _size) : v(_v), size(_size) {}
    string to_str() const {  
        string s; s +=  int2str(v) + "(" + int2str(size) + ")";
        return s;
    }
    uint getv() const { return v; } 
    uint get_size() const { return size; } 
    string get_bv_string() const {
        string s = Util::num_to_bv_string(v);
        if( size > s.size()) {
            string buf(size-s.size(), '0');
            s = buf + s;
        }
        return s;
    }
};

struct Const_to_str {
    string str;

    Const_to_str(const Const& c) { VAR::apply_visitor(*this,c);}

    void operator() (bool b) {
        str = (b)?"T":"F";
    }

    void operator() (Num n) {
        str = to_str_num(n);
    }
    void operator() (const Bv_const& bv) {
        str = bv.to_str();
    }
    template<class T>
    void operator() (const T& t) {
        assert(0);
    }
};

struct Const_to_num {
    Num num;

    Const_to_num(const Const& c) { VAR::apply_visitor(*this,c);}

    void operator() (bool b) { num = (b)?1:0; }
    void operator() (Num n) { num = n; }
    void operator() (const Bv_const& bv) { num = bv.getv(); }
};


struct Const : public VAR::variant<bool, Num, Bv_const>  {
    typedef VAR::variant<bool,Num,Bv_const> Super;

    Const() {}
    Const (const bool& b) : Super(b) {}
    Const (const Num& n) : Super(n) {}
    Const (const Bv_const& b) : Super(b) {}
    bool operator < (const Const& c) const  { return to_str() < c.to_str(); }
    string to_str() const {
        Const_to_str vis(*this);
        return vis.str;
    }
    Num to_num() const { Const_to_num vis(*this); return vis.num; }
};

typedef set<Const> Consts;

#endif
