#include"bdd_dp.h"
#include"naming.h"
#include"cuddObj.hh"

BDD bdd_list_lt ( list<BDD>& l1, list<BDD>& l2, Cudd& mgr) {
    CHK( l1.size() == l2.size() );
    BDD f = mgr.bddZero();
    list<BDD>::iterator it1 = l1.end(); it1--; 
    list<BDD>::iterator it2 = l2.end(); it2--; 
    while ( 1 ) {
        BDD b1 = *it1, b2 = *it2;
        f = ((!b1) & b2) | (b1.Xnor(b2) & f); 
        if( it1 == l1.begin() ) break;
        it1--; it2--;
    }
    f &= (!l1.front() & l2.front());
    //if ( !l1.front() & l2.front() ) return mgr.bddZero();
    return f;
}

BDD bdd_list_gt ( list<BDD>& l1, list<BDD>& l2, Cudd& mgr) {
    CHK( l1.size() == l2.size() );
    BDD f = mgr.bddZero();
    list<BDD>::iterator it1 = l1.end(); it1--;
    list<BDD>::iterator it2 = l2.end(); it2--;
    while ( 1 ) {
        BDD b1 = *it1, b2 = *it2;
        f = (b1 & (!b2)) | (b1.Xnor(b2) & f); 
        if( it1 == l1.begin() ) break;
        it1--; it2--;
    }
    //if ( !l1.front() & l2.front() ) return mgr.bddOne();
    f |= (!l1.front() & l2.front());
    return f;
}

void mk_bdd_list_sum ( list<BDD>& l1, list<BDD>& l2, BDD carry_in, Cudd& mgr, list<BDD>& res ) {
    CHK( l1.size() == l2.size() );
    res.clear();
    list<BDD>::iterator it1 = l1.end(); it1--;
    list<BDD>::iterator it2 = l2.end(); it2--;
    //BDD carry = mgr.bddZero();
    BDD carry = carry_in;
    while ( 1 ) {
        BDD b1 = *it1, b2 = *it2;
        BDD sum = b1 ^ b2 ^ carry; 
        carry = (b1 & b2) | (b1 & carry) | (b2 & carry); 
        res.push_front( sum );
        if( it1 == l1.begin() ) break;
        it1--; it2--;
    }
}

void match_lengths ( list<BDD>& l1, list<BDD>& l2, Cudd& mgr ) {
    //sign extension
    BDD zero = mgr.bddZero();
    uint diff=0;
    if ( l1.size() < l2.size() ) {
        diff = l2.size() - l1.size();
        BDD l1front = l1.front();
        for(uint i=0; i<diff; i++)
            //l1.push_front(l1front);
            l1.push_front(zero);
    }
    else if ( l1.size() > l2.size() ) {
        diff = l1.size() - l2.size();
        BDD l2front = l2.front();
        for(uint i=0; i<diff; i++)
            l2.push_front(zero);
            //l2.push_front(l2front);
    }
    CHK( l1.size() == l2.size() );
    // - over sign extension
}

list<BDD> apply_op ( op o, const list<BDD>& tl1, const list<BDD>& tl2, Cudd& mgr ) {
    list<BDD> l1 = tl1, l2 = tl2, res;
    match_lengths(l1,l2,mgr);

    CHK( l1.size() == l2.size() );
    list<BDD>::iterator it1 = l1.begin();
    list<BDD>::iterator it2 = l2.begin();
    switch(o) {
        case LorOp: {
                 CHK ( l1.size() == 1 && l2.size() == 1 );
                 res.push_back(l1.front() | l2.front() );
                 break;
            }
        case OrOp: {
            while ( it1 != l1.end() ) {
                res.push_back( (*it1) | (*it2) );
                it1++; it2++;
            }
            break;
        }

        case AndOp: {
                 CHK ( l1.size() == 1 && l2.size() == 1 );
                 res.push_back( l1.front() & l2.front() );
                 break;
        }
        case LandOp: {
            while ( it1 != l1.end() ) {
                res.push_back( (*it1) & (*it2) );
                it1++; it2++;
            }
            break;
        }
        case EqualOp: {
            BDD eq_res = mgr.bddOne();
            while ( it1 != l1.end() ) {
                BDD b1 = *it1, b2 = *it2;
                eq_res &= b1.Xnor(b2);
                it1++; it2++;
            }
            res.push_back( eq_res );
            break;
        }
        case NotEqualOp: {
            BDD neq_res = mgr.bddZero();
            while ( it1 != l1.end() ) {
                BDD b1 = *it1, b2 = *it2;
                neq_res |= b1.Xor(b2);
                it1++; it2++;
            }
            res.push_back( neq_res );
            break;
        }

        case LessOp: {
            res.push_back( bdd_list_lt(l1,l2,mgr) );
            break;
        }
        case GreaterOp: {
            res.push_back ( bdd_list_gt(l1,l2,mgr) );
            break;
        }
        case LessEqualOp: {
            res.push_back ( !bdd_list_gt(l1,l2,mgr) );
            break;
        }
        case GreaterEqualOp: {
            res.push_back ( !bdd_list_lt(l1,l2,mgr) );
            break;
        }
        case PlusOp: {
            mk_bdd_list_sum ( l1, l2, mgr.bddZero(), mgr, res );
            break;
        }
        case MinusOp: {
            list<BDD> l2_;
            const_foreach(it, l2) {
                l2_.push_back(!(*it));
            }
            mk_bdd_list_sum( l1, l2_, mgr.bddOne(), mgr, res);
            break;
        }
        case TimesOp: 
        case XorOp: 
        case DivOp: //return (v1 / v2);
        case LshiftOp:// return (v1 << v2);
        case RshiftOp: //return (v1 >> v2);
        case ModOp:// return (v1 % v2);
        default: assert(0);
    }

    return res;
}

list<BDD> apply_uop ( uop o, const list<BDD>& v, Cudd& mgr ) {
  switch(o) {
      case PlusUOp: return v;

      case NegUOp:
      case NotUOp:  {
        list<BDD> res;
        const_foreach(it, v) {
            res.push_back( !(*it) );
        }
        return res;
      }

      case MinusUOp: //return -v;
      default: assert(0);
  }
}

struct Const_sym_eval_bdd {
    Cudd& mgr;
    list<BDD> result;
    
    Const_sym_eval_bdd(const Const& c,  Cudd& _mgr) : mgr(_mgr) { VAR::apply_visitor(*this,c);}

    void operator() (bool b) {
        if(b) { result.push_back(mgr.bddOne()); }
        else { result.push_back(mgr.bddZero()); }
    }

    void operator() (Num n) {
        string s = Util::num_to_bv_string(n);
        for(uint i=0; i< s.size(); ++i) {
            if ( s[i] == '0' )
                result.push_back(mgr.bddZero());
            else if ( s[i] == '1')
                result.push_back( mgr.bddOne() );
            else assert(0);
        }
    }

    void operator() (const Bv_const& bv) {
        string s = bv.get_bv_string();
        for(uint i=0; i< s.size(); ++i) {
            if ( s[i] == '0' )
                result.push_back(mgr.bddZero());
            else if ( s[i] == '1')
                result.push_back( mgr.bddOne() );
            else assert(0);
        }
    }

    template<class T> void operator() (const T& t) {assert(0); }
};


list<BDD> eval_const_bdd(const Const& c, Cudd& mgr) { 
    Const_sym_eval_bdd vis_c(c,mgr);
    return vis_c.result;
}
  
struct Expr2BDDvis {
    Cudd &mgr;
    map<string, list<BDD> >& name2bdd;

    list<BDD> result;

    Expr2BDDvis (Expr* e, Cudd& _mgr, map<string, list<BDD> >& _name2bdd) : mgr(_mgr), name2bdd(_name2bdd) {
        CHK ( e != NULL );
        VAR::apply_visitor(*this,*e);
    }

    void operator() (const BinOp& bop) {
        const Expr* e1 = TUP::get<0>(bop);
        op o = TUP::get<1>(bop);
        const Expr* e2 = TUP::get<2>(bop);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;

        list<BDD> val1;

        apply_visitor(*this,et1);
        val1 = result;
        apply_visitor(*this,et2);
        result = apply_op(o,val1, result,mgr);
    }

    void operator() (const TernOp& e)  {
        op o = TUP::get<0>(e);
        CHK ( o == IteOp );
        const Expr* e1 = TUP::get<1>(e);
        const Expr* e2 = TUP::get<2>(e);
        const Expr* e3 = TUP::get<3>(e);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;
        const Exprt& et3 = *e3;

        apply_visitor(*this,et1);
        CHK ( result.size() == 1);
        BDD c = result.front();
        apply_visitor(*this,et2);
        list<BDD> t = result;
        apply_visitor(*this,et3);
        list<BDD> f = result;
        match_lengths(t,f,mgr);

        result.clear();
        list<BDD>::iterator itt = t.begin(), itf = f.begin();
        while ( itt != t.end() ) {
            result.push_back( c.Ite( *itt, *itf ) );
            itt++; itf++;
        }
    }

    void operator() (const UnOp& e) {
        //Util::msg(2,"unop: ");
        uop o = TUP::get<0>(e);
        const Expr* e1 = TUP::get<1>(e);
        const Exprt& et1 = *e1;

        apply_visitor(*this,et1);
        result = apply_uop(o, result,mgr);
        //DP::pp_expr(result);
    }

    void operator() (const Varref& vr) {
        //Util::msg(2,"{varref: ");
        string name = vr.to_str();
        CHK ( Key_In_Bag(name,name2bdd) );
        result = name2bdd[name];
    }

    void operator() (const Const& c) {
        result = eval_const_bdd(c,mgr);
    }

    void operator() (const Expr_Enum& ee) {
        assert(0);
    }

    void operator() (const ChanInit& ci) { //all channel inits are static
        assert(0);
    }
    template<class T>
        void operator() (const T& t) { assert(0); }
};

BDD BDD_DP::get_init_bdd () {
    CHK(initialized);
    Expr2BDDvis vis(init_e,mgr,name2bdd);
    CHK( vis.result.size() == 1);
    return *(vis.result.begin());
}


void BDD_DP::get_trans_bdd ( list<BDD>& trans_bdds ) {
    const_foreach(it, trans_vec) {
        Util::msg(2,"build trans iteration");
        Expr2BDDvis vis(*it,mgr,name2bdd);
        CHK( vis.result.size() == 1);
        trans_bdds.push_back( *(vis.result.begin()) );
    }
}

BDD BDD_DP::get_err_bdd() {
    CHK( err_vec.size() == 1);
    Expr2BDDvis vis(err_vec.front(),mgr,name2bdd);
    CHK( vis.result.size() == 1);
    return *(vis.result.begin());
}

void BDD_DP::add_var ( const string& name, const Ty& ty ) {
    if ( is_decl(name) ) {
        //assert(0);
        //Util::msg(2,name+" already declared");
        return;
    }
    //Util::msg(2,"adding var "+name+" of type "+ty2str(ty));
    string tystr = ty2str(ty);

    string oname; Pid pid; int depth;
    Naming::get_orig_name_pid_depth(name, oname, pid, depth);
    CHK (depth == 0);
    string name_ = Naming::get_glb_name(oname, pid, 1);
    var2ty[name] = ty;
    var2ty[name_] = ty;
    curr_next_var_list.push_back( pair<string,string>(name,name_) );

    if ( tystr == "bool" ) {
        BDD b = mgr.bddVar(bdd_var_cnt);
        BDD b_ = mgr.bddVar(bdd_var_cnt+1);
        name2bdd[name].push_back(b);
        name2bdd[name_].push_back(b_);
        bdd_var_cnt += 2;
    }
    else if ( tystr == "int" ) {
        //10 vars
        assert(0);
    }
    else if ( is_ty_bv(ty) ) {
        uint size = get_ty_size(ty);
        uint bits = (uint)log2(size)+1;
     //   Util::message(2,"add_var: size = %d, bits = %d, log=%f\n",size,bits,log(size));

        /*
        for(uint i=0; i<bits; i++) {
            BDD b = mgr.bddVar(bdd_var_cnt);
            name2bdd[name].push_back(b);
            bdd_var_cnt += 1;
        }

        for(uint i=0; i<bits; i++) {
            BDD b_ = mgr.bddVar(bdd_var_cnt);
            name2bdd[name_].push_back(b_);
            bdd_var_cnt += 1;
        }
        */
        for(uint i=0; i<bits; i++) {
            BDD b = mgr.bddVar(bdd_var_cnt);
            name2bdd[name].push_back(b);
            BDD b_ = mgr.bddVar(bdd_var_cnt+1);
            name2bdd[name_].push_back(b_);
            bdd_var_cnt += 2;
        }
    }
    else assert(0);
}

bool BDD_DP::check_err_reach (int& err_depth) {
    mgr.AutodynEnable(CUDD_REORDER_SYMM_SIFT);
    //Util::msg(2,"get init");
    BDD init = get_init_bdd();
    //Util::msg(2,"get trans");
    list<BDD> trans_bdds;
    get_trans_bdd(trans_bdds);
    //Util::msg(2,"get err");
    BDD err = get_err_bdd();

    BDD reach = init;
    BDD frontier = init;

// -- building the curr next map
    uint N = 0;
    BDD curr_cube = mgr.bddOne();
    const_foreach(it, curr_next_var_list) {
        string name = it->first;
        N += name2bdd[name].size();
        list<BDD> curr_bdd_list = name2bdd[name];
        foreach(it, curr_bdd_list) {
            curr_cube &= *it;
        }
    }
    BDDvector curr(N),next(N);
    uint i=0;
    const_foreach(it, curr_next_var_list) {
        //Util::message(2,"indices are %d,%d\n",it->first,it->second);
        string name = it->first, name_ = it->second;
        list<BDD> curr_bdd_list = name2bdd[name];
        list<BDD> next_bdd_list = name2bdd[name_];

        list<BDD>::iterator currit = curr_bdd_list.begin(), nextit = next_bdd_list.begin();
        while (1) {
            curr[i] = *currit;
            next[i] = *nextit;
            i++;
            currit++; nextit++;
            if( currit == curr_bdd_list.end() ) break;
        }
    }

    int num_it=0;;

    //Util::msg(2,"err is "); err.PrintMinterm(); Util::msg(2,"");

        //Util::msg(2,"err_reach is "); err_reach.PrintMinterm();
    while (1) {
        Util::msg(2,"computing fixpoint with BDDs..iteration = "+int2str(num_it));
        BDD err_reach = frontier & err;
        //Util::msg(2,"frontier is "); frontier.PrintMinterm(); Util::msg(2,"");
        if ( err_reach != mgr.bddZero() ) {
            Util::msg(2,"Error reached..");
            err_depth = num_it;
            return true;
        }
        const_foreach(it, trans_bdds) {
            //image = frontier.AndAbstract(trans,curr_cube);
            frontier &= *it;
        }
        BDD image = frontier.ExistAbstract(curr_cube);
        image = image.SwapVariables(next,curr);
        //Util::msg(2,"image is "); image.PrintMinterm(); Util::msg(2,"");
        
        frontier = image & !reach;
        if ( frontier == mgr.bddZero() ) break;
        reach |= frontier;
        num_it++;
    }
    return false;
}
