#ifndef _EXPR_DP_H
#define _EXPR_DP_H
#include"basic_vals.h"
#include"expr.h"
#include"yices_c.h"
#include"cuddObj.hh"


class BDD_DP {
    public:
        typedef Expr* expr ;
        typedef Ty DP_type;
    protected:
        map<string,Ty> var2ty;
        bool initialized;
        expr init_e;
        vector<expr> asserts;
        vector<expr> trans_vec, err_vec;

        Cudd mgr;
        uint bdd_var_cnt;
        map<string, list<BDD> > name2bdd;
        list< pair<string,string> > curr_next_var_list;

        bool is_decl ( const string& name ) const {
            //Util::msg(2,"is_decl: "+name);
            return Key_In_Bag(name,var2ty);

        }

        Ty get_ty ( const string& name ) const {
            CHK ( is_decl(name) );
            return var2ty.find(name)->second;
        }

        void add_var ( const string& name, const Ty& ty ) ;
    public:


        BDD_DP() : mgr(40,0) {
            initf();
        }

        void initf() {
            initialized = false;
            bdd_var_cnt = 0;
        }

        void set_init(expr inite) { init_e = inite; initialized = true; }
        expr get_init() { CHK ( initialized ); return init_e; }

        /*void set_trans(uint depth, expr e) {
            CHK(trans_vec.size() == depth);
            trans_vec.push_back( e );
        }*/
        //conj partitioned trans relation
        void add_trans(expr e) {
            trans_vec.push_back(e);
        }

        void set_err(uint depth, expr e) {
            CHK(err_vec.size() == depth);
            err_vec.push_back( e );
        }

        void mk_var_decl(const string& name, const Ty& ty) {
            add_var(name,ty);
        }

        expr mk_bool_var(const string& name) { 
            add_var(name, Ty(BoolTy) );
            Varref vr(name);
            return ExprManager::get_expr(vr);
        }

        DP_type mk_type (const Ty& ty) {
            return ty;
        }

        expr get_var_expr(const string& name) {
            CHK ( is_decl(name) );
            Varref vr(name);
            return ExprManager::get_expr(vr);
        }
        expr mk_const_expr(const Const& c) {
            return ExprManager::get_expr(c);
        }

        void create_model ( vector< pair<string,Const> >& model ) { assert(0); }
        void get_model(expr err, map<string,Const>& model, int& max_d, bool generalize, assertion_id ) {
            assert(0);
        }

        lbool check() { assert(0); }

        void show_sat () { assert(0); }

        expr mk_true() { return ExprManager::get_true(); }
        expr mk_false() { return ExprManager::get_false(); }
        expr mk_num(int n) { return ExprManager::get_expr(Const(n)); }
        expr eval_const(const Const& c) { return ExprManager::get_expr(c); }

        /*expr mk_sum(expr a[], int n) { return yices_mk_sum(dp_ct,a,n); }
        expr mk_sub(expr a[], int n) { return yices_mk_sub(dp_ct,a,n); }
        expr mk_mul(expr a[], int n) { return yices_mk_mul(dp_ct,a,n); }

        expr mk_and(expr a[], int n) { return yices_mk_and(dp_ct,a,n); }
        expr mk_or(expr a[], int n) { return yices_mk_or(dp_ct,a,n); }
        */
        expr mk_sum (expr a, expr b) { return ExprManager::get_sum(a,b); }
        expr mk_sub (expr a, expr b) { return ExprManager::get_sub(a,b); }
        expr mk_mul (expr a, expr b) { return ExprManager::get_mul(a,b); }

        expr mk_and(expr a, expr b) { return ExprManager::get_and(a,b); }
        expr mk_or(expr a, expr b) {  return ExprManager::get_or(a,b); }

        expr mk_diseq(expr a, expr b) { return ExprManager::get_diseq(a,b); }
        expr mk_eq(expr a, expr b) { return ExprManager::get_eq(a,b); }
        expr mk_ite( expr c, expr t, expr f) { return ExprManager::get_ite(c, t, f); }

        expr mk_not(expr a) { return ExprManager::get_not(a); }

        expr mk_le(expr a, expr b) { return ExprManager::get_le(a,b); }
        expr mk_lt(expr a, expr b) { return ExprManager::get_lt(a,b); }
        expr mk_ge(expr a, expr b) { return ExprManager::get_ge(a,b); }
        expr mk_gt(expr a, expr b) { return ExprManager::get_gt(a,b); }

        void push() { assert(0); }
        void pop() { assert(0); }

        void assertdp(expr e) { assert(0); }
        assertion_id assert_retract_dp(expr e) { assert(0); }

        void retract ( assertion_id id )  { assert(0); }
        void retract_all(vector<expr>& exprs ) { assert(0); }

        expr compute_interpolant ( expr A, expr B ) { assert(0); }
        expr compute_interpolant ( int depth ) { assert(0); }

        //bool inconsistent() { return yices_inconsistent(dp_ct); }

        BDD get_init_bdd ();
        void get_trans_bdd (list<BDD>&) ;
        BDD get_err_bdd ();

        bool check_err_reach (int& err_depth) ;


};

#endif
