#ifndef _CFA_H
#define _CFA_H


#include"action.h"
#include"fa.h"
#include"path.h"
#include"state.h"
#include"sys_state.h"
#include"eval_ea.h"
#include"rw_map.h"
#include"g_st_diff.h"
#include"pexpr.h"



typedef St_d<Const> Diff_acts;
class eCFA_act {

protected:
Diff_acts reads, writes;

public:
typedef eCFA_act Lab;

eCFA_act()  {}
eCFA_act(const Diff_acts& _reads, const Diff_acts& _writes) : reads(_reads), writes(_writes) {}

const Diff_acts& get_reads() const { return reads;}
const Diff_acts& get_writes() const { return writes;}

void get_glb_rw(RW& rw) const {
    reads.get_glb_rw(RW::R, rw);
    writes.get_glb_rw(RW::W, rw);
}

void set_val(size_t i, const Var_dcl& vd, const Const& c) {
        //Val v = eval_const(c);
        CHK( i ==0 || i== 1);
        if(i==0) reads.set_val(-1,vd,c);
        else writes.set_val(-1,vd,c);
}
void set_rw(const Diff_acts& _reads, const Diff_acts& _wr) { reads = _reads; writes = _wr; }

string to_str_const() const { 
    Util::msg(5,"cfa_action:to_str\n"); 
    string s = reads.to_str() + " " + writes.to_str(); 
    Util::msg(5,"cfa_action2:to_str\n"); 
    
    return s;
}
string to_str() { 
    Util::msg(5,"cfa_action:to_str\n"); 
    string s = reads.to_str() + " " + writes.to_str(); 
    Util::msg(5,"cfa_action2:to_str\n"); 
    
    return s;
}

eCFA_act get_eps() const { return eCFA_act(); }
bool is_eps() const { return (*this == get_eps()); }

bool operator == (const eCFA_act& act) const { return to_str_const() == act.to_str_const(); }
bool matches (const eCFA_act& act) const { return (*this == act); } 
bool operator < (const eCFA_act& act) const { return (to_str_const() < act.to_str_const()); }

void get_from_str(const string& str)  ;

void abstract_to_globals(const eCFA_act& rep, eCFA_act& abs) const {
    Diff_acts abs_reads = rep.get_reads();
    Diff_acts abs_writes = rep.get_writes();

    //compute abs_reads, abs_writes
    reads.abstract_to_globals(abs_reads);
    writes.abstract_to_globals(abs_writes);
    abs.set_rw(abs_reads, abs_writes);

}


};


typedef map<Pid,Pexpr> Pid2pexpr;

class sCFA_act {
protected:
Pid2pexpr pid2pexpr;
public:
typedef sCFA_act Lab;

    sCFA_act() { 
        //Expr* texp = ExprManager::get_expr( Const(true) );
        //reads = writes = texp;
    }

    sCFA_act ( const eCFA_act& eact ) {
        //Util::msg(2,"eact is "+eact.to_str());
        const Diff_acts& er= eact.get_reads();
        const Diff_acts& ew = eact.get_writes();

        //only allow globals
        CHK( er.get_psd().empty() && ew.get_psd().empty());
        const map<Var_dcl,Const>& gsdr = er.get_gsd().vars_diff;
        const_foreach(it, gsdr) {
            add_val(0, -1, Varref(Symbol(it->first.to_str())),it->second) ;
        }

        const map<Var_dcl,Const>& gsdw = ew.get_gsd().vars_diff;
        const_foreach(it, gsdw) {
            add_val(1, -1, Varref(Symbol(it->first.to_str())),it->second) ;
        }
        //Util::msg(2,"computed sact is "+to_str());
    }

    bool has_globals () const {
        return Key_In_Bag(-1,pid2pexpr);
    }
    bool has_globals_only() const {
        return (pid2pexpr.empty() || (has_globals() && pid2pexpr.size() ==1) );
    }

    Pexpr get_glb_pexpr () const { 
        if ( has_globals() ) return pid2pexpr.find(-1)->second;
        else return Pexpr();
    }
    const Pid2pexpr& get_pexpr() const { return pid2pexpr; }

    void get_glb_rw(RW& rw) const {
        get_glb_pexpr().get_rw(rw);
    }

    void set_glb_pexpr(const Pexpr& _pexpr) { pid2pexpr[-1] = _pexpr; }

    void set_val(size_t i, const Varref& vr, const Const& c) {
        //setting globals
        add_val(i,-1,vr,c);
    } 

    void add_val(size_t i, Pid pid, const Varref& vr, const Const& c) {
        pid2pexpr[pid].add_val(i,vr,c);
    }
    void add_neq_val(size_t i, Pid pid, const Varref& vr, const Const& c) {
        pid2pexpr[pid].add_neq_val(i,vr,c);
    }

    void set_r (Pid pid, BDD b) {
        pid2pexpr[pid].set_r(b);
    }
    void set_w (Pid pid, BDD b) {
        pid2pexpr[pid].set_w(b);
    }
    void set_r (Pid pid, const Pexpr& pexpr ) {
        //CHK ( pexpr.is_true_rw() );
        set_r (pid, pexpr.get_rw_bdd() );
    }
    void set_w (Pid pid, const Pexpr& pexpr ) {
        //CHK ( pexpr.is_true_rw() );
        set_w (pid, pexpr.get_rw_bdd() );
    }

    string to_str_const() const {
        string s;
        if (pid2pexpr.empty() ) { s = "(T)"; }
        const_foreach(it, pid2pexpr) {
            if ( it->first != -1 ) s += "[" + int2str(it->first) + "]" ;
            s +=  it->second.to_str_const() + " ";
        }
        return s;
    }

    string to_str() {
        string s;
        if (pid2pexpr.empty() ) { s = "(T)"; }
        foreach(it, pid2pexpr) {
            if ( it->first != -1 ) s += "[" + int2str(it->first) + "]" ;
            s +=  it->second.to_str() + " ";
        }
        return s;
    }

    string get_id() const {
        string s;
        const_foreach(it, pid2pexpr) {
            if ( it->first != -1 ) s += "[" + int2str(it->first) + "]" ;
            s +=  it->second.get_id() + " ";
        }
        return s;
    }

    bool operator == (const sCFA_act& act) const { return get_id() == act.get_id(); }
    void get_from_str(const string& str)  ;
    bool matches (const sCFA_act& act) const ;

    template<class DP>
    typename DP::expr get_c_expr_DP ( DP& dp, int depth ) const {
        typename DP::expr e = get_glb_pexpr().get_expr_DP(dp,depth);
        return e;
    }

    //use carefully
    bool operator < (const sCFA_act& act) const {  
        return get_id() < act.get_id();
    }
    

    void get_lifted ( set<sCFA_act>& acts, uint num ) {
        Pexpr p = get_glb_pexpr();
        set<Pexpr> lift_set;
        p.get_lifted(lift_set,num);
        foreach(it, lift_set) {
            sCFA_act act; act.set_glb_pexpr(*it);
            acts.insert(act);
        }
    }

    void abstract_to_globals(const sCFA_act& rep, sCFA_act& abs) const {
        //CHK ( has_globals() && rep.has_globals() ) ;
        CHK (  rep.has_globals() ) ;
        Pexpr my_pexpr = get_glb_pexpr();
        Pexpr rep_pexpr = rep.get_glb_pexpr();
        Pexpr abs_pexpr;

        my_pexpr.abstract_to_globals(rep_pexpr, abs_pexpr);
        abs.set_glb_pexpr(abs_pexpr);
    }
    
    void abstract_to_globals(sCFA_act& abs) const {
        CHK ( has_globals() );
        Pexpr my_pexpr = get_glb_pexpr();
        abs.set_glb_pexpr(my_pexpr);
    }

    set<sCFA_act> negate() const {
        CHK ( has_globals_only() );
        set<Pexpr> retps;
        set<sCFA_act> ret_acts;

        get_glb_pexpr().negate(retps);
        foreach(it, retps) {
            sCFA_act act; act.set_glb_pexpr(*it);
            ret_acts.insert(act);
        }
        return ret_acts;
    }


    sCFA_act and_with ( const sCFA_act& sact ) const {
        CHK ( has_globals_only() && sact.has_globals_only() );
        Pexpr res_pexpr;
        Pexpr sact_pexpr  = sact.get_glb_pexpr();
        get_glb_pexpr().and_with(sact_pexpr, res_pexpr );

        sCFA_act act; act.set_glb_pexpr(res_pexpr);
        return act;
    }

    static void union_set ( const set<sCFA_act>& sacts, sCFA_act& sact)  {
        Pexpr res_pexpr = Pexpr::get_false();
        const_foreach(it, sacts) {
            CHK ( it->has_globals_only() );
            Pexpr pexpr = it->get_glb_pexpr();
            res_pexpr.or_with(pexpr);
        }
        sact.set_glb_pexpr(res_pexpr);
    }


};

typedef set<eCFA_act> eCFA_actions;
typedef set<sCFA_act> sCFA_actions;

/* comm. finite automaton */
typedef FA<CFA_state, eCFA_act>  eCFA;
typedef DFA<CFA_state, eCFA_act>  eCDFA;

typedef FA<CFA_state, sCFA_act>  sCFA;
typedef DFA<CFA_state, sCFA_act>  sCDFA;
	
/*******  CFA_edge *********************/
typedef eCFA::Ev eCFA_edge ; 
typedef eCFA::Trace eCFA_trace;

typedef sCFA::Ev sCFA_edge ; 
typedef sCFA::Trace sCFA_trace;

/*********** define CFA for Prop ******/
//typedef eCFA PropFA;
//typedef eCDFA PropDFA;




#endif
