#ifndef _CFA_COMP_H
#define _CFA_COMP_H
#include"common.h"
#include"cfa.h"

template<class CDFAc>
struct CFA_compt {
    protected:
        const CDFAc* fa;
        Pid pid;
        bool fa_is_univ;

    public:
        typedef typename CDFAc::Sigma Sigma;
        typedef typename CDFAc::St St;
        typedef typename CDFAc::Ev Ev;

        CFA_compt() : fa(NULL) , pid (-1), fa_is_univ (false) {}
        CFA_compt ( const Sigma& univ_sigma)  {
            pid = Util::get_new_cfa_comp_pid(); 
            CDFAc* univ_fa = new CDFAc(true);
            CDFAc::get_univ_dfa(univ_sigma, *univ_fa);
            fa = univ_fa;
            fa_is_univ = true;
        }

        CFA_compt( const CDFAc& _fa) : fa(&_fa)  { 

            pid = Util::get_new_cfa_comp_pid(); 
            fa_is_univ = false; 
            //Util::msg(2,"CFA_comp: fa is "+ fa->to_str() );
        }

        //TODO
        //~CFA_comp() { if( fa_is_univ ) delete fa; }

        Pid get_pid() const { return pid; }
        //const CDFA& get_fa() const { return (fa); }

        const St& get_init() const { CHK( fa != NULL); return fa->get_init(); }

        void get_out_edges(const St& st, list<Ev>& ce_list) const {
            //Util::msg(2,"st is "+st.to_str());
            //fa.is_a_state(st);
            CHK( fa != NULL);
            fa->get_out_edges(st, ce_list);
        }

        //void get_comm_sigma(set<_Action_>& comm_sigma) const { fa.get_comm_sigma(comm_sigma); } 
        bool is_final(const St& s) const { CHK( fa!= NULL); return fa->is_final(s);}
        bool only_sink_is_final() const { return fa->only_sink_is_final(); }
        bool has_univ_fa() const { return fa_is_univ; }

        //is final formula for DP
        template<class DPt>
        typename DPt::expr is_final_f(DPt& dp, Pid pid, int depth) const {
            return fa_is_final_f((*fa),dp, pid, depth);
        }
        template<class DPt>
        typename DPt::expr get_init_f(DPt& dp, Pid pid, int depth) const {
            return fa_get_init_f((*fa),dp, pid, depth);
        }

        template<class DPt>
        typename DPt::expr get_trans_f(DPt& dp, Pid pid, int depth) const {
            return fa_get_trans_f((*fa), dp, pid, depth);
        }

        Ty get_st_var_type () const { return fa_get_st_var_type(*fa); }

};

typedef CFA_compt<eCDFA> eCFA_comp;
typedef CFA_compt<sCDFA> sCFA_comp;

#endif
