#ifndef _CFA_UTIL_H
#define _CFA_UTIL_H

#include"cfa.h"
#include"globals.h"
#include"eval_ea_sym.h"

template<class FAt>
void fa_get_glb_rw ( const FAt& fa, RW& rw ) {
    typedef typename FAt::Sigma Sigma;
    Sigma sigma; fa.get_sigma(sigma);
    typename_foreach(it, sigma) {
        it->get_glb_rw(rw);
    }
}

//CFA utils
template<class St, class DP>
typename DP::expr st2id ( const St& st, DP& dp, tag min_id ) {
    return dp.mk_num(st.tagi()- min_id);
}

template<class St, class DP>
typename DP::expr st2eq ( const St& st, DP& dp, typename DP::expr st_var, tag min_id ) {
    typedef typename DP::expr expr;
    expr id = st2id(st,dp,min_id);
    return dp.mk_eq(st_var,id);
}

template<class St, class DP>
typename DP::expr get_st_set_f ( const set<St>& sts, DP& dp, typename DP::expr st_var, tag min_id ) {
    typedef typename DP::expr expr;
    expr res = dp.mk_false();
    typename_const_foreach(it, sts) {
        const St& st = *it;
        expr eq = st2eq(st,dp,st_var,min_id);
        res = dp.mk_or(res,eq);
    }
    return res;
}

template<class DP>
typename DP::expr get_fa_state_var(DP& dp, int pid, int depth) {
    return Naming::get_state_var(dp,pid,depth);
}

template<class FAt>
Ty fa_get_st_var_type ( const FAt& fa ) {
    uint size = fa.get_num_nodes() ;
    uint bits = (uint)log2(size)+1;  //+1 just to be sure
    return BvTy(bits); 
}

/*   fa_is_final_f, fa_get_init_f */

template<class FAt, class DP>
typename DP::expr fa_is_final_f( const FAt& fa, DP& dp, int pid, int depth ) {
    typedef typename FAt::St St;
    typedef typename DP::expr expr;
    const set<St>& sts = fa.get_final();
    tag min_id = fa.get_min_st_id();
    expr st_var = get_fa_state_var(dp,pid,depth);
    return get_st_set_f(sts,dp,st_var,min_id);
}

template<class FAt, class DP>
typename DP::expr fa_get_init_f( const FAt& fa, DP& dp, int pid, int depth=0 ) {
    typedef typename FAt::St St;
    typedef typename DP::expr expr;
    const set<St>& sts = fa.get_inits();
    tag min_id = fa.get_min_st_id();
    expr st_var = get_fa_state_var(dp,pid,depth);
    return get_st_set_f(sts,dp,st_var,min_id);
}




/* fa_get_edges_f, fa_get_trans_f */


// NOTE: only global values are translated to symbolic form
template<class St, class Ev, class DP>
typename DP::expr fa_get_edges_f( const map<Ev,St>& edges, tag min_id, DP& dp, typename DP::expr serr, int depth) {

    typedef typename DP::expr expr;
    expr res = serr;

    typename_const_foreach( it, edges ) {
        const Ev& ev = it->first;
        const St& st_ = it->second;
        
        
        expr c_expr = ev.get_act().get_c_expr_DP(dp, depth);

        expr t_expr = st2id(st_,dp,min_id); 
        res = dp.mk_ite( c_expr, t_expr, res );
    }
    return res;

}


template<class FAt, class DP>
typename DP::expr fa_get_trans_f( const FAt& fa, DP& dp, Pid pid, int depth) {

    typedef typename FAt::St St;
    typedef typename FAt::Ev Ev;
    typedef typename DP::expr expr;
    typedef map<St, map<Ev,St> > trans_t; 

    tag min_id = fa.get_min_st_id();

    //expr sexp = Naming::get_state_var(dp,pid,depth);
    expr sexp = get_fa_state_var(dp,pid,depth);
    //expr sexp_ = Naming::get_state_var(dp,pid,depth+1);
    expr sexp_ = get_fa_state_var(dp,pid,depth+1);

    const St& err = fa.get_sink(); // All sCFA must have sink
    expr serr = st2id(err,dp,min_id);

    expr res = serr;
    const trans_t& trans = fa.get_trans();
    typename_const_foreach ( its, trans ) {
        const St& st = its->first;
        expr cond_expr = st2eq(st,dp,sexp,min_id);
        const map<Ev,St>& edges = its->second;
        expr t_expr = fa_get_edges_f(edges, min_id, dp, serr, depth); 
        res = dp.mk_ite(cond_expr, t_expr, res);
    }
    
    return dp.mk_eq(sexp_,res); 

    // R_k = ite( e_i.cond, opp_i, R_{k-1} ), R_0 = err, i ranges over edges[s_k]
    // S_k = ite( (s==s_k), R_k, S_{k-1} ), S_0 = err
    // s_ = S_n where n is no of states in trans
}


template<class DP>
typename DP::expr fa_get_trace_f( DP& dp, const sCFA_trace& t, int base_depth) {

    typedef typename DP::expr expr;
    expr res = dp.mk_true();
    int depth = base_depth;

    sCFA_trace::const_iterator it = t.begin();
    for(; it != t.end(); ++it ) {
        const sCFA_edge& ev = (*it);
        expr c_expr = ev.get_act().get_c_expr_DP(dp, depth);
        res = dp.mk_and(res,c_expr);
        depth++;
    }
    return res;
}


#endif
