#include"cfg.h"
#include"globals.h"

const CFG_guard CFG_stmt::guard_true = (const Expr*)ExprManager::get_true();
const CFG_guard CFG_stmt::guard_else = (const Expr*)ExprManager::get_expr(Else_expr);



string CFG_stmt::to_str_const() const {
    string op;
    //CFG_guard_tostr vg;
    //cout<<"called Stmt::to_str"<<endl;
    if( g == guard_true ) op = "(T)";
    else op = "("+g->to_str()+")";

    
    const_foreach(it, acts) {
      op += " "+ ::act_to_str<CFG_action>(*it) + ";";
      //cout<<"apply vis for action: "<<op<<endl;
    }
    return op;
}
string CFG_stmt::to_str() {
    string op;
    //CFG_guard_tostr vg;
    //cout<<"called Stmt::to_str"<<endl;
    if( g == guard_true ) op = "(T)";
    else op = "("+g->to_str()+")";

    
    foreach(it, acts) {
      op += " "+ ::act_to_str<CFG_action>(*it) + ";";
      //cout<<"apply vis for action: "<<op<<endl;
    }
    return op;
}

Edge_type CFG_stmt::get_type() const {
    if( acts.size() != 1 ) return OTHER;
    return ::act_get_type<CFG_action>(get_comm_act());
}

const CFG_action& CFG_stmt::get_comm_act() const {
    assert ( acts.size() == 1 ) ;
    return *(acts.begin());
}

void CFG_stmt::get_rw ( RW& rw ) const {
    if ( g != guard_true && g != guard_else ) { g->get_rw(RW::R, rw); }
    CHK ( acts.size() <= 1 ) ;
    if( acts.size() != 0 ) {
        (acts.begin())->get_rw(rw);
    }
}

template<class O>
void CFG_stmt::fix_const_types ( const O& oracle, Pid pid ) {
    if ( g != guard_true && g != guard_else ) {
        //Util::msg(2,"g is " + g->to_str());
        g = expr_fix_const_types(g,oracle,pid);
    }
    CHK ( acts.size() <= 1 ) ;
    if( acts.size() != 0 ) {
        (acts.begin())->fix_const_types(oracle,pid);
    }

}


// === CFG =========

template <class N, class E>
void Graph<N,E>::remove_inaccessible(const N& start) {
    check_N(start);
    DFS_explorer<N,E> D(*this);
    D.run(start);
    set<N> eraseList;
    typename_const_foreach(it, nodes) {
        if( !D.is_marked(*it) ) eraseList.insert(*it);
    }
    typename_foreach(it, eraseList) erase_N(*it);
    
}


void CFG::remove_unreach() { 
    remove_inaccessible(entry_node); 
    const set<CFG_node>& nodes = Graph<CFG_node, CFG_edge>::nodes;
    if( nodes.find(exit_node) == nodes.end() ) {
        cout<<"Exit node of cfg lost due to reduction\n";
        exit_node = CFG_node();
    }
//    assert( nodes.find(exit_node) != nodes.end() );
}

void CFG::get_rw ( RW& rw ) const {
    const CFG::Edges& edges = get_edges();
    const_foreach(it, edges) { it->get_rw(rw); }

}

template void CFG::fix_const_types<Globals>(const Globals&, Pid);

template<class O>
void CFG::fix_const_types ( const O& oracle, Pid pid )  {
    map<CFG_node, map<CFG_edge,CFG_node> >& trans = Super::trans; 
    map<CFG_node, map<CFG_edge,CFG_node> > new_trans;

    const_foreach(itn, trans) {
        const Node &n = itn->first;
        const map<CFG_edge, CFG_node>& es = itn->second;
        map<CFG_edge,CFG_node> new_es;
        const_foreach(it, es) {
            CFG_edge new_e = it->first;
            new_e.fix_const_types(oracle,pid);
            new_es[new_e] = it->second;
        }
        new_trans[n] = new_es;
    }
    trans = new_trans;

}
