#ifndef CFG_H
#define CFG_H

#include"graph.h"
#include"action.h"

/*typedef Expr_list Send_args;
  typedef VAR::variant<Const,Varref> Recv_arg;
  typedef list<Recv_arg> Recv_args;
  typedef TUP::tuple<Varref,Send_args> Send;
  typedef TUP::tuple<Varref,Recv_args> Recv;
 */

typedef _Action_ CFG_action;
typedef vector<CFG_action> CFG_actions;
typedef const Expr* CFG_guard; 

class CFG_stmt{
    public:
        typedef CFG_action Lab;

        static const CFG_guard guard_true;
        static const CFG_guard guard_else;

        CFG_guard g;
        CFG_actions acts;

        CFG_stmt(): g(guard_true)  { } 
        CFG_stmt(const CFG_guard& _g) : g(_g) {}
        CFG_stmt(const CFG_action& a) : g(guard_true) { acts.push_back(a);}
        CFG_stmt(const CFG_actions& _acts) : g(guard_true), acts(_acts) { }
        CFG_stmt(const CFG_guard& _g, const CFG_action& a) : g(_g) { acts.push_back(a);}
        CFG_stmt(const CFG_guard& _g, const CFG_actions& _acts) : g(_g), acts(_acts) { }
        /*CFG_stmt(const CFG_stmt& st) { g = st.g; fprintf(stderr,"a\n");
        //acts = st.acts; 
        for(uint i=0; i<st.acts.size(); i++) {
        fprintf(stderr,"%d\n",i);
        CFG_action act (st.acts[i]);
        fprintf(stderr,"n-%d\n",i);
        acts.push_back(act);
        }

        fprintf(stderr,"b\n");}
         */
        template<class O>
            void fix_const_types ( const O& oracle, Pid pid ) ;

        string to_str_const() const ;
        string to_str() ;
        Edge_type get_type() const;
        const CFG_action& get_comm_act() const;
        void get_rw ( RW& rw ) const ;

        bool is_eps() const { return acts.empty(); }
        bool has_else_guard() const { return (g == guard_else); }
        bool has_true_guard() const { return (g == guard_true); }
        bool operator < (const CFG_stmt& st) const { return (to_str_const() < st.to_str_const()); }

        bool operator == (const CFG_stmt& st) const { assert(0); }
        bool operator != (const CFG_stmt& st) const { assert(0); }
        bool operator > (const CFG_stmt& st) const { assert(0); }
        bool operator >= (const CFG_stmt& st) const { assert(0); }
        bool operator <= (const CFG_stmt& st) const { assert(0); }
};


class CFG_node: public Node {
    protected:
//        int pc;
    public:

        //CFG_node(const tag& i)  { id = i; pc = id;}
        int get_pc() const { return id;}
        //string to_str() const { return int2str(id)+ "("+int2str(if_begin)+","+int2str(if_end)+")"; }


};

class CFG_edge: public Lab_edge<CFG_node, CFG_stmt> {
    protected:
        typedef Lab_edge<CFG_node, CFG_stmt> Super;
        const CFG_stmt& get_stmt() const { return Super::get_lab(); }
    public:
        CFG_edge() {  }
        CFG_edge(const CFG_stmt& _s) : Super (_s)  {  }

        const CFG_guard& get_guard() const { return get_stmt().g; }
        const CFG_actions& get_acts() const { return get_stmt().acts; }
        const CFG_action& get_act() const { 
            const CFG_actions& acts = get_acts();
            CHK( acts.size() == 1);
            return *(acts.begin()); 
        }
        const CFG_action& get_comm_act() const { 
            assert(is_comm_type() && get_stmt().acts.size() == 1);
            return get_stmt().get_comm_act(); 
        }

        void set_act(const CFG_action& act) {
            CFG_stmt st(act); 
            Super::set_lab(st); 
        }
        bool has_else_guard() const { return get_stmt().has_else_guard(); }
        bool has_true_guard() const { return get_stmt().has_true_guard(); }
        void get_rw ( RW& rw ) const { return get_stmt().get_rw(rw); }
        template<class O>
            void fix_const_types ( const O& oracle, Pid pid ) {
                s.fix_const_types(oracle, pid);
            }

};


class CFG: public Graph<CFG_node, CFG_edge> {
    string id;
    CFG_node entry_node, exit_node;
    typedef Graph<CFG_node, CFG_edge> Super;

    bool initialized;
    set<CFG_node> if_begin, if_end;


    public:

    CFG(const string& _id="none") : id(_id) { initialized = false; }

    const CFG_node& initialize_entry_node() { 
        entry_node = new_node();
        initialized = true;
        return entry_node;
    }

    bool is_initialized() const { return initialized; }
    const CFG_node& get_entry_node() const { CHK( is_initialized() ); return entry_node; }
    //const CFG_node& get_exit_node() const { return exit_node;}
    void set_exit_node(const CFG_node& n) {exit_node = n;}
    void remove_unreach(); 

    void set_if_begin(const CFG_node& n) { if_begin.insert(n); }
    void set_if_end(const CFG_node& n) { if_end.insert(n); }
    bool is_if_begin(const CFG_node& n) const { return if_begin.find(n) != if_begin.end();}
    bool is_if_end(const CFG_node& n) const { return if_end.find(n) != if_end.end();}

    string to_str() const;
    friend ostream& operator << (ostream& os, const CFG& n) ;
    string get_id() const { return id; }
    void get_rw ( RW& rw ) const ;

    template<class O>
    void fix_const_types ( const O& oracle, Pid pid ) ;

};



#endif
