#ifndef _CFG_COMP_H
#define _CFG_COMP_H

#include"ast.h" //Var_decl, Type_decl, Decls
#include"sys_state.h" //Mem_map
#include"symtab.h"
#include"cfg.h"
#include"logic.h"


typedef map<Guard_expr_list, const Expr*> Case_map;

class Init_comp {
    CFG init_cfg;
    list<Assign> alist;
    map<int, CFG_node> ipc2node;

    bool is_cfg;

    public:

    /*Init_comp( const CFG& init ) : init_cfg(init), is_cfg(true) {}
    Init_comp( const list<Assign>& _alist ) : alist(_alist), is_cfg(false) {}
                    */
    Init_comp () : is_cfg(false) {}
    Init_comp( const CFG& init, const list<Assign>& _alist ) 
                    : init_cfg(init), alist(_alist), is_cfg(true) {}

    template<class O>
    void fix_const_types(const O& globals, Pid pid) {
        foreach(it, alist) {
            it->fix_const_types(globals,pid);
        }
    }

    bool has_cfg () const { return is_cfg; }
    bool is_initialized() const { CHK( has_cfg()); return init_cfg.is_initialized(); }

    void build_pc2node() {
        if( has_cfg() ) {
            const CFG::Nodes& inodes = init_cfg.get_nodes();
            const_foreach(it, inodes) { ipc2node[it->get_pc()] = *it;}  
        }
    }

    const CFG_node& get_entry_node() const { CHK( has_cfg()); return init_cfg.get_entry_node(); }
    void get_rw(RW& rw) { CHK( has_cfg()); return init_cfg.get_rw(rw); }

    void get_out_edges(int pc, list<CFG_edge>& ce_list) const {
        CHK ( has_cfg() );
        map<int, CFG_node>::const_iterator it;
        CHK( (it=ipc2node.find(pc)) != ipc2node.end() );
        init_cfg.get_out_edges(it->second, ce_list);
    }

    const list<Assign>& get_sym_form() const { return alist; }

    Expr* get_glb_sym_form ( const Symtab& symtab ) const {
        Expr * res = ExprManager::get_true();
        const_foreach(it, alist) {
            const Assign& a = *it;
            const Varref& vr = a.get_vr();
            const Var_dcl& vd = vr.to_vd_plain();
            //globals written to 
            if ( !symtab.has_var(vd) ) {
                res = ExprManager::get_and(res, a.to_Expr());
            }
        }
        return res;
    }

    string to_str() const {
        string s;
        if( has_cfg() ) {
            s = "init_cfg is "+init_cfg.to_str();
        }
        else {
            s = "assign list is " + list_to_str(alist);
        }
        return s;
    }

};



class CFG_comp {
    protected:
        string id;
        Pid pid;
        CFG main_cfg;
        Init_comp init_comp;
        Logic logic;
        RW rw;
        //map<string, Symtab> sym_tabs; //scope to sym tab, e.g. main -> own sym tab 
        Symtab symtab;

        set<Var_dcl>  loc_vars;
        Var_dcls loc_st_vars, glb_st_vars, loc_st_coi, glb_st_coi;
        map<int, CFG_node> pc2node;

        map<Var_dcl, Case_map> var2cases;

        void build_pc2node() { 
            const CFG::Nodes& nodes = main_cfg.get_nodes();
            const_foreach(it, nodes) { pc2node[it->get_pc()] = *it;}  
            init_comp.build_pc2node();
        }

        void build_state_vars( ) {
            //build state vars
            RW st_rw;
            main_cfg.get_rw(st_rw); init_comp.get_rw(st_rw); 
            const Var_dcls& wvars = st_rw.get_w();
            const_foreach(it, wvars) {
                if ( is_loc_var(*it) ) loc_st_vars.insert(*it);
                else glb_st_vars.insert(*it);
            }

            //build state coi vars;
          /*  Var_dcls pre,post; 
            pre.insert(loc_st_vars.begin(), loc_st_vars.end());
            pre.insert(glb_st_vars.begin(), glb_st_vars.end());
            logic.get_coi_vars(pre,post);

            loc_st_coi = loc_st_vars;
            glb_st_coi = glb_st_vars;
            const_foreach(it, post) {
                if ( is_loc_var(*it) ) loc_st_coi.insert(*it);
                else glb_st_coi.insert(*it);
            }
            */

        }




    public:
        /* assume init_cfg may access both local vars of main and global vars */

        CFG_comp() { pid = -1; build_pc2node(); }
        CFG_comp(const string& s, const CFG& _main_cfg, const Symtab& _symtab, 
                    const Logic& _logic, const Init_comp& _init_comp) 
                        : id(s), main_cfg(_main_cfg), init_comp(_init_comp), logic(_logic), symtab(_symtab) 
        {
            //string main_id = main_cfg.get_id();
            //sym_tabs[main_id] = _sym_tab;
            //const Symtab& sym_tab = sym_tabs[main_id];

            symtab.insert(Var_dcl(Symbol("pc")),IntTy);
            for(Symtab::Map_type::const_iterator it= symtab.vars.begin(); it!= symtab.vars.end(); ++it) {
                loc_vars.insert(it->first);
            }
            build_pc2node();
            pid = Util::get_new_cfg_comp_pid(); 


            main_cfg.get_rw(rw); init_comp.get_rw(rw); logic.get_rw(rw);
            //Util::msg(2,"RW is " + rw.to_str());
            build_state_vars();
            //build_sym_form() ;
        }

        template<class O>
        void fix_const_types ( const O& globals ) {
            main_cfg.fix_const_types(globals,pid);
            logic.fix_const_types(globals,pid);
            init_comp.fix_const_types(globals,pid);
            //Util::msg(2,"CFG_comp: main_cfg is "+main_cfg.to_str());
            //Util::msg(2,"CFG_comp: init_comp is "+init_comp.to_str());
            build_sym_form();
        }

        const Symtab& get_symtab() const { return symtab; }
        const Logic& get_logic() const { return logic; }
        /*void add_var(const Var_decl& d) { 
            //TODO : no queues allowed here
            sym_tab.insert(d); 
        }*/
        //TODO: change it for local vars of other cfgs (except init) if they exist
        bool is_loc_var ( const Var_dcl& vd ) const { return Key_In_Bag(vd,loc_vars); }
        const set<Var_dcl>& get_local_vars() const { return loc_vars; }
        int get_main_entry_pcval() const { const CFG_node& n = main_cfg.get_entry_node(); return n.get_pc(); }
        int get_init_entry_pcval() const { 
            const CFG_node& n = init_comp.get_entry_node(); return n.get_pc(); 
        } 
        Pid get_pid() const { return pid; }
        bool has_init_cfg() const { return init_comp.is_initialized(); } 
        bool has_main_cfg() const { return main_cfg.is_initialized(); } 

        void get_out_edges(int pc, list<CFG_edge>& ce_list) const {
            map<int, CFG_node>::const_iterator it;
            if( (it = pc2node.find(pc)) != pc2node.end() )
                main_cfg.get_out_edges(it->second, ce_list);
            else init_comp.get_out_edges(pc, ce_list);
        }

        void get_rw ( RW& _rw ) const {
            CHK ( !rw.empty() ) ;
            _rw.add(rw);
        }

        void get_state_vars(Var_dcls& _loc_st_vars, Var_dcls& _glb_st_vars) const {
            _loc_st_vars.insert(loc_st_vars.begin(), loc_st_vars.end() ) ;
            _glb_st_vars.insert(glb_st_vars.begin(), glb_st_vars.end());
        }

        /*void get_state_coi_vars(Var_dcls& _loc_st_coi, Var_dcls& _glb_st_coi) const {
            _loc_st_coi = loc_st_coi;
            _glb_st_coi = glb_st_coi;
        }*/

        void build_cfg_comp_mem_map(Mem_map& mem) const {
            const set < Var_dcl > &vardcls = get_local_vars ();
            //mem.add_loc_var (pid, Var_dcl(Symbol("pc")));
            const_foreach (itv, vardcls) {
                mem.add_loc_var (pid, *itv);
            }
        }

        void show_sym_form(const map<Var_dcl, Case_map>& var2cases) {
            string s;
            const_foreach(it, var2cases) {
                s += it->first.to_str() + "::\n";
                const map< Guard_expr_list, const Expr* >& case_map = it->second;
                const_foreach(it2, case_map) {
                    const Guard_expr_list& expr_list = it2->first;
                    const Expr* e = it2->second;
                    s += "\t" + list_to_str(expr_list) + " ----> " + e->to_str() + "\n";
                }
            }
            Util::msg(2,"sym form is \n"+ s);
        }

        void build_sym_form_rec(const CFG_node& n, Guard_expr_list& cond_list,
                                set<CFG_node>& visited, map<Var_dcl, Case_map>& var2cases) ;
        
        void build_sym_form() { //only for main cfg
            set<CFG_node> visited;
            const CFG_node& n = main_cfg.get_entry_node();
            Guard_expr_list cond_list;


            build_sym_form_rec(n, cond_list, visited,var2cases);
            Util::msg(2,"*********************\n");

           // show_sym_form(var2cases);

        }

        const list<Assign>& get_init_sym_form() const { return init_comp.get_sym_form(); }
        Expr* get_init_glb_sym_form() const { 
            Expr* res =  init_comp.get_glb_sym_form(symtab); 
            //res = ExprManager::get_and( res, logic.get_glb_sym_form(symtab) );
            return res;
        }
        const map<Var_dcl, Case_map>& get_sym_form() const { return var2cases; }

};

typedef vector<const CFG_comp*> CFG_comps; //const may be removed if necessary

#endif
