#include"common.h"
#include"dp.h"
#include"naming.h"



// DP functions ---------------------------

// Const eval -------------------------

struct DP_Const_sym_eval {
    DP& dp;
    DP::expr v;
    
    DP_Const_sym_eval(const Const& c,  DP& _dp) : dp(_dp) { VAR::apply_visitor(*this,c);}

    void operator() (bool b) {
        if(b) v = dp.mk_true();
        else v = dp.mk_false();
    }

    void operator() (Num n) {
        v = dp.mk_num(n);
    }

    void operator() (const Bv_const& bv) {
        //TODO: mk_bv
        v = dp.mk_num( bv.getv() );
    }

    template<class T> void operator() (const T& t) {assert(0); }
};


DP::expr DP::mk_const_expr(const Const& c) { 
    DP_Const_sym_eval vis_c(c,*this);
    return vis_c.v;
}


void DP::assertdp(yices_expr e) { 
    //last_assert = e; 
    yices_assert(dp_ct,e); 
}

assertion_id DP::assert_retract_dp(yices_expr e) { 
    //last_assert = e; 
    assertion_id id = yices_assert_retractable(dp_ct,e); 
    assert_retract_history[id] = e;
    return id;
}

void DP::retract ( assertion_id id ) {
    yices_retract(dp_ct,id);
    assert_retract_history.erase(id);
}

void DP::retract_all(vector<yices_expr>& exprs ) {
    foreach(it, assert_retract_history) {
        exprs.push_back(it->second);
        yices_retract(dp_ct,it->first);
    }
    assert_retract_history.clear();

}

Const show_val(yices_model m, yices_var_decl yvd) {
    /* todo: add for Bv_const */
    long res;
    int tmp1 = yices_get_int_value(m,yvd,&res);
    if (tmp1) {
        //Util::msg(2,var + " has int val = " + int2str(res) );
        return Const ( (int)res );
    }
    else {
        lbool b = yices_get_value(m, yvd);
        Const res_const;
        switch (b) {
            case l_true:  res_const = Const(true); break;
            case l_false:  res_const = Const(false); break;
            case l_undef:  break;
        }
       // Util::msg(2,var + " has bool val = " + res_const.to_str() );
        return res_const;
    }
}

/*void DP::create_model ( vector< pair<string,bool> >& model ) {
    yices_var_decl_iterator it = yices_create_var_decl_iterator(dp_ct);
    yices_model m              = yices_get_model(dp_ct);
    //yices_display_model(m);
    while (yices_iterator_has_next(it)) {
        yices_var_decl d         = yices_iterator_next(it);
        string var_name = yices_get_var_decl_name(d);
        //Util::msg(1,"var is "+var_name);
        switch(yices_get_value(m, d)) {
            case l_true: model.push_back( pair<string,bool> ( var_name, true) ); break;
            case l_false: model.push_back( pair<string,bool> ( var_name, false) ); break;
            case l_undef:  Util::msg(2,var_name + "is undef"); break;
        }
    }
    yices_del_iterator(it);
}*/

void DP::create_model ( vector< pair<string,Const> >& model ) {
    CHK ( check() == l_true );
    yices_model m = yices_get_model(dp_ct);
    //yices_display_model(m);
    const_foreach(it, var_decls) {
        const string& name = *it;
        yices_var_decl yvd = yices_get_var_decl_from_name(dp_ct,(char*)name.c_str());
        CHK ( yvd );
        Const c = show_val(m,yvd);
        if ( !c.is_singular() ) {
            //Util::msg(2,name + " has val = " + c.to_str() );
            model.push_back( pair<string,Const> ( name, c ) );
        }
    }
}

yices_expr DP::mk_and ( const vector<yices_expr>& assigns) {
    uint size = assigns.size();
    CHK ( size != 0);
    yices_expr *a; a = new yices_expr[size];
    for(uint i=0; i< size; ++i) {
        a[i] = assigns[i];
    }
    yices_expr res = yices_mk_and(dp_ct,a,size);
    delete a;
    return res;
}


yices_expr DP::get_proj_expr ( const vector<yices_expr>& assigns, const set<uint>& indices) {
    uint size = indices.size();
    vector<yices_expr> proj_assigns;
    CHK ( size != 0);
    const_foreach(it, indices) {
        CHK ( (*it) < assigns.size() );
        proj_assigns.push_back( assigns[(*it)] );
    }
    yices_expr res = mk_and(proj_assigns);
    return res;
}


void DP::eval_reduced_models ( const vector<yices_expr>& assigns, const set<uint>& keep_indices, set<uint>& indices) {
    for(uint i=0; i< assigns.size(); i++ ) indices.insert(i);

    yices_expr a = get_proj_expr(assigns, indices);
    CHK ( check() == l_true );
    //yices_expr be_unsat = mk_and(a,nF);
    assertion_id id = assert_retract_dp(a);
    if ( check() != l_false ) {
        show_sat(); assert(0);
    }
    retract(id);
    uint index = 0;

    while ( index < assigns.size() ) {
        if ( Key_In_Bag(index,keep_indices) ) {index++; continue;}
        a = get_proj_expr(assigns, indices);
        //yices_expr be_unsat = mk_and(a,nF);
        assertion_id id = assert_retract_dp(a);
        if ( check() != l_false ) {
            retract(id);
        }
        else { retract(id);
            indices.erase ( index );
        }
        index++;
    }
    if ( indices.empty() ) { assert(0); } 

}




void DP::generalize_model ( expr err, map<string,Const>& model, int& max_depth ) {
    vector<yices_expr> assigns;
    vector< pair<string,Const> > init_model;
    create_model ( init_model );

    set<uint> keep_indices;

    for(uint i=0; i< init_model.size(); i++) {
        string& s = init_model[i].first;
        //if ( Naming::is_st_var_name(s) ) keep_indices.insert(i);
        Const c = init_model[i].second;
        expr var = get_var_expr(s);
        expr rhs = mk_const_expr(c);
        expr e = mk_eq(var,rhs);
        assigns.push_back(e);

        string oname; Pid pid; int depth;
        Naming::get_orig_name_pid_depth(s, oname, pid, depth);
        if ( depth > max_depth ) max_depth = depth;
    }

    vector<yices_expr> dummy;
    retract_all(dummy);
    expr F = mk_and(dummy);

    assertion_id id = assert_retract_dp(F);
    CHK ( check() == l_true );
    //show_sat();
    retract(id);
    Util::msg(2,"----------");

    expr nF = mk_not(F);
    assertdp(nF);
    CHK ( check() == l_true );
    //show_sat();
    Util::msg(2,"----------");

    set<uint> indices;
    eval_reduced_models ( assigns, keep_indices, indices );

    for(uint i=0; i< init_model.size(); i++) {
        string &s = init_model[i].first;
        if ( Key_In_Bag(i,indices) ) {
            Const c = init_model[i].second;
            model[s] = c;
        }
        else { Util::msg(2,"deleted "+ s); }
    }

}

void DP::eval_feasible_reduced_models ( const vector<yices_expr>& assigns, const set<uint>& keep_indices, 
                                        const set<uint>& to_rem_indices, set<uint>& rem_indices) {
    vector<assertion_id> assign_ids;

    for(uint i=0; i< assigns.size(); i++ ) {
       // print(assigns[i]);
        assertion_id id = assert_retract_dp(assigns[i]);
        assign_ids.push_back(id);
        //indices.insert(i);
    }

    //BMC should be sat
    CHK ( check() == l_true );
    uint index = 0;

    while ( index < assigns.size() ) {
        if ( !Key_In_Bag(index,to_rem_indices) ) { index++; continue; }
        if ( Key_In_Bag(index,keep_indices) ) { index++; continue; }
        //Util::msg(2,"trying to remove index "+int2str(index));
        expr ea = assigns[index];
        //print(ea);
        retract(assign_ids[index]);
        CHK ( check() == l_true );
        //show_sat();

        assertion_id id = assert_retract_dp( mk_not(ea) );

        if ( check() == l_true ) {
            retract(id);
         //   indices.erase ( index );
            rem_indices.insert(index);
            //Util::msg(2,"removing index");
        }
        else {
            retract(id);
            assert_retract_dp(ea);
            //Util::msg(2,"unable to remove index");
        }
        index++;
    }
    //if ( indices.empty() ) assert(0);

}
// ---------- RS04 ---------------

void DP::eval_reduced_models_rs04_inc ( const vector<yices_expr>& assigns, const set<uint>& keep_indices, set<uint>& rem_indices) {
    vector<assertion_id> assign_ids;

    for(uint i=0; i< assigns.size(); i++ ) {
       // print(assigns[i]);
        assertion_id id = assert_retract_dp(assigns[i]);
        assign_ids.push_back(id);
     //   indices.insert(i);
    }

    CHK ( check() == l_false );
    uint index = 0;

    while ( index < assigns.size() ) {
        if ( Key_In_Bag(index,keep_indices) ) { index++; continue; }
        retract(assign_ids[index]);

        if ( check() != l_false ) {
            assert_retract_dp(assigns[index]);
        }
        else {
            rem_indices.insert ( index );
        }
        index++;
    }
    //if ( indices.empty() ) assert(0);

}

void DP::eval_reduced_models_rs04_plain ( const vector<yices_expr>& assigns, const set<uint>& keep_indices, set<uint>& indices) {
assert(0);
//change indices to remove_indices
    vector<assertion_id> assign_ids;

    for(uint i=0; i< assigns.size(); i++ ) {
        //print(assigns[i]);
        //assertion_id id = assert_retract_dp(assigns[i]);
        //assign_ids.push_back(id);
        indices.insert(i);
    }
    expr a = get_proj_expr(assigns, indices);
    push();
    assertdp(a);
    CHK ( check() == l_false ) ;
    pop();

    uint index = 0;

    while ( index < assigns.size() ) {
        if ( Key_In_Bag(index,keep_indices) ) { index++; continue; }
        expr a = get_proj_expr(assigns, indices);
        push();
        assertdp(a);
        if ( check() == l_false ) {
            indices.erase ( index );
        }
        pop();
        index++;
    }
    if ( indices.empty() ) assert(0);
}

void DP::get_proj_assigns(const vector<expr>& assigns, const set<uint>& indices, vector<expr>& proj) {
    for(uint i=0; i<assigns.size(); i++)
        if ( !Key_In_Bag(i,indices) ) {
            proj.push_back(assigns[i]);
        }
        else Util::message(2,"get_proj_assigns: deleted index %d",i);
}

void DP::generalize_model_rs04 ( expr err, map<string,Const>& model, int& max_depth, assertion_id err_id ) {
    vector<yices_expr> assigns;
    vector< pair<string,Const> > init_model;
    create_model ( init_model );
    //make sure push, assertdp(err) in check_err
    //pop(); 
    //retract(err_id);

    vector<yices_expr> dummy;
    retract_all(dummy);
    //expr init_uniq_err_expr = mk_and(dummy);
    //dummy.clear();

    set<uint> keep_indices;

    //foreach(it, init_model) {
    for(uint i=0; i< init_model.size(); i++) {
        string& s = init_model[i].first;
        //if ( Naming::is_st_var_name(s) ) keep_indices.insert(i);
        Const c = init_model[i].second;
        expr var = get_var_expr(s);
        expr rhs = mk_const_expr(c);
        expr e = mk_eq(var,rhs);
        assigns.push_back(e);

        string oname; Pid pid; int depth;
        Naming::get_orig_name_pid_depth(s, oname, pid, depth);
        if ( depth > max_depth ) max_depth = depth;
    }



    expr nerr = mk_not(err);
    //assert_retract_dp(nerr);
    assertdp(nerr);

    set<uint> to_rem_indices, rem_indices;

    eval_reduced_models_rs04_inc ( assigns, keep_indices, to_rem_indices );

    rem_indices = to_rem_indices;

/*    retract_all(dummy); //!err, rs04
    //assertdp(init_uniq_err_expr);
    CHK(init_e != NULL);
    assertdp(init_e); //assert init constraints also to make it feasible path
    assertdp(err);

    eval_feasible_reduced_models(assigns, keep_indices, to_rem_indices, rem_indices);
*/

    for(uint i=0; i< init_model.size(); i++) {
        string &s = init_model[i].first;
        if (!Key_In_Bag(i,rem_indices) ) {
            Const c = init_model[i].second;
            model[s] = c;
        }
        else { Util::msg(2,"deleted "+ s); }
    }

}


// ------- end RS04 -------------

void DP::get_model(expr err, map<string,Const>& model, int& max_depth, bool generalize, assertion_id err_id) {
    lbool check_res = check();
    CHK ( check_res == l_true );

    if ( generalize ) {
        Util::msg(2,"generalizing ce..");
        map<string,Const> gen_model;
        generalize_model_rs04( err, gen_model, max_depth, err_id );
        //generalize_model( err, gen_model, max_depth );
        foreach(it, gen_model ) {
            const string &s = it->first;
            if ( Naming::is_st_var_name(s) ) continue; 
            const Const& c = it->second;
            model[s] = c;
        }
    }
    else {
        vector< pair<string,Const> > init_model;
        create_model ( init_model );
        for(uint i=0; i< init_model.size(); i++) {
            string &s = init_model[i].first;

            string oname; Pid pid; int depth;
            Naming::get_orig_name_pid_depth(s, oname, pid, depth);
            if( depth > max_depth ) max_depth = depth;
            //Util::msg(2,"decomp oname = " + oname +", pid = "+ int2str(pid) + ", depth = "+int2str(depth));

            if ( Naming::is_st_var_name(s) ) continue; 
            Const c = init_model[i].second;
            model[s] = c;
        }

    }
}

void DP::get_one_model( list< map<string,Const> >& models, bool global ) {
    CHK ( check() == l_true );
    //show_sat();
    vector< pair<string,Const> > pair_model;
    create_model ( pair_model );
    map<string,Const> new_model;

    for(uint i=0; i< pair_model.size(); i++) {
        string& s = pair_model[i].first;
        Const c = pair_model[i].second;
        if ( Naming::is_st_var_name(s) ) continue;
        if ( global && !Naming::is_global_name(s) ) continue;
      //  Util::msg(2,"adding "+s+" = "+c.to_str()+" to new model");
        new_model[s] = c;
    }
    models.push_back(new_model);
}

//all sat
void DP::get_all_models( list< map<string,Const> >& models, bool global ) {
    CHK ( check() == l_true );
    while (1) {
        //show_sat();
        vector< pair<string,Const> > pair_model;
        create_model ( pair_model );
        map<string,Const> new_model;

        expr sat_expr = mk_true();
        for(uint i=0; i< pair_model.size(); i++) {
            string& s = pair_model[i].first;
            Const c = pair_model[i].second;
            if ( Naming::is_st_var_name(s) ) continue;
            if ( global && !Naming::is_global_name(s) ) continue;
            Util::msg(2,"adding "+s+" = "+c.to_str()+" to new model");
            new_model[s] = c;

            expr var = get_var_expr(s);
            expr rhs = mk_const_expr(c);
            expr e = mk_eq(var,rhs);
            sat_expr = mk_and(sat_expr,e);

        }
        models.push_back(new_model);
        Util::msg(2,"pushing new model..");

        assertdp( mk_not(sat_expr) );

        if ( check() != l_true ) break;
    }

}



void DP::lift_model_vars_unsat(const map<string,Const>& model, map<string,Const>& lifted_model) {
    
    vector< pair<string,Const> > model_list;
    
    vector<expr> assigns;
    vector<assertion_id> assign_ids;
    const_foreach(it, model) {
        const string& s = it->first;
        Const c = it->second;
        model_list.push_back( pair<string,Const>(s,c) );

        expr var = get_var_expr(s);
        expr rhs = mk_const_expr(c);
        expr e = mk_eq(var,rhs);
        assigns.push_back(e);
        assertion_id id = assert_retract_dp(e);
        assign_ids.push_back(id);
    }
    
    if ( check() == l_true ) Util::msg(2,"is true");
    CHK ( check() == l_false );


    for(uint i=0; i<assign_ids.size(); i++) {
        //Util::msg(2,"trying LIFT "); print(assigns[i]); 
        retract(assign_ids[i]);
        if ( check() != l_false ) {
            assertdp(assigns[i]);
            const string& s = model_list[i].first;
            const Const& c = model_list[i].second;
            lifted_model[s] = c;
         //   Util::msg(2,"unable to LIFT");
        }
        //else Util::msg(2,"LIFT success");
    }


}
