#ifndef _DP_H
#define _DP_H

#include"yices_c.h"
#include"basic_vals.h"
#include"tys.h"

class DP {
    public:
        typedef yices_expr expr ;
        typedef yices_context context ;
        typedef yices_type DP_type;
    protected:
        context dp_ct;
        yices_expr last_assert;
        map<assertion_id, yices_expr>  assert_retract_history;
        set<string> var_decls;

        context mk_context() const { return yices_mk_context() ; }
        void del_context(const context& ct) const { yices_del_context(ct) ; }

        expr init_e;
    public:


        DP() {
            init();
            dp_ct = mk_context();
        }
        ~DP() { del_context(dp_ct); }

        void init() {
            init_e = NULL;
            yices_enable_type_checker(1);
            yices_enable_log_file("yices.log");
        }

        void set_init_constraint(expr init) { init_e = init; }

        void mk_var_decl(const string& name, const Ty& ty) {
            CHK ( ! Key_In_Bag(name,var_decls) );
            var_decls.insert(name);
            DP_type dpty = mk_type(ty);

            yices_mk_var_decl(dp_ct,(char*)name.c_str(),dpty);
        }

        expr mk_bool_var(const string& name) { return yices_mk_bool_var_decl(dp_ct,(char*)name.c_str()); }

        /*DP_type mk_type (const string& ty_str) {
            return yices_mk_type(dp_ct, (char*)ty_str.c_str());
        }
        */
        DP_type mk_type (const Ty& ty) {
            string ty_str = ty2yices_str(ty);
            return yices_mk_type(dp_ct, (char*)ty_str.c_str());
        }

        /*void mk_var_decl(const string& name, const string& ty_str) {
            DP_type ty = mk_type(ty_str);
            mk_var_decl(name,ty);
        }*/

        expr get_var_expr(const string& name) {
            //Util::msg(2,"yices::get_var_expr: name is "+name);
            yices_var_decl var_decl = yices_get_var_decl_from_name(dp_ct,(char*)name.c_str());
            yices_expr var_expr = yices_mk_var_from_decl(dp_ct,var_decl);
            return var_expr;
        }

        void get_proj_assigns(const vector<expr>& assigns, const set<uint>& indices, vector<expr>& proj) ;
        yices_expr get_proj_expr ( const vector<yices_expr>& assigns, const set<uint>& indices) ;

        void eval_feasible_reduced_models ( const vector<expr>& , const set<uint>& , const set<uint>&, set<uint>& );

        void eval_reduced_models ( const vector<expr>& assigns, const set<uint>&, set<uint>& indices) ;
        void eval_reduced_models_rs04_inc ( const vector<expr>& assigns, const set<uint>&, set<uint>& indices ) ;
        void eval_reduced_models_rs04_plain ( const vector<expr>& assigns, const set<uint>&, set<uint>& indices ) ;

        void generalize_model ( expr err, map<string,Const>& model, int& max_d ) ;
        void generalize_model_rs04 ( expr err, map<string,Const>& model, int& max_d , assertion_id) ;

        void create_model ( vector< pair<string,Const> >& model ) ;
        void get_model(expr err, map<string,Const>& model, int& max_d, bool generalize, assertion_id ) ;
        void get_one_model( list< map<string,Const> >& models, bool global );
        void get_all_models( list< map<string,Const> >& models, bool global );
        void lift_model_vars_unsat(const map<string,Const>& model, map<string,Const>& lifted_model) ;

        static void print(const expr& e) {
            Util::msg(2,"print_yices: ");
            yices_pp_expr(e);
            Util::msg(2,"");
        }
        lbool check() { return yices_check(dp_ct); }

        void show_sat () {
            if ( check() != l_true ) { Util::msg(2,"dp::unsat"); return; }
            /*lbool check_res = check();
            if ( check_res == l_undef ) printf("undef\n");
            else if ( check_res == l_false ) printf("unsat\n");
            else if (check_res == l_true) {
                yices_var_decl_iterator it = yices_create_var_decl_iterator(dp_ct);
                yices_model m              = yices_get_model(dp_ct);
                while (yices_iterator_has_next(it)) {
                    yices_var_decl d         = yices_iterator_next(it);
                    char *         val;
                    switch(yices_get_value(m, d)) {
                        case l_true: val = "true"; break;
                        case l_false: val = "false"; break;
                        case l_undef: val = "unknown"; break;
                    }
                    printf("%s : %s, ", yices_get_var_decl_name(d), val);
                }
                yices_del_iterator(it);
            }
            else assert(0);
            printf("\n");*/

            yices_model m              = yices_get_model(dp_ct);
            yices_display_model(m);

        }


        expr mk_const_expr(const Const& c) ;

        expr mk_and ( const vector<yices_expr>& assigns) ;

        expr mk_true() { return yices_mk_true(dp_ct); }
        expr mk_false() { return yices_mk_false(dp_ct); }
        expr mk_num(int n) { return yices_mk_num(dp_ct,n); }
        expr mk_num_from_string(const string& s) { 
            return yices_mk_num_from_string( dp_ct, (char*)s.c_str() ); 
        }

        expr mk_sum(expr a[], int n) { return yices_mk_sum(dp_ct,a,n); }
        expr mk_sub(expr a[], int n) { return yices_mk_sub(dp_ct,a,n); }
        expr mk_mul(expr a[], int n) { return yices_mk_mul(dp_ct,a,n); }

        expr mk_sum(expr a, expr b) { DP::expr tmp[2]; tmp[0] = a; tmp[1] = b; return mk_sum(tmp,2); } 
        expr mk_sub(expr a, expr b) { DP::expr tmp[2]; tmp[0] = a; tmp[1] = b; return mk_sub(tmp,2); } 
        expr mk_mul(expr a, expr b) { DP::expr tmp[2]; tmp[0] = a; tmp[1] = b; return mk_mul(tmp,2); } 

        expr mk_and(expr a[], int n) { return yices_mk_and(dp_ct,a,n); }
        expr mk_or(expr a[], int n) { return yices_mk_or(dp_ct,a,n); }
        expr mk_and(expr a, expr b) { 
            expr tmp[2]; tmp[0] = a; tmp[1] = b;
            return mk_and(tmp,2);
        }
        expr mk_or(expr a, expr b) { 
            expr tmp[2]; tmp[0] = a; tmp[1] = b;
            return mk_or(tmp,2);
        }
        expr mk_diseq(expr a, expr b) { return yices_mk_diseq(dp_ct,a,b); }
        expr mk_eq(expr a, expr b) { return yices_mk_eq(dp_ct,a,b); }
        expr mk_ite( expr c, expr t, expr f) { 
            return yices_mk_ite(dp_ct, c, t, f); 
        }

        expr mk_not(expr a) { return yices_mk_not(dp_ct,a); }

        expr mk_le(expr a, expr b) { return yices_mk_le(dp_ct,a,b); }
        expr mk_lt(expr a, expr b) { return yices_mk_lt(dp_ct,a,b); }
        expr mk_ge(expr a, expr b) { return yices_mk_ge(dp_ct,a,b); }
        expr mk_gt(expr a, expr b) { return yices_mk_gt(dp_ct,a,b); }

        void push() { yices_push(dp_ct); }
        void pop() { yices_pop(dp_ct); }

        void assertdp(yices_expr e) ;
        assertion_id assert_retract_dp(yices_expr e) ;

        void retract ( assertion_id id ) ;
        void retract_all(vector<yices_expr>& exprs ) ;

        bool inconsistent() { return yices_inconsistent(dp_ct); }



};


// add types for add sub or and etc
// add yices_l with string : store list of expr locally












#endif
