#include"expr.h"
#include"naming.h"
#include"globals.h"

map<Expr, Expr*> ExprManager::exprMap;
int ExprManager::total_cnt = 0;
int ExprManager::found_cnt = 0;



/*
//Var_dcl::to_str

typedef VAR::variant<SimpleVarDcl, BvVarDcl, ArrayVarDcl> Var_dclt;
*/

struct Var_dcl_to_str_visitor {
  void operator() (const SimpleVarDcl& svd) { s = svd;}
  void operator() (const BvVarDcl& bvd) { s = bvd.first + "{" + int2str(bvd.second) + "}"; }
  void operator() (const ArrayVarDcl& avd) { s = TUP::get<0>(avd) + "[" + int2str(TUP::get<1>(avd)) + "]";}
  template<typename T> void operator() (const T& vd) {assert(0); }
  Symbol s;
};

string Var_dcl::to_str() const {
  Var_dcl_to_str_visitor vis;
  apply_visitor(vis, *this);
  return vis.s;
}

//Varref::to_str

/*typedef Symbol SymbolRef;
typedef TUP::tuple<Symbol,Expr*> ArrayRef;
typedef VAR::variant<SymbolRef, ArrayRef> Varreft;
*/

string BvRef::to_str() const {
    string s;
    s = TUP::get<0>(*this);
    s += (TUP::get<1>(*this))->to_str();
    return s;
}

struct Varref_tostr {
  void operator() (const SymbolRef& sr) { s = sr;}
  void operator() (const ArrayRef& ar) { 
      s = TUP::get<0>(ar);
      s += (TUP::get<1>(ar))->to_str();
  }
  void operator() (const BvRef& br) { 
    s = br.to_str();
  }
  template<class T> void operator() (const T& t) { assert(0); }
  string s;

};

string Varref::to_str() const {
  Varref_tostr vis;
  apply_visitor(vis,*this);
  return vis.s;
}


//Varref to Var_dcl

struct Varref_to_Var_dcl_plain {
  Var_dcl vd;
  
  Varref_to_Var_dcl_plain( const Varref& vr )   { VAR::apply_visitor(*this, vr); }

  void operator() (const SymbolRef& sr) { 
      vd = Var_dcl(Symbol(sr));
  }

  void operator() (const ArrayRef& ar) { 
      Symbol s = TUP::get<0>(ar);
      const Expr* e = TUP::get<1>(ar);
      CHK ( e->is_const() );
      Const c = VAR::get<Const>(*e);
      vd = Var_dcl(ArrayVarDcl(s,c.to_num()));
  }
  void operator() (const BvRef& br) { 
      Symbol s = br.get_symbol();
      const Expr* e = br.get_expr();
      CHK ( e->is_const() );
      Const c = VAR::get<Const>(*e);
      vd = Var_dcl(BvVarDcl(s,c.to_num()));
  }
  template<class T>
  void operator() (const T& t) { cout<<"Unsupported Var_dcl convertion for this type\n"; assert(0); }
};

Var_dcl Varref::to_vd_plain() const {
    Varref_to_Var_dcl_plain vis(*this);
    return vis.vd;
}

// expr is const

struct is_expr_const_vis {
	bool tf;
	void operator() (const Const& c) { tf = true;}
	template<class T>
	void operator() (const T& t) { tf = false;}
};


bool Expr::is_const () const {
	is_expr_const_vis vis;
	VAR::apply_visitor(vis,*this);
	return vis.tf;
}


// expr is varref

struct is_expr_varref_vis {
	bool tf;
	Varref& vr;
	is_expr_varref_vis(Varref& _vr) : vr(_vr) {}

	void operator() (const Varref& _vr) { tf = true; vr = _vr;}
	template<class T> void operator() (const T& t) { tf = false; }
};

//returns value in vr
bool Expr::is_varref(Varref& vr) const {
	is_expr_varref_vis vis(vr);
	VAR::apply_visitor(vis,*this);
	return vis.tf;
}


//expr::to_str


string to_str_num(Num n) {
  char s[100];
  sprintf(s,"%d",n);
  return string(s);
}

string to_str_op(op o) {
  switch(o) {
      case LorOp: return " || ";
      case LandOp: return (" && ");
      case XorOp: return ("^");
      case OrOp: return ("|");
      case AndOp: return ("&");
      case EqualOp: return (": ");
      case NotEqualOp: return (" != ");
      case LessOp: return ("<");
      case GreaterOp: return (">");
      case LessEqualOp: return ("<=");
      case GreaterEqualOp: return (">=");
      case LshiftOp: return ("<<");
      case RshiftOp: return (">>");
      case PlusOp: return ("+");
      case MinusOp: return ("-");
      case TimesOp: return ("*");
      case DivOp: return ("/");
      case ModOp: return ("%");
      case IteOp: return ("ite");
      default: assert(0);
  }
}

string to_str_uop(uop uo) {
  switch(uo) {
    case PlusUOp: return ("+");
    case MinusUOp: return ("-");
    case NegUOp: return ("!");
    case NotUOp: return ("!");
    /*case LenUOp: return ("len_");
    case FullUOp: return ("full_");
    case NFullUOp: return ("nfull_");
    case EmptyUOp: return ("empty_");
    case NEmptyUOp: return ("nempty_");
    */
    default: assert(0);
  }
}

struct expr_visitor_str {

  string str_val;


  void operator() (const BinOp& bop) {
    const Expr* e1 = TUP::get<0>(bop);
    op o = TUP::get<1>(bop);
    const Expr* e2 = TUP::get<2>(bop);
    const Exprt& et1 = *e1;
    const Exprt& et2 = *e2;

    string str1;
      apply_visitor(*this,et1);
      str1 = str_val;
      apply_visitor(*this,et2);
      if ( o == LandOp || o == LorOp )
      str_val =   "`" + str1 + to_str_op(o) + str_val + "'"  ;
      else 
      str_val =    str1 + to_str_op(o) + str_val  ;
  }
  
  void operator() (const UnOp& e) {
    uop o = TUP::get<0>(e);
    const Expr* e1 = TUP::get<1>(e);
    const Exprt& et1 = *e1;

      apply_visitor(*this,et1);
      str_val =  to_str_uop(o) + "(" + str_val + ")";
  }

  void operator() (const TernOp& e)  {
    op o = TUP::get<0>(e);
    CHK ( o == IteOp );
    const Expr* e1 = TUP::get<1>(e);
    const Expr* e2 = TUP::get<2>(e);
    const Expr* e3 = TUP::get<3>(e);
    const Exprt& et1 = *e1;
    const Exprt& et2 = *e2;
    const Exprt& et3 = *e3;

    string str1 = to_str_op(o) + "(";

    apply_visitor(*this,et1);
    str1 += str_val + ",";
    apply_visitor(*this,et2);
    str1 += str_val + ",";
    apply_visitor(*this, et3);
    str1 += str_val + ")";
    str_val = str1;
  }

  void operator() (const Varref& vr) {
          str_val = vr.to_str();
  }

  void operator() (const Const& c) {
    //Util::msg(2,"expr is const = "+c.to_str());
      str_val = c.to_str();
  }

  void operator() (const Expr_Enum& ee) {
    switch(ee) {
      case Else_expr:
                  str_val = "else";
                  break;
      default: assert(0);
    }
  }

  void operator() (const ChanInit& ci) { 
  }
  template<class T> void operator() (const T& t) { assert(0);}

};


string Expr::to_str() const {
  const Exprt& et = *this;
  expr_visitor_str ev;
  apply_visitor(ev,et);
  return ev.str_val;
}


//expr::get_rw
struct expr_visitor_rw {
    RW::RW_enum rwe;
    enum mode { ONE, BOTH };
    RW& rw;
    mode m;
    expr_visitor_rw ( const Exprt& et, RW::RW_enum _rwe, RW& _rw, mode _m = ONE) 
                                        : rwe(_rwe), rw(_rw), m(_m) { 
        VAR::apply_visitor(*this,et);
    }

    void operator() (const BinOp& bop) {
        const Expr* e1 = TUP::get<0>(bop);
        const Expr* e2 = TUP::get<2>(bop);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;
        apply_visitor(*this,et1);
        apply_visitor(*this,et2);
    }

    void operator() (const UnOp& e) {
        const Expr* e1 = TUP::get<1>(e);
        const Exprt& et1 = *e1;
        apply_visitor(*this,et1);
    }

    void operator() (const TernOp& e)  {
        op o = TUP::get<0>(e);
        CHK ( o == IteOp );
        const Expr* e1 = TUP::get<1>(e);
        const Expr* e2 = TUP::get<2>(e);
        const Expr* e3 = TUP::get<3>(e);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;
        const Exprt& et3 = *e3;

        apply_visitor(*this,et1);
        apply_visitor(*this,et2);
        apply_visitor(*this, et3);
    }
    void operator() (const Varref& vr) {
        if ( m == BOTH ) {
            string oname; Pid pid; int vr_depth; 
            Naming::get_orig_name_pid_depth(vr.to_str(), oname, pid, vr_depth) ;
            Var_dcl vd(oname);
            if( vr_depth == 0 ) rw.add(RW::R,vd);
            else if( vr_depth == 1 ) rw.add(RW::W,vd);
            else assert(0);
        }
        else rw.add( rwe, vr.to_vd_plain());
    }
    template<class T>
        void operator() (const T& t) { } //do nothing

};

void Expr::get_rw( RW::RW_enum rwe, RW& rw) const {
    const Exprt& et = *this;
    expr_visitor_rw vis(et, rwe, rw);
}

void Expr::get_rw_both(RW& rw) const {
    const Exprt& et = *this;
    RW::RW_enum rwe_dummy;
    expr_visitor_rw vis(et, rwe_dummy, rw, expr_visitor_rw::BOTH);
}

//expr::get_support

struct expr_visitor_support {
    set<Varref>& support;
    expr_visitor_support ( const Exprt& et, set<Varref>& _support) : support(_support) { 
        VAR::apply_visitor(*this,et);
    }

    void operator() (const BinOp& bop) {
        const Expr* e1 = TUP::get<0>(bop);
        const Expr* e2 = TUP::get<2>(bop);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;
        apply_visitor(*this,et1);
        apply_visitor(*this,et2);
    }

    void operator() (const UnOp& e) {
        const Expr* e1 = TUP::get<1>(e);
        const Exprt& et1 = *e1;
        apply_visitor(*this,et1);
    }

    void operator() (const TernOp& e)  {
        op o = TUP::get<0>(e);
        CHK ( o == IteOp );
        const Expr* e1 = TUP::get<1>(e);
        const Expr* e2 = TUP::get<2>(e);
        const Expr* e3 = TUP::get<3>(e);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;
        const Exprt& et3 = *e3;

        apply_visitor(*this,et1);
        apply_visitor(*this,et2);
        apply_visitor(*this, et3);
    }
    void operator() (const Varref& vr) {
        support.insert(vr);
    }
    template<class T>
        void operator() (const T& t) { } //do nothing

};

void Expr::get_support( set<Varref>& support ) const {
    const Exprt& et = *this;
    expr_visitor_support vis(et, support);
}


//project for small class of expressions
struct expr_visitor_project {
    const set<Varref>& support;
    Expr* res;

    expr_visitor_project ( const Exprt& et, set<Varref>& _support) : support(_support) { 
        //res = ExprManager::get_true();
        res = NULL;
        VAR::apply_visitor(*this,et);
        CHK( res != NULL);
    }

    void operator() (const BinOp& bop) {
        const Expr* e1 = TUP::get<0>(bop);
        op o = TUP::get<1>(bop);
        const Expr* e2 = TUP::get<2>(bop);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;

        if ( o == EqualOp ) {
            Varref vr;
            CHK ( e1->is_varref(vr) && e2->is_const() ) ;
            if ( support.find(vr) != support.end() )
                res = ExprManager::get_expr((Expr*)e1, o, (Expr*)e2);
            else res = ExprManager::get_true();
        }
        else {
            apply_visitor(*this,et1);
            Expr* res1 = res;
            apply_visitor(*this,et2);

            Expr* te = ExprManager::get_true();
            if ( res1 == te && res == te) res= te;
            else if ( res1 != te && res != te ) 
                res = ExprManager::get_expr(res1,o,res);
            else if ( res1 != te )
                res = res1;
            else if ( res != te )  {}
            else assert(0);
                
        }
    }

    void operator() (const UnOp& e) {
        assert(0);
    }

    void operator() (const TernOp& e)  {
        assert(0);
    }


    template<class T>
        void operator() (const T& t) { assert(0); } 

};

Expr* Expr::project_to (const Expr * e) const {
    CHK ( e != NULL);
    set<Varref> support;
    e->get_support(support);
    const Exprt& et = *this;
    //Util::msg(2,"before project expr = "+to_str());
    expr_visitor_project vis(et, support);
    //Util::msg(2,"after project expr = "+vis.res->to_str());
    return vis.res;
}


// ---------------------------- Fix const types in Expr ------------------ //

int get_expr_const_val(const Expr* e) {
    CHK( e->is_const() );
    const Const& c = VAR::get<Const>(*e);
    return c.to_num();
}

struct Ty_mk_const {
    int val;
    Const new_c;

    Ty_mk_const(const Ty& ty , int _val) : val(_val) { VAR::apply_visitor(*this,ty); }

    void operator() (const bTy& bty) {
        switch( bty ) {
            case BitTy:
            case BoolTy:
            case EvTy:
                bool b = val?true:false;
                new_c = Const ( b );
                break;
            default: assert(0);
        }
    }

    void operator() (const BvTy& bvty) {
        new_c = Const( Bv_const(val,bvty) );
    }

    template<typename T> void operator() (const T& ty) { assert(0); }

};

Const mk_ty_const(const Ty& ty, int val) {
    Ty_mk_const vis (ty,val);
    return vis.new_c;
}

template<class O>
struct Expr_fix_consts_type_vis {
    const O& oracle;
    Pid pid;
    Ty ety;
    Expr* res;

    Expr_fix_consts_type_vis ( const Expr* e, const O& o, Pid _pid, const Ty& _ety ) : oracle(o), pid(_pid), ety(_ety) {
        res = NULL;
        //Util::msg(2,"expr is  "+e->to_str());
        VAR::apply_visitor(*this,*e);
        CHK ( res != NULL );
    }

    void operator() (const BinOp& bop) {
        const Expr* e1 = TUP::get<0>(bop);
        op o = TUP::get<1>(bop);
        const Expr* e2 = TUP::get<2>(bop);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;

        bool e1_const = e1->is_const();
        bool e2_const = e2->is_const();
        if ( e1_const || e2_const ) {
            //Util::msg(2,"e1 is "+ e1->to_str());
            //Util::msg(2,"e2 is "+ e2->to_str());
            const Expr* econst, *other;
            int val;
            if ( e1_const ) { 
                econst = e1; val = get_expr_const_val(econst);  
                other = e2; CHK( !other->is_const() ); 
            }
            if ( e2_const ) { 
                econst = e2; val = get_expr_const_val(econst);  
                other = e1; CHK( !other->is_const() ); 
            }
            Varref vr;
            CHK ( other->is_varref(vr) );

            Var_dcl vd = vr.to_vd_plain();
            Ty ty = oracle.get_ty(vd, pid);

            Const new_c = mk_ty_const(ty,val);
            Expr* c_expr = ExprManager::get_expr( new_c );
            res = ExprManager::get_expr( (Expr*)other, o, c_expr );
        }
        else {
            apply_visitor(*this,et1);
            Expr* res1 = res;
            apply_visitor(*this,et2);
            res = ExprManager::get_expr( res1, o, res );
        }
    }

    void operator() (const UnOp& e) {
        uop o = TUP::get<0>(e);
        const Expr* e1 = TUP::get<1>(e);
        const Exprt& et1 = *e1;

        apply_visitor(*this,et1);
        res = ExprManager::get_expr(o,res);
    }

    void operator() (const TernOp& e)  {
        op o = TUP::get<0>(e);
        CHK ( o == IteOp );
        const Expr* e1 = TUP::get<1>(e);
        const Expr* e2 = TUP::get<2>(e);
        const Expr* e3 = TUP::get<3>(e);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;
        const Exprt& et3 = *e3;

        apply_visitor(*this,et1);
        Expr* res1 = res;
        apply_visitor(*this,et2);
        Expr* res2 = res;
        apply_visitor(*this,et3);
        res = ExprManager::get_expr(o, res1, res2, res);
    }

    void operator() (const Varref& vr) {
        res = ExprManager::get_expr(vr);
    }

    void operator() (const Const& c) {
        //Util::msg(2,"expr is const = "+c.to_str());
        CHK ( ! ety.is_singular() );
        int val = c.to_num();
        Const new_c = mk_ty_const(ety,val);
        res = ExprManager::get_expr( new_c );
    }

    void operator() (const Expr_Enum& ee) {
        res = ExprManager::get_expr(ee);
    }

    void operator() (const ChanInit& ci) { 
        assert(0);
    }
    template<class T> void operator() (const T& t) { assert(0);}

};

template Expr* expr_fix_const_types<Globals> (const Expr* e, const Globals&, Pid, const Ty&);

template<class O>
Expr* expr_fix_const_types ( const Expr* e, const O& oracle, Pid pid, const Ty& ety = Ty() ) {
    Expr_fix_consts_type_vis<O> vis (e,oracle,pid,ety);
    return vis.res;
}













