#ifndef EXPR_H
#define EXPR_H

#include"common.h"
#include"ops.h"
#include"basic_vals.h"
#include"tys.h"
using namespace std;

//typedef Symbol Mtype_const;
//typedef VAR::variant<bool, Num> Const;





class Expr;

//---------------------------------------------------
// Var_dcl
typedef Symbol SimpleVarDcl;
typedef pair<Symbol, Num> BvVarDcl; //just to differentiate from array
typedef TUP::tuple<Symbol, Num> ArrayVarDcl ;
typedef VAR::variant<SimpleVarDcl, BvVarDcl, ArrayVarDcl> Var_dclt;


class Var_dcl : public Var_dclt {
    public:
        Var_dcl() {}
        Var_dcl (const Var_dclt& v) : Var_dclt(v) {}
        bool operator < (const Var_dcl& v) const  {
            return (to_str() < v.to_str());
        }
        string to_str() const;
        bool operator == (const Var_dcl& vd) const { return (to_str() == vd.to_str()); }
};

typedef set<Var_dcl> Var_dcls;

//---------------------------------------------------
// RW
class RW {
    protected:
        Var_dcls reads, writes;
    public:
        enum RW_enum {R, W};
        RW() {}
        RW(const Var_dcls& r, const Var_dcls& w) : reads(r), writes(w) { }
        const Var_dcls& get_r() const { return reads;}
        const Var_dcls& get_w() const { return writes;}

        void add_read(const Var_dcl& vd) { reads.insert(vd); }
        void add_write(const Var_dcl& vd) { writes.insert(vd); }
        void add( RW_enum rwe, const Var_dcl& vd) {
            if( rwe == R ) add_read(vd);
            else if (rwe == W) add_write(vd);
            else assert(0);
        }
        void add ( const RW& rw ) {
            reads.insert(rw.reads.begin(), rw.reads.end());
            writes.insert(rw.writes.begin(), rw.writes.end());
        }
        
        bool has_read (const Var_dcl& vd) const { return (reads.find(vd) != reads.end()); }
        bool has_write (const Var_dcl& vd) const { return (writes.find(vd) != writes.end()); }
        bool has_read_or_write (const Var_dcl& vd) const { return ( has_read(vd) || has_write(vd) ); }

        void project_to ( const Var_dcls& vds ) {
            Var_dcls rs, ws;
            set_intersection( reads.begin(), reads.end(), vds.begin(), vds.end(), inserter(rs,rs.begin()) ); 
            set_intersection( writes.begin(), writes.end(), vds.begin(), vds.end(), inserter(ws,ws.begin()) ); 
            reads = rs;
            writes = ws;
        }
        bool empty() const { return reads.empty() && writes.empty(); }
        string to_str() {
            return set_to_str(reads) + " | " + set_to_str(writes);
        }

};



//---------------------------------------------------
//Varref
typedef Symbol SymbolRef;
typedef TUP::tuple<Symbol,Expr*> ArrayRef;
struct BvRef : TUP::tuple<Symbol,Expr*> {
    typedef TUP::tuple<Symbol,Expr*> Super;
    BvRef () {}
    BvRef (const Symbol& s, Expr* e) : Super(s,e) {} 
    string to_str() const ;
    const Symbol& get_symbol () const { return TUP::get<0>(*this); }
    const Expr* get_expr () const { return TUP::get<1>(*this); }
};

typedef VAR::variant<SymbolRef, BvRef, ArrayRef> Varreft;

class Varref : public Varreft {
    public:
        Varref() {}
        Varref(const Varreft& v) : Varreft(v) {}
        Var_dcl to_vd_plain() const; 
        string to_str() const;
        bool operator == (const Varref& vr) const { return (to_str() == vr.to_str()); }
        bool operator < (const Varref& vr) const  {
            return (to_str() < vr.to_str());
        }
};
//---------------------------------------------------

//Expr
enum Expr_Enum {Timeout, Else_expr}; //else added specifically for ast2sys_visitor
typedef TUP::tuple<op, Expr*, Expr* , Expr*> TernOp;
typedef TUP::tuple<Expr*, op , Expr*> BinOp;
typedef TUP::tuple<uop, Expr*> UnOp;
typedef TUP::tuple<Ty_list>  ChanInit;

typedef VAR::variant<BinOp,UnOp,TernOp,Varref,Const,Expr_Enum,ChanInit> Exprt;

class Expr: public Exprt {
  public:
      //Expr() {}
      Expr(const Exprt& e) : Exprt(e) {}
      bool is_const() const;
      bool is_varref(Varref& vr) const;
      string to_str() const;
      void get_rw(RW::RW_enum rwe, RW& rw) const;
      void get_rw_both(RW& rw) const ;
      void get_support(set<Varref>& support) const;
      Expr* project_to (const Expr * e) const ;
      bool operator < (const Expr& e) const { return to_str() < e.to_str(); }

      //bool operator == (const Expr& e) const { return to_str() == e.to_str(); } 
};


template<class O>
Expr* expr_fix_const_types ( const Expr* e, const O& oracle, Pid pid, const Ty& ety = Ty() ) ;
// Visitors
enum expr_visitor_mode {EVAL,TO_STR}; //evaluation expr_visitor_mode of expr
string to_str(const Varref& vr) ;

class ExprManager {

    private:
    static map<Expr, Expr*> exprMap;
    static int found_cnt, total_cnt; 

    //static Expr* simplify ( const Expr& e );

    static Expr* lookup(const Expr& e) {

      total_cnt++;
      if( exprMap.find(e) == exprMap.end() ) {
        found_cnt++;
        exprMap[e] = new Expr(e);
        //Util::msg(2,"not found expr "+e.to_str());
      }
      else { 
        //Util::msg(2,"found expr "+e.to_str());
      }
      return exprMap[e];

    }

    static Expr* simplify_lookup( const Expr& _e ) {
      //Expr* e = simplify (_e);
      Expr* e = lookup (_e);
      return e;
    }

    public:

    ExprManager() { }

    static Expr* get_expr (Expr* l, op o, Expr* r) { return simplify_lookup(Exprt(BinOp (l, o, r)));}
    static Expr* get_expr (op o, Expr* e1, Expr* e2, Expr* e3) { return simplify_lookup(Exprt(TernOp (o, e1, e2, e3)));}
    static Expr* get_expr (uop u, Expr* e) { return simplify_lookup(Exprt(UnOp (u, e)));}
    static Expr* get_expr (const Varref& v) { return simplify_lookup(Exprt(v));}
    static Expr* get_expr (Const c) { return simplify_lookup(Exprt(c));}
    static Expr* get_expr (Expr_Enum t) { return simplify_lookup(Exprt(t));}
    static Expr* get_expr (const Ty_list& t) { return simplify_lookup(Exprt(ChanInit (t)));}

    static Expr* get_sum ( Expr* a, Expr* b ) { return get_expr(a, PlusOp, b); }
    static Expr* get_sub ( Expr* a, Expr* b ) { return get_expr(a, MinusOp, b); }
    static Expr* get_mul ( Expr* a, Expr* b ) { return get_expr(a, TimesOp, b); }

    static Expr* get_eq(const Varref& vr, Expr* rhs) {
        Expr* vdexpr = get_expr(vr);
        return get_expr(vdexpr,EqualOp,rhs);
    }

    static Expr* get_diseq(const Varref& vr, Expr* rhs) {
        Expr* vdexpr = get_expr(vr);
        return get_expr(vdexpr,NotEqualOp,rhs);
    }
    static Expr* get_eq(const Varref& vr, const Const& c) 
    {
        Expr* cexpr = get_expr(c);
        return get_eq(vr,cexpr);
    }

    static Expr* get_diseq(const Varref& vr, const Const& c) 
    {
        Expr* cexpr = get_expr(c);
        return get_diseq(vr,cexpr);
    }

    static Expr* get_eq(Expr *a, Expr* b) { return ExprManager::get_expr(a, EqualOp, b); }
    static Expr* get_diseq(Expr *a, Expr* b) { return ExprManager::get_expr(a, NotEqualOp, b); }

    static Expr* get_ite ( Expr* c, Expr* t, Expr* f) { return get_expr(IteOp, c, t, f); }

    static Expr* get_and(Expr* e1, Expr* e2) {
        if ( e1 == get_true() ) return e2;
        if ( e2 == get_true() ) return e1;
        return get_expr(e1,LandOp,e2);
    }
    static Expr* get_or(Expr* e1, Expr* e2) {
        if ( e1 == get_false() ) return e2;
        if ( e2 == get_false() ) return e1;
        return get_expr(e1,LorOp,e2);
    }

    static Expr* get_true() { return lookup(Exprt(Const(true))); }
    static Expr* get_false() { return lookup(Exprt(Const(false))); }
    static Expr* get_not ( Expr* e ) { 
        if ( e == get_true() ) return get_false();
        if ( e == get_false() ) return get_true();
        return get_expr(NotUOp,e); 
    }

    static Expr* get_le ( Expr*a, Expr*b ) { return ExprManager::get_expr(a, LessEqualOp, b); }
    static Expr* get_lt ( Expr*a, Expr*b ) { return ExprManager::get_expr(a, LessOp, b); }
    static Expr* get_ge ( Expr*a, Expr*b ) { return ExprManager::get_expr(a, GreaterEqualOp, b); }
    static Expr* get_gt ( Expr*a, Expr*b ) { return ExprManager::get_expr(a, GreaterOp, b); }

    static void show_counts() {
        Util::message(2,"total count = %d, found count = %d\n", total_cnt, found_cnt);
    }

    friend struct Expr_simplify_vis ;
};


struct Guard_expr {
    list<const Expr*> exprs;
    Guard_expr() {}
    Guard_expr(const Expr* e) { exprs.push_back(e); }
    Guard_expr(list<const Expr*>& el) { exprs.splice(exprs.end(),el); }
    //void push_back( const Expr* e ) 
    string to_str() const { return list_ptr_to_str(exprs); }
    void push_back(const Expr* e) { exprs.push_back(e); }
    bool operator < (const Guard_expr& gex) const { return exprs < gex.exprs; }

};

typedef list<Guard_expr> Guard_expr_list;

#endif 
