#ifndef _FA_H
#define _FA_H
#include<fstream>
#include<sstream>
#include<deque>
#include"graph.h"
#include"path.h"
//#include"action.h"


template<class _St, class _Sig> 
class FA : public Graph< _St, Lab_edge<_St, _Sig> > {

    public:
        typedef _St St;
        typedef _Sig Sig;
        typedef set<Sig> Sigma;
        typedef set<St> Sts;
        typedef list<St> St_list;
        typedef Lab_edge<St,Sig> Ev;
        typedef Path<Ev> Trace;

        typedef typename Sts::const_iterator sts_const_it;
        typedef typename Sigma::const_iterator sig_const_it;

    protected:
        Sts I;
        Sts F;
        Sigma sigma;

        bool has_sink_st; St sink; //has a separate "sink" state
        typedef Graph<St,Ev> Super;
    public:
        bool has_sink() const { return has_sink_st; }
        void create_sink() { has_sink_st = true; sink = new_state(); }

        FA(bool mk_sink_st) { 
            has_sink_st = false;
            if( mk_sink_st ) { create_sink(); }
        }

        St new_state() { return Super::new_node(); }

        void set_init(const St& s) { CHK(I.empty()); I.insert(s); }
        void set_final(const St& s) { F.insert(s); }
        void set_sigma(const Sigma& _sigma) { 
            //Util::msg(2,"set sigma : sigma is "+set_to_str(_sigma));
            sigma = _sigma; }
        void add_to_sigma(const Sig& sig) { sigma.insert(sig); }

        void add_trans(const St& st1, const Sig& sig, const St& st2) {
            Ev ev(sig);
            Super::add_edge(st1, ev, st2);
            add_to_sigma(sig);
        }

        const map< St,map<Ev,St> >& get_trans() const { return Super::trans; }


        /* *                          ***                       * */

        const Sts& get_inits() const { return I;} 
        const Sts& get_final() const { return F;} 
        const St& get_sink() const { CHK( has_sink_st ); return sink; }
        size_t get_num_states() const { return Super::get_num_nodes(); }
        void is_a_state(const St& s) const { Super::check_N(s); } 
        bool is_final(const St& s) const { is_a_state(s); return (F.find(s) != F.end());}
        bool only_sink_is_final() const { 
            if (F.size() == 0) return true; 
            if ( has_sink_st && F.size() == 1 ) {
                const St& st = *(F.begin());
                if ( st == sink ) return true;
            }
            return false;
        }
        bool sigma_has(const Sig& sig) const { 
            
            bool ret = sigma.find(sig) != sigma.end(); 
            //Util::msg(2,"sigma_has: sigma = "+ set_to_str(sigma) + "\n" );
            //string res = ret?"has ":"does not have ";
            //Util::msg(5,"Sigma "+ res + sig.to_str()+"\n");
            return ret;
        }

        void delta_plain(const St& st, const Sig& sig, St_list& succs, bool check_sig=true) const {
            if ( check_sig ) CHK(sigma_has(sig));
            //Util::msg(2,"delta_plain: "+st.to_str()+" "+sig.to_str_const() );
            Ev ev(sig); 
            Super::get_opp(st, ev, succs);
            //Util::msg(2,"succs = "+list_to_str(succs));
            //Util::msg(2," succs.size() = "+int2str(succs.size())+"\n");
        }

        void delta(const St& st, const Sig& sig, St_list& succs, bool check_sig = true) const {
            if ( check_sig ) CHK(sigma_has(sig));
            else CHK ( has_sink_st);
            if( has_sink_st && st == sink && check_sig) { succs.push_back(sink); return; }

            delta_plain(st,sig,succs, check_sig);

            if( succs.empty() && check_sig ) { //check_sig == !delta_sym
                if( has_sink_st ) { succs.push_back(sink); }
                else Util::msg(1,"FA::delta: No successor and no sink state, assert(0) here\n");
            }
        }

        void delta_set(const St_list& st_set, const Sig& sig, St_list& opp_st_list, bool check_sig = true) 
                                                                                            const {
            typename_const_foreach(it, st_set) {
                St_list opp;
                delta(*it, sig, opp, check_sig);
                opp_st_list.splice( opp_st_list.end(), opp );
            }
        }

        void delta_sym(const set<St>& sts, const Sig& sig, set<St>& opp_sts) const {
            St_list st_list, opp_st_list;
            st_list.insert(st_list.begin(),sts.begin(),sts.end());
            delta_set(st_list,sig,opp_st_list,false);
            opp_sts.insert(opp_st_list.begin(), opp_st_list.end());
        }

        /* *                           ***                    * */

        void compute_eps_closure(const St_list& st_set, St_list& st_closure ) {
            Sig sig = Sig::get_epsilon(); //Epsilon letter
            delta_set(st_set, sig, st_closure);
        }

        void compute_act_succs(const St_list& st_set, map<Sig, St_list>& act_succ_map) const {
            //FIXed: better implementation - don't iterate over Sigma
            /*
               sig_const_it it = sigma.begin();
               for(; it != sigma.end(); ++it) {
               const Sig& sig = *it;
               Sts opp;
               delta_set(st_set, sig, opp);
               act_succ_map[sig].insert(opp.begin(), opp.end());
               }*/

            //better
            typename_const_foreach(it, st_set) {
                list<Ev> out_edges;
                get_out_edges(*it,out_edges);
                typename_const_foreach(ite, out_edges) {
                    Sig sig = ite->get_lab();
                    act_succ_map[sig].push_back(ite->target());
                }
            }
        }



        St get_state(const string& str, map<string, St>& str2st) {
            if( str2st.find(str) == str2st.end() ) {
                St s = new_state();
                str2st[str] = s;
            }
            return str2st[str];
        }


        void get_comm_sigma(Sigma& comm_sigma) const { comm_sigma = sigma; } 
        void get_sigma(Sigma& comm_sigma) const { comm_sigma = sigma; } 

        Sig get_sig(const string& str) {
            Sig sig; 
            sig.get_from_str(str);
            //sig = get_act_from_str<Sig>(str);
            return sig;
        }

        void get_opp_map(const St& st, const Sig& sig, map<Ev,St>& e2st) const {
            Super::get_opp_map(st, Ev(sig), e2st);
        }

        void read_from_file(const string& fname) {
            ifstream in(fname.c_str());
            if ( !in.is_open() ) {
                Util::msg(2,"unable to read from file "+fname);
                assert(0);
            }
            string line;
            map<string, St> str2st;
            while( !getline(in,line).eof() ) {
                istringstream ls(line);
                string str;
                if ( line == "" || line =="\n" ) continue;
                if( line[0] == '#' ) {
                    ls >> str;
                    if( str == "#INIT" ) {
                        while( ls >> str ) {
                            set_init(get_state(str,str2st));
                        }
                    }
                    else if ( str == "#FINAL" ) {
                        while ( ls >> str ) {
                            set_final(get_state(str,str2st));
                        }
                    }
                    else assert(0);
                    continue;
                }

                ls >> str;
                St s1 = get_state(str,str2st);
                ls >> str;
                CHK ( str == "(" );
                ls >> str;
                string sig_str;
                while ( str != ")" ) {
                    sig_str += " " + str;
                    ls >> str;
                }
                Sig sig = get_sig(sig_str);
                ls >> str;
                St s2 = get_state(str,str2st);
                add_trans(s1,sig,s2);
                CHK( !(ls >> str) );
            }
            in.close();
        }

        string to_str_const() const {
            string s = "\n"+Super::to_str_const();
            s += "\nInit: ";
            typename_const_foreach(it, I) {
                s += it->to_str();
            }

            s += "\nFinal: ";
            typename_const_foreach(it, F) {
                s += it->to_str();
            }
            return s;
        }

        string to_str() {
            string s = "\n"+Super::to_str();
            s += "\nInit: ";
            typename_foreach(it, I) {
                s += it->to_str();
            }

            s += "\nFinal: ";
            typename_foreach(it, F) {
                s += it->to_str();
            }
            return s;
        }

};

template<class St, class Sig> 
class DFA: public FA<St,Sig> {
    protected:
        typedef FA<St,Sig> Super;
        int max_depth;

        void set_depth( int depth ) { max_depth = depth; }

    public:
        typedef typename Super::Sigma Sigma;
        typedef typename Super::Sts Sts;
        typedef typename Super::Trace Trace;
        typedef typename Super::St_list St_list;
        typedef typename Super::Ev Ev;

        DFA (bool mk_sink_st) : Super(mk_sink_st) { set_depth(-1); }
        // Not allowing this - always make this explicit by calling construct_from
        // DFA (const Super& fa) : Super(fa.has_sink()) {assert(0); construct_from(fa); }

        void set_init(const St& s) { CHK(Super::I.empty()); Super::set_init(s); }
        int get_depth() const { return max_depth; }

        void add_trans(const St& st1, const Sig& sig, const St& st2) {
            Super::add_trans(st1,sig,st2);
            //St_list succs;
            //Super::delta_plain(st1, sig, succs);
            //if( Super::has_sink_st ) { CHK ( succs.size() == 1 && *(succs.begin()) == Super::sink );}
            //else
            //CHK( succs.size() == 1);
        }

        DFA( const pair<Trace,Sigma>& path) : Super(true) {

            Super::sigma = path.second;
            const Trace& t = path.first;
            set_depth( t.size() );

            St st = Super::new_state();
            Super::set_init(st);
            St prev = st;
            Super::set_final(prev);
            typename_const_foreach(it, t) {
                typename Trace::const_iterator itn = it;
                itn++;
                St next = Super::new_state();
                Super::add_trans(prev, it->get_act(), next);
                prev = next;
                Super::set_final(prev);
                if( itn == t.end() ) break;
            }
        }

        void construct_from(const DFA<St,Sig>& dfa) { *this = dfa; }

        void construct_from(const FA<St,Sig>& fa) {
            set_depth(-1);
            map<St_list,St> q_map;

            const Sts& fa_init_set = fa.get_inits();
            St_list fa_init; 
            copy(fa_init_set.begin(), fa_init_set.end(), inserter(fa_init, fa_init.begin()));

            //St_list fa_init_closure;
            //compute_eps_closure(fa_init, fa_init_closure);
            //FIX later for epsilon actions
            const St_list& fa_init_closure = fa_init;

            St s = Super::new_state();
            q_map[fa_init_closure] = s;
            Super::set_init(s);

            deque<St_list> worklist;
            worklist.push_back(fa_init_closure);

            while( !worklist.empty() ) {
                St_list curr_list = worklist.front();
                worklist.pop_front();
                CHK( q_map.find(curr_list) != q_map.end());
                const St& curr_state = q_map[curr_list];

                bool is_st_final = false;
                typename_const_foreach(it, curr_list) {
                    if( fa.is_final(*it) ) {is_st_final = true;}
                }
                if(is_st_final) set_final(curr_state);

                map<Sig,St_list> act_succ_map;
                fa.compute_act_succs(curr_list, act_succ_map);

                typename_const_foreach(it, act_succ_map) {
                    const Sig& sig = it->first;
                    const St_list& opp_set = it->second;

                    //St_list opp_set_closure;
                    //compute_eps_closure(opp_set, opp_set_closure);
                    // FIX later for epsilon actions
                    const St_list& opp_set_closure = opp_set;

                    St opp_state;
                    if( q_map.find(opp_set_closure) == q_map.end() )  {
                        opp_state = Super::new_state();
                        q_map[opp_set_closure] = opp_state; 
                        worklist.push_back(opp_set_closure);
                    }
                    else opp_state = q_map[opp_set_closure];

                    add_trans(curr_state, sig, opp_state);
                }
            }

        }

        const St& get_init() const { return *(Super::I.begin());} 
        void delta(const St& st, const Sig& sig, St& succ) const {
            CHK( sigma_has(sig) );
            const St& sink = Super::sink;

            if( Super::has_sink_st && st == sink ) { succ = sink; return; }

            St_list succs;
            Super::delta(st,sig,succs);

            if( succs.empty() ) {
                if( Super::has_sink_st ) { succ = sink; }
                else assert(0);
            }
            else  { CHK( succs.size() == 1); succ = *(succs.begin()); }
        }

        void negate() {
            const Sts&  nodes = Super::nodes;
            Sts rest;
            Sts& F = Super::F;
            set_difference(nodes.begin(), nodes.end(),F.begin(), F.end(), inserter(rest, rest.begin()));
            F = rest;
        }

        template<class Run>
            void run(Run& t, uint run_len, St& st) const {
                st = get_init();
                //candidate->project(candidate->get_init(),ce.GetHead(),stSet);
                //CHK(stSet.size() == 1);
                typename Run::const_iterator it = t.begin();
                St opp; 
                for(size_t j = 0; j < run_len; ++j,++it) {
                    CHK( it != t.end() );
                    //Util::msg(3, it->get_act().to_str());
                    delta(st,Sig(it->get_act()),opp);
                    st = opp;
                    //candidate->Project(x,it->second,stSet); 
                    //CHK(stSet.size() == 1);
                }
            }

        template<class Run>
            void run_sym(Run& t, uint run_len, set<St>& sts) const {
                sts.insert(get_init());
                typename Run::const_iterator it = t.begin();
                for(size_t j = 0; j < run_len; ++j,++it) {
                    set<St> opp; 
                    CHK( it != t.end() );
                    //Util::msg(3, it->get_act().to_str());
                    Super::delta_sym(sts,Sig(it->get_act()),opp);
                    sts = opp;
                }
                CHK( !sts.empty());

            }


        // run up to run_len, then build emap from sts[run_len] to sts[run_len+1]
        //pass run_len = mid to this one
        template<class Run>
            void run_sym_ext(Run& t, uint run_len, map<St, map<Ev,St> >& emap) const {
                CHK ( run_len >= 0 && run_len + 1 <= t.size() );
                Sts sts;
                run_sym(t,run_len,sts);
                typename Run::const_iterator itr = t.begin();
                for(size_t j = 0; j < run_len; ++j,++itr) ;
                Sig sig = itr->get_act();

                typename_const_foreach(it, sts) {
                    St st = *it;
                    map<Ev,St> e2st;
                    Super::get_opp_map(st,sig,emap[st]);
                }
            }

        static void get_univ_dfa( const Sigma& sigma, DFA& dfa ) {
            dfa.set_sigma(sigma);
            St st = dfa.new_state();
            dfa.set_init(st);
            dfa.set_final(st);
            
            typename_const_foreach(it, sigma) {
                dfa.add_trans(st, *it, st);
            }
        
        } 
};

template<class St, class Sig>
class Trace : public list< Lab_edge<St,Sig> > {

};

class FA_manager {
    //map< pair<FA*,DFA*>, Quotient_map> qmap_cache;
    public:
        //void determinize(const FA& M, Quotient_map& qmap, DFA& N) ;
};





#endif 
