#ifndef _FOCI_DP_H
#define _FOCI_DP_H
#include"basic_vals.h"
#include"tys.h"
#include"foci_c.h"
#include"yices_c.h"

struct Foci_variant {
    foci_pred p;
    foci_expr e;
    enum form {NONE, E, P};
    form f;

    //static int tmp_cnt;
    
    Foci_variant() { f = NONE; }
    Foci_variant ( foci_pred _p ) : p(_p) { f = P; }
    Foci_variant ( foci_expr _e ) : e(_e) { f = E; }

    bool is_pred() { return (f== P); }
    bool is_expr() { return (f== E); }

    //string get_new_tmp() { return "____tmp"+int2str(tmp_cnt++); }

    string to_str() {
        if( is_pred() ) return foci_pred_to_str(p);
        else if ( is_expr() ) return foci_expr_to_str(e);
        else assert(0);
    }

    Foci_variant operator + ( Foci_variant& o ) ;
    Foci_variant operator - ( Foci_variant& o ) ;
    Foci_variant mk_mul ( Foci_variant& o ) ;

    Foci_variant operator && ( Foci_variant& o ) ;
    Foci_variant operator || ( Foci_variant& o ) ;
    
    Foci_variant mk_ite ( Foci_variant& t, Foci_variant& f );
    Foci_variant operator ! (); 

    Foci_variant operator != ( Foci_variant& o ) ;
    Foci_variant operator == ( Foci_variant& o ) ;

    Foci_variant operator <= ( Foci_variant& o );
    Foci_variant operator < ( Foci_variant& o );
    Foci_variant operator >= ( Foci_variant& o );
    Foci_variant operator > ( Foci_variant& o );

    bool interpolate ( Foci_variant& o, Foci_variant& res ) ;
    lbool check();
    Foci_variant shift_to_zero ();
    Foci_variant elim_ite();



};

class FociDP {

    public:
        typedef Foci_variant expr;
        //typedef value expr;
        //typedef expr foci_pred;
        //typedef expr foci_expr;
        typedef Ty DP_type;

    protected:

        map<Ty,string> ty2var;
        map<string,Ty> var2ty;
        expr init_e;
        vector<expr> asserts;
        vector<expr> trans_vec, err_vec;

        bool initialized;


        bool is_decl ( const string& name ) const {
            //Util::msg(2,"is_decl: "+name);
            return Key_In_Bag(name,var2ty);

        }

        Ty get_ty ( const string& name ) const {
            CHK ( is_decl(name) );
            return var2ty.find(name)->second;
        }

        void add_var ( const string& name, const Ty& ty , bool dup_check = true) {
            if ( dup_check && is_decl(name) ) {
                assert(0);
                return;
            }
            //  Util::msg(2,"adding var "+name+" of type "+ty2str(ty));
            var2ty[name] = ty;
            //           ty2var[ty] = name;
        }

    public:

        FociDP() { initf(); }

        void initf() {
            initialized = false;
        }

        void set_init(expr inite) { init_e = inite.elim_ite(); initialized = true; }
        expr get_init() { CHK ( initialized ); return init_e; }

        void set_trans(uint depth, expr e) {
            CHK(trans_vec.size() == depth);
            trans_vec.push_back( e.elim_ite() );
        }
        void set_err(uint depth, expr e) {
            CHK(err_vec.size() == depth);
            err_vec.push_back( e.elim_ite() );
        }

        void mk_var_decl(const string& name, const Ty& ty) {
            add_var(name,ty);
        }

        expr mk_bool_var(const string& name) { 
            add_var(name, Ty(BoolTy) );
            return foci_mk_atom_pred(name);
        }

        DP_type mk_type (const Ty& ty) {
            return ty;
        }

        expr get_var_expr(const string& name) {
            Ty ty = get_ty(name);
            string tystr = ty2str(ty);
            if ( tystr == "bool" ) return foci_mk_atom_pred(name);
            else return foci_mk_var(name); 
            //TODO: handle bv
        }

        //void create_model ( vector< pair<string,Const> >& model ) ;
        //void get_model(expr err, map<string,Const>& model, int& max_d, bool generalize, assertion_id ) ;
        lbool check();
        //lbool check_BMC();
        //void show_sat () { assert(0); }

        expr mk_num(int n) { return foci_mk_const(n); }
        expr mk_const_expr(const Const& c);

        expr mk_true () { return foci_mk_true(); }
        expr mk_false () { return foci_mk_false(); }

        expr mk_sum (expr a, expr b) { return a + b; }
        expr mk_sub (expr a, expr b) { return a - b; }
        expr mk_mul (expr a, expr b) { return a.mk_mul(b); }

        expr mk_and(expr a, expr b) { return a && b; }
        expr mk_or(expr a, expr b) {  return a || b; }

        expr mk_diseq(expr a, expr b) { return a != b; }
        expr mk_eq(expr a, expr b) { return a == b; }
        expr mk_ite( expr c, expr t, expr f) { return c.mk_ite(t, f); }

        expr mk_not(expr a) { return !a; }

        expr mk_le(expr a, expr b) { return a <= b; }
        expr mk_lt(expr a, expr b) { return a < b; }
        expr mk_ge(expr a, expr b) { return a >= b; }
        expr mk_gt(expr a, expr b) { return a > b; }

        void push() { assert(0); }
        void pop() { assert(0); }

        void assertdp(expr& e) ;
        assertion_id assert_retract_dp(expr& e) ;

        void retract ( assertion_id id )  { assert(0); }
        void retract_all(vector<expr>& exprs ) { assert(0); }

        void mk_bmc_split (uint depth, expr& A, expr& B ) ;
        //expr interpolate ( expr& A, expr& B ) ;
        bool interpolate ( int depth, expr& A_ );

};


#endif
