#ifndef G_ST_DIFF
#define G_ST_DIFF
#include"expr.h"

template<class V>
struct Psd {
    map<Pid, map<Var_dcl,V> > vars_diff; //pid -> (pos -> val)

    const V& get_val(Pid pid, const Var_dcl& vd) const { 
        CHK( vars_diff.find(pid) != vars_diff.end() );
        CHK( vars_diff.find(pid)->second.find(vd) != vars_diff.find(pid)->second.end() );
        return vars_diff.find(pid)->second.find(vd)->second; 
    }

    void set_val(Pid pid, const Var_dcl& vd, const V& val) { vars_diff[pid][vd]= val; }
    bool empty() const { return vars_diff.empty(); }
    void clear() { vars_diff.clear(); }

    template<class L>
    void add_vals (Pid pid, const Var_dcl& vd, const L& vs ) {
        vars_diff[pid][vd].insert(vs.begin(),vs.end());
    }

    void abstract_to( Psd& psd ) const {
        const map<Pid, map<Var_dcl,V> >& vars_diff2 = psd.vars_diff;
        typename_const_foreach (it, vars_diff2) {
            Pid pid = it->first;
            const map<Var_dcl,V>& pos2val = it->second; 
            typename_const_foreach(it2, pos2val) {
                const Var_dcl& vd = it2->first;
                psd.set_val( pid, vd, get_val(pid,vd) );
            }
        }
    }

    template<class T>
        void dup_by_ranges(list<T>& S) {
            list<T> S_;

            typename_const_foreach (it, vars_diff) {
                Pid pid = it->first;
                const map<Var_dcl,V>& pid_vars_diff = it->second; 
                typename_const_foreach(it, pid_vars_diff) {
                    const Var_dcl& vd = it->first;
                    const V& conc_values = it->second;

                    typename_const_foreach(itl, S) {
                        typename_const_foreach(itv, conc_values) {   
                            T new_s = *itl;
                            new_s.set_val(pid, vd, *itv);
                            S_.push_back(new_s);
                        }
                    }
                    S = S_;
                    S_.clear();
                    Util::msg(2,"S's size = "+int2str(S.size()));
                }
            }
        }

    string to_str() const {
        string s;
        typename_const_foreach(it, vars_diff) {
            Pid pid = it->first;
            s+="["+int2str(pid)+"] ";

            const map<Var_dcl,V>& var_map = it->second;
            typename_const_foreach(it2, var_map) {
                s += it2->first.to_str() + ":" + it2->second.to_str() + " ";
            }
        }
        return s;
    }
};

template<class V>
struct Gsd {
    typedef map<Var_dcl,V> Val_map;
    Val_map vars_diff; //pos to new val

    bool has_var(const Var_dcl& vd) const { return vars_diff.find(vd) != vars_diff.end(); }
    bool empty() const { return (vars_diff.empty()); }
    const V& get_glb_val(const Var_dcl& vd) const { 
        CHK( vars_diff.find(vd) != vars_diff.end());
        return vars_diff.find(vd)->second; 
    }
    void set_val(const Var_dcl& vd, const V& glb_val) { vars_diff[vd] = glb_val; }
    template<class L>
    void add_vals ( const Var_dcl& vd, const L& vs ) {
        vars_diff[vd].insert(vs.begin(),vs.end());
    }
    void add_vals ( const Gsd& gsd ) {
        const Val_map& vm = gsd.vars_diff;
        typename_const_foreach(it, vm) {
            add_vals(it->first, it->second);
        }
    }

    void abstract_to( Gsd& gsd ) const {
        const map<Var_dcl, V>& vd2gv = gsd.vars_diff;
        typename_const_foreach(it, vd2gv) {
            CHK ( vars_diff.find(it->first) != vars_diff.end());
            gsd.set_val( it->first, get_glb_val(it->first) );
        }
    }

    void get_rw ( RW::RW_enum mode, RW& rw) const {
        typename_const_foreach(it, vars_diff) {
            rw.add(mode, it->first);
        }
    }

    template<class T>
        void dup_by_ranges(list<T>& S) {
            list<T> S_;
            typename_const_foreach(it, vars_diff) {
                const Var_dcl& vd = it->first;
                const V& conc_values = it->second;

                typename_const_foreach(itl, S) {
                    typename_const_foreach(itv, conc_values) {   
                        T new_s = *itl;
                        new_s.set_val(-1, vd, *itv);
                        S_.push_back(new_s);
                    }
                }
                S = S_;
                S_.clear();
            }
        }

    string to_str() const {
        string s;
        if( vars_diff.empty() ) return s;
        typename_const_foreach(it, vars_diff) {
            s += it->first.to_str() + ":" + it->second.to_str() + " ";
        }
        return s;
    }

    string to_str_first() const {
        string s;
        if( vars_diff.empty() ) return s;
        typename_const_foreach(it, vars_diff) {
            s += it->first.to_str() + " ";
        }
        return s;
    }
};


template<class V>
class St_d {
    protected:
        Psd<V> psd;
        Gsd<V> gsd;
    public:
        St_d() {}
        Psd<V>& get_psd() { return psd; }
        Gsd<V>& get_gsd() { return gsd; }
        const Psd<V>& get_psd() const { return psd; }
        const Gsd<V>& get_gsd() const { return gsd; }

        void set_loc_val(Pid pid, const Var_dcl& vd, const V& val) {
            return psd.set_val(pid,vd,val);
        }
        void set_glb_val(const Var_dcl& vd, const V& val) {
            return gsd.set_val(vd,val);
        }

        void set_val(Pid pid, const Var_dcl& vd, const V& val) {
            if( pid ==-1 ) set_glb_val(vd,val);
            else set_loc_val(pid,vd,val);
        }
        template<class L> //list
        void add_vals (Pid pid, const Var_dcl& vd, const L& vs ) {
            if ( pid == -1 ) gsd.add_vals(vd,vs);
            else psd.add_vals(pid,vd,vs);
        }

        bool empty() const { return (psd.empty() && gsd.empty()); }

        void abstract_to ( St_d& sd ) const {
            psd.abstract_to(sd.psd);
            gsd.abstract_to(sd.gsd);
        }
        template<class T>
            void dup_by_ranges ( const T& s, list<T>& S_ ) {
                list<T> S; S.push_back(s);
                psd.dup_by_ranges(S);
                gsd.dup_by_ranges(S);

                S_ = S;
                CHK( !S_.empty() );
            }

        void abstract_to_globals ( St_d& sd ) const { gsd.abstract_to(sd.gsd); }
        //void get_glb_rwm (RW_map::RW_enum mode, RW_map& rw) const { gsd.get_rwm (mode, rw); }
        void get_glb_rw (RW::RW_enum mode, RW& rw) const { gsd.get_rw (mode, rw); }

        string to_str() const { 
            string s = "(" + psd.to_str() + "| "+ gsd.to_str() +")"; 
            return s;
        }
        bool operator < (const St_d& sd) const { return (to_str() < sd.to_str()); }

};



typedef St_d<Consts> Conc_sd; //larger version of Glb_val_map


#endif
