#ifndef _GLOBALS_H
#define _GLOBALS_H

#include"symtab.h"
#include"cfg_comp.h"
#include"naming.h"


struct Globals {

    CFG init_cfg; 
    //glb_sym_tab stores vars separately
    Symtab glb_sym_tab;
    map<Pid,const CFG_comp*> pid2cfg;

    void init() {
        init_cfg = CFG("_global_init_");
    }
    void add_ev_var(const Var_decl& d) { glb_sym_tab.insert(d); }
    void add_type_decl(const Type_decl& td) { glb_sym_tab.insert_type(td); }
    void add_glb_var(const Var_decl& d) { glb_sym_tab.insert(d);}


    const CFG_comp& get_cfg_comp(Pid pid) const {
        CHK( pid2cfg.find(pid) != pid2cfg.end() );
        return *(pid2cfg.find(pid)->second);
    }

    const Symtab& get_symtab_pid(Pid pid) const {
        if(pid == -1) return glb_sym_tab;
        else {
            const CFG_comp& cfg_comp = get_cfg_comp(pid);
            return cfg_comp.get_symtab();
        }
    }

    /*bool cfg_comp_has_var (Pid pid, const Var_dcl& vd) const {
        const Symtab& symtab = get_symtab_pid(pid);
        return symtab.has_var(vd);
    }*/

    Pid get_pid(const Var_dcl& vd, Pid hint_pid) const {
        if ( get_symtab_pid(hint_pid).has_var(vd) ) return hint_pid;
        else if ( glb_sym_tab.has_var(vd) ) return -1;
        else { Util::msg(2,"Globals::get_pid - unable to locate var "+vd.to_str()); assert(0); }
        /*
        const_foreach(it, pid2cfg) {
            const Symtab& symtab_i = (it->second)->get_symtab();
            if ( symtab_i.has_var(vd) ) return it->first;
        }
        //one of the sym tabs should have it
        assert(0);
        */
    }

    //must know the pid before calling get_range_values
    void get_range_values (Pid pid, const Var_dcl& vd, list<Const>& clist) const {
        const Symtab& symtab = get_symtab_pid(pid);
        symtab.get_range_values(vd, clist);
        CHK( !clist.empty() );
    }

    Ty get_ty( const Var_dcl& vd, Pid hint_pid) const {
        Pid pid = get_pid(vd,hint_pid);
        const Symtab& symtab = get_symtab_pid(pid);
        return symtab.get_type(vd);
    }
    /*Ty get_ty ( const Var_dcl& vd) const {
        Pid pid = get_pid(vd);
        return get_ty(pid,vd);
    }*/


    string get_glb_name( const Var_dcl& vd, Pid hint_pid, int depth) const {
        Pid pid = get_pid(vd,hint_pid);
        if( pid != -1 ) {
            return Naming::get_prefix(pid) + vd.to_str() + Naming::get_suffix(depth);
        }
        else return vd.to_str() + Naming::get_suffix(depth);
    }
/*
    template<class DP>
    typename DP::DP_type get_DP_ty ( DP& dp, const Var_dcl& vd, Pid hint_pid ) const {
        Ty ty = get_ty(vd,hint_pid);
        return  dp.mk_type(ty);
    }
*/
};

#endif
