#ifndef GRAPH_H
#define GRAPH_H

#include"common.h"
#include"util.h"
using namespace Util;

class Node {
    protected:
        tag id; //long int?
    public:
        Node() { id = -1; }
       // Node(const tag& i): id(i) {}
        //Node(const Node& n): id(n.id) {}
        virtual ~Node() {}

        bool operator < (const Node& n) const { 
            //  Util::message(2,"node< op addresses, this = %x, n=%x\n",this,&n);
            return (id < n.id); }
        bool operator == (const Node& n) const { return (id == n.id); }
        friend ostream& operator << (ostream& os, const Node& n) ;

        virtual string to_str() const;
        void set_tag( tag _id ) { id = _id; }
        tag tagi() const {return id;}

        //template<class N, class E> friend class Graph;
        //friend class Graph<class N, class E>;
        //friend class CFG;
};

template<class N>
class Edge {
    protected:
        tag id;
        N sou, tar;

    public:
        Edge() {id =-1;}
        //Edge(const tag& t): id(t) {}
        Edge(const Edge& e) : id(e.id), sou(e.sou), tar(e.tar) {}
        virtual ~Edge() {}

        void set_tag() { id = Util::get_new_etag();}
        bool operator < (const Edge& e) const { return (id < e.id); }
        bool operator == (const Edge& e) const { return (id == e.id); }
        template<class n>
            friend ostream& operator << (ostream& os, const Edge<n>& e);

        virtual string to_str_const() const = 0;
        virtual string to_str() = 0;
        void set_ends(const N& _s, const N& _t) {sou = _s; tar = _t;}
        const N& source() const { return sou;}
        const N& target() const { return tar;}

};

template<class N>
ostream& operator << (ostream& os, const Edge<N>& e) {os << e.to_str(); return os;}

template<class N, class E> 
class Graph {

    protected:
        typedef map<N, map<E,N> > trans_t; //node entry in this transition map only if node has any transitions
        typedef typename map<E,N>::iterator mm_it;
        typedef typename map<E,N>::const_iterator mm_cit;
        trans_t trans;
        set<N> nodes;
        set<E> edges;

        void add_N(const N& n) { 
            assert(nodes.find(n) == nodes.end()); 
            nodes.insert(n);
            trans[n] = map<E,N>();
        }

        void erase_N(const N& n) {
            check_N(n);
            trans_iterator it = trans.find(n);
            for( mm_it it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
                edges.erase(it2->first); 
            }
            it->second.clear(); //clear <E,N> pairs
            trans.erase(n); // erase n from trans
            nodes.erase(n); // erase n from nodes
        }

    public:
        typedef E Edge;
        typedef set<E> Edges;
        typedef N Node;
        typedef set<N> Nodes;
        typedef typename trans_t::const_iterator trans_const_iterator;
        typedef typename trans_t::iterator trans_iterator;

        Graph() {}
        virtual ~Graph() {}

        N new_node() { tag id = Util::get_new_tag(); N n; n.set_tag(id); add_N(n); return n;}

        void check_N(const N& n) const { assert(nodes.find(n) != nodes.end()); }

        void add_edge(const N& n1, E& e, const N& n2)
        {
            check_N(n1);
            check_N(n2);
            e.set_tag();
            e.set_ends(n1,n2);
            trans[n1].insert(pair<E,N>(e,n2));
            edges.insert(e);
            Util::message(5,"%%%%%%%%%%%%%%%% added edge: %s%s%s\n",n1.to_str().c_str(), e.to_str().c_str(),n2.to_str().c_str());
        }

        void add_edge(const N& n1, const N& n2) {
            E e;
            add_edge(n1, e, n2);
        }

        //void get_out_edges(const N& n, set< pair<E,N> >& E_list) const 
        void get_in_edges(const N& n, list< E >& E_list) const { assert(0); }

        template<class E_>
            void get_out_edges(const N& n, list< E_ >& E_list) const {
                check_N(n);
                trans_const_iterator it = trans.find(n);
                assert( it != trans.end());
                //weaker         if( it == trans.end()) return;
                if( it->second.size() == 0) return;
                for( mm_cit it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
                    E_list.push_back( E_(it2->first) );
                }
            }


        void get_opp(const N& n, const E& e_, list<N>& opp) const {
            Util::msg(5,"inside get_opp\n");
            check_N(n);
            trans_const_iterator it = trans.find(n);
            assert( it != trans.end());
            if( it->second.size() == 0) { Util::msg(5,"return\n"); return;}
            for( mm_cit it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
                const E& e = it2->first;
                if( e.matches(e_) ) {
                    Util::msg(5,"e matches e_\n");
                    opp.push_back(e.target());
                }
            }
        }

        void get_opp_map( const N& n, const E& e_, map<E,N>& e2st) const {
            check_N(n);
            trans_const_iterator it = trans.find(n);
            assert( it != trans.end());
            if( it->second.size() == 0) return;
            for( mm_cit it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
                const E& e = it2->first;
                if( e.matches(e_) ) {
                    e2st[e] = e.target();
                }
            }
        }

        template<class Bag>
        void get_succs (const N& n, Bag& succ) const {
                check_N(n);
                trans_const_iterator it = trans.find(n);
                assert( it != trans.end());
                if( it->second.size() == 0) return;
                for( mm_cit it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
                    //maybe use a generalized "insert" here
                    succ.push_back( it2->second );
                }

        }

        size_t get_num_nodes() const { return nodes.size(); }

        tag get_min_st_id () const {
            CHK( !nodes.empty());
            tag min_id = nodes.begin()->tagi();
            typename_const_foreach(it, nodes) {
                if ( min_id > it->tagi() ) min_id = it->tagi();
            }
            return min_id;
        }


        const Edges& get_edges() const { return edges; }
        const Nodes& get_nodes() const { return nodes; }

        //mutator utilities
        void remove_inaccessible(const N& start);

        string to_str_const() const {
            string s;
            trans_const_iterator it;
            for(it = trans.begin(); it != trans.end(); ++it) {
                const N& n = it->first;
                s += n.to_str();
                const map<E,N>& en = it->second;
                for (mm_cit it2 = en.begin(); it2 != en.end() ; ++it2) {
                    s += "\t" + it2->first.to_str_const() + it2->second.to_str() + "\n";
                }
                s +="\n";
            }
            return s;
        }

        string to_str() {
            string s;
            trans_iterator it;
            for(it = trans.begin(); it != trans.end(); ++it) {
                const N& n = it->first;
                s += n.to_str();
                map<E,N>& en = it->second;
                for (mm_it it2 = en.begin(); it2 != en.end() ; ++it2) {
                    E e = it2->first;
                    s += "\t" + e.to_str() + it2->second.to_str() + "\n";
                }
                s +="\n";
            }
            return s;
        }
};

//define node, edge types, graph dfs adapters

template<class N, class E> 
class DFS_explorer {
    protected:
        const Graph<N,E>& g; 
        map<N,bool> marked;

        void dfs_rec(const N& n) {
            marked[n] = true;
            list<E> elist;
            g.get_out_edges(n, elist);
            for(typename list<E>::const_iterator it = elist.begin(); it != elist.end(); ++it) {
                const N& opp = it->target();
                if( !is_marked(opp) )
                    dfs_rec(it->target());
            }
        }
    public:

        DFS_explorer(const Graph<N,E>& _g) : g(_g) {}
        bool is_marked(const N& n) const { return (marked.find(n) != marked.end()); }

        void run(const N& n) { marked.clear(); dfs_rec(n); }
};


#endif
