#ifndef IDD_H
#define IDD_H

#include"sys_state.h"

class ID_diff { //immediate-delta diff tuple
    protected:
        Sys_state_diff imm_diff, delta_diff;

    public:
        ID_diff( const Mem_map& mem) : imm_diff(mem), delta_diff(mem) {}
        const Sys_state_diff& imm() const { return imm_diff; }
        const Sys_state_diff& delta() const { return delta_diff; }
        bool empty() const { return imm_diff.empty() && delta_diff.empty(); }

        void get_diff( Sys_state_diff& sd ) const { //apply imm_diff, then delta_diff
            sd = imm_diff;
            //Util::msg(2,"id_diff::get_diff: add_check{");
            sd.add_check(delta_diff) ; ;
            //Util::msg(2,"}");
        }

        string to_str() const { return "[imm] "+ imm_diff.to_str() + "\n[delta]" + delta_diff.to_str(); }
        bool operator < (const ID_diff& id) const { return to_str() < id.to_str(); }

        bool apply_imm_diff( const Sys_state_diff& sd, Upd_mode U = NO_OW ) { 
            //Util::msg(2,"id_diff::apply_imm_diff::add_check{");
            return imm_diff.add_check(sd,U); 
            //Util::msg(2,"}");
        }
        bool apply_delta_diff( const Sys_state_diff& sd, Upd_mode U = NO_OW) { 
            //Util::msg(2,"id_diff::apply_delta_diff::add_check{");
            return delta_diff.add_check(sd,U); 
            //Util::msg(2,"}");
        }
        //get rid of add_check
        bool add_check ( const ID_diff& idd, Upd_mode U = NO_OW)  {
            bool ret = apply_imm_diff(idd.imm(),U);
            if ( !ret ) return false;
            return apply_delta_diff(idd.delta(),U);
        }

        void set_imm_pcval (Pid pid, int pc) { imm_diff.set_pcval(pid,pc); } 
        //void set_delta_pcval (Pid pid, int pc) { delta_diff.set_pcval(pid,pc); } 



};

typedef set<ID_diff> ID_diffs;

#endif
