/**************************** CPPHeaderFile ***************************

* FileName [LLearner2.h]

* PackageName []

* Synopsis [Header file for LLearner2 class.]

* Description [This class encapsulates a generalized MN learner]

* SeeAlso []

* Author [Nishant Sinha]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University.  All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: nishants+@cs.cmu.edu. ]

**********************************************************************/

#ifndef __LLEARNER2_H_
#define __LLEARNER2_H__

//#include"teacher.h"
//#include"fa.h"

template<class Trace>
class Follow2 {

    public:
        typedef typename Trace::Prefix Row_el;
        typedef typename Row_el::Ext Row_el_ext;
        typedef typename Row_el_ext::Act Sig;

    protected:
        typedef set<Sig> Sigma;
        map<Row_el, map<Sig,Sig> > follow_map;
        //map<Row_el, Sigma> follow_map_refine; //sig used to refine follow_map
        Sig basis;


        map<Sig,Sig>& get_follow_int( const Row_el& rel ) {
            CHK ( Key_In_Bag(rel,follow_map) );
            return follow_map.find(rel)->second;
        }
    public:

        Follow2( const Sig& _basis ) : basis(_basis) { }

        bool has_rel ( const Row_el& rel ) const {
            return Key_In_Bag(rel, follow_map);
        }

        void init_row_el ( const Row_el& rel ) {
            if ( !has_rel(rel) ) {
                follow_map[rel][basis] = Sig();
            }
            // follow_map[rel] = basis_set;
            //follow_map_refine[rel].clear();
        }

        const map<Sig,Sig>& get_follow( const Row_el& rel ) const {
            CHK ( has_rel(rel) );
            return follow_map.find(rel)->second;
        }


        void set_follow( const Row_el& rel, const map<Sig,Sig>& follow_sigs) {
            CHK ( !has_rel(rel) );
            follow_map[rel] = follow_sigs;
        }


        // split sigs with refine_sig
        // refine_sig is ceA, add ceA_enl too

        void refine_with( Row_el& rel, const Sig& rep_sig, const Sig& add_sig, const Sig& refine_sig  ) {
            Util::msg(3,"refining follow of"+ rel.to_str()+ " with "+ refine_sig.to_str_const());
            //CHK ( !sigs.empty() );
            //Sigma rem_sigs, add_sigs;
            //CHK ( sigs.size() == 1);


            //Util::msg(2,"before refine_with: follow is "+ set_to_str(follow));
            Util::msg(2,"rep_sig is " + rep_sig.to_str_const() + ", add_sig = "+add_sig.to_str_const() );

            map<Sig,Sig>& follow = get_follow_int(rel);
            CHK ( !Key_In_Bag(add_sig, follow) );
            CHK ( Key_In_Bag(rep_sig, follow) );
            Sig old_fa_sig = follow[rep_sig];
            Util::msg(2,"old_fa_sig is "+ old_fa_sig.to_str() );

            set<Sig> refine_sigs = refine_sig.negate();
            Sig neg_refine_sig = *(refine_sigs.begin());

            CHK ( old_fa_sig.matches(refine_sig) && old_fa_sig.matches(neg_refine_sig) );
            Sig new_fa_sig1 = old_fa_sig.and_with (refine_sig);
            Sig new_fa_sig2 = old_fa_sig.and_with(neg_refine_sig);

            if ( new_fa_sig1.matches(add_sig) ) { 
                follow[add_sig] = new_fa_sig1;
                follow[rep_sig] = new_fa_sig2;
            }
            else {
                CHK ( new_fa_sig1.matches(rep_sig) );
                follow[rep_sig] = new_fa_sig1;
                follow[add_sig] = new_fa_sig2;
            }

            //Util::msg(2,"after refine_with: follow is "+ set_to_str_const(follow));

        }


        /*void merge ( const Row_el& rel, const set<Sig>& sigs, Sig& sig ) {
            if ( sigs.size() == 1) {
                sig = *(sigs.begin());
            }
            else {
                Sig::union_set(sigs,sig);

                Sigma& follow = get_follow_int(rel);
                typename_const_foreach(it, sigs) {
                    follow.erase(*it);
                }
                //follow.erase(sigs.begin(),sigs.end());
                follow.insert(sig);

            }
        }*/


};

template<class Trace>
class Split_t2 {
public:
    typedef typename Trace::Prefix Row_el;
    typedef typename Row_el::Ext Row_el_ext;
    typedef typename Row_el_ext::Act Sig;
    typedef typename Trace::Suffix Col_el;
    typedef typename Col_el::Ext Col_el_ext;

    Row_el u;
    Col_el v;
    set<Sig> follow_sigs;
    Sig add_sig, refiner_sig;

    /*Split_t ( const Row_el& _u, const Col_el& _v, const set<Sig>& fsigs, const Sig& rsig) 
            : u(_u), v(_v), follow_sigs(fsigs), refiner_sig(rsig) {  }
            */
    Split_t2 ( const Row_el& _u, const Col_el& _v) : u(_u), v(_v) { }

    Split_t2 ( const Row_el& _u, const Col_el& _v, const Sig& fsig, const Sig& asig, const Sig& rsig) 
        : u(_u), v(_v), add_sig(asig), refiner_sig(rsig) { 

            follow_sigs.insert(fsig); 
            //follow_sig = fsig;
        }

};

//sigma refine learner
template <typename Teacher>
class LLearner2
{
 public:
 
	typedef typename Teacher::_DFA DFA;
	typedef typename DFA::St St;
	typedef typename DFA::Sig Sig;
	typedef typename DFA::Ev Ev;

	typedef typename Teacher::Sigma Sigma;

	typedef typename Teacher::Trace Trace;
	typedef typename Trace::Prefix Row_el;
	typedef set<Row_el> Row_els;
	typedef typename Row_el::Ext Row_el_ext;
        typedef set<Row_el_ext> Row_el_exts;
	typedef typename Trace::Suffix Col_el;
	typedef typename Col_el::Ext Col_el_ext;

	typedef string Bv;
	typedef std::map<Row_el,Bv> row_type;
	typedef typename row_type::iterator row_it;
	typedef typename vector<Col_el>::iterator col_it;
        typedef Split_t2<Trace> Split;

  protected:
  Teacher *teacher;
  DFA candidate;
  row_type rowS;
  row_type rowSA;
  vector<Col_el> cols;
  Follow2<Trace> follow;

  //map from candidate state to Row_el - needed in learn_from_ce
  map<St,Row_el> state2rowel;
  map<Trace,int> ce_history;
  

  void add_extensions_to_rowsa(const Row_el& S, const map<Sig,Sig>& exts) {

      typename_const_foreach(it, exts) {
          Row_el SA(S); SA += Row_el_ext(it->first);
          Bv bv = get_bv_by_mem_q(SA);
          if ( Key_In_Bag(SA,rowSA) ) continue;
          add_row_SA(SA,bv);
      }

  }

  /*void compact_rows() {
    set< pair<Row_el, Bv> > rowS_add, rowSA_add;
    set< Row_el > rowS_rem, rowSA_rem;
    Util::msg(3,"begin compact_rows, table is ");
    Util::msg(3,"----------------------------------------------------");
    display_table(3);

    typename_const_foreach(it, rowS) {
        const Row_el& S = it->first;
        const Sigma& exts = follow.get_follow(S);

        map< Bv, Sigma  > bv2sigs;

        typename_const_foreach(it, exts) {
            Row_el SA(S); SA += Row_el_ext(*it);
            Bv bv;
            if ( Key_In_Bag(SA,rowSA) )  {
                bv = rowSA[SA];
                bv2sigs[bv].insert( *it );
            }
            else {
                CHK ( Key_In_Bag(SA,rowS) );
              //  bv = rowS[SA];
            }
        }

        typename_const_foreach(it, bv2sigs) {
            const Sigma& sigs = it->second;
            if ( sigs.size() > 1) {
                Sig sig; follow.merge(S,sigs,sig);

                bool add_to_rowsa = true;

                typename_const_foreach(its, sigs) {
                    Row_el SA(S); SA += Row_el_ext(*its);
                    if ( Key_In_Bag(SA,rowSA) )
                        rowSA_rem.insert(SA);
                    else { rowS_rem.insert(SA); add_to_rowsa = false; }
                  //  else assert(0);
                }

                Row_el SA(S); SA += Row_el_ext(sig);
                pair<Row_el,Bv> rbv(SA,it->first);
                if ( add_to_rowsa ) rowSA_add.insert( rbv );
                else rowS_add.insert( rbv );
                //rowSA_add.insert( rbv );
            }
        }
    }

    typename_const_foreach(it, rowS_add) {
        const Row_el& rel = it->first;
        const Bv& bv = it->second;
        //remove from rem list
        if ( Key_In_Bag(rel, rowS_rem) ) { rowS_rem.erase(rel); }
        //dont add if it is already there
        if ( Key_In_Bag(rel,rowS) ) continue;
        add_row_S(rel,bv);
    }
    typename_const_foreach(it, rowS_rem) {
        rowS.erase(*it);
    }

    typename_const_foreach(it, rowSA_add) {
        const Row_el& rel = it->first;
        const Bv& bv = it->second;
        if ( Key_In_Bag(rel, rowSA_rem) ) { rowSA_rem.erase(rel); }
        if ( Key_In_Bag(rel,rowSA) ) continue;
        add_row_SA(rel,bv);
    }
    typename_const_foreach(it, rowSA_rem) {
        rowSA.erase(*it);
    }

    Util::msg(3,"end compact_rows, table is ");
    Util::msg(3,"----------------------------------------------------");
    display_table(3);


  }*/

  public:

/*********************************************************************/
//constructors and destructors
/*********************************************************************/
LLearner2() : teacher(NULL), candidate(true), follow(Sig()) {}

LLearner2(Teacher *t) : teacher(t), candidate(true), follow(t->get_basis_sig())
{
  cols.push_back(Col_el());
  add_row_by_mem_q_S(Row_el());

  Row_el S = Row_el();
  const map<Sig,Sig>& exts = follow.get_follow(S);
  add_extensions_to_rowsa(S,exts);

  compute_closure();
}

const Row_el& get_rep_el ( const Row_el& t ) const {
    if ( Key_In_Bag(t,rowS) ) {
        return t;
    }
    else if ( Key_In_Bag(t,rowSA) ) {
        Bv bv = rowSA.find(t)->second;
        typename_const_foreach(it, rowS) {
            if( it->second == bv ) 
                return it->first;
        }
        assert(0);
    }
    else assert(0);
}


const Row_el& get_rep_el ( const St& st ) const {
    typename map<St,Row_el>::const_iterator j = state2rowel.find(st);
    CHK (j != state2rowel.end());
    return get_rep_el(j->second);
    
    
}

void get_rep_elements ( const Trace &ce, size_t index, set<Row_el>& rep_els ) const {
    set<St> sts; 
    candidate.run_sym(ce,index,sts);

    typename_const_foreach(it, sts) {
        const St& st = *it;
        const Row_el& prefix = get_rep_el(st);
        rep_els.insert(prefix); //merges same rep_el for diff prefixes
    }
    CHK ( !rep_els.empty() );
}

const Sig& get_rep_A (const Row_el& rel, const Sig& faA) const {
  const map<Sig,Sig>& exts = follow.get_follow(rel);
  typename_const_foreach(it, exts) {
    if ( it->second == faA ) return it->first;
  }
  assert(0);
}


void refine_rows ( Row_el& t, const Sig& refine_sig ) {
    
    Row_els rowS_rem, rowSA_rem, rowSA_add;

    Row_el r(t); r += Row_el_ext(refine_sig);

    if ( !Key_In_Bag(r,rowSA) ) {
        add_row_by_mem_q_SA(r);
    }
}

void learn_ce_update_rows ( const Trace& ce, size_t mid , const Split& split) {
    //get prefix, ext
    Util::msg(3,"begin learn_ce_update_rows, table is ");
    Util::msg(3,"----------------------------------------------------");
    display_table(3);

    Row_el t = split.u;
    CHK ( split.follow_sigs.size() == 1 );
    const Sig& follow_sig = *(split.follow_sigs.begin());
    follow.refine_with(t, follow_sig, split.add_sig, split.refiner_sig );
    refine_rows ( t, split.add_sig ) ;

    

    Util::msg(3,"end learn_ce_update_rows, table is ");
    Util::msg(3,"----------------------------------------------------");
    display_table(3);
}

bool st_accepts_suffix ( const St& st, const Trace& ce, size_t index) const {
        const Row_el& prefix = get_rep_el(st);
        Col_el suffix; ce.get_suffix(index,suffix);
        Row_el _pr = prefix;
        Util::msg(3,"st_accept_suffix " + st.to_str() + ", split at [index]="+int2str(index)+",\n [prefix] "+_pr.to_str()+ ",\n [suffix] "+suffix.to_str());
        Util::message(2,"ce is %s\n",ce.to_str_const().c_str());
        Trace t(prefix,suffix); 
        return teacher->ask_mem_q(t);
}

int compute_alpha(const Trace &ce,const size_t index) const
{
    Util::message(3,"---> { begin compute_alpha[index=%d] \n",index);
    CHK (index <= ce.size());
    //St st;
    set<St> sts; 
    candidate.run_sym(ce,index,sts);
    CHK( sts.size() != 0);
    Util::message(3,"sts.size after run_sym = %d\n",sts.size());
    int VL = Util::VERBOSITY_LEVEL; 
    Util::VERBOSITY_LEVEL = 0;

    map<St,bool> st2acc;

    typename_foreach(it, sts) {
        const St& st = *it;
        bool mem_q_res = st_accepts_suffix (st, ce, index);
        st2acc[st] = mem_q_res;
    }

    int ret = -1;
    bool b = st2acc.begin()->second;
    ret = b?1:0;
    typename_const_foreach(it, st2acc) {
        if ( it->second != b ) {
            ret = -1;
            break;
        }
    }

    Util::VERBOSITY_LEVEL = VL;
    Util::message(3,"} end compute_alpha[index=%d], res=%d \n",index,ret);
    return ret;
}

bool has_dis_col (bool left, const Row_el& u, const Sig& repA, const Sig& ceA, Col_el& v_ ) const {
    Row_el ua1(u); ua1 += Row_el_ext(repA);
    Row_el ua2(u); ua2 += Row_el_ext(ceA);

    for(size_t i = 0; i < cols.size();++i) {
        const Col_el& v = cols[i];
        Trace t1 (ua1); t1 += v;
        bool q1 = teacher->ask_mem_q(t1);
        Trace t2 (ua2); t2 += v;
        bool q2 = teacher->ask_mem_q(t2);
        if ( q1 != q2 ) {
            v_ = v;
            return true;
        }
    }
    return false;
}

void handle_split ( bool left, const Row_el& u, const Sig& repA, const Row_el& u_, const Col_el& v, const Sig& ceA, list<Split>& splits) const {
    Util::msg(2,"handle split: u = "+ u.to_str_const() + "repA = "+repA.to_str_const()+ ", ceA = "+ ceA.to_str_const());
    Col_el v_ = v;
    if ( !cols_has(v) ) {
        if ( !has_dis_col(left, u, repA, ceA, v_) ) {
            v_ = v;
            Util::msg(2,"new v is same as old v");
        }
        else {
            Util::msg(2,"old v is "+ v.to_str_const() + "\nnew v is "+ v_.to_str_const() );
        }
    }

    Sig refiner_sig;
    Util::msg(2,"checking ceA or repA is accepted");
    Trace t1(u); t1+= Row_el_ext(ceA); t1+=v_;
    Trace t2(u); t2+= Row_el_ext(repA); t2+=v_;
    bool q1 = teacher->ask_mem_q(t1);
    bool q2 = teacher->ask_mem_q(t2);
    CHK ( q1 != q2 );
    if ( q1 ) {
        Util::msg(2,"ceA is accepted");
        refiner_sig = ceA;
    }
    else if ( q2 ) {
        Util::msg(2,"repA is accepted");
        refiner_sig = repA;
    }
    else assert(0);
    Sig dummy_faA;
    bool enl = teacher->enlarge_split(u, v_, dummy_faA, refiner_sig);
    if ( !enl ) refiner_sig = ceA;
    

    Split split(u, v_, repA, ceA, refiner_sig );
    splits.push_back(split);

}


void compute_alpha_pair( const Trace& ce, size_t index, int iloCE, int& acceptVal1, int& acceptVal2, list<Split>& splits) const {
    Util::message(3,"---> { begin compute_alpha_pair[index=%d] \n",index);
    CHK (index <= ce.size());

    bool loCE;
    if ( iloCE == 0 ) loCE = false;
    else if ( iloCE == 1 ) loCE = true;
    else assert(0);

    map<St, map<Ev,St> > emap; 
    candidate.run_sym_ext(ce,index,emap);
    CHK( emap.size() != 0 );
    Util::message(3,"emap.size after run_sym = %d\n",emap.size());

    map<St,bool> st2acc, st2acc2;
    set<St> stsi, stsii;

    typename map<St, map<Ev,St> >::const_iterator it;
    for(it=emap.begin(); it != emap.end(); ++it) {
        const St& st = it->first;
        bool mem_q_res = st_accepts_suffix (st, ce, index);
        st2acc[st] = mem_q_res;
        stsi.insert(st);
    }

    

    for(it=emap.begin(); it != emap.end(); ++it) {
        const St& st = it->first;
        const Row_el& U = get_rep_el(st);

        const map<Ev,St>& e2st = it->second;
        if ( e2st.size() != 1 ) {
            Util::message(2,"WWWWWWWWWWWW : e2st.size = %d\n",e2st.size() );

        }
        
        CHK( Key_In_Bag(st,st2acc) );
        bool st_acc = st2acc[st];

        typename_const_foreach(it2, e2st) {
            const St& opp = it2->second;
            bool opp_acc = st_accepts_suffix (opp, ce, index+1);
            st2acc2[opp] = opp_acc;
            stsii.insert(opp);

            Util::msg(2," st = "+st.to_str()+", opp = "+opp.to_str());
            Util::message(2,"st_acc = %d, opp_acc = %d", st_acc, opp_acc );

            if ( st_acc == loCE && st_acc != opp_acc ) {
                Row_el ceU, v; Row_el_ext ceA;
                ce.split(index, ceU, ceA, v );
                const Sig& ceA_sig = ceA.get_act();
                
                const Sig& faA = it2->first.get_act();
                const Row_el& u_ = get_rep_el(opp);
                const Sig& repA = get_rep_A(U,faA);

                Util::msg(2,"ceA is "+ceA_sig.to_str_const() + ", repA = "+repA.to_str_const() + "\nv is "+ v.to_str() );

                Trace t1(U); t1+= Row_el_ext(ceA_sig); t1+=v;
                Trace t2(U); t2+= Row_el_ext(repA); t2+=v;
                //check if ceA and repA actually belong to same class - then v must not be present
                Util::msg(2,"Checking if ceA and repA belong to same class");
                if ( teacher->ask_mem_q(t1)  == teacher->ask_mem_q(t2) ) {
                    // if (  (ceA_sig == repA) ) {
                    CHK ( !cols_has(v) ) ;
                    Split split(U, v );
                    splits.push_back(split);
                }
                else {
                    Util::msg(2,"No..ceA, repA go to different classes..Handle split");
                    handle_split(st_acc, U, repA, u_, v, ceA_sig, splits); 
                }
            }

        }

    }

    acceptVal1 = -1;
    typename_const_foreach(it, stsi) {
        CHK( Key_In_Bag(*it,st2acc) );
        int val = st2acc[*it]?1:0;
        if ( acceptVal1 == -1 ) { 
            acceptVal1 = val;
        }
        else if ( acceptVal1 != val ) {
            acceptVal1 = -1;
            break;
        }
    }

    
    acceptVal2 = -1;
    typename_const_foreach(it, stsii) {
        CHK( Key_In_Bag(*it,st2acc2) );
        int val = st2acc2[*it]?1:0;
        if ( acceptVal2 == -1 ) { 
            acceptVal2 = val;
        }
        else if ( acceptVal2 != val ) {
            acceptVal2 = -1;
            break;
        }
    }
    Util::message(2,"accval1=%d, accval2=%d",acceptVal1,acceptVal2);
    Util::message(3,"---> { end compute_alpha_pair[index=%d] \n",index);

}
                    //TODO: check if act split by e_sig leads to new state
                    /*
                    Sig refiner_sig = e.get_act();
                    Sigma negs = refiner_sig.negate();
                    CHK ( negs.size() == 1);
                    Sig neg_refiner_sig = *(negs.begin());

                    {
                        if ( refinee_sig.matches(refiner_sig) && refinee_sig.matches(neg_refiner_sig) )  {
                            as.insert( act );
                            Util::msg(2,"inserted in as");
                        }
                        else {
                            Util::msg(2,"not inserted in as");
                        }
                    }*/

/*********************************************************************/
//learn a counterexample
/*********************************************************************/


bool check_alpha_hi_lo ( const Trace& ce, int& loCE, int& hiCE ) {
	size_t ceLen = ce.size();
	size_t lo = 0,hi = ceLen;
	loCE = compute_alpha(ce,lo); 
        CHK ( loCE != -1 );
	Util::message(3,"----------over compute alpha lo\n");
	hiCE = compute_alpha(ce,hi);
        //CHK ( hiCE != -1 );
	Util::message(3,"----------over compute alpha hi\n");
        Util::message(3,"--------------  loCE = %d, hiCE=%d\n",loCE,hiCE);
        if ( loCE == 1 ) CHK( hiCE != -1 );
	if (loCE != hiCE) return true;
        else return false;
}

size_t find_splits(const Trace &ce, int loCE, list<Split>& splits)
{
    size_t ceLen = ce.size();
    size_t low = 0,high = ceLen;
    int acceptVal1, acceptVal2;
    /*while(low <= high) {    
        size_t mid = (low+high) / 2;
        if( mid + 1 > ceLen ) break;
        //int acceptVal1 = compute_alpha(ce,mid);
        //int acceptVal2 = compute_alpha(ce,mid + 1);
        int acceptVal1, acceptVal2;
        compute_alpha_pair(ce, mid, loCE, acceptVal1, acceptVal2, splits);
        //if(acceptVal1 != -1 && acceptVal2 != -1 && acceptVal1 != acceptVal2) {
        //if(acceptVal1 != acceptVal2) {
        if ( !splits.empty() ) {
            Trace tmp_ce = ce;
            Util::msg(2,"#### learn_ce: ce is "+ tmp_ce.to_str() );
            Util::message(3,"#### learn_ce: Split point found at mid = %d\n",mid);
            return mid;
        }
        if(loCE == acceptVal1) low = mid + 1;
        else high = mid - 1;

    }*/

    //for (uint i= low; i < high; i++) {
    for (uint i= high-1; i >= low; i--) {
        compute_alpha_pair(ce, i, loCE, acceptVal1, acceptVal2, splits);
        if ( !splits.empty() ) {
            Trace tmp_ce = ce;
            Util::msg(2,"#### learn_ce: ce is "+ tmp_ce.to_str() );
            Util::message(3,"#### learn_ce: Split point found at pos = %d\n",i);
            return i;
        }
    }
    assert (false);
}

void refine_from_splits ( const Trace& ce, size_t mid, const list<Split>& splits )  {
    
    typename_const_foreach(it, splits) {
        add_col_by_mem_q(it->v);
        if ( ! it->follow_sigs.empty() )
         learn_ce_update_rows ( ce, mid, *it );
    }

}

void learn_ce(const Trace &ce) {
    int loCE, hiCE;
    Util::message(2,"\n####### begin learnCE ...ce = %s\n",ce.to_str_const().c_str());
    if ( Key_In_Bag(ce,ce_history) ) {
        Util::message(2,"ERRORRRR !!!! : ce already learned from: prev_val = %d\n", ce_history[ce]);
        check_alpha_hi_lo(ce,loCE,hiCE);
        Util::message(2,"curr val is %d\n",loCE);
        assert(0);
    }

    size_t num_it=0;

    //Trace ce_;
    //  generalize_split(ce,mid,acceptVal1,acceptVal2,ce_);
    while ( 1 ) {
        Util::msg(2,"--------------------------------------------------");
        Util::message(2,"Checking if prev CE learnt...num_it=%d\n",num_it);
        bool hi_lo = check_alpha_hi_lo(ce,loCE,hiCE);
        if ( !hi_lo ) break;
        Util::msg(2,"----------- no");
        num_it++;
        Util::msg(2,"learn_ce: ce is "+ce.to_str_const() + ", num_it = "+int2str(num_it));
        list<Split> splits;
        size_t mid = find_splits(ce, loCE, splits);
        refine_from_splits(ce, mid, splits);
/*
        Col_el distinguish; ce.get_suffix(mid + 1,distinguish);
        add_col_by_mem_q(distinguish);
        learn_ce_update_rows ( ce, mid );
        */
        compute_closure();
        //compact_rows();
        compute_candidate();
    }
    Util::msg(2,"---------yes");

    CHK ( num_it != 0 );
    ce_history[ce] = loCE;

    Util::msg(2,"-------------- begin history check --------------");
    bool update = true;
    while ( update ) {
        update = false;
        typename_foreach(it, ce_history) {
            const Trace& old_ce = it->first;
            while ( 1 ) {
                bool hi_lo = check_alpha_hi_lo(old_ce,loCE,hiCE);
                if ( hi_lo ) {
                    Util::msg(2,"history check failed for ce = "+old_ce.to_str_const());
                    update = true;
                    list<Split> splits;
                    size_t mid = find_splits(old_ce, loCE, splits);
                    refine_from_splits(old_ce, mid, splits);
                    compute_closure();
                    //compact_rows();
                    compute_candidate();
                    num_it++;
                }
                else break;
            }

        }
    }
    Util::msg(2,"-------------- end history check --------------");
    

    display_table();
    Util::message(2,"####### end learnCE ...num_it=%d\n",num_it);

    return;


}

/*********************************************************************/
//add a row to the table
/*********************************************************************/

void add_row_S(const Row_el& rel, const Bv& bv) {
    CHK ( rowS.find(rel) == rowS.end() );
    follow.init_row_el(rel);
    rowS[rel] = bv;
}

void add_row_SA(const Row_el& rel, const Bv& bv) {
    CHK  ( rowSA.find(rel) == rowSA.end() ) ;
    /*{
        Row_el _rel = rel;
        Util::msg(2,"add_row_SA:   " + _rel.to_str() + " is already in rowSA.");
        //assert(0);
    }*/
    rowSA[rel] = bv;
}

Bv get_bv_by_mem_q(const Row_el& access)
{
	//Util::msg(1,"get_bv_by_mem_q: "+access.to_str()+"\n"); 
	Bv row_bv;
	for(size_t i = 0;i< cols.size();++i) {
		Trace member(access, cols[i]);
		bool flag = teacher->ask_mem_q(member);   
		row_bv.append(flag ? "1" : "0");
	}
        return row_bv;
	//Util::msg(1,"end get_bv_by_mem_q \n"); 
}

void add_row_by_mem_q_S (const Row_el& access) {
    //Util::msg(1,"add_row_by_mem_q_S: "+access.to_str()+"\n"); 
    Bv bv = get_bv_by_mem_q(access);
    add_row_S(access,bv); //display_table();
    //Util::msg(1,"end add_row_by_mem_q_S \n"); 
}

void add_row_by_mem_q_SA (const Row_el& access) {
    Bv bv = get_bv_by_mem_q(access);
    add_row_SA(access,bv);
}

/*********************************************************************/
//add a column to the table
/*********************************************************************/
bool cols_has ( const Col_el& distinguish ) const {
    for(size_t i = 0;i < cols.size();++i) {
        if ((cols[i] == distinguish)) {
            Util::message(2,"cols[%d] is %s\n",i,cols[i].to_str_const().c_str());
            return true;
        }
    }
    return false;
}

void add_col_by_mem_q(const Col_el &distinguish)
{
  Util::message(2,"### add_col_by_mem_q: distinguish is %s\n",distinguish.to_str_const().c_str());
  if ( cols_has(distinguish) ) {
      Util::msg(2,"add_col_by_mem_q: No state added, should add transition here\n");
      return;
  }
  cols.push_back(distinguish);  
  //add in rowS
  for(row_it it = rowS.begin();it != rowS.end();++it) {
    Trace t(it->first, distinguish);
    bool flag = teacher->ask_mem_q(t);
    it->second.append(flag? "1" : "0");
  }
  //add in rowSA
  for(row_it it = rowSA.begin();it != rowSA.end();++it) {
    Trace t(it->first,distinguish);
    bool flag = teacher->ask_mem_q(t);        
    it->second.append(flag? "1" : "0");
  }
}


/*********************************************************************/
//check if the table has changed from closed to non-closed state
/*********************************************************************/
bool check_closed_change()
{
  Util::message(3,"begin ClosedChange ...\n");
  for(row_it itR = rowSA.begin();itR != rowSA.end();++itR) {    
    const Row_el &SA = itR->first;
    const Bv &bv = itR->second;
    bool isClosed = false;
    for(row_it it1 = rowS.begin();it1 != rowS.end();++it1) {
      if((it1->second == bv)) {
        isClosed = true;
        break;
      }
    }
    //this row "bv" does not exist in rowS 
    if(!isClosed) { 
        Row_el _SA = SA;
      Util::message(3,"checkClosedChange : rowS[%s] = %s does not exist ...\n",_SA.to_str().c_str(), bv.c_str());
      //rowS[SA] = bv;
      add_row_S(SA,bv);
      
      const map<Sig,Sig>& exts = follow.get_follow(SA);
      add_extensions_to_rowsa(SA,exts);
      rowSA.erase(SA);
      return true;
    }           
  }             
  Util::message(3,"end ClosedChange ...\n");
  return false;
}

void delete_cols ( const set<uint>& del_indices ) {
    
    for(row_it it = rowS.begin();it != rowS.end();++it) {
        Bv currBv(it->second);
        Bv newBv;
        for(size_t i=0; i<currBv.size(); i++) 
            if ( !Key_In_Bag(i,del_indices) ) 
                newBv += currBv[i];
        it->second = newBv;
    }

    for(row_it it = rowSA.begin();it != rowSA.end();++it) {
        Bv currBv(it->second);
        Bv newBv;
        for(size_t i=0; i<currBv.size(); i++) 
            if ( !Key_In_Bag(i,del_indices) ) 
                newBv += currBv[i];
        it->second = newBv;
    }

    vector<Col_el> new_cols;
    for( size_t i =0; i< cols.size(); ++i ) {
        if ( !Key_In_Bag(i,del_indices) ) 
            new_cols.push_back( cols[i] );
    }
    cols = new_cols;


}

void dup_col_compress( bool compress ) {
    vector<string> col_bv (cols.size());

    for(row_it it = rowS.begin();it != rowS.end();++it) {
        Bv currBv(it->second);
        CHK ( cols.size() == currBv.size() );
        for(size_t i=0; i<currBv.size(); i++) 
            col_bv[i] += currBv[i];
    }

    set<uint> remove_col_indices;

    for( size_t i=0; i<col_bv.size(); ++i)
        for( size_t j=i+1; j<col_bv.size(); ++j)
            if( col_bv[i] == col_bv[j]) {
                Util::message(1,"***** matching cols %d and %d\n",i,j);
                remove_col_indices.insert(j);
            }

    if ( remove_col_indices.empty() ) return;
    if (compress) delete_cols(remove_col_indices);

}


/*********************************************************************/
//make the table closed and consistent
/*********************************************************************/
void compute_closure()
{
  Util::message(3,"start compute_closure ...\n");
  bool closedChanged = check_closed_change();
  while(closedChanged) closedChanged = check_closed_change();
  dup_col_compress(true);
  Util::message(3,"end compute_closure ...\n");
}


/*********************************************************************/
//compute the next candidate
/*********************************************************************/
void compute_candidate()
{
    CHK (!check_closed_change());
    //display_table();
    Util::message(2,"begin compute_candidate ...\n");
    Util::msg(3,"Curr table is ..\n"); display_table();
    const Sigma& sigma = teacher->get_sigma();
    Util::message(3,"teacher is %x\n",teacher);
    //if(candidate != NULL) delete candidate;
    //candidate = new DFA(false); //with no sink state
    candidate = DFA(true); //allow sink state
    candidate.set_sigma(sigma);
    state2rowel.clear();
    map<Bv,St> bv2state;

    //make states in DFA corr to rows
    for(row_it it = rowS.begin();it != rowS.end();++it) {
        Bv currBv(it->second);
        CHK (bv2state.find(currBv) == bv2state.end());
        //create state
        St s = candidate.new_state();
        bv2state[currBv] = s;
        Util::msg(3, currBv + " is state "+ s.to_str());
        state2rowel[s] = it->first;
        //final states
        if(currBv[0] == '1') candidate.set_final(bv2state[currBv]);
    }

    //start state
    CHK (rowS.find(Row_el()) != rowS.end());
    Bv bv_start(rowS[Row_el()]);
    candidate.set_init(bv2state[bv_start]);

    //transitions
    for(row_it itR = rowS.begin();itR != rowS.end();++itR) {
        Bv currBv(itR->second);
        CHK (bv2state.find(currBv) != bv2state.end());
        const St &state1 = bv2state[currBv];    
        const map<Sig,Sig>& exts = follow.get_follow(itR->first);
        typename_const_foreach(it, exts) {
            Row_el r(itR->first); r += Row_el_ext(it->first);              
            if(rowSA.find(r) != rowSA.end()) {
                Bv nextBv(rowSA[r]);
                CHK (bv2state.find(nextBv) != bv2state.end());
                const St &state2 = bv2state[nextBv];
                candidate.add_trans(state1,it->second,state2);
            } else if(rowS.find(r) != rowS.end()) {
                Bv nextBv(rowS[r]);
                CHK (bv2state.find(nextBv) != bv2state.end());
                const St &state2 = bv2state[nextBv];
                candidate.add_trans(state1,it->second,state2);
            }
            else Util::error(true,"ERROR : %s not found in rowS or rowSA\n",r.to_str().c_str());
        }
    }
    Util::message(2,"new candidate now has %d states\n",candidate.get_num_states()-1); //-1 for extra sink 
    Util::msg(2,candidate.to_str());
    Util::message(2,"\nend compute_candidate ...\n");
}

/*********************************************************************/
//return the new candidate DFA
/*********************************************************************/
const DFA& get_new_candidate()
{
  compute_candidate();
  return (candidate);
}

/*********************************************************************/
//return the previous candidate
/*********************************************************************/
const DFA& get_prev_candidate() const { return (candidate); }








/*********************************************************************/
//Remove duplicate rows in rowS, either by moving them to rowSA
//or completely removing them
//Note: The cache manager at teacher's end should still have the removed entries
/*********************************************************************/
void compact()
{
	map< Bv,set<Row_el> > dup_entries;
	for(typename row_type::iterator it = rowS.begin();it != rowS.end();++it) {
		dup_entries[it->second].insert(it->first);
	}		
	for(typename map< Bv,set<Row_el> >::iterator it = dup_entries.begin(); it!=dup_entries.end(); ++it) {
		if( it->second.size() == 1) continue;
		else {
			set<Row_el>& curr_entries = it->second;
			//move all dup entries except smallest from rowS to rowSA
			//find smallest trace entry
			typename set<Row_el>::iterator itE = curr_entries.begin();
			Row_el min_trace = *itE;
			while(itE != curr_entries.end()) {
				if(min_trace.size() > itE->size()) min_trace = *itE;
				++itE;
			}
			curr_entries.erase(min_trace);
			//move rest to rowSA	
			for(itE = curr_entries.begin(); itE!= curr_entries.end(); ++itE) {
				rowSA[*itE]= rowS[*itE];
				rowS.erase(*itE);
			}
		}
	}	 	
	set<Row_el> to_eliminate;
	//mark elements in rowSA which extend those in rowS
	for(row_it it = rowSA.begin();it != rowSA.end();++it) 
		to_eliminate.insert(it->first);

	const set<Sig>& sigma = teacher->get_sigma();
	for(row_it it = rowS.begin();it != rowS.end();++it) 
		for(typename set<Sig>::iterator itA = sigma.begin(); itA!= sigma.end(); ++itA) {
				Row_el SAA(it->first); SAA += Row_el_ext(*itA);
				if(rowSA.find(SAA) != rowSA.end()) to_eliminate.erase(SAA);
		}
	//remove rest from rowSA completely   
	for(typename set<Row_el>::iterator it = to_eliminate.begin(); it!= to_eliminate.end(); ++it) {
		rowSA.erase(*it);
	}
}


/*********************************************************************/
//resetting to a new teacher - TODO: note that in the general case, both
//language and alphabet may change and learner must take care of it.
/*********************************************************************/
void reset_mat(Teacher *t) 
{ 
	Util::message(2,"Begin reset_mat...");
  teacher = t; 
  //Validate the Table
  for(size_t i = 0;i< cols.size();++i) {
    const Col_el& distinguish = cols[i];
    //add in rowS
    for(row_it it = rowS.begin();it != rowS.end();++it) {
      Trace t(it->first, distinguish);
      bool flag = teacher->ask_mem_q(t);
      it->second[i]= (flag? '1' : '0');
    }
    //add in rowSA
    for(row_it it = rowSA.begin();it != rowSA.end();++it) {
      Trace t(it->first, distinguish);
      bool flag = teacher->ask_mem_q(t);        
      it->second[i]= (flag? '1' : '0');
    }      
  }	
  //compact the Table
  compact();
  Util::message(2,"End reset_mat...\n");
}

/*********************************************************************/
//display the table
/*********************************************************************/
void display_table(int prio=2)
{
  Util::message(prio,"*********************************\n");
  for(col_it it = cols.begin(); it!= cols.end();++it) {
    Util::message(prio,"   %s",it->to_str().c_str());
  }
  Util::message(prio,"\n");
  for(row_it it = rowS.begin(); it!= rowS.end();++it) {
    Row_el rel = it->first;
    Util::message(prio,"%s    :%s\n",rel.to_str().c_str(),it->second.c_str());
  }
  Util::message(prio,"\n-------------------------\n");
  for(row_it it = rowSA.begin(); it!= rowSA.end();++it) {
    Row_el rel = it->first;
    Util::message(prio,"%s    :%s\n",rel.to_str().c_str(),it->second.c_str());
  }
  Util::message(prio,"*********************************\n");
}


};

#endif

// misc functions

/*void generalize_split ( const Trace& ce, int mid, bool prev_val1, bool prev_val2, Trace& ce_ ) {
    
    Row_el access;
    Row_el_ext ext; 
    Col_el distinguish; 
    
    ce.split ( mid, access, ext, distinguish );
    Sig curr_sig = ext.get_act();
    Sig gen_sig;
    bool can_gen = true;

    while( can_gen ) {
        Sigma lifted_sigs;
        curr_sig.get_lifted(lifted_sigs,1);
        can_gen = false;
        gen_sig = curr_sig;

        typename_foreach(it, lifted_sigs) {
            Trace tmp_ce; 
            tmp_ce += access;
            tmp_ce += Row_el_ext(*it);
            tmp_ce += distinguish;
            bool acceptVal1 = compute_alpha(tmp_ce,mid);
            bool acceptVal2 = compute_alpha(tmp_ce,mid + 1);
            if(acceptVal1 == prev_val1 && acceptVal2 == prev_val2) {
                curr_sig = *it;
                can_gen = true;
                break;
            }
        }
    }

    ce_ += access;
    ce_ += Row_el_ext(gen_sig);
    ce_ += distinguish;
    Util::msg(2,"orig sig = "+ext.get_act().to_str() + " gen_sig = "+gen_sig.to_str());

}*/
