#include"expr.h"
#include"stmts.h"
#include"logic.h"
#include"eval_ea_sym.h"
#include"foci_dp.h"
#include"bdd_dp.h"

/*
struct expr2logic_base {
    Logic_type get_logic_op (op o) {
        switch (o) {
            case XorOp: return XOR;
            case OrOp: return OR;
            case AndOp: return AND;
            default : assert(0);
        }
    }

    Logic_type get_logic_uop (uop o) {
        switch (o) {
            case NegUOp: 
            case NotUOp: return NOT;
            default: assert(0);
        }
    }
};

struct expr2logic_o_vis : expr2logic_base{

    Logic_own<Var_dcl>& logic;
    Lnode result;

    //  expr_visitor(expr_visitor_mode _m, const Sys_state& _s, int _pid): m(_m), s(_s), pid(_pid), result(0) {}
    expr2logic_o_vis ( const Expr& e, Logic_own<Var_dcl>& _logic ) : logic(_logic)
        { VAR::apply_visitor(*this,e); }



    void operator() (const BinOp& bop) {
        const Expr* e1 = TUP::get<0>(bop);
        op o = TUP::get<1>(bop);
        const Expr* e2 = TUP::get<2>(bop);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;

        Lnode n1;

        apply_visitor(*this,et1);
        n1 = result;
        apply_visitor(*this,et2);

        Logic_type l = get_logic_op (o);
        result = logic.add_constraint (n1, l, result); 
    }

    void operator() (const TernOp& e)  {
        op o = TUP::get<0>(e);
        CHK ( o == IteOp );
        const Expr* e1 = TUP::get<1>(e);
        const Expr* e2 = TUP::get<2>(e);
        const Expr* e3 = TUP::get<3>(e);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;
        const Exprt& et3 = *e3;

        apply_visitor(*this,et1);
        Lnode c = result;
        apply_visitor(*this,et2);
        Lnode tn = result;
        apply_visitor(*this, et3);
        Lnode fn = result;

        //c & tn
        Lnode n1 = logic.add_constraint( c, AND, tn);
        // not (c) & fn
        Lnode n2 = logic.add_constraint(NOT, c);
        Lnode n3 = logic.add_constraint(n2, AND, fn);
        //or
        result = logic.add_constraint (n1, OR, n3);
    }

    void operator() (const UnOp& e) {
        uop o = TUP::get<0>(e);
        const Expr* e1 = TUP::get<1>(e);
        const Exprt& et1 = *e1;

        apply_visitor(*this,et1);
        Logic_type l = get_logic_uop (o);
        result = logic.add_constraint (l, result); 
    }

    void operator() (const Varref& vr) {
        Var_dcl vd = vr.to_vd_plain();
        result = logic.get_node(vd);
    }

    template<class T>
        void operator() (const T& t) { assert(0); }

};

struct expr2logic_yices_vis : expr2logic_base{

    Logic& logic;
    yices_expr result;

    expr2logic_yices_vis ( const Expr& e, Logic& _logic ) : logic(_logic)
        { VAR::apply_visitor(*this,e); }


    void operator() (const BinOp& bop) {
        const Expr* e1 = TUP::get<0>(bop);
        op o = TUP::get<1>(bop);
        const Expr* e2 = TUP::get<2>(bop);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;

        yices_expr n1;

        apply_visitor(*this,et1);
        n1 = result;
        apply_visitor(*this,et2);

        Logic_type l = get_logic_op (o);
        result = logic.add_constraint (n1, l, result); 
    }

    void operator() (const TernOp& e)  {
        op o = TUP::get<0>(e);
        CHK ( o == IteOp );
        const Expr* e1 = TUP::get<1>(e);
        const Expr* e2 = TUP::get<2>(e);
        const Expr* e3 = TUP::get<3>(e);
        const Exprt& et1 = *e1;
        const Exprt& et2 = *e2;
        const Exprt& et3 = *e3;

        apply_visitor(*this,et1);
        DP::expr c = result;
        apply_visitor(*this,et2);
        DP::expr tn = result;
        apply_visitor(*this, et3);
        DP::expr fn = result;

        result = logic.add_ite(c, tn, fn);
    }

    void operator() (const UnOp& e) {
        uop o = TUP::get<0>(e);
        const Expr* e1 = TUP::get<1>(e);
        const Exprt& et1 = *e1;

        apply_visitor(*this,et1);
        Logic_type l = get_logic_uop (o);
        result = logic.add_constraint (l, result); 
    }

    void operator() (const Varref& vr) {
        result = logic.get_node(vr);
    }

    void operator() (const Const& c) { 
        result = logic.get_node(c);
    }

    template<class T>
        void operator() (const T& t) { assert(0); }

};
*/


//explicit instantiation
template void Logic_yices<Varref,DP::expr>::build_from(const list<Assign>&);

template<class Id, class N>
void Logic_yices<Id,N>::build_from ( const list<Assign>& _alist ) {
    alist = _alist;
    const_foreach(it, alist) {
        const Assign& assign = *it;
        const Varref& vr = assign.get_vr();
        Var_dcl vd = vr.to_vd_plain();
        const Expr* e = assign.get_rhs();

        compute_add_rw(e,vd);
    }
}

template void Logic_yices<Varref,DP::expr>::get_coi_vars(Var_dcls&, Var_dcls&) const;

template<class Id, class N>
void Logic_yices<Id,N>::get_coi_vars ( Var_dcls& pre, Var_dcls& post) const {
    bool pre_changed = true;

    while ( pre_changed ) {
        const_foreach(it, alist) {
            const Assign& assign = *it;
            const Varref& vr = assign.get_vr();
            Var_dcl vda = vr.to_vd_plain();
            if ( Key_In_Bag(vda,pre) ) continue;

            const Expr* e = assign.get_rhs();
            set<Varref> vrs; e->get_support(vrs);
            foreach( itv, vrs ) {
                Var_dcl vd = itv->to_vd_plain();
                if ( Key_In_Bag(vd,pre) ) {
                    post.insert(vda);
                    break;
                }
            }
        }
        if ( !post.empty() ) {
            pre.insert(post.begin(),post.end());
            post.clear();
        }
        else pre_changed = false;
    }


}

template DP::expr Logic_yices<Varref,DP::expr>::eval_sym(DP& dp, Pid pid, int depth) const ;
template FociDP::expr Logic_yices<Varref,DP::expr>::eval_sym(FociDP& dp, Pid pid, int depth) const ;
template BDD_DP::expr Logic_yices<Varref,DP::expr>::eval_sym(BDD_DP& dp, Pid pid, int depth) const ;


template<class Id, class N>
template<class DP>
typename DP::expr Logic_yices<Id,N>::eval_sym(DP& dp, Pid pid, int depth) const {
    return eval_sym_alist(dp, alist, pid, depth);
}

