#ifndef _LOGIC_H
#define _LOGIC_H


/*#include"action.h"
#include"fa.h"
#include"path.h"
#include"state.h"
#include"sys_state.h"
#include"eval_ea.h"
#include"rw_map.h"
 */
#include "common.h"
#include "graph.h"
#include "stmts.h"
#include "dp.h"

enum Logic_type { AND, OR, NOR, XOR, NOT, EQ, INP };

struct Lnode : Node {
    Logic_type l;

    Lnode() : Node() { l = INP; }

    void set_logic ( Logic_type _l)  { l = _l; }
    Logic_type get_logic () const { return l; }

};

struct Ledge : Edge<Lnode> {
    string to_str_const() const { return " -> "; }
    string to_str() { return " -> "; }
};

class Logic_base {
    
    protected:
        RW rw;
        list<Assign> alist;
    public:
        Logic_base ( )  {} 
        void add_rw ( RW& _rw ) { rw.add(_rw); }

        const RW& get_rw() const { return rw;}
        void get_rw ( RW& _rw ) const { _rw.add(rw); } 
};

template<class Id>
class Logic_own : public Logic_base {
    protected:
        typedef Graph<Lnode, Ledge > Super;
        typedef set<Lnode> Lnodes;
        Super g;

        map<Id, Lnode> id2node;
        map<Lnode,Id> node2id;
        //map<Lnode, Logic_type> node2logic;
        Lnodes inputs;
        Lnodes outputs;



        void mark_input(const Lnode& n) { inputs.insert(n); }
        void mark_output(const Lnode& n) { outputs.insert(n); }
        void unmark_input(const Lnode& n) { CHK( Key_In_Bag(n,inputs) ); inputs.erase(n); }
        void unmark_output(const Lnode& n) { if ( Key_In_Bag(n, outputs) ) outputs.erase(n); }

        bool has_id ( const Lnode& n) const { return ( Key_In_Bag(n,node2id) ); }
        const Id& get_id (const Lnode& n) const { CHK ( has_id(n) ); return node2id[n]; }

        bool has_node ( const Id& id) const { return ( Key_In_Bag(id,id2node) ); }

        void set_node_type(Lnode& n, Logic_type l) { n.set_logic(l); }

        //edges go from outputs towards inputs, in keeping with the order of eval
        void add_constraint_bop (const Lnode& o, const Lnode&i1, const Lnode& i2) {
            g.add_edge(o, i1);
            g.add_edge(o, i2);

            unmark_input(o);
            unmark_output(i1);
            unmark_output(i2);
        }

        void add_constraint_uop (const Lnode& o, const Lnode& i) {
            g.add_edge(o, i);
            unmark_input(o);
            unmark_output(i);
        }


        Lnode get_node() {
            Lnode n = g.new_node();
            mark_input(n);
            mark_output(n);
            return n;
        }


    public:

        Logic_own()   {}


        Lnode get_node(const Id& id) {
            if ( !has_node(id) ) {
                Lnode n = get_node();
                set_id(n,id);
            }
            return id2node[id];
        }

        void set_id( const Lnode& n, const Id& str) {
            g.check_N(n) ;
            id2node[str] = n;
            node2id[n] = str;
        }

        Lnode add_constraint ( Logic_type l, const Lnode& i) {
            CHK ( l == NOT || l == EQ );
            Lnode o = get_node();
            set_node_type(o,l);
            add_constraint_uop( o, i);
            return o;
        }

        Lnode add_constraint( const Lnode&i1, Logic_type l, const Lnode& i2) {
            CHK ( l != NOT && l != EQ );
            Lnode o = get_node();
            set_node_type(o,l);
            add_constraint_bop( o, i1, i2);
            return o;
        }

        
        bool is_consistent() const {
            // check that inputs and outputs don't overlap
            const_foreach(it, inputs) {
                if ( outputs.find(*it) != outputs.end() ) return false;
            }
            return true;
        }

        template<class Val>
            Val eval_op(Logic_type l, const Val& v1, const Val& v2) {
                switch(l) {
                    case AND: return (v1 & v2);
                    case OR: return (v1 | v2);
                    case NOR: return !(v1 | v2);
                    case XOR: return (v1 ^ v2);
                    default: assert(0);
                }
            }


        template<class Val>
            Val eval_op(Logic_type l, const Val& v) {
                switch(l) {
                    case EQ: return v;
                    case NOT: return !v;
                    default: assert(0);
                }
            }

        template<class Val>
            Val eval_rec ( map<Lnode,Val>& vals, const Lnode& n)  {
                //const Id& id = get_id(n);
                if ( Key_In_Bag(n, vals) ) return vals[n];

                Logic_type l = n.get_logic();
                Val ret;
                vector<Lnode> in_vec;
                g.get_succs(n, in_vec);

                switch (l) {
                    case AND:
                    case OR:
                    case NOR:
                    case XOR:
                        {
                            CHK(in_vec.size() == 2);

                            Val v1 = eval_rec( vals, in_vec[0]);
                            Val v2 = eval_rec( vals, in_vec[1]);
                            ret = vals[n] = eval_op<Val>(l,v1,v2);
                            break;
                        }
                    case NOT:
                    case EQ:
                        CHK(in_vec.size() == 1);
                        Val v = eval_rec( vals, in_vec[0]);
                        ret = vals[n] = eval_op<Val>(l,v);
                        break;
                    case INP: {
                                  CHK ( Key_In_Bag (n, vals) );  
                                  ret = vals[n];
                              }
                              break;
                    default: assert(0);
                }
                return ret;
            }

        template<class Val>
            Val eval ( const map<Id, Val>& in_vals, map<Id,Val>& out_vals)  {
                CHK( is_consistent() );
                map<Lnode,Val> node_vals; //compute intermed node values

                typename_const_foreach( it, in_vals ) {
                    const Id& id = it->first;
                    if ( has_node(id) ) {
                        const Node& n = get_node(id);
                        node_vals[n] = it->second;
                    }
                }

                const_foreach(it, outputs) {
                    eval_rec<Val>( node_vals, *it);
                }

                typename_const_foreach( it, node_vals ) {
                    const Lnode& n = it->first;
                    if ( has_id(n) ) {
                        const Id& id = get_id(n);
                        if( !Key_In_Bag(id,in_vals) )
                            out_vals[id] = it->second;
                    }
                }
            }

};


//varref, yices_expr
template<class Id, class N>
class Logic_yices : public Logic_base {
    protected:
        //yices_context ct;
        //yices_ty bool_ty;
        DP local_dp;

        map<string, yices_expr> id2node;
        map<DP::expr,string> node2id;

        bool has_node ( const string& str) const { return ( Key_In_Bag(str,id2node) ); }

    public:
        void set_id( const N& n, const string& str) {
            id2node[str] = n;
            node2id[n] = str;
        }

        Logic_yices() {
           // ct = yices_mk_context(); 
            //bool_ty = yices_mk_ty(ct,"bool");
        }
        /*
        ~Logic_yices() {
            yices_del_context(ct);
        }*/

        
        void build_from ( const list<Assign>& _alist ) ;
        void get_coi_vars ( Var_dcls& pre, Var_dcls& post) const;
        template<class DP>
        typename DP::expr eval_sym(DP& dp, Pid pid, int depth) const; 

        template<class O>
        void fix_const_types ( const O& oracle, Pid pid ) {
            foreach(it, alist) {
                it->fix_const_types(oracle,pid);
            }
        }

        void compute_add_rw ( const Expr* e, const Var_dcl& vd) {
            RW rw;
            e->get_rw(RW::R, rw);
            rw.add_write(vd);
            add_rw(rw);
        }

        N get_node(const Id& id) {
            string str = id.to_str();
            if ( !has_node(str) ) {
                //yices_var_decl id_decl = yices_mk_var_decl ( ct, id.to_str(), bool_ty );
                //yices_expr id_expr = yices_mk_var_from_decl (ct, id_decl);
                DP::expr id_expr = local_dp.mk_bool_var(str);
                //yices_expr id_expr = yices_mk_bool_var(ct, (char*)str.c_str());
                set_id(id_expr, str);
            }
            return id2node[str];
        }
        N get_node ( const Const& c ) {
            int n = c.to_num();
            CHK( n==0 || n==1);
            return (n==0)? local_dp.mk_false() : local_dp.mk_true() ;

        }


        N add_constraint ( Logic_type l, const N& i) {
            CHK ( l == NOT || l == EQ );
            /*N o = get_node();
            set_node_type(o,l);
            add_constraint_uop( o, i);
            return o;
            */
            yices_expr o;
            if ( l == NOT )
             o = local_dp.mk_not(i);
            else if ( l == EQ ) assert(0); //NMT
            else assert(0);
            return o;
        }

        N add_constraint( const N&i1, Logic_type l, const N& i2) {
            CHK ( l != NOT && l != EQ );
            /*N o = get_node();
            set_node_type(o,l);
            add_constraint_bop( o, i1, i2);
            */
            DP::expr o;
            DP::expr inp[2];
            inp[0] = i1; inp[1] = i2;
            switch (l) {
                    case AND:
                        o = local_dp.mk_and(inp,2);
                        break;
                    case OR:
                        o = local_dp.mk_or(inp,2);
                        break;
                    case NOR: {
                        DP::expr o1 = local_dp.mk_or(inp, 2);
                        o = local_dp.mk_not(o1);
                    }
                        break;
                    case XOR: {
                                DP::expr s[2];
                                s[0] = add_constraint( local_dp.mk_not(i1) , AND, i2);
                                s[1] = add_constraint( i1 , AND, local_dp.mk_not(i2) );
                                o = local_dp.mk_or(s, 2);
                              }
                        break;
                    default: assert(0);

            }
            return o;
        }
        N add_ite( const N&c, const N& tn, const N& fn) {
            return local_dp.mk_ite(c,tn,fn);
        }

/*
        template<class T>
            void eval(const T& t) {
                yices_push(ct);
                //TODO: add constraints due to input state

                lbool sat = yices_check(ct);
                CHK ( sat == l_true) ;
                yices_model m= yices_get_model(ct);


                yices_var_decl_iterator it = yices_create_var_decl_iterator(ct);
                while (yices_iterator_has_next(it)) {
                    yices_var_decl d   = yices_iterator_next(it);
                    string val;
                    switch(yices_get_value(m, d)) {
                        case l_true: val = "true"; break;
                        case l_false: val = "false"; break;
                        case l_undef: val = "unknown"; break;
                    }
                    printf("%s = %s\n", yices_get_var_decl_name(d), val.c_str());
                }
                yices_del_iterator(it);

                yices_pop(ct);

            }*/

};



//typedef Logic_own<Var_dcl> Logic;
typedef Logic_yices<Varref,DP::expr> Logic;

//void build_logic_from_assign ( const Assign& assign,  Logic& logic ) ;
#endif
