#include"ast.h"
#include"system.h"
#include"sys_expl.h"
#include"sys_expl_sym.h"
#include"config.h"
#include"stats.h"
extern "C" {
#include<caml/callback.h>
}

Ast::Decls ast;

Config config;

bool check_ag(const eSystem& sys, const ePropFA& _spec, eSys_CE& ce) ;
bool check_ag(const sSystem& sys, const sPropFA& _spec, sSys_CE& ce) ;

void process_cmdline(int argc, char* argv[]) {
	Util::VERBOSITY_LEVEL = 3; //FIX: set it based on cmd line options
        vector<string> cmd_line;
        for(int i=1; i< argc; i++) {
            cmd_line.push_back(argv[i]);
        }
        config.init(cmd_line);
}


void process_model_files() {
/*    extern int sprm_debug;
    sprm_debug=1;
    */
    extern FILE* sprm_in;
    sprm_in = fopen(config.model_file_name.c_str(), "r");
    if ( !sprm_in ) {
        Util::err(true, "Unable to open model file "+ config.model_file_name);
    }
    int sprm_parse();
    if ( sprm_parse() ) {
        Util::err(true,"Parse Error..Exiting..");
    }
    fclose(sprm_in);

}
/*
void check_prop ( const eSystem& sys, const string& fname) {
    Util::message(1,"Checking Prop..\n");

    ePropFA prop(true); //has sink state
    prop.read_from_file( fname );
    Util::msg(1,"read prop from file..\n"+prop.to_str()+"\n");


    if ( config.cp_type == Config::CP_BASIC ) {
        Sys_explorer se(sys);

        ePropDFA dprop(true); dprop.construct_from(prop);
        dprop.negate();
        Util::msg(1,"determinized and negated prop..\n"+dprop.to_str());

        eSys_CE ce;
        bool ret = se.check_prop(dprop,ce);
        if( ret ) Util::message(1,"*** Property verified **\n");
        else {
            Util::message(1,"*** Property does not hold, CE is\n");
            Util::message(1,"%s\n",ce.to_str().c_str());
        }
    }
    else if ( config.cp_type == Config::CP_AGR ) {
        Util::message(1,"Begin AG verif..\n");
        eSys_CE ce;
        bool ret = check_ag (sys,prop,ce);

        if( ret ) Util::message(1,"*** Property verified **\n");
        else {
            Util::message(1,"*** Property does not hold, CE is\n");
            Util::msg(1,ce.to_str()+"\n");
        }
    }
    else assert(0);
}
*/
void check_prop ( const sSystem& sys, const string& fname) {
    Util::message(1,"Checking Prop..\n");

    sPropFA prop(true); //has sink state
    prop.read_from_file( fname );
    Util::msg(1,"read prop from file..\n"+prop.to_str()+"\n");


    if ( config.cp_type == Config::CP_BASIC ) {
        Sys_expl_sym se(sys);

        sPropDFA dprop(true); dprop.construct_from(prop);
        dprop.negate();
        Util::msg(1,"determinized and negated prop..\n"+dprop.to_str());

        sSys_CE ce;
        bool ret = se.check_prop(dprop,ce);
        if( ret ) Util::message(1,"*** Property verified **\n");
        else {
            Util::message(1,"*** Property does not hold, CE is\n");
            Util::message(1,"%s\n",ce.to_str().c_str());
        }
    }
    else if ( config.cp_type == Config::CP_AGR ) {
        //TODO: take care of depth-bound in mem query
        Util::message(1,"Begin AG verif..\n");
        sSys_CE ce;
        bool ret = check_ag (sys,prop,ce);

        if( ret ) Util::message(1,"*** Property verified **\n");
        else {
            Util::message(1,"*** Property does not hold, CE is\n");
            Util::msg(1,ce.to_str()+"\n");
        }
    }
    else assert(0);
}
int main(int argc, char* argv[]){
	std::set_terminate (__gnu_cxx::__verbose_terminate_handler);
        //Caml_init caml_init;
        caml_startup(argv);
	process_cmdline (argc, argv);
        string cmdline;
        for(int i=0; i< argc; i++ ) {
            cmdline += string(argv[i]) +" ";
        }
        Util::msg(1,cmdline);
	process_model_files();
	sSystem s_sys(ast); 
//	eSystem e_sys(ast); 

        Sys_expl_sym sem (s_sys);
        Expr* e = sem.get_init_glb_state();
        s_sys.set_init_state(e);

        //Pexpr::mgr.makeVerbose();
        Pexpr::mgr.AutodynEnable(CUDD_REORDER_RANDOM);
        //Pexpr::mgr.AutodynEnable(CUDD_REORDER_SYMM_SIFT);

        switch ( config.run_type ) {

            case Config::EXPLORE: 
                {
                    Sys_expl_sym se(s_sys);
                    Timer explore_timer; 
                    explore_timer.start();
                    se.explore();
                    explore_timer.stop();
                    Util::message(1, "Time to explore the state space = %f seconds\n",explore_timer.read());
                    break;
                }

            case Config::CHECK_PROP: 
                {
                    //get system init state
                    /*
                    Sys_explorer sys_expl (e_sys);
                    Sys_state s = sys_expl.get_init_X_state() ;
                    e_sys.set_init_state ( s );
                    */
                    Timer timer; timer.start();
                    //check_prop (e_sys, config.prop_file_name);
                    check_prop (s_sys, config.prop_file_name);
                    timer.stop();
                    Util::message(1, "Time to check property = %f seconds\n",timer.read());
                    break;
                }

            default: assert(0); break;
        }

        //Pexpr::mgr.info();
        Statistics::display();
}

