#ifndef _NAMING_H
#define _NAMING_H
#include"common.h"
#include"tys.h"

class Naming {
    
    protected:

    static string get_prop_state_var_name(int depth) {
        return "__PST" + get_suffix(depth); 
    }
    //get var names
    static string get_state_var_name(int pid, int depth) {
        if ( pid == -1 ) return get_prop_state_var_name(depth);
        else return get_prefix(pid) + "ST" + get_suffix(depth); 
    }


    public:

    static string get_prefix(Pid pid) {
        CHK( pid != -1 );
        return "__t"+int2str(pid) + "_";
    }
    static string get_suffix(int depth) {
        CHK( depth>= -1 );
        if ( depth == -1) return "";
        else return "_"+int2str(depth);
    }


    //declare state vars
    template<class DP>
    static void declare_state_var( DP& dp, int pid, int depth, Ty ty=IntTy) {
        string name = get_state_var_name(pid,depth);
        dp.mk_var_decl(name,ty);
    }


    //get state var exprs
    template<class DP>
    static typename DP::expr get_state_var( DP& dp, int pid, int depth) {
        string name = get_state_var_name(pid,depth);
        return dp.get_var_expr(name);
    }
    static void get_orig_name_pid_depth(const string& name, string& oname, Pid& pid, int& depth) ;

    static bool is_st_var_name ( const string& name ) ;
    static bool is_global_name ( const string& name ) ;

    //only called when pid is known for sure
    static string get_glb_name( const string& oname, Pid pid, int depth)  ;

};


#endif
