#ifndef PATH_H_
#define PATH_H_
#include"common.h"

template<class E>
class Path : public list<E> {

typedef list<E> Super;

public:
	typedef Path Prefix;
	typedef Path Suffix;
	typedef E Ext;

protected:
	void append(const Path& _p) { Path p(_p); splice(this->end(),p); }


public:
	typedef typename Super::iterator iterator;
	typedef typename Super::const_iterator const_iterator;

	Path() {}
	Path( const Super& s) : Super(s) {}

	template<class Prefix>
	Path(const Prefix& pr) {
		typename_foreach(it, pr) { append(Ext(*it)); }
	}

	template<class Prefix, class Suffix> 
	Path( const Prefix& pr, const Suffix& suff) {
		(*this) = pr;
		(*this)+= suff;
	}

	template<class Suffix>
	const Path& operator += (const Suffix& suff) {
		//FIX? assert(can_append_suffix(suff));
		typename_const_foreach(its, suff) { append(Ext(*its)); }
		return (*this);
	}
	
	const Path& operator += (const Ext& ext) {
		//FIX? assert(can_append_suffix(suff));
		append(ext);
		return (*this);
	}

        Path operator + (const Ext& ext) const {
            Path p = *this;
            p += ext;
            return p;
        }

	//size_t size() { Super::size(); }
	iterator begin() { return Super::begin(); }
	const_iterator begin() const { return Super::begin(); }
	iterator end() { return Super::end(); }
	const_iterator end() const { return Super::end(); }
	
	size_t size() const { return Super::size(); }
	bool empty() const { return Super::empty(); }

	bool operator == (const Path& p) const {
		// FIX? : do epsilon compression, trimming and then check size
		// or check online in loop
		if( size() != p.size() ) return false;
		const_iterator it = begin(), itp = p.begin();
		for(; it != end() ; ++it,++itp) { 
                    //was matches earlier
			if( ! ((*it) == (*itp)) ) {
				return false;
			}
		}
		return true;
	}
		
	bool operator < (const Path& p) const {  
		//if(size() < p.size() ) return true;
		//if(size() > p.size() ) return false;
		bool ret = ::operator < (Super(*this), Super(p));
		Util::msg(5,"Path< : "+ this->to_str_const() + ", "+ p.to_str_const() +" ret = "+ int2str(ret) + "\n");
		return ret;
	}

	bool operator != (const Path& p) const { assert(0); return false; }
	bool operator > (const Path& p) const { assert(0); return false;}
	bool operator <= (const Path& p) const { assert(0); return false;}
	bool operator >= (const Path& p) const { assert(0); return false;}

	bool is_epsilon () const { 
		if( empty() ) return true;
		//FIX?: also if all actions are epsilon
	}

	void append(const Ext& e) { push_back(e); }
	void prepend(const Ext& e) { push_front(e); }


	template<class Act>
	void project(const set<Act>& acts, Path& p) const {
		const_iterator it = begin();
		for(; it != end() ; ++it) { 
			if( it->has_act_in(acts) ) p.append(*it);
		}
	}


	template<class Prefix>
	void get_prefixes(const set<Prefix>& prefixes) const {
		Prefix pr;
		for(const_iterator it = begin(); it != end(); ++it) {
			pr.append(*it);
			prefixes.insert(pr); 
		}
	}

	template<class Suffix>
	void get_suffix(size_t pos, Suffix& suff) const {
		CHK( pos <= size() );
		size_t i=0;
		const_iterator it;
		for(it = begin(); i < pos; ++i,++it);
		for(; it != end(); ++it)
			suff.append(*it);
	}

        void get_ext(size_t pos, Ext& ext) const {
		CHK( pos <= size() );
		size_t i=0;
		const_iterator it;
		for(it = begin(); i < pos; ++i,++it);
                ext = (*it);
        }

        void split ( size_t pos, Prefix& prefix, Ext& ext, Suffix& suffix ) const {
            CHK( pos <= size() );
            size_t i=0;
            const_iterator it;
            for(it = begin(); i < pos; ++i,++it) prefix.append(*it);
            ext = (*it);
            it++;
            for(; it != end(); ++it) suffix.append(*it);
        }

	
	string to_str_const() const {
		string sep;
		if( size() > 4 ) sep = "\n";
		else sep = " ";
		string s = sep+"#"+sep;
		for(const_iterator it = begin(); it != end(); ++it ) {
			s += it->to_str_const() + sep;
		}
		s+="#"+sep;
		return s;
	}

	string to_str() {
		string sep;
		if( size() > 4 ) sep = "\n";
		else sep = " ";
		string s = sep+"#"+sep;
		for(iterator it = begin(); it != end(); ++it ) {
			s += it->to_str() + sep;
		}
		s+="#"+sep;
		return s;
	}

};


#endif
