#ifndef _PEXPR_H
#define _PEXPR_H
#include"common.h"
#include"expr.h"
#include"cuddObj.hh"

class Pexpr_bdd { //Predicate expr

    protected:
        BDD rw;


        static uint bdd_var_count;

        static map<uint,Expr*> bool2pred;
        static map<Expr*,uint> pred2bool;

        //support map
        static map< Varref, set<uint> > vr2bools; 
        //static map< Varref, BDD > vr2constraints; //constraints from underlying theory on bool vars sharing vr
        static BDD vr2constraints; //constraints from underlying theory on bool vars sharing vr
        static map<uint, BDD > shared_sup;

        //bdd 2 Expr
        static map<DdNode*,Expr*> bdd2expr_map;
        //pairs of curr n next indices
        static set< pair<uint,uint> > pair_indices;


        Expr* Pexpr_bdd::get_expr(BDD b) ;
        //Expr* Pexpr_bdd::get_expr_const(BDD b) const ;


        void add_to_rw( BDD b ) ;

        Expr* bdd2expr(BDD b) const ;
        void update_bdd2expr(BDD b);

        BDD and_restrict(BDD a, BDD b) ;
        BDD and_restrict_const(BDD a, BDD b) const ;

        BDD get_new_bdd_var() ;
        BDD get_bool_var(const Varref& vr, const Const& c) ;
        BDD make_var_bdd ( size_t i, const Varref& vr, const Const& c) ;

        void get_vecs( BDDvector& curr, BDDvector& next ) const;


        void insert_in_shared_sup(uint i, uint j) ;
        void update_support_and_add_constraints ( const Varref& vr, uint index ) ;
        const set<uint>& get_shared_indices(uint i) const ;
        BDD get_sup_closure_bdd ( BDD sup ) const ;

    public:
        typedef BDD Expr_type;
        static Cudd mgr;

        Pexpr_bdd() ;
        Pexpr_bdd(BDD _rw) ;

        BDD get_rw_bdd() const; 
        Expr* get_rw_expr() ;
        //Expr* get_rw_expr_const() const ;
        bool is_true_rw()  ;

        static Pexpr_bdd get_false() { BDD f = mgr.bddZero(); return Pexpr_bdd(f); }

        void set_r(BDD e) ;
        void set_w(BDD e) ;

        void get_rw(RW& rw) ;

        string to_str_const() const ;
        string to_str() ;
        string get_id() const ;

        bool operator < (const Pexpr_bdd& p ) const;
        bool matches (const Pexpr_bdd& p) const ;


        void add_val(size_t i, const Varref& vr, const Const& c) ;
        void add_neq_val(size_t i, const Varref& vr, const Const& c) ;

        void negate ( set<Pexpr_bdd>& rets ) ;
        void and_with (Pexpr_bdd& p, Pexpr_bdd& res) ;
        void or_with (Pexpr_bdd& p) ;

        BDD abstract_to_globals ( BDD rw, BDD rep_rw ) const; 
        void abstract_to_globals ( const Pexpr_bdd& rep, Pexpr_bdd& abs ) const ;
        void get_lifted ( set<Pexpr_bdd>& l, uint num ) const ;

        template<class DP>
            typename DP::expr get_expr_DP ( DP& dp, int depth) ;
};

class Pexpr_bdd_old { //Predicate expr

    protected:
        BDD reads;
        BDD writes;


        static Cudd mgr;
        static map<uint,Expr*> bool2pred;
        static map<Expr*,uint> pred2bool;

        //support map
        static map< Varref, set<uint> > vr2bools; 
        static map< Varref, BDD > vr2constraints; //constraints from underlying theory on bool vars sharing vr
        static map<uint, BDD > shared_sup;

        static map<DdNode*,Expr*> bdd2expr_map;

        //Expr* re;
        //Expr* we;

        Expr* get_reads_expr() const ;
        Expr* get_writes_expr() const ;


        void add_to_reads( BDD b ) ;
        void add_to_writes( BDD b ) ;

        Expr* Pexpr_bdd_old::bdd2expr(BDD b) const ;
        void update_bdd2expr(BDD b);
        BDD Pexpr_bdd_old::and_restrict(BDD a, BDD b) const ;
        BDD get_bool_var(const Varref& vr, const Const& c, int i) ;


        void insert_in_shared_sup(uint i, uint j) ;
        void update_support_and_add_constraints ( const Varref& vr, uint index ) ;
        const set<uint>& get_shared_indices(uint i) const ;
        BDD get_sup_closure_bdd ( BDD sup ) const ;

    public:
        typedef BDD Expr_type;

        Pexpr_bdd_old() ;
        Pexpr_bdd_old(BDD r, BDD w) ;

        BDD get_reads() const { return reads; }
        BDD get_writes() const { return writes; }
        bool has_true_writes() { return (writes == mgr.bddOne()); }

        void set_r(BDD e) ;
        void set_w(BDD e) ;

        void get_rw(RW& rw) const ;

        string to_str() const ;
        string get_id() const ;

        bool operator < (const Pexpr_bdd_old& p ) const { return get_id() < p.get_id(); }
        bool matches (const Pexpr_bdd_old& p) const ;


        void add_val(size_t i, const Varref& vr, const Const& c) ;

        void negate ( set<Pexpr_bdd_old>& rets ) const ;


        void and_with (Pexpr_bdd_old& p, Pexpr_bdd_old& res) ;

        BDD abstract_to_globals ( BDD rw, BDD rep_rw ) const; 

        void abstract_to_globals ( const Pexpr_bdd_old& rep, Pexpr_bdd_old& abs ) const ;

        template<class DP>
            typename DP::expr get_expr_DP ( DP& dp, int depth) const ;
};

class Pexpr_e { //Predicate expr

    protected:
        Expr* reads;
        Expr* writes;

        Expr* get_reads() const { 
            if( reads == NULL ) return ExprManager::get_true();
            else return reads;
        }
        Expr* get_writes() const { 
            if( writes == NULL ) return ExprManager::get_true();
            else return writes;
        }

        void add_to_reads( Expr* e ) {
            CHK ( e!= NULL);
            if( reads != NULL )
                reads = ExprManager::get_and(reads,e);
            else reads = e;
        }

        void add_to_writes( Expr* e ) {
            CHK ( e!= NULL);
            if( writes != NULL )
                writes = ExprManager::get_and(writes,e);
            else writes = e;
        }

    public:
        typedef Expr* Expr_type;

        Pexpr_e() { reads =  writes = ExprManager::get_true() ; }
        Pexpr_e(Expr* r, Expr* w) { CHK(r&&w); reads = r; writes = w; }

        void set_r(Expr* e) { CHK(e); reads = e; } 
        void set_w(Expr* e) { CHK(e); writes = e; } 

        void get_rw(RW& rw) const {
            get_reads() ->get_rw(RW::R,rw);
            get_writes() ->get_rw(RW::W,rw);
        }

        string to_str() const {
            string s;
            s = "(";
            s += get_reads()->to_str();
            s += ") (";
            s += get_writes()->to_str();
            s += ")";
            return s;
        }

        bool operator < (const Pexpr_e& p ) const { return to_str() < p.to_str(); }
        bool matches (const Pexpr_e& p) const ;


        void set_val(size_t i, const Varref& vr, const Const& c) {
            //setting globals
            CHK( i ==0 || i== 1);
            Expr* eq = ExprManager::get_eq(vr,c);
            if(i==0)  { add_to_reads(eq); }
            else { add_to_writes(eq); }
        } 

        void negate ( set<Pexpr_e>& rets ) const {
            Expr* r = get_reads();
            Expr* w = get_writes();
            Expr* not_r =  ExprManager::get_not( r );
            Expr* not_w =  ExprManager::get_not( w );

            Pexpr_e ret;

            // !r & w
            ret = Pexpr_e(not_r,w);
            rets.insert(ret);

            // r & !w
            ret = Pexpr_e(r,not_w);
            rets.insert(ret);

            //!r & !w
            ret = Pexpr_e(not_r,not_w);
            rets.insert(ret);
        }


    void and_with (const Pexpr_e& p, Pexpr_e& res) const {
        Expr *and_r, *and_w ;
        
        //if ( reads == NULL ) and_r = p.reads;
        //else 
        and_r = ExprManager::get_and( get_reads(), p.get_reads() );

        //if ( writes == NULL ) and_r = p.writes;
        //else 
        and_w = ExprManager::get_and( get_writes(), p.get_writes() );

        res = Pexpr_e(and_r,and_w);
    }

    Expr* abstract_to_globals ( Expr* rw, Expr* rep_rw ) const {
        CHK ( rw != NULL && rep_rw != NULL);
        return rw->project_to(rep_rw);
    }

    void abstract_to_globals ( const Pexpr_e& rep, Pexpr_e& abs ) const {
        abs.reads = abstract_to_globals ( reads, rep.reads );
        abs.writes = abstract_to_globals ( writes, rep.writes );
    }


    template<class DP>
    typename DP::expr get_expr_DP ( DP& dp, int depth) const ;

};


typedef Pexpr_bdd Pexpr;

#endif
