#include<stdio.h>
#include"pexpr.h"
#include"dp.h"
#include"foci_dp.h"
#include"bdd_dp.h"
#include"eval_ea_sym.h"

#include"cuddInt.h"
#include"util.h"
#include"naming.h"

//TODO: remove instances of get_rw_expr_const

Cudd Pexpr_bdd::mgr(100,0);
uint Pexpr_bdd::bdd_var_count = 0;
map<uint,Expr*> Pexpr_bdd::bool2pred; //maps bool var indices to corr predicates
map<Expr*,uint> Pexpr_bdd::pred2bool; //reverse of above
map< Varref, set<uint> > Pexpr_bdd::vr2bools; // maps vr to bool var indices i, where vr \in conc(i)
//map< Varref, BDD > Pexpr_bdd::vr2constraints; //constraints from underlying theory on bool vars sharing vr
BDD Pexpr_bdd::vr2constraints = mgr.bddOne(); //constraints from underlying theory on bool vars sharing vr
map<uint, BDD > Pexpr_bdd::shared_sup; //maps bool var indices to conjunct BDD of index vars that share concrete varrefs
set< pair<uint,uint> > Pexpr_bdd::pair_indices;
map<DdNode*, Expr* > Pexpr_bdd::bdd2expr_map; //maps each bdd computed to its expr; useful for caching



static	DdNode	*background, *zero;
 // ---------------- bdd2expr -----------------

static Expr*
my_ddPrintMintermAux(
  DdManager * dd /* manager */,
  DdNode * node /* current node */,
 // int * list /* current recursion path */,
  //map<uint,bool>& varmap,
  const map<uint,Expr*>& b2p,
  map<DdNode*,Expr*>& node2expr)
{
    DdNode	*N,*Nv,*Nnv;
    Expr * res = ExprManager::get_false();

    N = Cudd_Regular(node);
    if ( node2expr.find(node) != node2expr.end() ) {
      //  Util::msg(2,"found in node2expr");
        return node2expr[node];
    }
    //else Util::msg(2,"not found in node2expr");

    if (cuddIsConstant(N)) {
	/* Terminal case: Print one cube based on the current recursion
	** path, unless we have reached the background value (ADDs) or
	** the logical zero (BDDs).
	*/
	if (node != background && node != zero) {
            res = ExprManager::get_true();
	}
        //else res = ExprManager::get_false();
        else res = NULL;
    } else {
	Nv  = cuddT(N);
	Nnv = cuddE(N);
	if (Cudd_IsComplement(node)) {
	    Nv  = Cudd_Not(Nv);
	    Nnv = Cudd_Not(Nnv);
	}
	uint index = N->index;
        CHK ( Key_In_Bag(index,b2p) ) ;
        Expr * var = b2p.find(index)->second;

        //var = 0
        Expr * e = my_ddPrintMintermAux(dd,Nnv,b2p,node2expr); 
        if ( e != NULL ) {
            Expr *var_ = ExprManager::get_not(var);
            e = ExprManager::get_and(var_,e);
            res = ExprManager::get_or(res,e);
        }

        //var = 1
            e = my_ddPrintMintermAux(dd,Nv,b2p,node2expr);
        if ( e != NULL ) {
            e = ExprManager::get_and(e,var);
            res = ExprManager::get_or(res,e);
        }
    }

    node2expr[node]  = res;
    if ( res == ExprManager::get_false() || res == NULL ) return NULL;
    return res ;

} /* end of ddPrintMintermAux */

Expr *
my_Cudd_PrintMinterm(
  DdManager * manager,
  DdNode * node,
  const map<uint,Expr*>& b2p)
  //map<DdNode*,Expr*>& node2expr)
{

    background = manager->background;
    zero = Cudd_Not(manager->one);
    //map<uint,bool> varmap;
    map<DdNode*,Expr*> node2expr;
    Expr* res =  my_ddPrintMintermAux(manager,node, b2p,node2expr);
    if ( res == NULL ) res = ExprManager::get_false();
    return res;

} /* end of Cudd_PrintMinterm */
//void my_Cudd_PrintMinterm( DdManager * manager, DdNode * node, const map<uint,Expr*>& b2p, set<Expr*>& res_list);

Expr* Pexpr_bdd::bdd2expr(BDD b) const {
    //set<Expr*> res_list;
    //Timer tim; tim.start();
    Util::msg(2,"bdd2expr : bdd is ");
    //b.PrintMinterm();
    //CHK ( !bool2pred.empty() );
    Expr * res = my_Cudd_PrintMinterm(mgr.getManager(),b.getNode(),bool2pred);

    /*Expr *res = ExprManager::get_false();
    foreach(it, res_list) {
        res = ExprManager::get_or(res,*it);
    }*/
    //tim.stop();
    //Util::msg(2,"expr is "+res->to_str());
    //Util::message(2,"bdd2expr: time taken is %f\n",tim.read());
    Util::message(2,"bdd2expr over");
    assert( res );
    return res;
}


//--------------- end bdd2expr

void Pexpr_bdd::update_bdd2expr(BDD b) {
    DdNode *node = b.getNode();
    CHK(node != NULL);
    if ( !Key_In_Bag(node,bdd2expr_map) ) {
        bdd2expr_map[node] = bdd2expr(b);
    }
}

Expr* Pexpr_bdd::get_expr(BDD b) {
    DdNode* node = b.getNode();
    //Util::msg(2,"get_expr :bdd is "); b.PrintMinterm();
    if ( !Key_In_Bag(node,bdd2expr_map) )  update_bdd2expr(b); 
    //Util::msg(2,"get_expr : over "); 
    return bdd2expr_map.find(node)->second;
}

/*Expr* Pexpr_bdd::get_expr_const(BDD b) const {
    DdNode* node = b.getNode();
    //Util::msg(2,"get_expr :bdd is "); b.PrintMinterm();
    if ( Key_In_Bag(node,bdd2expr_map) )
        return bdd2expr_map.find(node)->second;
    else return bdd2expr(b); 
} */

Pexpr_bdd::Pexpr_bdd() { rw = mgr.bddOne() ; } //update_bdd2expr(rw); }
Pexpr_bdd::Pexpr_bdd(BDD _rw) : rw(_rw) { }//update_bdd2expr(rw); }


BDD Pexpr_bdd::get_rw_bdd() const { return rw; }
bool Pexpr_bdd::is_true_rw() { return (rw == mgr.bddOne()); }
Expr* Pexpr_bdd::get_rw_expr() { return get_expr(rw); }
//Expr* Pexpr_bdd::get_rw_expr_const() const { return get_expr_const(rw); }


void Pexpr_bdd::get_rw(RW& rw_val) {
    Expr * e = get_expr(rw);
    e->get_rw_both(rw_val);
}

string Pexpr_bdd::to_str_const() const {
    //return get_id();
    Pexpr_bdd b = *this;
    return b.to_str();
}

string Pexpr_bdd::to_str() {
    string s;
    s = "(";
   // s += get_id();
    s += get_rw_expr()->to_str();
    s += ")";
    return s;
}

string Pexpr_bdd::get_id() const {
    string s;
    char a[100];
    s = "(";
    sprintf(a,"%p",get_rw_bdd().getNode());
    s += a;
    s += ")";
    //Util::msg(2,"pexpr_bdd::get_id = "+ s);
    
    return s;
}

bool Pexpr_bdd::operator < (const Pexpr_bdd& p ) const { return get_id() < p.get_id(); }

bool Pexpr_bdd::matches (const Pexpr_bdd& p) const {
    bool ret;
  
    Pexpr_bdd me = *this;
    Pexpr_bdd _p = p;

   // Util::msg(2,"me is "); me.rw.PrintMinterm();
    //Util::msg(2,"p is "); _p.rw.PrintMinterm();

    /*BDD res = and_restrict_const(me.rw,_p.rw);
    if ( res == mgr.bddZero() ) ret = false;
    else ret = true;
    */

    //Util::msg(2,"after BDD and restrict, me is "); me.rw.PrintMinterm();
    //Util::msg(2,"p is "); _p.rw.PrintMinterm();
    //Util::msg(2,"res is "); res.PrintMinterm();

    ///*
    Expr* e = me.get_rw_expr();
    Expr* pe = _p.get_rw_expr();

    DP dp;
    set<Varref> sup;
    e->get_support(sup); pe->get_support(sup);
    DP_declare_varrefs_ext(dp,sup,-1,0);

    DP::expr e_expr = eval_sym_expr_pair(dp,e,-1,0);
    DP::expr pe_expr = eval_sym_expr_pair(dp,pe,-1,0);
    DP::expr c = dp.mk_and(e_expr, pe_expr);

    dp.assertdp(c);
   /* if ( dp.check() != l_false ) { CHK ( ret ); ret = true;}
    else { CHK( !ret); ret = false; }
    */
    if ( dp.check() != l_false ) { ret = true;}
    else { ret = false; }
    //*/
    

    /*if ( res == mgr.bddZero() ) CHK( ret == false );
    else  CHK ( ret == true );
    */
    
    return ret;
    

}

template DP::expr Pexpr_bdd::get_expr_DP ( DP& dp, int depth )  ;
template FociDP::expr Pexpr_bdd::get_expr_DP ( FociDP& dp, int depth )  ;
template BDD_DP::expr Pexpr_bdd::get_expr_DP ( BDD_DP& dp, int depth )  ;


template<class DP>
typename DP::expr Pexpr_bdd::get_expr_DP ( DP& dp, int depth ) {
    typedef typename DP::expr expr;
    const Expr* rw_expr = get_rw_expr();
    expr c_expr = eval_sym_expr_pair(dp,rw_expr,-1,depth);
    return c_expr;
}

// ----  add_val, update support, add constraints  -------------

void Pexpr_bdd::insert_in_shared_sup(uint i, uint j) {
    BDD bj = mgr.bddVar(j);
    if( !Key_In_Bag(i,shared_sup) ) 
        shared_sup[i] = bj;
    else shared_sup[i] = shared_sup[i] & bj;

}


void Pexpr_bdd::update_support_and_add_constraints (const Varref& vr, uint index ) {
    BDD curr_diff = mgr.bddZero();
    //Util::msg(2,"adding for var "+vr.to_str() );
    if ( !Key_In_Bag(vr,vr2bools)) {
            //vr2constraints[vr] = curr_diff;
    }
    else {

        set<uint>& indices = vr2bools[vr];
        //relate all previous indices *it with index
        const_foreach(it, indices) {
            uint i = *it;
            //update support sharing info
            insert_in_shared_sup(index,i);
            insert_in_shared_sup(i,index);

            //add pred constraints
            BDD bdd_i = mgr.bddVar(i);
            curr_diff = curr_diff | bdd_i;
            //BDD bdd_index = mgr.bddVar(index);
            //BDD diff = bdd_i.Xor(bdd_index); // !( bdd_i /\ bdd_index)
            //curr_diff = curr_diff & diff;
            //CHK ( curr_diff != mgr.bddZero() );
        }

        BDD bdd_index = mgr.bddVar(index);
        curr_diff = curr_diff & bdd_index;
        curr_diff = !curr_diff;

        //Util::msg(2,"vr2constraints is "); vr2constraints.PrintMinterm();
        //Util::msg(2,"curr_diff is "); curr_diff.PrintMinterm();
        vr2constraints = vr2constraints & curr_diff;
        CHK ( vr2constraints != mgr.bddZero() );
    }
    vr2bools[vr].insert(index);
}


BDD Pexpr_bdd::get_new_bdd_var() {
    return mgr.bddVar(bdd_var_count++);
    //return mgr.bddVar();
}

BDD Pexpr_bdd::get_bool_var( const Varref& vr, const Const& c ) {
    //add suffix to vr depending on i
    Expr* eq = ExprManager::get_eq(vr,c);
    //Util::msg(2,"\nbegin get bool var for Expr "+eq->to_str());
    uint index;
    if( pred2bool.find(eq) == pred2bool.end() ) {
        BDD b = get_new_bdd_var();
        index = b.NodeReadIndex();
        bool2pred[index] = eq;
        pred2bool[eq] = index;
        update_support_and_add_constraints(vr,index);
     //   Util::msg(2,"Expr "+eq->to_str()+" gets index "+int2str(index));
    }
    else {
        index  = pred2bool[eq];
     //   Util::msg(2,"Expr "+eq->to_str()+" has index "+int2str(index));
    }
    return mgr.bddVar(index);
}


BDD Pexpr_bdd::make_var_bdd ( size_t i, const Varref& vr, const Const& c) {
    string s = vr.to_str() + Naming::get_suffix(i);
    Varref vr_(s);  
    BDD v_bdd = get_bool_var(vr_,c);
    //update_bdd2expr(v_bdd);
    return v_bdd;
}


void Pexpr_bdd::add_val(size_t i, const Varref& vr, const Const& c) {
    CHK( i ==0 || i== 1);
    BDD v_bdd = make_var_bdd(i,vr,c);
    add_to_rw(v_bdd);
    BDD ov_bdd = make_var_bdd(1-i,vr,c);

    uint index0, index1;
    if ( i==0 ) {
        index0 = v_bdd.NodeReadIndex();
        index1 = ov_bdd.NodeReadIndex();
    }
    else {
        index1 = v_bdd.NodeReadIndex();
        index0 = ov_bdd.NodeReadIndex();
    }
    pair_indices.insert( pair<uint,uint>(index0,index1) );
} 

void Pexpr_bdd::add_neq_val(size_t i, const Varref& vr, const Const& c) {
    CHK( i ==0 || i== 1);
    BDD v_bdd = make_var_bdd(i,vr,c);
    add_to_rw(!v_bdd);
    BDD ov_bdd = make_var_bdd(1-i,vr,c);

    uint index0, index1;
    if ( i==0 ) {
        index0 = v_bdd.NodeReadIndex();
        index1 = ov_bdd.NodeReadIndex();
    }
    else {
        index1 = v_bdd.NodeReadIndex();
        index0 = ov_bdd.NodeReadIndex();
    }
    pair_indices.insert( pair<uint,uint>(index0,index1) );
} 
//----- over add_val

BDD Pexpr_bdd::and_restrict(BDD a, BDD b) { 
    /*BDD ct = mgr.bddOne();
    const_foreach(it, vr2constraints ) {
        ct = ct & it->second;
    }*/
    BDD ct = vr2constraints;
    BDD res =  a & b;
    //Util::msg(2,"before restrict : bdd is "); res.PrintMinterm();
    //Util::msg(2,"restrict bdd is "); ct.PrintMinterm();
    BDD res2 = res.Restrict(ct);
    //update_bdd2expr(res);
    //Util::msg(2,"after restrict : bdd is "); res.PrintMinterm();
    return res2;
}
BDD Pexpr_bdd::and_restrict_const(BDD a, BDD b) const { 
    BDD ct = vr2constraints;
    BDD res =  a & b;
    BDD r2 = res.Restrict(ct);
    return r2;
}

void Pexpr_bdd::add_to_rw( BDD b ) {
    rw = and_restrict(rw, b);
}

void Pexpr_bdd::get_vecs( BDDvector& curr, BDDvector& next ) const {
    int N = pair_indices.size();
    //curr = BDDvector(N); // crashes cudd
    //next = BDDvector(N);
    int i=0;

    const_foreach(it, pair_indices) {
        //Util::message(2,"indices are %d,%d\n",it->first,it->second);
        curr[i] = mgr.bddVar(it->first);
        next[i] = mgr.bddVar(it->second);
        i++;
        CHK (i<=N);
    }

}

void Pexpr_bdd::set_r(BDD e) {
    //Util::msg(2,"pexpr_bdd:set_r");
    //CHK ( is_true_rw() );
    rw = e;
    //Util::msg(2,"end pexpr_bdd:set_r");
}

void Pexpr_bdd::set_w(BDD e) {
    //Util::msg(2,"pexpr_bdd:set_w");
    BDDvector v0_vec(pair_indices.size()), v1_vec(pair_indices.size());
    get_vecs(v0_vec,v1_vec);
    //Util::msg(2,"1pexpr_bdd:set_w");

    //Util::msg(2,"set_w : e = "); e.PrintMinterm();
    BDD e_ = e.SwapVariables(v0_vec,v1_vec);
    //Util::msg(2,"2pexpr_bdd:set_w");
    //Util::msg(2,"set_w : e_ = "); e_.PrintMinterm();
    add_to_rw(e_);
    //Util::msg(2,"end pexpr_bdd:set_w");
}

// ----------- boolean negate, and

void Pexpr_bdd::negate ( set<Pexpr_bdd>& rets ) {
    BDD not_rw = !rw;
    //update_bdd2expr(not_rw);
    rets.insert(not_rw);
}

void Pexpr_bdd::and_with (Pexpr_bdd& p, Pexpr_bdd& res) {
    BDD and_rw = and_restrict(rw, p.rw) ;
    res = Pexpr_bdd(and_rw);
}

//performs or_restrict
void Pexpr_bdd::or_with (Pexpr_bdd& p) {
    rw |= p.rw;
    rw.Restrict(vr2constraints);
    //update_bdd2expr(rw);
}

//----- abstract_to_globals

BDD Pexpr_bdd::get_sup_closure_bdd ( BDD sup ) const {
    BDD tmp_sup = sup;
    //TODO: maybe can extract indices from sup and then check
    //or maybe have a relation and directly compute closure
    const_foreach(it, shared_sup) {
        BDD b = mgr.bddVar(it->first);
        if ( (tmp_sup & !b) == mgr.bddZero() ) { //b's var is in tmp_sup
            tmp_sup = tmp_sup & it->second;
        }
    }
    return tmp_sup;
}

BDD Pexpr_bdd::abstract_to_globals ( BDD rw, BDD rep_rw ) const {
    //return rw->project_to(rep_rw);
    BDD common, only_me, only_rep;
    BDD sup = rep_rw.Support();
    BDD all_sup = get_sup_closure_bdd(sup);
    
    rw.ClassifySupport(all_sup,&common,&only_me,&only_rep);
    return rw.ExistAbstract(only_me);
}

void Pexpr_bdd::abstract_to_globals ( const Pexpr_bdd& rep, Pexpr_bdd& abs ) const {
    abs.rw = abstract_to_globals ( rw,  rep.rw );
   // Pexpr_bdd me = *this;
   // Pexpr_bdd _rep = rep;
    //Util::msg(1,"abstract to globals: me is "+me.to_str()+ ", rep is "+_rep.to_str() + ", abs is "+ abs.to_str());
}

void Pexpr_bdd::get_lifted ( set<Pexpr_bdd>& l, uint num ) const {
    BDD sup = rw.Support();
    BDD tmp_sup = sup;
    set<uint> sup_indices;
    rw.PrintMinterm();

    for( uint i=0; i < bdd_var_count; i++) {
        BDD b = mgr.bddVar(i);
        tmp_sup = sup & !b;
        if ( tmp_sup == mgr.bddZero() ) { //b's var is in tmp_sup
            sup_indices.insert(i);
        }
    }

    foreach(it, sup_indices) {
        BDD b = mgr.bddVar(*it);

        BDD abst = rw.ExistAbstract(b);
        l.insert(abst);
        abst.PrintMinterm();
    }


}
