#include<stdio.h>
#include"pexpr.h"
#include"dp.h"
#include"eval_ea_sym.h"

#include"cuddInt.h"
#include"util.h"
//#include"cudd.h"

Cudd Pexpr_bdd_old::mgr;
map<uint,Expr*> Pexpr_bdd_old::bool2pred; //maps bool var indices to corr predicates
map<Expr*,uint> Pexpr_bdd_old::pred2bool; //reverse of above
//map<uint, set<Varref> > Pexpr_bdd_old::bool2sup; //maps bool var indices to var they rep
map< Varref, set<uint> > Pexpr_bdd_old::vr2bools; // maps vr to bool var indices i, where vr \in conc(i)
map< Varref, BDD > Pexpr_bdd_old::vr2constraints; //constraints from underlying theory on bool vars sharing vr
map<uint, BDD > Pexpr_bdd_old::shared_sup; //maps bool var indices to conjunct BDD of index vars that share concrete varrefs
map<DdNode*, Expr* > Pexpr_bdd_old::bdd2expr_map; //maps each bdd computed to its expr; useful for caching

static	DdNode	*background, *zero;

static void
my_ddPrintMintermAux(
  DdManager * dd /* manager */,
  DdNode * node /* current node */,
 // int * list /* current recursion path */,
  map<uint,bool>& varmap,
  const map<uint,Expr*>& b2p,
  set<Expr*>& res_list)
{
    DdNode	*N,*Nv,*Nnv;

    N = Cudd_Regular(node);

    if (cuddIsConstant(N)) {
	/* Terminal case: Print one cube based on the current recursion
	** path, unless we have reached the background value (ADDs) or
	** the logical zero (BDDs).
	*/
        Expr* te = ExprManager::get_true();
        Expr* res = te;
	if (node != background && node != zero) {
            const_foreach(it, varmap) {
                uint i = it->first;
                CHK ( Key_In_Bag(i,b2p) ) ;
                Expr* e = b2p.find(i)->second;

                bool v = it->second;
		if (!v) {
                    e = ExprManager::get_not(e);
                    res = ExprManager::get_and(res,e);
                }
		else if (v) {
                    res = ExprManager::get_and(res,e);
                }
		else {} 

            }
            res_list.insert(res);
	}
    } else {
	Nv  = cuddT(N);
	Nnv = cuddE(N);
	if (Cudd_IsComplement(node)) {
	    Nv  = Cudd_Not(Nv);
	    Nnv = Cudd_Not(Nnv);
	}
	uint index = N->index;
//	list[index] = 0;
        varmap[index] = false;
	my_ddPrintMintermAux(dd,Nnv,varmap,b2p,res_list); 
//	list[index] = 1;
        varmap[index] = true;
	my_ddPrintMintermAux(dd,Nv,varmap,b2p,res_list);
//	list[index] = 2;
        varmap.erase(index);
    }
    return ;

} /* end of ddPrintMintermAux */

void
my_Cudd_PrintMinterm(
  DdManager * manager,
  DdNode * node,
  const map<uint,Expr*>& b2p,
  set<Expr*>& res_list)
{

    background = manager->background;
    zero = Cudd_Not(manager->one);
    map<uint,bool> varmap;
    my_ddPrintMintermAux(manager,node,varmap, b2p,res_list);

} /* end of Cudd_PrintMinterm */

BDD Pexpr_bdd_old::and_restrict(BDD a, BDD b) const { 
    BDD ct = mgr.bddOne();
    const_foreach(it, vr2constraints ) {
        ct = ct & it->second;
    }
    BDD res =  a & b;
    //Util::msg(2,"before restrict : bdd is "); res.PrintMinterm();
    //Util::msg(2,"restrict bdd is "); ct.PrintMinterm();
    res = res.Restrict(ct);
    //Util::msg(2,"after restrict : bdd is "); res.PrintMinterm();
    return res;
}

void Pexpr_bdd_old::add_to_reads( BDD b ) {
    reads = and_restrict(reads, b);
    update_bdd2expr(reads);
}

void Pexpr_bdd_old::add_to_writes( BDD b ) {
    writes = and_restrict(writes, b);
    update_bdd2expr(writes);
}


Expr* Pexpr_bdd_old::bdd2expr(BDD b) const {
    set<Expr*> res_list;
    Timer tim; tim.start();
    //Util::msg(2,"bdd2expr : bdd is ");
    //b.PrintMinterm();
    //CHK ( !bool2pred.empty() );
    my_Cudd_PrintMinterm(mgr.getManager(),b.getNode(),bool2pred,res_list);

    Expr *res = ExprManager::get_false();
    foreach(it, res_list) {
        res = ExprManager::get_or(res,*it);
    }
    tim.stop();
    //Util::msg(2,"expr is "+res->to_str());
    //Util::message(2,"bdd2expr: time taken is %f\n",tim.read());
    assert( res );
    return res;
}

Expr* Pexpr_bdd_old::get_reads_expr() const{ 
    //return bdd2expr(reads);

    DdNode *node = reads.getNode();
    CHK ( Key_In_Bag(node,bdd2expr_map) );
    return bdd2expr_map.find(node)->second;
}
Expr* Pexpr_bdd_old::get_writes_expr() const { 
    //return bdd2expr(writes);
    DdNode *node = writes.getNode();
    CHK ( Key_In_Bag(node,bdd2expr_map) );
    return bdd2expr_map.find(node)->second;
}

void Pexpr_bdd_old::update_bdd2expr(BDD b) {
    DdNode *node = b.getNode();
    if ( !Key_In_Bag(node,bdd2expr_map) ) {
        bdd2expr_map[node] = bdd2expr(b);
        //Util::msg(2,"update_bdd2expr: added");
    }
        //Util::msg(2,"update_bdd2expr: b is "); b.PrintMinterm();
        //Util::msg(2,"expr is "+bdd2expr_map[node]->to_str());
}

Pexpr_bdd_old::Pexpr_bdd_old() { reads =  writes = mgr.bddOne() ; update_bdd2expr(reads); }
Pexpr_bdd_old::Pexpr_bdd_old(BDD r, BDD w) { reads = r; writes = w; update_bdd2expr(reads); update_bdd2expr(writes);}

void Pexpr_bdd_old::set_r(BDD e) {  reads = e; update_bdd2expr(reads); } 
void Pexpr_bdd_old::set_w(BDD e) {  writes = e; update_bdd2expr(writes); } 


void Pexpr_bdd_old::get_rw(RW& rw) const {
    get_reads_expr()->get_rw(RW::R,rw);
    get_writes_expr()->get_rw(RW::W,rw);
}

string Pexpr_bdd_old::to_str() const {
    string s;
    s = "(";
    s += get_reads_expr()->to_str();
    s += ") (";
    s += get_writes_expr()->to_str();
    s += ")";
    return s;
}

string Pexpr_bdd_old::get_id() const {
    string s;
    char a[100];
    s = "(";
    sprintf(a,"%p",get_reads().getNode());
    s += a;
    s += ") (";
    sprintf(a,"%p",get_writes().getNode());
    s += a; 
    s += ")";
    //Util::msg(2,"pexpr_bdd::get_id = "+ s);
    
    return s;
}

void Pexpr_bdd_old::insert_in_shared_sup(uint i, uint j) {
    BDD bj = mgr.bddVar(j);
    if( !Key_In_Bag(i,shared_sup) ) 
        shared_sup[i] = bj;
    else shared_sup[i] = shared_sup[i] & bj;

}


void Pexpr_bdd_old::update_support_and_add_constraints ( const Varref& vr, uint index ) {
    BDD curr_diff = mgr.bddOne();
    if ( !Key_In_Bag(vr,vr2bools)) {
            vr2constraints[vr] = curr_diff;
    }
    else {

        set<uint>& indices = vr2bools[vr];
        //relate all previous indices *it with index
        const_foreach(it, indices) {
            uint i = *it;
            //update support sharing info
            insert_in_shared_sup(index,i);
            insert_in_shared_sup(i,index);

            //add pred constraints
            BDD bdd_i = mgr.bddVar(i);
            BDD bdd_index = mgr.bddVar(index);
            BDD diff = bdd_i.Xor(bdd_index); // !( bdd_i /\ bdd_index)
            curr_diff = curr_diff & diff;
        }

        CHK ( Key_In_Bag(vr,vr2constraints) ) ;
        vr2constraints[vr] = vr2constraints[vr] & curr_diff;
    }
    vr2bools[vr].insert(index);
}

/*
//Expr* is of particular form v == c, a predicate
const Varref& get_varref_from_eq(Expr* e) {
    BinOp& ebop = VAR::get<BinOp>(*e);
    op o = TUP::get<1>(ebop);
    CHK ( o == EqualOp ); 
    Expr* vre = (TUP::get<0>(ebop));
    return VAR::get<Varref>(*vre);
}

const Const& get_const_from_eq(Expr* e) {
    BinOp& ebop = VAR::get<BinOp>(*e);
    op o = TUP::get<1>(ebop);
    CHK ( o == EqualOp ); 
    Expr* ce = TUP::get<2>(ebop);
    return VAR::get<Const>(*ce);
}

const Varref& Pexpr_bdd_old::get_varref_for_index(uint i) const {
    CHK ( Key_In_Bag(i,bool2pred) );
    Expr* e = bool2pred.find(i)->second;
    return get_varref_from_eq(e);
}
const Const& Pexpr_bdd_old::get_const_for_index(uint i) const {
    CHK ( Key_In_Bag(i,bool2pred) );
    Expr* e = bool2pred.find(i)->second;
    return get_const_from_eq(e);
}

const set<uint>& Pexpr_bdd_old::get_shared_indices(uint i) const {
    const Varref& vr = get_varref_for_index(i);
    CHK( Key_In_Bag(vr,vr2bools));
    return vr2bools.find(vr)->second;
}
*/




BDD Pexpr_bdd_old::get_bool_var(const Varref& vr, const Const& c, int i) {
    //add suffix to vr depending on i
    Expr* eq = ExprManager::get_eq(vr,c);
    uint index;
    if( pred2bool.find(eq) == pred2bool.end() ) {
        BDD b = mgr.bddVar();
        index = b.NodeReadIndex();
        bool2pred[index] = eq;
        pred2bool[eq] = index;
        update_support_and_add_constraints(vr,index);
        Util::msg(2,"Expr "+eq->to_str()+" gets index "+int2str(index));
    }
    else {
        index  = pred2bool[eq];
     //   Util::msg(2,"Expr "+eq->to_str()+" has index "+int2str(index));
    }
    return mgr.bddVar(index);
}

void Pexpr_bdd_old::add_val(size_t i, const Varref& vr, const Const& c) {
    //setting globals
    CHK( i ==0 || i== 1);
    BDD v_bdd = get_bool_var(vr,c,i);
    if(i==0)  { add_to_reads(v_bdd); }
    else { add_to_writes(v_bdd); }
} 

void Pexpr_bdd_old::negate ( set<Pexpr_bdd_old>& rets ) const {
    BDD r = reads;
    BDD w = writes;
    BDD not_r =  !( r );
    BDD not_w =  !( w );

    Pexpr_bdd_old ret;

    // !r & w
    ret = Pexpr_bdd_old(not_r,w);
    rets.insert(ret);

    // r & !w
    ret = Pexpr_bdd_old(r,not_w);
    rets.insert(ret);

    //!r & !w
    ret = Pexpr_bdd_old(not_r,not_w);
    rets.insert(ret);
}



void Pexpr_bdd_old::and_with (Pexpr_bdd_old& p, Pexpr_bdd_old& res) {
    BDD and_r, and_w ;

    and_r = and_restrict(reads, p.reads) ;
    and_w = and_restrict(writes, p.writes) ;
    res = Pexpr_bdd_old(and_r,and_w);
}

BDD Pexpr_bdd_old::get_sup_closure_bdd ( BDD sup ) const {
    BDD tmp_sup = sup;
    //TODO: maybe can extract indices from sup and then check
    //or maybe have a relation and directly compute closure
    const_foreach(it, shared_sup) {
        BDD b = mgr.bddVar(it->first);
        if ( (tmp_sup & !b) == mgr.bddZero() ) { //b's var is in tmp_sup
            tmp_sup = tmp_sup & it->second;
        }
    }
    return tmp_sup;
}

BDD Pexpr_bdd_old::abstract_to_globals ( BDD rw, BDD rep_rw ) const {
    //return rw->project_to(rep_rw);
    BDD common, only_me, only_rep;
    BDD sup = rep_rw.Support();
    BDD all_sup = get_sup_closure_bdd(sup);
    
    //rw.ClassifySupport(rep_rw,&common,&only_me,&only_rep);
    rw.ClassifySupport(all_sup,&common,&only_me,&only_rep);
    /*Util::msg(2,"rw is "); rw.PrintMinterm();
    Util::msg(2,"rep_rw is "); rep_rw.PrintMinterm();
    Util::msg(2,"all_sup is "); all_sup.PrintMinterm();
    Util::msg(2,"common is "); common.PrintMinterm();
    Util::msg(2,"only_me is "); only_me.PrintMinterm();
    Util::msg(2,"only_rep is "); only_rep.PrintMinterm();
    */
    //CHK(only_rep == mgr.bddZero());
    return rw.ExistAbstract(only_me);
}

void Pexpr_bdd_old::abstract_to_globals ( const Pexpr_bdd_old& rep, Pexpr_bdd_old& abs ) const {
    abs.reads = abstract_to_globals ( reads, rep.reads );
    abs.writes = abstract_to_globals ( writes, rep.writes );
    Util::msg(2,"abstract to globals: me is "+to_str()+ " rep is "+rep.to_str() + "abs is "+ abs.to_str());
}


bool Pexpr_bdd_old::matches (const Pexpr_bdd_old& p) const {
    Expr* r = get_reads_expr();
    Expr* w = get_writes_expr();
    Pexpr_e p1(r,w);

    Expr* p_r = p.get_reads_expr();
    Expr* p_w = p.get_writes_expr();
    Pexpr_e p2(p_r,p_w);
    return p1.matches(p2);
}

template DP::expr Pexpr_bdd_old::get_expr_DP ( DP& dp, int depth ) const ;


template<class DP>
typename DP::expr Pexpr_bdd_old::get_expr_DP ( DP& dp, int depth ) const {
    typedef typename DP::expr expr;
    const Expr* r_expr = get_reads_expr();
    const Expr* w_expr = get_writes_expr();
    //expr cr_expr = dp.mk_true();
    expr cr_expr = eval_sym_expr(dp,r_expr,-1,depth);
    expr cw_expr = eval_sym_expr(dp,w_expr,-1,depth+1);
    expr c_expr = dp.mk_and(cr_expr, cw_expr);
    return c_expr;
}

// ----   Pexpr_e -------

bool Pexpr_e::matches ( const Pexpr_e& p ) const {
    const Expr* r = get_reads();
    const Expr* w = get_writes();

    const Expr* p_r = p.get_reads();
    const Expr* p_w = p.get_writes();

    DP dp;
    set<Varref> sup;
    r->get_support(sup); p_r->get_support(sup);
    DP_declare_varrefs(dp,sup,-1,0);

    sup.clear();
    w->get_support(sup); p_w->get_support(sup);
    DP_declare_varrefs(dp,sup,-1,1);

    DP::expr r_expr = eval_sym_expr(dp,r,-1,0);
    DP::expr p_r_expr = eval_sym_expr(dp,p_r,-1,0);
    DP::expr and_r_expr = dp.mk_and(r_expr, p_r_expr);

    DP::expr w_expr = eval_sym_expr(dp,w,-1,1);
    DP::expr p_w_expr = eval_sym_expr(dp,p_w,-1,1);
    DP::expr and_w_expr = dp.mk_and(w_expr, p_w_expr);

    DP::expr c = dp.mk_and(and_r_expr,and_w_expr);
    dp.assertdp(c);
    if ( dp.check() != l_false ) { return true;}
    else { return false; }

}

template DP::expr Pexpr_e::get_expr_DP ( DP& dp, int depth ) const;

template<class DP>
typename DP::expr Pexpr_e::get_expr_DP ( DP& dp, int depth ) const {
    typedef typename DP::expr expr;

    const Expr* r_expr = get_reads();
    const Expr* w_expr = get_writes();
    expr cr_expr = dp.mk_true();
    if ( r_expr) cr_expr = eval_sym_expr(dp,r_expr,-1,depth);
    expr cw_expr = dp.mk_true();
    if ( w_expr ) cw_expr = eval_sym_expr(dp,w_expr,-1,depth+1);
    expr c_expr = dp.mk_and(cr_expr, cw_expr);
    return c_expr;
}




//misc

/*BDD get_sup_closure( const set<uint>& indices ) {
    BDD res = mgr.bddOne();
    const_foreach(it, indices) {
        uint i = *it;
        res = res & mgr.bddVar(i);
        if ( Key_In_Bag(i,shared_sup) ) {
            BDD sup = shared_sup.find(i)->second;
            res = res & sup;
        }
    }
}*/

