
#ifndef PE_H
#define PE_H

#include"action.h"
#include"cfg.h"
#include"cfa.h"
#include"prop_fa.h"
#include"sys_state.h"
#include"eval_ea.h"

/*******  Proc_edge *********************/

struct Proc_edge_type_vis {
	Edge_type et;
	void operator() (const CFG_edge& e) { et = e.get_type(); }
	//void operator() (const CFA_edge& e) { et = e.get_type(); }
	template<class T>
	void operator() (const T&) { assert(0); }
};

struct Proc_edge_comm_type_vis {
	bool is_comm_type;
	void operator() (const CFG_edge& e) { is_comm_type = e.is_comm_type(); }
	//void operator() (const CFA_edge& e) { is_comm_type = e.is_comm_type(); }
	template<class T>
	void operator() (const T&) { assert(0); }
};

struct Proc_edge_to_str_vis {
	string s;
	void operator() (CFG_edge& e) { s = e.to_str(); }
	void operator() (eCFA_edge& e) { s = e.to_str(); }
	template<class T>
	void operator() (T&) { assert(0); }
};

template<class Action>
struct Proc_edge_comm_act_vis {

	const Action* act_ptr;
	void operator() (const CFG_edge& e) { act_ptr = & (VAR::get<Action>(e.get_comm_act()) ); }
	//void operator() (const CFA_edge& e) { act_ptr = & (VAR::get<Action>(e.get_comm_act()) ); }
	template<class T>
	void operator() (const T&) { assert(0); }
};

template<class Act>
struct Proc_edge_get_act_vis {

	Act act;
	template<class T1>
	void operator() (const T1& e) { 
		if( e.is_eps() ) act = Act::get_eps();
		else act = e.get_act() ; 
	}
};

struct Proc_edge_chan_name_vis  {
	const Sys_state& s;
	int pid;

	Proc_edge_chan_name_vis (const Sys_state& _s, int _pid) : s(_s), pid(_pid) {}

	Var_dcl vd;

	void operator() (const CFG_edge& e) { vd = eval_edge_chan_name(e, s, pid); }
	//void operator() (const CFA_edge& e) { vd = eval_edge_chan_name(e, s, pid); }
	template<class T>
	void operator() (const T&) { assert(0); }
};

//WONDER: can't these visitors be avoided? the direct template versions would 
//lead to a variant on Proc_edge<CFG> and Proc_edge<Sys> at higher level

class Proc_edge {
protected:
typedef VAR::variant<CFG_edge,eCFA_edge> Super;
Super e;
Pid pid;

template<class Action> 
	const Action& get_comm_act() const {
		Proc_edge_comm_act_vis<Action> vis;
		VAR::apply_visitor(vis,e);
		return *(vis.act_ptr);
	}

public:
	Proc_edge(const CFG_edge& _e, size_t _pid) : e(_e), pid(_pid) {}
	Proc_edge(const eCFA_edge& _e, size_t _pid) : e(_e), pid(_pid) {}
	Proc_edge(const eProp_edge& _e) : e(_e), pid(-1) {}
	
	bool is_comm_type () const {
		Proc_edge_comm_type_vis vis;
		VAR::apply_visitor(vis,e);
		return vis.is_comm_type;
	}

	Edge_type get_type() const {
		Proc_edge_type_vis vis;
		VAR::apply_visitor(vis,e);
		return vis.et;
	}

	size_t get_pid() const { return pid; }
	const Super& get_edge() const { return e;}
	const Send& get_send_act () const {
		CHK(get_type() == SEND);
		return get_comm_act<Send>();
	}

	const Recv& get_recv_act () const {
		CHK(get_type() == RECV);
		return get_comm_act<Recv>();
	}

        template<class T>
	T get_act() const {
		Proc_edge_get_act_vis<T> vis;
		VAR::apply_visitor(vis,e);
		return (vis.act);
	}

	Var_dcl eval_chan_name(const Sys_state& s) const {
		Proc_edge_chan_name_vis vis(s, pid);
		VAR::apply_visitor(vis,e);
		return vis.vd;
	}
        /*
	size_t eval_chan_size(const Sys_state& s) const {
		assert( is_comm_type() );
		Var_dcl vd = eval_chan_name(s);
		size_t num = s.get_mem().get_q_size( vd ); //Send|Recv ::Varref -> Var_dcl
		return num;
	}

	bool eval_chan_full(const Sys_state& s) {
		assert( is_comm_type() );
		Var_dcl vd = eval_chan_name(s);
		return s.is_qvar_full(vd);
	}

	bool eval_chan_empty(const Sys_state& s) {
		assert( is_comm_type() );
		Var_dcl vd = eval_chan_name(s);
		return s.is_qvar_empty(vd);
	}
	*/
	string to_str() {
		Proc_edge_to_str_vis vis;
		VAR::apply_visitor(vis,e);
		string s = "["+ int2str(pid) + "] " + vis.s;
		return s;
	}

};

#endif
