#ifndef RW_MAP_H
#define RW_MAP_H
#include"expr.h"
#include"g_st_diff.h"

//typedef map<Var_dcl, Consts> Glb_val_map;
typedef Gsd<Consts> Glb_val_map;

class RW_map {
    protected:
        Glb_val_map reads;
        Glb_val_map writes;
    public:
        enum RW_enum {R,W};
        RW_map() { }
        RW_map(const Glb_val_map& _reads, const Glb_val_map& _writes) : reads(_reads), writes(_writes) {}

        const Glb_val_map& get_r () const { return reads; }
        Glb_val_map& get_r () { return reads; }
        const Glb_val_map& get_w () const { return writes; }
        Glb_val_map& get_w () { return writes; }

        string to_str() const {
            string s = reads.to_str_first() + " | " + writes.to_str_first();
            return s;
        }

        void add_val(RW_enum m, const Var_dcl& vd, const Const& c) {
            /*if ( m == R) reads[vd].insert(c);
            else if( m == W) writes[vd].insert(c);
            else assert(0);
            */
            list<Const> cs; cs.push_back(c);
            add_vals(m, vd, cs);
        }

        template< class T>
        void add_vals(RW_enum m, const Var_dcl& vd, const T& clist) {
            /*if ( m == R) reads[vd].insert( clist.begin(), clist.end() );
            else if (m == W) writes[vd].insert(clist.begin(), clist.end() );
            else assert(0);
            */

            if ( m == R) reads.add_vals(vd, clist );
            else if (m == W) writes.add_vals(vd, clist);
            else assert(0);
        }

        bool has_read (const Var_dcl& vd) const {
            //return (reads.find(vd) != reads.end());
            return reads.has_var(vd);
        }

        bool has_write (const Var_dcl& vd) const {
            //return (writes.find(vd) != writes.end());
            return writes.has_var(vd);
        }

        bool has_read_or_write (const Var_dcl& vd) const {
            return ( has_read(vd) || has_write(vd) );
        }

        void add( const RW_map& ext_rw ) {
            const Glb_val_map& ext_r = ext_rw.get_r();
            const Glb_val_map& ext_w = ext_rw.get_w();

            reads.add_vals(ext_r);
            writes.add_vals(ext_w);
        }

        bool empty() const { return reads.empty() && writes.empty(); }

        template<class T>
        void build_one ( const Glb_val_map& gsd, RW_enum rw_mode, list<T>& curr_list ) const {
            list<T> tmp_list;
            const map<Var_dcl, Consts>& val_map = gsd.vars_diff;
            const_foreach(it, val_map ) {
                const Var_dcl& vd = it->first;
                const Consts& conc_values = it->second;
                //Util::msg(2,"build_one: vd is "+vd.to_str()+": "+set_to_str(conc_values)+"\n"); 

                typename_const_foreach(itl, curr_list) {
                    const_foreach(itv, conc_values) {   
                        T new_cfa_act = *itl;
                        new_cfa_act.set_val(rw_mode, vd, *itv);
                        tmp_list.push_back(new_cfa_act);
                    }
                }
                curr_list = tmp_list;
                tmp_list.clear();
            }
        }

        template<class T>
            void build_acts(list<T>& init_list, set<T>& sigma) const {

                list<T> curr_list = init_list;

                build_one( reads, R, curr_list );
                build_one( writes, W, curr_list );

                sigma.insert(curr_list.begin(), curr_list.end() );
                CHK( !sigma.empty() );
            }

        template<class T>
        void build_rep_one ( const Glb_val_map& gsd, RW_enum rw_mode, list<T>& curr_list ) const {
            list<T> tmp_list;
            const map<Var_dcl, Consts>& val_map = gsd.vars_diff;
            const_foreach(it, val_map ) {
                const Var_dcl& vd = it->first;
                const Consts& conc_values = it->second;
                Const c = *(conc_values.begin());
                //Util::msg(2,"build_one: vd is "+vd.to_str()+": "+set_to_str(conc_values)+"\n"); 

                typename_const_foreach(itl, curr_list) {
                    T new_cfa_act = *itl;
                    new_cfa_act.set_val(rw_mode, vd, c);
                    tmp_list.push_back(new_cfa_act);
                }

                curr_list = tmp_list;
                tmp_list.clear();
            }
        }

        template<class T>
            void build_rep_acts(list<T>& init_list, set<T>& sigma) const {

                list<T> curr_list = init_list;

                build_rep_one( reads, R, curr_list );
                build_rep_one( writes, W, curr_list );

                sigma.insert(curr_list.begin(), curr_list.end() );
                CHK( !sigma.empty() );
            }
};


#endif
