#include"ast.h"
#include"logic.h"
#include"cfg_comp.h"
#include"globals.h"

//Symtab::Map_type vars; //a[0] -> (int, global)

class cfg_context {
    CFG_node begin_node;
    CFG_node end_node;
    //const CFG_node& end;
    CFG_node break_node;
    CFG_node continue_node;
    string scope;


    public:

    cfg_context()   { } 
    cfg_context(const CFG_node& _begin_node) : begin_node(_begin_node)  {  } 

    void set_scope_param() {scope = "P";}
    void set_scope_local() {scope = "L";}
    void set_scope_global() {scope = "G";}

    void set_begin(const CFG_node& _begin_node) { begin_node = _begin_node;  } 
    void set_begin_end_break_continue(const CFG_node& _begin_node, const CFG_node& _end_node, 
                                        const CFG_node& _break_node, const CFG_node& _cont_node) {
        begin_node=_begin_node; end_node = _end_node; break_node = _break_node; continue_node = _cont_node; 
    } 

    string get_scope() const { return scope;}
    bool is_scope_global() const { return scope == "G";}
    bool is_scope_local() const { return scope == "L";}
    bool is_scope_param() const { return scope == "P";}

    CFG_node get_begin() const { return begin_node;}
    CFG_node get_end() const { return end_node;}
    CFG_node get_break() const { return break_node;}
    CFG_node get_continue() const { return continue_node;}

};

struct ast2sys_visitor;

//---- main cfg_add_edge functions ----//
void cfg_add_edge(CFG& cfg, const CFG_node& n1, CFG_edge& t, const CFG_node& n2, cfg_context& ct) {
    //Util::msg(2,"adding edge " + t.to_str()+ " from "+n1.to_str()+" to "+n2.to_str());
      cfg.add_edge(n1,t,n2);
      ct.set_begin(n2);
}

void cfg_add_epsilon_edge(CFG& cfg, const CFG_node& n1, const CFG_node& n2, cfg_context& ct) {
    //Util::msg(2,"adding eps edge from "+n1.to_str()+" to "+n2.to_str());
  CFG_edge t;
  cfg_add_edge(cfg,n1,t,n2,ct);
}
//only used in cfg_fix_gotos
void cfg_add_epsilon_edge(CFG& cfg, const CFG_node& n1, const CFG_node& n2) {
  CFG_edge t;
  cfg.add_edge(n1,t,n2);
}

//=== cfg_add_assign, break, guard. else, continue ===//
void cfg_add_action(CFG& cfg, const CFG_action& a, cfg_context& ct) {
    CFG_stmt s = a;
    CFG_edge t = CFG_edge(s);
    CFG_node end_node = cfg.new_node(); 
    cfg_add_edge(cfg,ct.get_begin(),t,end_node,ct);
}

void cfg_add_guard(CFG& cfg, const CFG_guard& g, cfg_context& ct) {
    CFG_edge t = CFG_edge(CFG_stmt(g));
    CFG_node end_node = cfg.new_node(); 
    cfg_add_edge(cfg,ct.get_begin(),t,end_node,ct);
}

void cfg_add_break(CFG& cfg, cfg_context& ct) {
    CFG_edge t;
    CFG_node end_node = cfg.new_node(); 
    cfg_add_edge(cfg,ct.get_begin(), t, ct.get_break(), ct);
    ct.set_begin(end_node);
}

void cfg_add_goto(CFG& cfg, cfg_context& ct) {
    CFG_node end_node = cfg.new_node(); 
    ct.set_begin(end_node);
}

void cfg_add_else(CFG& cfg, cfg_context& ct) {
    CFG_guard g = CFG_stmt::guard_else;
    cfg_add_guard(cfg,g,ct);
}

void cfg_add_pause(CFG& cfg, cfg_context& ct) {
    CFG_action act = Pause();
    cfg_add_action(cfg, act, ct);
}

void cfg_add_continue(CFG& cfg, cfg_context& ct) {
    CFG_edge t;
    cfg_add_edge(cfg, ct.get_begin(), t, ct.get_continue(), ct);
}

void cfg_postprocess(CFG& cfg, ast2sys_visitor& vis);

/* build cfg directly from a set of assigns in INIT */
void cfg_alist_build_from_sstmts(CFG& cfg, list<Assign>& alist, const SStmts& stmts ) {
    if( stmts.empty() ) return;
    CFG_node n = cfg.initialize_entry_node();
    cfg_context ct (n);
    const_foreach(it, stmts) {
        const Assign& assign = VAR::get<Assign>(*it);
        alist.push_back(assign);
        CFG_action act (assign);
        cfg_add_action(cfg, act, ct);
    }
}

void build_logic_from_sstmts(const SStmts& stmts, Logic& logic) {
    list<Assign> alist;
    const_foreach(it, stmts) {
        const Assign& assign = VAR::get<Assign>(*it);
        alist.push_back(assign);
    }
    logic.build_from(alist);
}

struct ast2sys_visitor {
  //void operator() (CFG& cfg, const Decl& d ) { VAR::apply_visitor(*this,d); }
  //System& sys;
  CFG_comps& cfg_comps;
  CFG& sys_init_cfg;
  Globals& globals;

  cfg_context ct;
  CFG cfg;
  Symtab curr_sym_tab;
  map<CFG_node,Symbol> goto_map;
  map<Symbol, CFG_node> lab2node;
  bool else_flag; //hack for not implementing cfg_fix_else
  

  ast2sys_visitor( CFG_comps& _cfg_comps, CFG& _sys_init_cfg, Globals& _globals)
                    : cfg_comps(_cfg_comps), sys_init_cfg(_sys_init_cfg), globals(_globals){ 
      CFG_node n = sys_init_cfg.initialize_entry_node();
      ct = cfg_context(n);
      ct.set_scope_global();
      else_flag = false;
  }

  void operator() (const Var_decl& _d ) {  
    Var_decl d(_d);

    //fprintf(stderr,"Handling Var_decl\n");
    Ty ty = TUP::get<0>(d);
    Varspec& vs = TUP::get<1>(d);  
    Var_dcl& vd = TUP::get<0>(vs);
    Expr* e = TUP::get<1>(vs);
    
    Symbol var_name = vd.to_str(); 
    //fprintf(stderr,"-- Var name is %s\n",var_name.c_str());
    //TODO match_type(varref,e);

    if( is_ty_chan(ty) ) {
        assert(ct.is_scope_global());
        const Exprt& et = *e;
        const ChanInit& ci = VAR::get<ChanInit>(et);
        assert(0); //no chan vars in symoda
    }
    else if( is_ty_ev(ty) ) {
        assert(ct.is_scope_global());
        globals.add_ev_var(d);
    }
    else  {
        if(ct.is_scope_global()) globals.add_glb_var(d);
        else curr_sym_tab.insert(d,ct.get_scope());

        if( e != NULL) {
            Assign a = Assign(Varref(var_name),e);
            const VAR::variant<Assign>& var_a = a;
            apply_visitor(*this,var_a);
        }
    }
  }

  void operator() (const Var_decls& vds) {
    //fprintf(stderr,"Handling Var_decls, size is %d\n",vds.size());
    const_foreach (it, vds) { 
     // printf("var_decl\n");
      const VAR::variant<Var_decl>& var_vd = *it;
      apply_visitor(*this,var_vd);
    }
  }

  void operator() (const Send& send) {
      CFG_action a(send);
      assert(!ct.is_scope_global());
      cfg_add_action(cfg,a,ct);
  }

  void operator() (const Recv& recv) {
      CFG_action a(recv);
      assert(!ct.is_scope_global());
      cfg_add_action(cfg,a,ct);
  }

  void operator() (const Assign& assign) {
    CFG_action a(assign);
    if(ct.is_scope_global()) //global initializations not handled right now
	 	{ assert(0); cfg_add_action(sys_init_cfg,a,ct); }
    else cfg_add_action(cfg,a,ct);
  }

  void operator() (const DAssign& dassign) {
    CFG_action a(dassign);
    CHK(!ct.is_scope_global()); //no global dassign initializations allowed
    cfg_add_action(cfg,a,ct);
  }

  void operator() (const Inc& inc) {
    const Varref& v = inc.v;
    Expr* e = ExprManager::get_expr(
                mk_expr_varref(v),mk_op_plus(),mk_expr_const(1) 
              );
    Assign a(v,e);
    const VAR::variant<Assign>& var_a = a;
    apply_visitor(*this,var_a);
  }

  void operator() (const Dec& dec) {
    const Varref& v = dec.v;
    Expr* e = ExprManager::get_expr(
                mk_expr_varref(v),mk_op_minus(),mk_expr_const(1) 
              );
    Assign a(v,e);
    const VAR::variant<Assign>& var_a = a;
    apply_visitor(*this,var_a);
  }

  void operator() (const Goto& gt) {
    goto_map[ct.get_begin()] = (Symbol)gt;
    cfg_add_goto(cfg, ct);
  }

  void operator() (const Label& lb) {
    lab2node[(Symbol)lb] = ct.get_begin();
  }

  void operator() (const Xr& xr) {}
  void operator() (const Xs& xs) {}

  void operator() (const Expr* e) {
     if( ct.is_scope_global() ) assert(0);
     //fprintf(stderr,"Handling Expr*: %s\n",e->to_str().c_str());
     CFG_guard g = CFG_guard(e);
     cfg_add_guard(cfg,g,ct);
  }

  void operator() (const Stmt_enum& se) {
      //fprintf(stderr,"*Handling Stmt_enum*\n");
      switch (se) {
          case ELSE_STMT:
              else_flag = true;
              cfg_add_else(cfg,ct);
              break;
          case PAUSE_STMT:
                cfg_add_pause(cfg,ct);
                break;
          case Break:
              cfg_add_break(cfg,ct);
              break;
          case Continue:
              cfg_add_continue(cfg,ct);
              break;
          default: assert(0);
                   break;
      }
  }

/*typedef VAR::variant<Xr,Xs,Expr*,If,Do,Stmt_enum,Assign,DAssign,Inc,Dec,Goto,Label,Assert,Atomic,Send,Recv> SStmt; */
  void operator() (const Opts& opts) {
      //fprintf(stderr,"Handling Opts\n");
      //fix start, end if nodes
      const_foreach(it, opts) {
          const VAR::variant<Sequence>& var_seq = *it;
          cfg_context ct_old = ct;
          apply_visitor(*this,var_seq);
      //    Util::msg(2,"adding eps edge from "+ct.get_begin().to_str()+" to "+ct_old.get_end().to_str());
          cfg_add_epsilon_edge(cfg, ct.get_begin(), ct_old.get_end(), ct);
          ct = ct_old;
          //get opt end node
          //make epsilon from opt end node to if end node
      }

  }

  void operator() (const If& ifstmt) {
     //fprintf(stderr,"Handling If\n");
     const Opts* opts_ptr = (TUP::get<0>(ifstmt));
     const Opts& opts = (*opts_ptr);
     const VAR::variant<Opts>& var_opts = opts;

     CFG_node start = ct.get_begin(); 
     cfg.set_if_begin(start);
     //Util::msg(2,"set if begin : "+ start.to_str());
     CFG_node end = cfg.new_node(); 
     cfg.set_if_end(end);
     //Util::msg(2,"set if end : "+ end.to_str());

     ct.set_begin_end_break_continue(start,end,CFG_node(),CFG_node()); //TODO: break, continue should not occur inside if, check
     apply_visitor(*this,var_opts);
     ct = cfg_context(end); //begin
     
  }

  void operator() (const Do& dostmt) {
     //fprintf(stderr,"Handling Do\n");
     const Opts* opts_ptr = (TUP::get<0>(dostmt));
     const Opts& opts = (*opts_ptr);
     const VAR::variant<Opts>& var_opts = opts;

     CFG_node start = ct.get_begin(); 
     cfg.set_if_begin(start);
     //Util::msg(2,"set do-if begin : "+ start.to_str());
     CFG_node brk = cfg.new_node(); 
     cfg.set_if_end(brk);
     //Util::msg(2,"set do-if end : "+ brk.to_str());
     ct.set_begin_end_break_continue(start,start,brk,start);
     apply_visitor(*this,var_opts);
     ct = cfg_context(brk); //begin
  }

  void operator() (const SStmt& sstmt) { 
     // fprintf(stderr,"Handling SStmt\n"); 
      apply_visitor(*this, sstmt); 
  }

  void operator() (const Sequence* seq) {
    //fprintf(stderr,"Handling Sequence*, size is %d\n", seq->size());
    const Sequence& seq_ = *seq;
      const_foreach(it, seq_) {
        //fprintf(stderr,"step\n");
        const Step& step_ = *it;
        apply_visitor(*this, step_);
      }
  }

  void operator() (const Sequence& seq) {
    //fprintf(stderr,"Handling Sequence, size is %d\n", seq.size());
      const_foreach(it, seq) {
        //fprintf(stderr,"step\n");
        const Step& step_ = *it;
        apply_visitor(*this, step_);
      }
  }

  void operator() (const Stmt& stmt) { apply_visitor(*this, stmt);}


  //void operator() (const Init_decl& id) {assert(0);}

  void operator() (const Type_decl& d ) {
    assert(ct.is_scope_global());
    globals.add_type_decl(d);
  }

  void operator() (const Proc_decl& d ) { 
    //int num = TUP::get<0>(d);
    Symbol proc_name = TUP::get<1>(d);
    //Util::message(2,"proc_name=%s, number=%d\n",proc_name.c_str(), num);
    Util::message(2,"#################  proc_name=%s  ##############\n",proc_name.c_str());
    
    cfg = CFG(proc_name);
    CFG_node n = cfg.initialize_entry_node();
    ct = cfg_context(n);
    curr_sym_tab = Symtab();
    goto_map.clear(); lab2node.clear();

    const Var_decls& vds = TUP::get<2>(d);
    const Body& b = TUP::get<3>(d);

    /* build local CFG */
    ct.set_scope_param(); //parameter
    const VAR::variant<Var_decls>& var_vds = vds;
    apply_visitor(*this,var_vds);

    ct.set_scope_local();
    const VAR::variant<Sequence>& var_seq = TUP::get<2>(b);
    apply_visitor(*this,var_seq);

    cfg.set_exit_node(ct.get_begin());
    //cout<< cfg;
    cfg_postprocess(cfg, *this);
    Util::msg(3,"Finished processing main cfg");
    //cout<< "after postprocessing..\n" << cfg;
    /* end build local CFG */

    Util::msg(3,"Processing logic..");
    const SStmts& logic_stmts = TUP::get<0>(b); 
    Logic logic;
    build_logic_from_sstmts( logic_stmts, logic);
    Util::msg(3,"Finished processing logic");

    Util::msg(3,"Building init comp..");
    const SStmts& init_stmts = TUP::get<1>(b); 
    list<Assign> alist;
    CFG init_cfg(proc_name+"_init");
    cfg_alist_build_from_sstmts(init_cfg,alist,init_stmts);
    Init_comp init_comp (init_cfg,alist);
    Util::msg(3,"Finished building init comp");
    ct.set_scope_global();

    CFG_comp* c = new CFG_comp(proc_name,cfg,curr_sym_tab, logic, init_comp);
    //add_cfg_component
    cfg_comps.push_back(c);
    globals.pid2cfg[c->get_pid()] = c;
      
  }

  template<typename T> void operator() (const T& t) {printf("Not handled\n"); assert(0); }
};


//--- cfg_post_process ---//
void cfg_fix_gotos(CFG& cfg, const ast2sys_visitor& vis) {
    const map<CFG_node,Symbol>& goto_map = vis.goto_map;
    const map<Symbol, CFG_node>& lab2node = vis.lab2node;

    const_foreach(it, goto_map) {
        const CFG_node& source = it->first;
        const Symbol& l = it->second;
        assert(lab2node.find(l) != lab2node.end());
        const CFG_node& target = lab2node.find(l)->second;
        cfg_add_epsilon_edge(cfg, source, target);
    }
}

void cfg_fix_else(CFG& cfg) {
//TODO: implemented in proc_explorer, maybe could implement it here later with expr sharing
    //assert(0);
}
void cfg_postprocess(CFG& cfg, ast2sys_visitor& vis) {
  cfg_fix_gotos(cfg, vis);
  if( vis.else_flag )  {
      cfg_fix_else(cfg);
      vis.else_flag = false;
  }
  cfg.remove_unreach();
}

void build_sys_from_ast( const Decls& ast, CFG_comps& cfg_comps, CFG& sys_init_cfg, Globals& globals) {
    ast2sys_visitor v(cfg_comps,sys_init_cfg,globals);
    const_foreach (it, ast) { VAR::apply_visitor(v,*it); }

    //now cfg_comps, globals are built
    foreach(it, cfg_comps) {
        CFG_comp* cfg_comp = (CFG_comp*)(*it);
        cfg_comp->fix_const_types(globals);
    }

}
