letter  [A-Za-z]
underscore  _
dot "."
digit  [0-9]
number {digit}+ 
name ({letter}|{underscore})({letter}|{underscore}|{digit}|{dot})* 
comment_start  "/*"
comment_end  "*/"
star  '*'
not_star  [^'*']
not_slash  [^'/']
not_newline  [^'\n']

%option noyywrap nounput
%x  COMMENT

%{

  /* Exception used to indicate that an invalid character has been encountered
   * in the input source */
  //exception InvalidCharacter of Location.loc * char

  /* Exception used to indicate that an unbalance in the comment delimiters has
   * been detected */
  //exception UnbalancedComment of Location.loc
#include<cstdio>
#include<cassert>
#include"parser_union.h"
#include"sprm_parser.h"

  enum lexEx_t {InvChar, UnbCom};
  int lineNo = 1, colNo = 0;

  void raise_lex_exception(lexEx_t lexEx) {
    printf("line %d, column %d: ",lineNo,colNo);
    printf("(yytext = %s) \n",yytext);
    switch(lexEx) {
      case InvChar:
        printf("lex exception: invalid character\n");
        break;
      case UnbCom:
        printf("lex exception: unbalanced comment\n");
        break;
      default:
        printf("unknown lex exception\n");
        assert(0);
    }
    assert(0);
    printf("\n");
  }


  /* Current nesting level of comments. Equal to 0 if we are not currently in a
   * comment */
  //let comment_nesting_level : int ref = ref 0
  int comment_nesting_level = 0;

  /* Predicate that is true if we are currently in a comment */
  bool in_comment () { if(comment_nesting_level != 0) return true;
                        else return false; }

  /* Starts a new comment */
  void begin_comment () { comment_nesting_level++; BEGIN(COMMENT);}

  /* Ends an existing comment */
  void end_comment ()  {
    if  (!in_comment()) raise_lex_exception(UnbCom);
    else comment_nesting_level--;
  }
/*<INITIAL> BEGIN(TOKEN);*/
%}


%%

"proctype"                    { return PROCTYPE  ; }
"instance"                    { return INSTANCE  ; }
"active"                      { return ACTIVE  ; }
"LOGIC"                       { return LOGIC  ; }
"INIT"                        { return INIT  ; }
"of"                          { return OF  ; }
"bit"                         { return BIT  ; }
"bool"                        { return BOOL  ; }
"byte"                        { return BYTE  ; }
"bv"                          { return BV  ; }
"event"                       { return EVENT  ; }
"chan"                        { return CHAN  ; }
"int"                         { return INT  ; }
"typedef"                     { return TYPEDEF  ; }
"pid"                         { return PID  ; }
"short"                       { return SHORT  ; }
"unsigned"                    { return UNSIGNED  ; }
"if"                          { return IF  ; }
"fi"                          { return FI  ; }
"ite"                         { return ITE  ; }
"do"                          { return DO  ; }
"od"                          { return OD  ; }
"else"                        { return ELSE  ; }
"timeout"                     { return TIMEOUT  ; }
"break"                       { return BREAK  ; }
"goto"                        { return GOTO  ; }
"pause"                       { return PAUSE  ; }
"assert"                      { return ASSERT  ; }
"atomic"                      { return ATOMIC  ; }
"true"                        { return TRUE  ; }
"false"                       { return FALSE  ; }
{name}                          { sprm_lval.symbol=yytext; return SYMBOL; }
{number}                        { sprm_lval.number = atoi(sprm_text); return NUMBER; }
"+"                           { return PLUS  ; }
"-"                           { return MINUS  ; }
"*"                           { return TIMES  ; }
"/"                           { return DIV  ; }
"%"                           { return MOD  ; }
"="                           { return ASGN  ; }
"{="                          { return D_ASGN  ; }
","                           { return COMMA  ; }
":"                           { return COLON  ; }
";"                           { return SEMI; }
"("                           { return LPAR  ; }
")"                           { return RPAR  ; }
"["                           { return LBRACKET  ; }
"]"                           { return RBRACKET  ; }
"{"                           { return LBRACE  ; }
"}"                           { return RBRACE;  }
"&"                           { return AND  ; }
"|"                           { return OR  ; }
"^"                           { return XOR  ; }
"!"                           { return BANG  ; }
"?"                           { return QUESTION  ; }
"~"                           { return NEG  ; }
"<"                           { return LESS  ; }
">"                           { return GREATER  ; }
"=="                          { return EQUAL  ; }
"!="                          { return NOT_EQUAL  ; }
"<="                          { return LESS_EQUAL  ; }
">="                          { return GREATER_EQUAL  ; }
"<<"                          { return LSHIFT  ; }
">>"                          { return RSHIFT  ; }
"&&"                          { return LOGICAL_AND  ; }
"||"                          { return LOGICAL_OR  ; }
"++"                          { return INC  ; }
"--"                          { return DEC  ; }
"::"                          { return COLON_COLON  ; }
"->"                          { return SEMI  ; } 
{comment_start}                 { begin_comment(); }
{comment_end}                   { raise_lex_exception (UnbCom ) ; }
[ \t\r]                     { colNo++; }
<*>\n                          { lineNo++; colNo = 0; }
.                             { raise_lex_exception (InvChar) ; }




<COMMENT>{
    {comment_start}                  { begin_comment();  }
    {comment_end}                    {
                                end_comment();
                                if (!in_comment()) BEGIN(INITIAL);
                               }
[^*\n]*                        {/* eat anything that's not a '*' or '\n' */}
}


