%{
#include<cstdio>
#include<cassert>
#include"parser_union.h"
#include"ast.h"


int yylex();
extern int lineNo, colNo;
extern Ast::Decls ast;

void yyerror (char* str) {
 fprintf(stderr,"Error: %s at line number %d, col. %d\n",str,lineNo,colNo);
}

%}


%token PROCTYPE INSTANCE ACTIVE XR XS OF BIT BOOL BYTE BV EVENT INT 
%token LOGIC INIT
%token TYPEDEF PID SHORT UNSIGNED  CHAN
%token ITE IF FI DO OD ELSE TIMEOUT BREAK GOTO PAUSE ASSERT ATOMIC TRUE FALSE
%token SYMBOL NUMBER PLUS MINUS TIMES DIV MOD ASGN D_ASGN
%token COMMA COLON COLON_COLON SEMI LPAR RPAR LBRACKET RBRACKET LBRACE RBRACE 
%token AND OR XOR BANG QUESTION NEG LESS GREATER EQUAL NOT_EQUAL LESS_EQUAL GREATER_EQUAL LSHIFT RSHIFT
%token LOGICAL_AND LOGICAL_OR INC DEC COLON2 


%start parse
%type <expr> expr
%type <expr> chan_init
%type <expr> unary_expr
%type <expr> primary_expr
%type <decls> parse
%type <decls> decls
%type <decls> decls_r
%type <decls> decl
%type <proc_decl> proc_decl
%type <number> instance
%type <var_decls> var_decls
%type <var_decls> ne_var_decls
%type <var_decls> ne_var_decls_r
%type <var_decls> var_decl
%type <type_decl> type_decl
%type <ty> ty
%type <ty> ty2
%type <symbol_list> ne_name_list
%type <symbol_list> ne_name_list_r
%type <body> body new_body
%type <sequence> sequence 
%type <step> step
%type <stmt> stmt
%type <sstmt> simple_stmt
%type <sstmt> label
%type <ty_list> ne_ty_list
%type <ty_list> ne_ty_list_r
%type <varref> varref
%type <varspec_list> ne_var_list
%type <varspec_list> ne_var_list_r
%type <varspec> ivar
%type <vardcl> vardcl
%type <send_args> ne_send_args
%type <send_args> ne_send_arg_list
%type <send_args> ne_send_arg_list_r
%type <expr> send_arg
%type <recv_args> ne_recv_args
%type <recv_args> ne_recv_arg_list
%type <recv_args> ne_recv_arg_list_r
%type <recv_arg> recv_arg
%type <opts> opts
%type <opts> opts_r
%type <opt> opt
%type <sstmt> assign
%type <sstmts> assigns opt_assigns logic init_proc
%type <constant> const
%type <symbol> symbol
%type <number> number

%left LOGICAL_OR LOGICAL_AND
%left OR XOR AND
%left EQUAL NOT_EQUAL
%left LESS GREATER LESS_EQUAL GREATER_EQUAL
%left LSHIFT RSHIFT
%left PLUS MINUS
%left TIMES DIV MOD

%%

parse: decls                    { ast = $1; printf("%%INFO%%: Parsing over: line %d\n",lineNo);  };

decls: decls_r                  { $$ = $1; };

decls_r: decl                   { $$ = $1; }
       | decls_r decl           { $$ = $1; $$.splice($$.end(),$2); }
;

decl: var_decl SEMI     { $$ = Ast::mk_decl_var($1); }
    | type_decl SEMI     { $$.push_back(Ast::mk_decl_type($1)); }
    | proc_decl                 { $$.push_back(Ast::mk_decl_proc($1)); }
    | SEMI                 {; }
;

proc_decl: instance PROCTYPE symbol LPAR var_decls RPAR body
                                { $$ = Ast::mk_proc_decl($1, $3, $5, $7) }
;

instance:                       { $$ = Ast::mk_number(0); }
        | ACTIVE                { $$ = Ast::mk_number(1); }
        | ACTIVE LBRACKET number RBRACKET
                                { $$ = $3; }
;

var_decls:                      {; }
         | ne_var_decls         { $$ = $1; }
;

ne_var_decls: ne_var_decls_r    { $$ = $1; }
            ;

ne_var_decls_r: var_decl        { $$.splice($$.end(), $1); }
              | ne_var_decls_r SEMI var_decl
                                { $$.splice($$.end(), $1); $$.splice($$.end(), $3); }
;

var_decl: ty ne_var_list        { $$ = Ast::declare_vars ($1, $2); }
        | symbol ne_var_list    { $$ = Ast::declare_vars (Ast::mk_ty_named ($1), $2); }
        //| MTYPE ne_var_list     { $$ = Ast::declare_vars (Ast::mk_ty_mtype (), $2); }
;

type_decl: TYPEDEF symbol LBRACE ne_name_list RBRACE
                                { $$ = Ast::mk_type_decl ($4, $2) }
;

ty: BIT                         { $$ = Ast::mk_ty_bit (); }
  | BOOL                        { $$ = Ast::mk_ty_bool (); }
  | BYTE                        { $$ = Ast::mk_ty_byte (); }
  | INT                         { $$ = Ast::mk_ty_int (); }
  | CHAN                        { $$ = Ast::mk_ty_chan (); }
  | EVENT                       { $$ = Ast::mk_ty_ev(); }
  | PID                         { $$ = Ast::mk_ty_pid (); }
  | SHORT                       { $$ = Ast::mk_ty_short (); }
  | UNSIGNED                    { $$ = Ast::mk_ty_unsigned (); }
  | BV LESS number GREATER     { $$ = Ast::mk_ty_bv ($3); }
;

    
ne_name_list: ne_name_list_r    { $$ = $1; }
            ;
ne_name_list_r: symbol          { $$.push_back($1); }
              | ne_name_list_r symbol
                                { $$ = $1; $$.push_back($2); }
              | ne_name_list_r COMMA symbol
                                { $$ = $1; $$.push_back($3); }
;

body: LBRACE new_body RBRACE    { $$ = $2; }
    ;

/* better grammar: sequence logic init_proc sequence */
new_body:  logic init_proc sequence OS  { $$ = Ast::mk_body($1,$2,$3); }
;

    
logic:                          {; }     
    | LOGIC LBRACE opt_assigns RBRACE  {  $$ = $3;}
    ;

init_proc:                      {;}
         | INIT  LBRACE opt_assigns RBRACE  { $$ = $3;}
        ;

opt_assigns:           {;}
            | assigns OS { $$ = $1; }
            ;

assigns: assign { $$.push_back($1);}
        | assigns MS assign { $$ = $1; $$.push_back($3); } 
        ;

/*
sequence:                       {; }
        | ne_sequence           { $$ = $1; }
;

ne_sequence: ne_sequence_r      { $$ = $1; }
           ;
*/

sequence: step             { $$.push_back($1); }
        | sequence MS step
                                { $$ = $1; $$.push_back($3); }
;

ne_var_list: ne_var_list_r      { $$ = $1; }
           ;

ne_var_list_r: ivar             { $$.push_back($1); }
             | ne_var_list_r COMMA ivar
                                { $$ = $1; $$.push_back($3); }
;

ivar: vardcl                    { $$ = Ast::mk_varspec($1); }
    | vardcl ASGN expr          { $$ = Ast::mk_varspec($1, $3); }
    | vardcl ASGN chan_init  { $$ = Ast::mk_varspec($1, $3); }
;


chan_init : LBRACE ne_ty_list RBRACE 
                                { $$ = Ast::mk_expr_chan_init($2); }
                                ;

ne_ty_list: ne_ty_list_r        { $$ =$1; }
          ;

ty2: ty {$$=$1;}
   | symbol {$$ = Ast::mk_ty_named($1);}
   ;

ne_ty_list_r: ty2                { $$.push_back($1); }
            | ne_ty_list_r COMMA ty2 { $$ = $1; $$.push_back($3); }
;

vardcl: symbol                  { $$ = Ast::mk_simplevardcl($1); }
     /* | symbol COLON number     { $$ = Ast::mk_bitvardcl($1, $3); } */
      | symbol LBRACKET number RBRACKET
                                { $$ = Ast::mk_arrayvardcl($1, $3); }
;

OS:      {/* empty */}
    | MS { /* one redundant semicolon at end of a sequence*/}
    ;

MS: SEMI { /*one semi */}
    | MS SEMI { /* more than one semi */ }
    ;

expr: expr LOGICAL_OR expr      { $$= Ast::mk_expr_binop($1, Ast::mk_op_lor(),$3); }
    | expr LOGICAL_AND expr     { $$= Ast::mk_expr_binop($1, Ast::mk_op_land(), $3); }
    | expr OR expr              { $$= Ast::mk_expr_binop($1, Ast::mk_op_or(), $3); }
    | expr XOR expr             { $$= Ast::mk_expr_binop($1, Ast::mk_op_xor(), $3); }
    | expr AND expr             { $$= Ast::mk_expr_binop($1, Ast::mk_op_and(), $3); }
    | expr EQUAL expr           { $$= Ast::mk_expr_binop($1, Ast::mk_op_equal(), $3); }
    | expr NOT_EQUAL expr       { $$= Ast::mk_expr_binop($1, Ast::mk_op_not_equal(), $3); }
    | expr LESS expr            { $$= Ast::mk_expr_binop($1, Ast::mk_op_less(), $3); }
    | expr GREATER expr         { $$= Ast::mk_expr_binop($1, Ast::mk_op_greater(), $3); }
    | expr LESS_EQUAL expr      { $$= Ast::mk_expr_binop($1, Ast::mk_op_less_equal(), $3); }
    | expr GREATER_EQUAL expr   { $$= Ast::mk_expr_binop($1, Ast::mk_op_greater_equal(), $3); }
    | expr LSHIFT expr          { $$= Ast::mk_expr_binop($1, Ast::mk_op_lshift(), $3); }
    | expr RSHIFT expr          { $$= Ast::mk_expr_binop($1, Ast::mk_op_rshift(), $3); }
    | expr PLUS expr            { $$= Ast::mk_expr_binop($1, Ast::mk_op_plus(), $3); }
    | expr MINUS expr           { $$= Ast::mk_expr_binop($1, Ast::mk_op_minus(), $3); }
    | expr TIMES expr           { $$= Ast::mk_expr_binop($1, Ast::mk_op_times(), $3); }
    | expr DIV expr             { $$= Ast::mk_expr_binop($1, Ast::mk_op_div(), $3); }
    | expr MOD expr             { $$= Ast::mk_expr_binop($1, Ast::mk_op_mod(), $3); }
    | ITE LPAR expr COMMA expr COMMA expr RPAR { $$ = Ast::mk_expr_ternop(Ast::mk_op_ite(), $3, $5, $7); }
    | unary_expr                { $$= $1; }
;

unary_expr: PLUS unary_expr     { $$= Ast::mk_expr_unop(Ast::mk_uop_plus(), $2); }
          | MINUS unary_expr    { $$= Ast::mk_expr_unop(Ast::mk_uop_minus(), $2); }
          | NEG unary_expr      { $$= Ast::mk_expr_unop(Ast::mk_uop_neg(), $2); }
          | BANG unary_expr     { $$= Ast::mk_expr_unop(Ast::mk_uop_not(), $2); }
          | primary_expr        { $$= $1; }
;

primary_expr: varref            { $$= Ast::mk_expr_varref($1); }
            | const             { $$= Ast::mk_expr_const($1); }
            | LPAR expr RPAR    { $$= $2; }
            | TIMEOUT           { $$= Ast::mk_expr_timeout (); }
            /*| LEN LPAR expr RPAR
                                { $$= Ast::mk_expr_unop(Ast::mk_uop_len(), $3); }
            | FULL LPAR expr RPAR
                                { $$= Ast::mk_expr_unop(Ast::mk_uop_full(), $3); }
            | EMPTY LPAR expr RPAR
                                { $$= Ast::mk_expr_unop(Ast::mk_uop_empty(), $3); }
            | NFULL LPAR expr RPAR
                                { $$= Ast::mk_expr_unop(Ast::mk_uop_nfull(), $3); }
            | NEMPTY LPAR expr RPAR
                                { $$= Ast::mk_expr_unop(Ast::mk_uop_nempty(), $3); }
                                */
;

varref: symbol                  { $$ = Ast::mk_varref_symbol($1); }
      | symbol LBRACKET expr RBRACKET
                                { $$ = Ast::mk_varref_array($1, $3); }
      | symbol LBRACE expr RBRACE     { 
                                    $$ = Ast::mk_varref_bv($1,$3); 
                                }
;

step: var_decl                  { $$ = $1; }
    | stmt                      { $$ = $1; }
;

stmt: simple_stmt               { $$ = $1; }
    | LBRACE sequence OS RBRACE    { $$ = new Ast::Sequence($2); }
    | label simple_stmt          { Ast::Sequence* seq = new Ast::Sequence();
                                  seq->push_back(Ast::Step(Ast::Stmt($1))); 
                                  seq->push_back(Ast::Step(Ast::Stmt($2))); 
                                  $$ = seq;
                                }
;

simple_stmt: 
            expr               { $$= Ast::mk_stmt_expr ($1) ; }
           | IF opts FI      { $$= Ast::mk_stmt_if ($2) ; }
           | DO opts OD      { $$= Ast::mk_stmt_do ($2) ; }
           | ELSE               { $$= Ast::mk_stmt_else () ; }
           | BREAK              { $$= Ast::mk_stmt_break () ; }
           | assign             { $$= $1 ; }
           | PAUSE              { $$= Ast::mk_stmt_pause() ; }
           | GOTO symbol        { $$= Ast::mk_stmt_goto ($2) ; }
           | label              { $$= $1 ; }
           | ASSERT expr        { $$= Ast::mk_stmt_assert ($2) ; }
           /*| ATOMIC LBRACE sequence OS RBRACE { $$= Ast::mk_stmt_atomic ($3) ; } */
           | varref BANG ne_send_args
                                { $$= Ast::mk_stmt_send ($1, $3) ; }
           | varref QUESTION ne_recv_args
                                { $$= Ast::mk_stmt_recv ($1, $3) ; }
;

label: symbol COLON             { $$ = Ast::mk_stmt_label ($1); }
     ;

ne_send_args: ne_send_arg_list  { $$ = ($1); }
            | send_arg LPAR ne_send_arg_list RPAR
                                { $$ = $3; $$.insert($$.begin(),$1); }
;

ne_send_arg_list: ne_send_arg_list_r
                                { $$ = $1; }
;

ne_send_arg_list_r: send_arg    { $$.push_back($1); }
                  | ne_send_arg_list_r COMMA send_arg
                                { $$ = $1; $$.push_back($3); }
;

send_arg: expr                  { $$ = $1; }
        ;

ne_recv_args: ne_recv_arg_list
                                { $$ = $1; }
               | recv_arg LPAR ne_recv_arg_list RPAR
                                { $$ = $3; $$.insert($$.begin(),$1); }
;

ne_recv_arg_list: ne_recv_arg_list_r { $$ = $1; };
ne_recv_arg_list_r: recv_arg
                                { $$.push_back($1); }
                     | ne_recv_arg_list_r COMMA recv_arg
                                { $$ = $1; $$.push_back($3); }
;

recv_arg: varref             {$$ = Ast::mk_recv_arg_varref ($1); }
           | const              { $$ = Ast::mk_recv_arg_const ($1); }
;

opts: opts_r                    { $$ = $1; }
    ;
opts_r: opt                     { $$.push_back($1); }
      | opts_r opt              { $$ = $1; $$.push_back($2); }
;

opt: COLON_COLON sequence OS
                                { $$ = Ast::mk_opt ($2); }
;

assign: varref ASGN expr        { $$ = Ast::mk_stmt_assign ($1, $3); }
      | varref D_ASGN expr      { $$ = Ast::mk_stmt_dassign ($1, $3); }
      | varref INC              { $$ = Ast::mk_stmt_inc ($1); }
      | varref DEC              { $$ = Ast::mk_stmt_dec ($1); }
;

const: TRUE                     { $$ = Ast::mk_const_bool (true); }
     | FALSE                    { $$ = Ast::mk_const_bool (false); }
     | number                   { $$ = Ast::mk_const_int ($1); }
;

/*init_decl: INIT LBRACE run_sequence RBRACE
                                { $$ = Ast::mk_init_decl ($3); }
         | INIT LBRACE RBRACE   { $$ = Ast::mk_init_decl (); }
        ;
run_sequence: run_sequence_r    { $$ = $1; }
            ;
run_sequence_r: run_step        { $$.push_back($1); }
              | run_sequence_r SEMI run_step
                                { $$ = $1; $$.push_back($3); }
;

run_step: RUN symbol LPAR arguments RPAR
                                { $$ = Ast::mk_run_step ($2, $4) }
;

arguments:                      { ; }
         | ne_arguments         { $$ = $1 }
;
ne_arguments: ne_arguments_r    { $$ = $1; };
ne_arguments_r: expr            { $$.push_back($1); }
              | ne_arguments_r COMMA expr
                                { $$ = $1; $$.push_back($3); }
;

 */

symbol: SYMBOL                    { $$ = ($1.symbol); 
                                      
                                      }
      ;

number: NUMBER                  { $$ = $1.number; }
      ;

%%

