#ifndef SSYS_H
#define SSYS_H

#include"cfg.h"
#include"ast.h" //Var_decl, Type_decl, Decls
//#include"sys_state.h"
#include"fa.h"
#include"cfa.h"
#include"cfa_util.h"
#include"rw_map.h"

#include"cfg_comp.h"
#include"cfa_comp.h"
#include"globals.h"


//symbolic
#include"eval_ea_sym.h"


void build_sys_from_ast( const Decls& ast, CFG_comps& cfg_comps, CFG& sys_init_cfg, Globals& globals) ;


class sSystem {
    public:
        typedef sCDFA DFAt;
        typedef sCDFA::Sig CFA_action;
        typedef set<CFA_action> CFA_actions;

        typedef CFA_compt<sCDFA> CFA_comp;
        typedef vector<CFA_comp> CFA_comps;
        static Globals globals;


    protected:

        //Local to each system
        CFG_comps cfg_comps; 
        CFA_comps cfa_comps; //FAs participating in system, e.g., assumption in A-G
        set<Comm_act> comm_sigma;
        RW my_glb_rw, env_rw;
        Mem_map mem;

        bool has_init_glb_state;
        Expr* init_glb_state;

        int max_depth; //depth to which system should be unrolled, if known


        void get_cfg_state_vars(size_t i, Var_dcls& loc_vars, Var_dcls& glb_vars) const {
            CHK( i < num_cfg_components() );
            cfg_comps[i]->get_state_vars(loc_vars, glb_vars);
        }

        void get_state_vars( map <Pid, Var_dcls >& pid2vars) const {
            for(size_t i=0; i< num_cfg_components(); ++i) {
                Pid pid = cfg_comps[i]->get_pid();
                get_cfg_state_vars(i,pid2vars[pid],pid2vars[-1]);
            }
        }

        /*void get_cfg_state_coi_vars(size_t i, Var_dcls& loc_vars, Var_dcls& glb_vars) const {
            CHK( i < num_cfg_components() );
            cfg_comps[i]->get_state_coi_vars(loc_vars, glb_vars);
        }

        void get_state_coi_vars( map <Pid, Var_dcls >& pid2vars) const {
            for(size_t i=0; i< num_cfg_components(); ++i) {
                Pid pid = cfg_comps[i]->get_pid();
                get_cfg_state_coi_vars(i,pid2vars[pid],pid2vars[-1]);
            }
        }*/

        //set state vars same as sys
        void set_state_vars( const sSystem& sys) {
            //build_fa_mem_map();
            map<Pid, Var_dcls > pid2stvars;
            sys.get_state_vars(pid2stvars);
            mem.set_state_vars( pid2stvars );
        }
        void add_state_vars( Pid pid, const Var_dcls& vds) {
            mem.add_state_vars( pid, vds );
        }

        void build_glb_var_mem_map() {
            //global vars
            const set < Var_dcl > &glb_vardcls = globals.glb_sym_tab.get_glb_vars ();
            const_foreach (it, glb_vardcls) {
                if ( globals.glb_sym_tab.is_ev(*it) )
                    mem.add_ev_var (*it);
                else mem.add_glb_var (*it);
            }
        }

        void build_cfg_mem_map() {
            //local vars
            const_foreach(it, cfg_comps) {
                (*it)->build_cfg_comp_mem_map(mem);
            }

        }


        void build_mem_map(bool build_cfg=true) {
            if ( build_cfg ) build_cfg_mem_map();
            build_glb_var_mem_map();

        }

        bool no_cfa_comps() const {
            if (!cfa_comps.empty())
                return( num_fa_components() == 1 && cfa_comps[0].has_univ_fa() );
            else return true;
        }

        void copy_globals( const sSystem& sys) {
            //since globals is static, dont do anything here
        }


    public:
        //remove
        sSystem() { has_init_glb_state = false; set_depth(-1); }//has_init_sym_state = false;  

        void ast2sys(const Decls& ast) {
            CFG init_cfg; //not used
            build_sys_from_ast( ast, cfg_comps, init_cfg, globals );
        }

        sSystem(const Decls& ast)  { 
            globals.init();

            ast2sys(ast); 
            build_mem_map();
            set_state_vars( *this );
            has_init_glb_state = false;
            //has_init_sym_state = false;
            set_depth (-1);
        }


        //probably shud have called this project_sys
        sSystem(const sSystem& sys, const set<size_t>& cfg_indices, const RW& _env_rw) 
            : env_rw (_env_rw) {
                //project env_rw_vals to this_rw_vals?
                copy_globals(sys);
                //univ_dfa = NULL;

                const CFG_comps& _cfg_comps = sys.get_cfg_components();
                const_foreach(it, cfg_indices) {
                    size_t i=*it; CHK( i < _cfg_comps.size() );
                    cfg_comps.push_back(_cfg_comps[i]);
                }

                build_mem_map();
                set_state_vars( *this );
                add_state_vars ( -1, env_rw.get_w() );

                //TODO: make univ fa for rw_vals.build_acts(), get_concrete_comm_sigma
                if ( !env_rw.empty() ) {
                    CFA_actions univ_sigma;
                    build_univ_actions(univ_sigma);
                    CFA_comp cfa_comp (univ_sigma);
                    add_fa_component( cfa_comp );
                }

                //build_fa_mem_map();
                fix_init_state(sys);
                set_depth (-1);

            }

        sSystem(const sSystem& sys, const sCDFA& cdfa) { //union of sys and fa_comp
            CHK( no_cfa_comps() );
            copy_globals(sys);

            cfg_comps = sys.get_cfg_components();
            CFA_comp fa_comp (cdfa);
            add_fa_component(fa_comp);
            build_mem_map();

            set_state_vars( sys );
            RW rw; fa_get_glb_rw(cdfa,rw);
            add_state_vars ( -1, rw.get_w() );

            fix_init_state(sys);

            CHK ( sys.get_depth() == -1 );
            set_depth ( cdfa.get_depth() );
        }


        sSystem(const vector<sCDFA>& _cfas, const sSystem& sys) { 
            //use sys as reference, e.g., to get glb_sym_tab
            CHK( no_cfa_comps() );
            copy_globals(sys);

            //cfg_comps = sys.get_cfg_components();
            //cfa_comps = _cfa_comps;
            const_foreach (it, _cfas) { 
                CHK ( it->get_depth() == -1);
                RW rw; fa_get_glb_rw(*it,rw);
                add_state_vars ( -1, rw.get_w() );
                add_fa_component( CFA_comp(*it) ); 
            }
            build_mem_map(false);
            //build_fa_mem_map();
            set_state_vars( sys );
            fix_init_state(sys);
            set_depth (-1);
        }

        const sSystem& operator = ( const sSystem& sys) {
            copy_globals(sys);

            cfg_comps = sys.cfg_comps; 
            cfa_comps = sys.cfa_comps; 
            comm_sigma = sys.comm_sigma;
            my_glb_rw = sys.my_glb_rw;
            env_rw = sys.env_rw;

            has_init_glb_state = sys.has_init_glb_state;
            //has_init_sym_state = sys.has_init_sym_state;
            mem = sys.mem;
            if ( has_init_glb_state ) {
                //init_glb_state.set_mem(mem);
                init_glb_state = sys.init_glb_state;
            }
            set_depth ( sys.get_depth() );
            return *this;
        }


        void fix_init_state ( const sSystem& sys ) { //fix init state using sys
            if ( sys.has_init_state() )
                set_init_state ( sys.get_init_state() );
            else has_init_glb_state = false;

        }

        void set_init_state ( Expr* _init_glb_state ) { //some init_glb_state from top level
            //Util::msg(2,"set_init_state: sys_init_glb_state = "+ _init_glb_state->to_str());
            //init_glb_state.set_mem (mem);
            //init_glb_state.initialize();
            //Util::msg(2,"set_init_state: init_glb_state = "+ init_glb_state.to_str());
            //sys_init_glb_state.abstract_using_mem_map(mem, init_glb_state);
            init_glb_state = _init_glb_state; //copy+abstract
            has_init_glb_state = true;
            //Util::msg(2,"set_init_state: init_glb_state = "+ init_glb_state->to_str());
        }

        bool has_init_state() const { return has_init_glb_state; }
        //bool has_init_sym_state() const { return has_init_sym_state; }
        Expr* get_init_state() const { 
            CHK( has_init_glb_state); 
            //Util::msg(2,"system:: init_state = "+ init_glb_state.to_str());
            return init_glb_state; 
        }
        void show_init_state() const {
            if ( has_init_glb_state ) {
                Util::msg(2,"init state is "+get_init_state()->to_str());
            }
        }

        void set_depth(int depth) { max_depth = depth; }
        int get_depth() const { return max_depth; }

        void add_cfg_component(const CFG_comp* c) { 
            (globals.pid2cfg)[c->get_pid()] = c; cfg_comps.push_back(c);
        }
        void add_fa_component(const CFA_comp& fa) { cfa_comps.push_back(fa);}

        size_t num_cfg_components() const { return cfg_comps.size();}
        size_t num_fa_components() const { return cfa_comps.size();}
        size_t size() const { return (num_cfg_components() + num_fa_components()); }


        const CFG_comps& get_cfg_components() const { return cfg_comps;}
        const CFA_comps& get_fa_components() const { return cfa_comps; }
        //both should return const but ...

        const Mem_map& get_mem_map() const { return mem;}
        CFG& get_init_cfg() { return globals.init_cfg;}

        const RW& get_env_rw() const { return env_rw; }





        void get_range_values(Pid pid, const Var_dcl& vd, list<Const>& clist) const {
            globals.get_range_values(pid,vd,clist);
        }

        void get_cfg_rw(size_t i, RW& rw_i) {
            CHK( i < num_cfg_components() );
            cfg_comps[i]->get_rw(rw_i);
        }


        void get_vmap(const RW& rw, RW_map& rwm) const {
            const Var_dcls& r = rw.get_r();
            const_foreach(it, r) {
                list<Const> clist; get_range_values(-1,*it,clist);
                rwm.add_vals(RW_map::R, *it, clist );
            }
            const Var_dcls& w = rw.get_w();
            const_foreach(it, w) {
                list<Const> clist; get_range_values(-1,*it,clist);
                rwm.add_vals(RW_map::W, *it, clist );
            }
        }

        void get_glb_rw ( RW& glb_rw ) {
            const set < Var_dcl > &glb_vars = globals.glb_sym_tab.get_all_glb_vars ();
            for(size_t i=0; i< num_cfg_components(); ++i) {
                RW rw;
                get_cfg_rw(i,rw);
                rw.project_to(glb_vars);
                glb_rw.add(rw);
            }
        }

        /*void get_glb_rwm(RW_map& rwm) {
            //only sigma from cfg components
            CHK( no_cfa_comps() );

            RW glb_rw;
            get_glb_rw(glb_rw);
            get_vmap(glb_rw,rwm);
        }

        const RW_map& get_glb_rwm() {
            if( my_glb_rwm.empty() ) {
                get_glb_rwm(my_glb_rwm);
            }
            return my_glb_rwm;
        }*/

        const RW& get_glb_rw() {
            if( my_glb_rw.empty() ) {
                get_glb_rw(my_glb_rw);
            }
            return my_glb_rw;
        }


        void eact2sact ( const eCFA_actions& eacts, sCFA_actions& sacts ) const {
            const_foreach(it, eacts) {
                sCFA_act act(*it);
                sacts.insert(act);
                set<sCFA_act> neg_act = act.negate();
                CHK ( neg_act.size() == 1);
                sacts.insert( *(neg_act.begin()) );
            }
        }


        void build_univ_actions(sCFA_actions& univ_sigma) {
            //TODO: make it simply "true"
            eCFA_actions e_univ_sigma;
            build_univ_actions(e_univ_sigma);
            eact2sact(e_univ_sigma,univ_sigma);
            //Util::message(1,"build_univ_actions: size is %d",univ_sigma.size());
        }

        void build_univ_actions(eCFA_actions& univ_sigma) {
            get_concrete_comm_sigma(env_rw, univ_sigma);
            CHK( !univ_sigma.empty() );
            //Util::msg(2,"univ sigma is "+ set_to_str(univ_sigma) +"\n");
        }

        void get_concrete_comm_sigma(const RW& rw,
                                     set<sCFA_act>& comm_sigma) const {
            eCFA_actions e_comm_sigma;
            get_concrete_comm_sigma(rw,e_comm_sigma);
            eact2sact(e_comm_sigma,comm_sigma);
            //Util::message(1,"get_concrete_comm_sigmaS: size is %d",comm_sigma.size());
        }

        void get_concrete_comm_sigma(const RW& rw,
                                     set<eCFA_act>& comm_sigma) const {
            RW_map rwm;
            get_vmap(rw,rwm);
            list<eCFA_act> init; init.push_back( eCFA_act() );
            rwm.build_rep_acts<eCFA_act>(init, comm_sigma); //only adds one action
            //Util::message(1,"get_concrete_comm_sigmaE: size is %d",comm_sigma.size());
        }



        // friend struct ast2sys_visitor;
};


#endif
