#ifndef SEND_RECV_H
#define SEND_RECV_H
#include"expr.h"
#include"util.h"

enum Edge_type {SEND=1000, RECV, SEND_RECV, OTHER};


template<typename T> 
ostream& operator << (ostream& os, const vector<T>& v) { os << vec_to_str(v); return os; }

class Pause {
    public:
    Pause() {}
    string to_str() const { return "pause"; }
    bool operator == (const Pause& p) const { return true; }
};

class Assign : public TUP::tuple<Varref,Expr*> {
	typedef TUP::tuple<Varref,Expr*> Super;

	public:
	Assign() {}
	Assign (const Varref& vr, Expr* e) : Super(vr,e) {}
	Assign( const Super& s ) : Super(s) {}
	const Varref& get_vr() const { return TUP::get<0>(*this); }
	const Expr* get_rhs() const { return TUP::get<1>(*this); }
	string to_str() const { return (get_vr().to_str() + "=" + get_rhs()->to_str()); }
	bool operator == (const Assign& a) const { 
		return (to_str() == a.to_str());
	}
        void get_rw(RW& rw) const {
            Var_dcl lhs = get_vr().to_vd_plain();
            rw.add_write(lhs);
            get_rhs()->get_rw(RW::R,rw);
        }
        Expr* to_Expr() const {
            const Varref& vr = get_vr();
            Expr* rhs = (Expr*) get_rhs();
            return ExprManager::get_eq(vr,rhs);
        }

        template<class O>
        void fix_const_types (const O& oracle, Pid pid) {
            const Expr* e = get_rhs();
            const Varref& vr = get_vr();
            Var_dcl vd = vr.to_vd_plain();
            Ty ty = oracle.get_ty(vd,pid);
            Expr* e_ = expr_fix_const_types(e,oracle,pid,ty);
            (*this) = Super(vr,e_);
        }
        

};

class DAssign : public TUP::tuple<Varref,Expr*> {
	typedef TUP::tuple<Varref,Expr*> Super;

	public:
	DAssign() {}
	DAssign (const Varref& vr, Expr* e) : Super(vr,e) {}
	DAssign( const Super& s ) : Super(s) {}
	const Varref& get_vr() const { return TUP::get<0>(*this); }
	const Expr* get_rhs() const { return TUP::get<1>(*this); }
	string to_str() const { return (get_vr().to_str() + "<=" + get_rhs()->to_str()); }
	bool operator == (const DAssign& a) const { 
		return (to_str() == a.to_str());
	}
        void get_rw(RW& rw) const {
            Var_dcl lhs = get_vr().to_vd_plain();
            rw.add_write(lhs);
            get_rhs()->get_rw(RW::R, rw);
        }

        template<class O>
        void fix_const_types (const O& oracle, Pid pid) {
            const Expr* e = get_rhs();
            const Varref& vr = get_vr();
            Var_dcl vd = vr.to_vd_plain();
            Ty ty = oracle.get_ty(vd,pid);
            Expr* e_ = expr_fix_const_types(e,oracle,pid,ty);
            (*this) = Super(vr,e_);
        }

};

//typedef map<Symbol,int> Mtype_map;

class Send_arg {
const Expr* e;
public:
	Send_arg( const Expr* _e ) : e(_e) {}
        Send_arg( const Const& c) { e = ExprManager::get_expr(c); }
	Send_arg( const Varref& vr )  { e = ExprManager::get_expr(vr); }

	bool is_const() const { return e->is_const(); }
	string to_str() const { return e->to_str(); }

	const Expr* get_expr() const {return e;}
	const Const& get_const () const {
		CHK(is_const());
		return VAR::get<Const>(*e);
	}
/*	bool is_mtype(const Mtype_map& mtype_indices) const ;
	Varref get_mtype_varref() const ;
	int get_mtype_val(const Mtype_map& mtype_indices) const ;
 */

	bool operator == (const Send_arg& sa) const { return to_str() == sa.to_str(); }
        void get_rw(RW& rw) const { e->get_rw(RW::R,rw); } //reads all args before writing them to channel

};

typedef vector<Send_arg> Send_args;

class Send : public TUP::tuple<Varref,Send_args> {
	typedef TUP::tuple<Varref,Send_args> Super;

	Varref& get_vr() { return TUP::get<0>(*this); }
	Send_args& get_sas() { return TUP::get<1>(*this); }
	public:
	Send() {}
	Send(const Varref& vr, const Send_args& sas) : Super(vr,sas) {}
        Send(const Varref& vr, int val) : Super(vr, Send_args() ) {
            Send_arg sa = Send_arg(Const(val));
            get_sas().push_back(sa);
        } 
	Send( const Super& s ) : Super(s) {}

        bool is_concrete() const { 
            const Send_args& sas = get_sas();
            const_foreach(it, sas) { if( ! it->is_const() ) return false; }
            return true;
        }

        Const get_const_arg() const {
            const Send_args& sas = get_sas();
            CHK( sas.size() == 1 );
            return sas[0].get_const();
        }

	const Varref& get_vr() const { return TUP::get<0>(*this); }
	const Send_args& get_sas() const { return TUP::get<1>(*this); }
	string to_str() const { return (get_vr().to_str() + "!" + vec_to_str(get_sas())); }
	bool operator == (const Send& s) const { 
		return (get_vr() == s.get_vr() && get_sas() == s.get_sas());
	}

        void get_rw(RW& rw) const {
            rw.add_write(get_vr().to_vd_plain());
            const Send_args& sas = get_sas();
            const_foreach(it, sas) { it->get_rw(rw); }
        }

};

/*  --------------- RECV ----------------------------- */

class Recv_arg : public VAR::variant<Const,Varref>  {
	typedef VAR::variant<Const,Varref>  Super;

	public:
	Recv_arg() {}
	Recv_arg(const Super& s) : Super(s) {}
        Recv_arg(const Const& c) : Super(c) {  }

	bool is_const() const ;

	const Const& get_const () const {
		CHK(is_const());
		return VAR::get<Const>(*this);
	}

	const Varref& get_varref () const {
		CHK(!is_const());
		return VAR::get<Varref>(*this);
	}

	string to_str() const;
	bool operator == (const Recv_arg& ra) const { return to_str() == ra.to_str(); }
        void get_rw(RW& rw) const { 
            if( !is_const() ) rw.add_write( get_varref().to_vd_plain() );
        }

};


typedef vector<Recv_arg> Recv_args;

class Recv: public TUP::tuple<Varref,Recv_args> {
	typedef TUP::tuple<Varref,Recv_args> Super;

	public:
	Recv() {}
	Recv ( const Varref& vr, const Recv_args& ras) : Super(vr,ras) {}
	Recv( const Super& s ) : Super(s) {}
        Recv(const Varref& vr, int val) : Super(vr, Recv_args() ) {
            Recv_arg ra = Const(val);
            TUP::get<1>(*this).push_back(ra);
        } 

        bool is_concrete() const { 
            const Recv_args& ras = get_ras();
            const_foreach(it, ras) { if( ! it->is_const() ) return false; }
            return true;
        }

	const Varref& get_vr() const { return TUP::get<0>(*this); }
	const Recv_args& get_ras() const { return TUP::get<1>(*this); }
	string to_str() const { return (get_vr().to_str() + "?" + vec_to_str(get_ras()) ); }
	bool operator == (const Recv& r) const { 
		return (get_vr() == r.get_vr() && get_ras() == r.get_ras());
	}
        Const get_const_arg() const {
            const Recv_args& ras = get_ras();
            CHK( ras.size() == 1 );
            return ras[0].get_const();
        }

        void get_rw(RW& rw) const {
            rw.add_read(get_vr().to_vd_plain());
            const Recv_args& ras = get_ras();
            const_foreach(it, ras) { it->get_rw(rw);}
        }

};




#endif
