#ifndef SYS_EXPL_H
#define SYS_EXPL_H

#include"path.h"
#include"esystem.h"
#include"sys_state.h"
#include"upd_map.h"
#include"proc_edge.h"
#include"eval_ea.h"
#include"sys_fa.h"
#include"prop_fa.h"

bool eval_guard_cfa_edge (const Sys_state& s, const eCFA_edge& ce, int pid) ;
void eval_cfa_act(const Sys_state& s, const eCFA_act& act, Sys_state_diff& s_diff) ;

template<class Diff>
class State_diff_pair {
    protected:
        typedef set<Diff> Diffs;
        Diffs ds;
        Sys_state s;
    public:
        State_diff_pair (const Diffs& _ds, const Sys_state& _s) 
            : ds(_ds), s(_s) {}
        State_diff_pair (const Sys_state& _s) 
            :  s(_s) {}
        string to_str() const { return set_to_str<Diff>(ds) + " " + s.to_str(); }

        const Diffs& diffs() const { return ds; }
        const Sys_state& state() const { return s; }
        void insert (const Diff& d) { ds.insert(d); }

        bool operator < (const State_diff_pair& sp) const { 
            if( diffs().to_str() < sp.diffs().to_str() ) return true;
            if( diffs().to_str() > sp.diffs().to_str() ) return false;

            if(  state().to_str() < sp.state().to_str() ) return true;
            else return false;
        }

        bool empty() const { return ds.empty(); }
};


/*
   struct eval_recv_arg_vis {
   Recv_arg ra_res;


   void operator() (const Const& c) { 
   ra_res = Recv_arg(c);	
   }
   void operator() (const Varref& vr) { 
   Val v;
   if ( is_varref_mtype(vr, mtype_indices) ) {
   v = eval_varref_mtype(vr, mtype_indices);
//cout<<"varref is "<< vr.to_str() <<", v is "<<v <<endl;
ra_res = Recv_arg(Const(v));
}
else ra_res = Recv_arg(vr);
}

};
 */

//-------------------------------------------------------//
/* defined in eval_ea.cpp */
void concretize_cfg_edge(CFG_edge& e, const Sys_state& s, int pid) ;
bool eval_proc_edge(const Proc_edge& pe, const Sys_state& s, Sys_state_diff& s_diff);
bool eval_proc_edge_ID(const Proc_edge& pe, const Sys_state& s, ID_diff& idd);
Val eval_send_arg (const Send_arg& sa, const Sys_state& s, int pid) ;

class CFG_comp_explorer {
    protected:
        const CFG_comp& c;
        Pid pid;
        typedef State_diff_pair<Sys_state_diff> Sd_pair;
    public:
        typedef Upd_map<Sys_state_diff> Sd_upd_map;

        CFG_comp_explorer ( const CFG_comp& _c ) : c(_c)
        { pid = c.get_pid(); }

        void get_enabled_acts (const Sys_state & s, list < Proc_edge > &en_acts) const
        {
            int pc = s.get_pcval (pid);
            Util::msg(5,"cfg_comp_expl::get_enabled_acts: pcval is "+int2str(pc));
            list< CFG_edge > en_edges;
            c.get_out_edges(pc, en_edges);
            foreach(it, en_edges) {
                //eval guard
                Util::msg(5,"Checking edge "+ it->source().to_str() + " " + it->to_str()+ " " + it->target().to_str()+"\n");
                tbool lb = eval_guard_cfg_edge(s, *it, pid);
                CHK ( lb.is_t() || lb.is_f());
                if( lb.is_f() ) continue;

                if( !it->is_comm_type() ) { /* push in non-comm statements at beginning */
                    Proc_edge pe (*it, pid);
                    en_acts.push_back(pe);
                    continue;
                }

                if( !eval_cfg_edge_chan_act_enabled(*it, s, pid) ) continue;

                //CONCRETIZE send (evaluate) and recv ( do nothing)
                concretize_cfg_edge(*it,s,pid);
                Proc_edge pe (*it, pid);
                Util::msg(5,"Pushing pe "+pe.to_str()+"\n");
                en_acts.push_back(pe);
            }
            Util::msg(5,"en acts size = "+int2str(en_acts.size()) );
        }


        void explore ( const set<Sys_state>& from, Sd_upd_map& upd_map) {
            deque<Sd_pair> frontier;
            Sd_upd_map local_upd;

            const_foreach(it, from) { frontier.push_back( Sd_pair(*it) ); }

            while( !frontier.empty()) {
                Sd_pair sp = frontier.front();
                const Sys_state& s = sp.state();
                const Sys_state_diffs& sds = sp.diffs();  
                Sys_state_diff s_diff(s.get_mem());
                CHK( sds.size() <= 1); //tmp hack
                if ( !sds.empty() ) s_diff = *(sds.begin());

                frontier.pop_front();

                Sys_state s_ = s;
                s_.apply_diff_check(s_diff);
                // check if there is a loop: same state s is revisited
                //CHK( s_diff.empty() || s != s_);

                list<Proc_edge> en_acts;
                list<Proc_edge>::iterator it;
                get_enabled_acts(s_,en_acts);

                if( en_acts.empty() ) { 
                    //to.insert(sp);
                    if( !s_diff.empty() )
                        local_upd.add_union(s_diff,s);
                }

                //better impl: keep a map from each new state to original s
                //remove and add to the map on each eval_proc_edge
                //finally for each mapped s: add s_diff to its upd_map
                for( it = en_acts.begin(); it != en_acts.end(); ++it) {
                    //NOTE: s_diff is written multiple times here
                    Util::msg(5,"before eval_proc_edge, pe is "+ it->to_str() +"\n");
                    bool ret = eval_proc_edge(*it,s_,s_diff);
                    if (!ret) {
                    // for pause edges, don't push into frontier, add to upd_map
                        local_upd.add_union(s_diff,s);
                    }
                    else {
                        Sys_state_diffs sds; sds.insert(s_diff);
                     frontier.push_back( Sd_pair(sds,s) );
                    }
                }
            }

            upd_map.add_intersect(local_upd);
        }



};




class Proc_explorer
{
    protected:
        //typedef System::CFG_comp CFG_comp;
        //const System* sys;
        const CFG_comp* c;
        eCFA_comp fa;
        Pid pid;
        //const Mem_map* mem;

        enum Expl_type {_CFG_, _CFA_};
        Expl_type expl_type;

        typedef State_diff_pair<ID_diff> St_IDs_pair;

    public:
        Proc_explorer (const CFG_comp* _c):  c (_c) 
        {
            expl_type = _CFG_;
            pid = c->get_pid();
        }

        Proc_explorer ( const eCFA_comp& _fa): fa (_fa)
        {
            c = NULL; expl_type = _CFA_;
            pid = fa.get_pid();
        }

        Pid get_pid () const {  return pid; }

        Local_val get_init_pcval() const { 
            CHK(expl_type == _CFG_);
            return c->get_main_entry_pcval();
        }
        const CFA_state& get_init_cfa_state() const {
            /*const CFA_state inits = fa.get_init(); 
              CHK(inits.size() == 1);
              return *(inits.begin());
             */
            return fa.get_init();
        }


        // run initialization sequence
        void get_init_diff(Sys_state& start_st, Sys_state& init_st) {
            if ( expl_type == _CFA_ ) {
                init_st = start_st;
                CFA_state st = get_init_cfa_state();
                init_st.set_cfa_state(pid, st);
            }
            else if ( expl_type == _CFG_ ) {
                // set pc to init entry
                if( c->has_init_cfg() ) {
                    int init_entry_pcval = c->get_init_entry_pcval();
                    //Util::msg(2,"get_init_diff: setting pc val to "+int2str(init_entry_pcval)); 
                    start_st.set_pcval(pid, init_entry_pcval);

                    //explore
                    CFG_comp_explorer cfg_expl(*c);
                    set<Sys_state> start_sts;
                    CFG_comp_explorer::Sd_upd_map cfg_comp_upd;
                    start_sts.insert(start_st);
                    cfg_expl.explore(start_sts, cfg_comp_upd);

                    //get all states from cfg_comp_upd
                    //start_sts.clear();
                    cfg_comp_upd.get_image_states(start_sts); 
                    CHK( start_sts.size() == 1);
                    init_st = *(start_sts.begin());
                }
                else init_st = start_st;

                //set pc to main entry
                //Util::msg(2,"get_init_diff2: setting pc val to "+int2str(c->get_main_entry_pcval())); 
                init_st.set_pcval(pid, c->get_main_entry_pcval());

            }
            else assert(0);

        }

        void get_enabled_acts_cfg_X(const Sys_state& s, list<Proc_edge>& en_acts) const ;
        void get_enabled_acts_cfg(const Sys_state& s, list<Proc_edge>& en_acts) const ;

        //check if guard is true; send, recvs on len-0 queues are always enabled,
        void get_enabled_acts (const Sys_state & s, list < Proc_edge > &en_acts) const;

        void explore_succs ( const Sys_state& s, const ID_diff& d,
                             St_IDs_pair& sp_, ID_upd_map& local_upd,
                             ID_diffs& vis_ids) {
                             //sp_[s] goes into frontier, while local_upd[s] is not explored
                             //to_be_explored, not to be explored respectively

            Sys_state s_ = s;
            s_.apply_diff_check(d.imm());
            //Util::message(2,"at s_.pc = %d,",s_.get_pcval(pid));
            //s_.apply_diff_check( eval_logic_diff (s_) ); TODO

            list<Proc_edge> en_acts; list<Proc_edge>::iterator it;
            get_enabled_acts(s_,en_acts);
            //Util::msg(2,"explore_cfg: en_acts.size() = "+int2str(en_acts.size()));

            if( en_acts.empty() ) { 
                if( !d.empty() ) local_upd.add_union(d,s);
                return;
            }

            for( it = en_acts.begin(); it != en_acts.end(); ++it) {
                ID_diff d_ = d;
           //     Util::msg(2,"before eval_proc_edge, pe is "+ it->to_str() +"\n");
                bool ret = eval_proc_edge_ID(*it,s_,d_);
                if ( vis_ids.find(d_) == vis_ids.end() ) { vis_ids.insert(d_); }
                else continue;
                
                if (!ret) {
                    // for pause edges, don't push into frontier, add to upd_map
                    local_upd.add_union(d_,s);
                }
                else sp_.insert(d_);
            }
        }


        
        void explore_cfg ( const set<Sys_state>& from, ID_upd_map& local_upd) {
            deque< St_IDs_pair > frontier;
            //ID_upd_map local_upd;

            const_foreach(it, from) { 
                const Sys_state& s = *it;
                // for semantics where imm upd of one proc will affect another proc
                //ID_diffs ds; upd_map.get_diffs(s,ds);
                //frontier.push_back( St_IDs_pair(ds,s) ); 
                frontier.push_back( St_IDs_pair(s) );
            }

            map<Sys_state, ID_diffs> vis_map;

            while( !frontier.empty()) {
                const St_IDs_pair& sp = frontier.front();
                const Sys_state& s = sp.state();
                const ID_diffs& ds = sp.diffs();  
             //   Util::message(2,"at s.pc = %d,",s.get_pcval(pid));
              //  Util::message(2,"ds size = %d\n",ds.size());

                St_IDs_pair sp_(s);

                if ( ds.empty() ) {
                    ID_diff d(s.get_mem());
                    explore_succs(s,d,sp_,local_upd,vis_map[s]);
                }
                else {
                    const_foreach(it, ds) {
                        const ID_diff& d = *it;
                        explore_succs(s,d,sp_,local_upd,vis_map[s]);
                    }
                }

                if( !sp_.empty() ) frontier.push_back( sp_ );

                frontier.pop_front();
            }//end while

            //TODO: (if necessary later) upd_vars_that_stay_same
            //local_upd.add_equals on unmodified write variables of this cfg
            //c->get_rw(RW);

            //upd_map.add_intersect(local_upd);

        }

        void explore_cfa ( const set<Sys_state>& from, ID_upd_map& local_upd) {
            //ID_upd_map local_upd;

            const_foreach(itS, from) {
                const Sys_state& s = *itS;
                ID_diff idd(s.get_mem());
                list<Proc_edge> en_acts;
                get_enabled_acts(s,en_acts);

                if( en_acts.empty() )  { 
                    //to.insert(sp);
                }

                foreach(it, en_acts) {
                    ID_diff tmp_idd (idd);
                   // Util::msg(2,"idd------------------");
                    eval_proc_edge_ID(*it,s,tmp_idd);
                    //Util::msg(2,"------------------");
                    //to.insert( State_diff_pair(idd,s));
                    if( !tmp_idd.empty() )
                        local_upd.add_union(tmp_idd,s);
                }
            }

            //upd_map.add_intersect(local_upd);
            
        }


        void explore ( const set<Sys_state>& from, ID_upd_map& loc_upd_map) {

            if( expl_type == _CFG_ ) {
                //TODO: if( c->is_method() ) init_pcs();
                // do bfs assuming there are no cycles in the graph
                explore_cfg(from ,loc_upd_map);
            }
            else if( expl_type == _CFA_) {
                explore_cfa(from, loc_upd_map);
            }
            else assert(0);

            Util::msg(2,"loc_upd_map is "+loc_upd_map.to_str());

        }


        bool is_final(const CFA_state& st) const {
            CHK( expl_type == _CFA_ );
            return fa.is_final(st);
        }


};


//-- System explorer

class Sys_explorer
{
    protected:
        const eSystem & sys;
        const Mem_map& mem;
        vector < Proc_explorer > procs;
        map<Sys_state, Sys_state> reach_st2pre_st; //stores reachable states; map to their preecessors
        const ePropDFA* P;
        size_t fa_start_index;
        

        bool CHECK_REACH_ERR, CHECK_PRP; //run time flags

        void build_proc_explorers() {
            const CFG_comps& cfg_comps = sys.get_cfg_components();
            size_t index=0;
            const_foreach(it, cfg_comps) {
                Proc_explorer pexp(*it);
                procs.push_back(pexp);
                index++;
            }

            fa_start_index = index;

            const eSystem::CFA_comps& fa_comps = sys.get_fa_components();
            const_foreach(it2, fa_comps) {
                Proc_explorer pexp(*it2);
                procs.push_back(pexp);
            }
        }

        bool are_fa_P_in_err ( const Sys_state& s) const {
            if ( !P->is_final(s.get_prop_state()) ) return false;
            //if( mem.get_num_fas() != 0 ) {
            if ( fa_start_index != procs.size() ) { //no fa comps
                size_t index = fa_start_index;
                for(; index < procs.size(); ++index) {
                    Pid pid = procs[index].get_pid();
                    const CFA_state& st = s.get_cfa_state(pid);
                    if( ! procs[index].is_final(st) ) return false;
                }
            }
            return true;
        }

    public:

        Sys_explorer (const eSystem & _sys): sys (_sys), mem(_sys.get_mem_map()) {
            build_proc_explorers();
            P = NULL;
            CHECK_REACH_ERR = false; //TODO: handle case when it is true
            CHECK_PRP = false;
        } 


        Sys_state get_init_X_state () {
            Sys_state init_s(mem);
            init_s.initialize();

            size_t index =0;

            if ( sys.has_init_state() ) {//precomputed init state for cfg_comps
                init_s = sys.get_init_state();
                index = fa_start_index;
                Util::msg(2,"init_s1 is "+init_s.to_str());
            }

            for (; index< procs.size(); ++index) {
                Sys_state next_s(init_s.get_mem());
                procs[index].get_init_diff(init_s, next_s); //no no-det assumed for init cfg
                init_s = next_s;
            }
            init_s.clear_events();

            //concretize Xs in init_s, TODO (only globals concretized now)
            Util::msg(2,"init_s2 is "+init_s.to_str());
            return init_s;
        }



        void get_init_states ( Sys_states& inits) {
            //TODO: store init states as member; check if empty, then do this, else do nothing

            Sys_state init_s = get_init_X_state();
/*            Mem_map imem;
            init_s.get_X_mem_map(imem);
            if( !imem.empty() ) {
                Conc_sd csd;
                imem.get_conc_sd(sys, csd);
                CHK (!csd.empty());
                list<Sys_state> ls;
                csd.dup_by_ranges( init_s, ls);
                inits.insert(ls.begin(), ls.end());
            }
            else inits.insert(init_s);
*/
            inits.insert(init_s); //only for working with X
            CHK( !inits.empty() );
        }


        const Mem_map& get_memory() { return mem;}


        void explore() {
            set<Sys_state>  d_frontier;
            //Sys_state init_s = get_init_state();
            //reach_st2pre_st[*it] = *it; //put in reach set, pre to init is init itself 
            //d_frontier.insert(init_s);

            get_init_states(d_frontier);
            const_foreach(it, d_frontier) {
                reach_st2pre_st[*it].set_mem(it->get_mem()); 
                reach_st2pre_st[*it] = *it; //put in reach set, pre to init is init itself 
            }
            Util::msg(2,"init states  " + set_to_str<Sys_state>(d_frontier)+"\n");

            while(!d_frontier.empty()) {
                Util::msg(2,"begin D-cycle..\n");

                ID_upd_map upd_map;
                vector<ID_upd_map> loc_upd_maps(procs.size());

                for(size_t i=0; i< procs.size(); ++i) {//one for loop is a delta-cycle 
                    //EVALUATE: all procs run from the same state
                    //and collect UPDATES in upd_map
                    Util::msg(2,"Exploring pid "+int2str(procs[i].get_pid()));
                    procs[i].explore(d_frontier, loc_upd_maps[i]); 
                }

                Util::msg(2,"Intersecting loc_upd_maps...");

                for(size_t i=0; i<procs.size(); ++i)
                    upd_map.add_intersect(loc_upd_maps[i]);

                //apply UPDATES
                d_frontier.clear();
                upd_map.get_new_states(d_frontier, reach_st2pre_st);
                Util::msg(2,"end D-cycle..\n");
            }

        }

        void upd_prop_states ( const Sys_states& sts, ID_upd_map& upd_map ) {
//            CHK ( upd_map.empty() );

            ID_upd_map loc_upd_map;
            //bool flag = false; // at least one prop state matches
            const_foreach( it, sts ) {
                Sys_state s = *it;
                Prop_state p = s.get_prop_state();
                list<eCFA_edge> pes;
                P->get_out_edges(p,pes);
                CHK ( pes.size() > 0 );


                const_foreach ( itp, pes ) {
                    //eval_prop_edge : better to make Prop_edge another type and
                    //as a component of Proc_edge
                    if( !eval_guard_cfa_edge(s, *itp, -1) ) continue;
                    //if (!flag) flag = true;

                    Sys_state_diff s_diff(s.get_mem());
                    eval_cfa_act(s, itp->get_act(), s_diff);
                    s_diff.set_prop_state(itp->target());
                    //if( !s_diff.empty() )
                    ID_diff idd(s.get_mem());
                    idd.apply_delta_diff(s_diff);
                    loc_upd_map.add_union(idd,s);
                }
            }
            //CHK( flag ); no need since algo takes care of it
            upd_map.add_intersect(loc_upd_map);

        }

        void get_prop_diffs ( const Sys_states& sts, map<Sys_state, Sys_state_diffs>& prop_diffs ) {
            const_foreach( it, sts ) {
                Sys_state s = *it;
                Prop_state p = s.get_prop_state();
                list<eCFA_edge> pes;
                P->get_out_edges(p,pes);
                CHK ( pes.size() > 0 );


                const_foreach ( itp, pes ) {
                    if( !eval_guard_cfa_edge(s, *itp, -1) ) continue;

                    Sys_state_diff s_diff(s.get_mem());
                    eval_cfa_act(s, itp->get_act(), s_diff);
                    s_diff.set_prop_state(itp->target());
                    CHK( !s_diff.empty() );
        /*            ID_diff idd(s.get_mem());
                    idd.apply_delta_diff(s_diff);
                    loc_upd_map.add_union(idd,s);
                    */
                    
                    prop_diffs[s].insert(s_diff);
                }
            }
        }

        //void gen_ce(const Sys_state& err, const Sys_state& init, Sys_CE& ce) {
        void gen_ce(const Sys_state& err,  eSys_CE& ce) {

            Sys_state post = err;
            Util::msg(2,"post is "+ post.to_str());

            //remove last transition; artifact of check prop
            /*
            CHK( reach_st2pre_st.find(post) != reach_st2pre_st.end() );
            Sys_state pre = reach_st2pre_st[post];
            if( pre == post ) return; //all init states s have pre = s
            Util::msg(2,"pre is "+pre.to_str());

            post = pre;
            */
            while(1) {
                CHK( reach_st2pre_st.find(post) != reach_st2pre_st.end() );
                Sys_state pre = reach_st2pre_st[post];
                if( pre == post ) break; //all init states s have pre = s
                Util::msg(2,"pre is "+pre.to_str());
                eSys_edge e (eSys_act(pre,post));
                ce.prepend(e);
                post = pre;
            }
            CHK( ce.size() > 0);
            
        }

        //P is the negated property
        bool check_prop (const ePropDFA& _P, eSys_CE& ce, bool must_gen_ce = true) {
            P = &_P;
            set<Sys_state>  inits, d_frontier;
            
            //Sys_state init_s = get_init_state();
            //init_s.set_prop_state( P->get_init() );
            //Util::msg(2,"init state is " + init_s.to_str()+"\n");

            get_init_states(inits);
            //foreach(it, d_frontier) 
            for( set<Sys_state>::iterator it = inits.begin(); it != inits.end(); ++it) {
                Sys_state init_s (*it);
                init_s.set_prop_state( P->get_init() );
                d_frontier.insert(init_s);
                //TODO: make mem_map_manager
                Sys_state& pre_s = reach_st2pre_st[init_s];
                pre_s.set_mem(init_s.get_mem()); 
                pre_s = init_s;//put in reach set, pre to init is init itself 
            }
            Util::msg(2,"init states  " + set_to_str<Sys_state>(d_frontier)+"\n");


            //Sys_state err_st(init_s.get_mem());
            Sys_state err_st( d_frontier.begin()->get_mem()); //pick any one state

            while(!d_frontier.empty()) {
                Util::msg(2,"begin D-cycle..\n");

                ID_upd_map upd_map;
                //apply P : force the sys exploration to be guided by P: try later
                //upd_prop_states(d_frontier, upd_map); 

                vector<ID_upd_map> loc_upd_maps(procs.size());
                //evaluate procs
                for(size_t i=0; i< procs.size(); ++i) {//one for loop is a delta-cycle 
                    procs[i].explore(d_frontier, loc_upd_maps[i]); 
                }

                Util::msg(2,"Intersecting loc_upd_maps...");
                for(size_t i=0; i<procs.size(); ++i)
                    upd_map.add_intersect(loc_upd_maps[i]);

               // upd_prop_states(d_frontier, upd_map);
                map<Sys_state, Sys_state_diffs> prop_diffs;
                get_prop_diffs(d_frontier, prop_diffs);

                d_frontier.clear();
                //apply proc + prop UPDATES

               // upd_map.get_new_states(d_frontier, reach_st2pre_st);
                upd_map.get_new_states_matching_prop(prop_diffs, d_frontier, reach_st2pre_st);

                const_foreach(it, d_frontier) {
                    const Sys_state& s = *it;
                    if( are_fa_P_in_err(s) ) { 
                        err_st = s;  
                        Util::msg(2,"CE found\n");
                        if( must_gen_ce ) gen_ce(err_st,ce);
                        return false;
                    }
                }

                Util::msg(2,"end D-cycle..\n");
            }
            return true;
        }

/*


        void upd_prop_states ( const Sys_states& sts, ID_upd_map& upd_map ) {
            CHK ( upd_map.empty() );

            //bool flag = false; // at least one prop state matches
            const_foreach( it, sts ) {
                Sys_state s = *it;
                Prop_state p = s.get_prop_state();
                list<CFA_edge> pes;
                P->get_out_edges(p,pes);
                CHK ( pes.size() > 0 );


                const_foreach ( itp, pes ) {
                    //eval_prop_edge : better to make Prop_edge another type and
                    //as a component of Proc_edge
                    if( !eval_guard_cfa_edge(s, *itp, -1) ) continue;
                    //if (!flag) flag = true;

                    Sys_state_diff s_diff(s.get_mem());
                    eval_cfa_act(s, itp->get_act(), s_diff);
                    s_diff.set_prop_state(itp->target());
                    //if( !s_diff.empty() )
                    ID_diff idd(s.get_mem());
                    idd.apply_delta_diff(s_diff);
                    upd_map.add_union(idd,s);
                }
            }
            //CHK( flag ); no need since algo takes care of it

        }

        bool check_prop (const ePropDFA& _P, Sys_CE& ce) {
            P = &_P;
            set<Sys_state>  inits, d_frontier;
            
            //Sys_state init_s = get_init_state();
            //init_s.set_prop_state( P->get_init() );
            //Util::msg(2,"init state is " + init_s.to_str()+"\n");

            get_init_states(inits);
            //foreach(it, d_frontier) 
            for( set<Sys_state>::iterator it = inits.begin(); it != inits.end(); ++it) {
                Sys_state init_s (*it);
                init_s.set_prop_state( P->get_init() );
                d_frontier.insert(init_s);
                //TODO: make mem_map_manager
                Sys_state& pre_s = reach_st2pre_st[init_s];
                pre_s.set_mem(init_s.get_mem()); 
                pre_s = init_s;//put in reach set, pre to init is init itself 
            }
            Util::msg(2,"init states  " + set_to_str<Sys_state>(d_frontier)+"\n");


            //Sys_state err_st(init_s.get_mem());
            Sys_state err_st( d_frontier.begin()->get_mem()); //pick any one state

            while(!d_frontier.empty()) {
                Util::msg(2,"begin D-cycle..\n");

                ID_upd_map upd_map;
                //apply P : force the sys exploration to be guided by P
                upd_prop_states(d_frontier, upd_map);

                //evaluate procs
                for(size_t i=0; i< procs.size(); ++i) {//one for loop is a delta-cycle 
                    procs[i].explore(d_frontier, upd_map); 
                }

                //apply proc UPDATES
                d_frontier.clear();
                upd_map.get_new_states(d_frontier, reach_st2pre_st);
                const_foreach(it, d_frontier) {
                    const Sys_state& s = *it;
                    if( are_fa_P_in_err(s) ) { 
                        err_st = s;  
                        Util::msg(2,"CE found\n");
                        gen_ce(err_st,ce);
                        return false;
                    }
                }

                Util::msg(2,"end D-cycle..\n");
            }
            return true;
        }
        */
};


#endif
