
#ifndef _SYS_FA_H
#define _SYS_FA_H

#include"path.h"
#include"fa.h"
#include"cfa.h"
#include"sys_state.h"


class eSys_act: public eCFA_act {
    
    public:
    eSys_act() {}
    eSys_act(const Sys_state& pre, const Sys_state& post) {
        Diff_acts reads, writes;
        //pre.get_diff_eqv(post, writes, reads);
        pre.convert_to_diff(reads);
        post.convert_to_diff(writes);
        eCFA_act::set_rw(reads, writes);
    }
};

class eSysFA : public FA<Sys_state, eSys_act> { };
typedef eSysFA::Ev eSys_edge;


class eSys_CE : public Path<eSys_edge> {
    typedef Path<eSys_edge> Super;
    public:
    typedef Super::const_iterator const_iterator;
    typedef Super::iterator iterator;

    /*t.abstract(sigma,t_abs); //remove composites
      bool res = check_M_t_satisfy_spec(t_abs,ce);
      ce.concretize_using(t); //add back composites at appropriate places
     */
    //Only for Sys_CE and Sys_act
    void abstract_to_globals(const eCFA_act& rep, eCFA_trace& p) const {
        for( const_iterator it = begin(); it!= end(); ++it ) {
            const eSys_act& sys_act = it->get_act();
            eCFA_act cfa_act;
            sys_act.abstract_to_globals(rep, cfa_act);
            p += eCFA_edge(cfa_act);
        }
    }


};


//symbolic 
//typedef map<Pid, Expr*> Pid2expr;

class sSys_act : public sCFA_act {
public:
    sSys_act() {}
    /*sSys_act ( const Pid2expr& pre, const Pid2expr& post ) {
        const_foreach(it, pre ) {
            sCFA_act::set_r(it->first, it->second);
        }
        const_foreach(it, post ) {
            sCFA_act::set_w(it->first, it->second);
        }
    }
    */
    sSys_act ( const sCFA_act& pre, const sCFA_act& post ) {
        //both pre and post have only reads
        //TODO: CHK
        const Pid2pexpr& pid2reads = pre.get_pexpr();
        const Pid2pexpr& pid2writes = post.get_pexpr();
        const_foreach(it, pid2reads) {
            sCFA_act::set_r(it->first, it->second);
        }
        const_foreach(it, pid2writes) {
            sCFA_act::set_w(it->first, it->second);
        }
    }

    void and_with ( const sSys_act& sact ) {
        
        const Pid2pexpr& _pp = sact.get_pexpr();
        const_foreach(it, _pp) {
            Pid pid = it->first;
            Pexpr _p = it->second;
            Pexpr res;
            pid2pexpr[pid].and_with(_p,res);
            pid2pexpr[pid] = res;
        }
    }


};



class sSysFA : public FA<CFA_state, sSys_act> { };
typedef sSysFA::Ev sSys_edge;

class sSys_CE : public Path<sSys_edge> {
public:
    void abstract_to_globals(const sCFA_act& rep, sCFA_trace& p) const {
        for( const_iterator it = begin(); it!= end(); ++it ) {
            const sSys_act& sys_act = it->get_act();
            sCFA_act cfa_act;
            sys_act.abstract_to_globals(rep, cfa_act);
            p += sCFA_edge(cfa_act);
        }
        Util::msg(2,"after abstract_to_globals: p is" + p.to_str()); 
    }

    //don't project to rep variables
    void abstract_to_globals(sCFA_trace& p) const {
        for( const_iterator it = begin(); it!= end(); ++it ) {
            const sSys_act& sys_act = it->get_act();
            sCFA_act cfa_act;
            sys_act.abstract_to_globals(cfa_act);
            p += sCFA_edge(cfa_act);
        }
        Util::msg(2,"after abstract_to_globals: p is" + p.to_str()); 
    }

    void merge_with( const sSys_CE& ce ) {
        iterator it_me = begin();
        const_iterator it_ce = ce.begin();
        while ( 1 ) {
            if ( it_me == end() ) {
                while ( it_ce != ce.end() )  { append( *it_ce ); ++it_ce; }
                break;
            }

            if ( it_ce == ce.end() ) {
                //nothing to do for the rest ce_acts
                break;
            }

            sSys_act& my_sys_act = it_me->get_act();
            const sSys_act& ce_sys_act = it_ce->get_act();

            my_sys_act.and_with(ce_sys_act);

            ++it_me;
            ++it_ce;
        }
    }


};

#endif
