#ifndef SYS_STATE_H
#define SYS_STATE_H

#include<deque>
#include"expr.h"
#include"state.h"
#include"rw_map.h"
#include"val.h"
#include"g_st_diff.h"
#include"eval_ea_sym.h"

#include"yices_c.h"

int eval_const(const Const& c);
enum Upd_mode { NO_OW, OW }; // no overwrite, overwrite

template<class State>
class Storage : public set<State> {
    public:
        Storage() {}
        void store(const State& s) { insert(s); }
        bool lookup(const State& s) { return this->find(s) != this->end(); }
        void erase(const State& s) { assert(lookup(s)); erase(find(s)); }

};

template<class State>
class State_manager {
    protected:
        Storage<State> states;
    public:
        State_manager() {}
        void store(const State& s) { states.store(s);}
        bool is_visited(const State& s) { return states.lookup(s);}
        void clear() { states.clear();}

};

typedef Val Local_val ;
typedef Val Global_val ;

//typedef Val Chan_val; //should be a vector of several vals like Q_el
//typedef Val Ev_val; //should be a bool : hack for eval_chanvar

class Mem_map {
    protected:
        map<Pid, map<Var_dcl,int> >  pid2pos; //index of var in Proc_state::loc_vals
        map<Pid, Var_dcls > pid2stvars;
        map<Var_dcl,int>  glbvar2pos; //position in glbs
        set<int> evvar_pos; //pos of ev vars in glbvar2pos

    public:
        string to_str() const {
            string s;
            const_foreach(it, pid2pos) {
                Pid pid = it->first;
                s += int2str(pid) + ": ";
                const map<Var_dcl,int>& var2pos = it->second;
                const_foreach(it2, var2pos) { 
                    s += it2->first.to_str() + ":" + int2str(it2->second) + " "; 
                }
                s +="\n";
            }
            const_foreach(it, glbvar2pos) { 
                s += it->first.to_str() + ":" + int2str(it->second) + "\n";
            }
            const_foreach(it, evvar_pos) {
                s += int2str(*it);
            }
            s += "\n";
            return s;
        }

        void show_pid2stvars() const {
            Util::msg(1,"---- pid2stvars ----");
            string s;
            const_foreach(it, pid2stvars) {
                Util::message(1,"%d: ",it->first);
                const Var_dcls& stvars = it->second;
                const_foreach(it2, stvars) {
                    s += it2->to_str() + " ";
                }
                Util::msg(1,s);
            }
            Util::msg(1,"---- over pid2stvars ----");
        }

        Mem_map() {  }
        void add_loc_var(int pid, const Var_dcl& vd) {
            int no_vars;
            no_vars = pid2pos[pid].size(); 
            no_vars++;
            pid2pos[pid][vd]= no_vars-1;
        }

        void add_glb_var(const Var_dcl& vd) {
            int no_vars;
            no_vars = glbvar2pos.size(); 
            no_vars++;
            glbvar2pos[vd]= no_vars-1;
        }

        void add_ev_var(const Var_dcl& vd) {
            int no_vars;
            no_vars = glbvar2pos.size(); 
            glbvar2pos[vd]= no_vars;
            evvar_pos.insert(no_vars);
            no_vars++;
        }
        void get_pids( list<Pid>& l) const { 
            const_foreach(it, pid2pos) { l.push_back(it->first); }
        }

        int get_num_loc_vars(int pid) const { 
            CHK(pid2pos.find(pid) != pid2pos.end());
            return (pid2pos.find(pid)->second).size();
        } 
        int get_num_glb_vars() const { return glbvar2pos.size();}
 //       int get_num_ev_vars() const { return evvar2pos.size();}
 //      int get_num_chan_vars() const { return chanvar2pos.size();}

        int get_loc_pos(int pid, const Var_dcl& vd) const {
            map<Var_dcl,int>::const_iterator it = (pid2pos.find(pid)->second).find(vd);
            assert(it!= (pid2pos.find(pid)->second).end());
            return it->second; 
        }

        int get_glb_pos(const Var_dcl& vd) const{
            map<Var_dcl,int>::const_iterator it = glbvar2pos.find(vd);
            assert(it!= glbvar2pos.end());
            return it->second; 
        }

        const set<int>& get_evvar_pos() const { return evvar_pos; }
        const map<Var_dcl,int>& get_loc_var_map(int pid) const { return pid2pos.find(pid)->second; }
        const map<Var_dcl,int>& get_glb_var_map() const { return glbvar2pos; }

        bool is_glb_var(const Var_dcl& vd) const { return glbvar2pos.find(vd) != glbvar2pos.end(); }

        bool is_loc_var(int pid, const Var_dcl& vd) const { 
            if( pid2pos.find(pid) == pid2pos.end() ) return false;
            const map<Var_dcl,int>& pos_map = (pid2pos.find(pid)->second);
            if ( pos_map.find(vd) != pos_map.end() ) return true;
            return false;
        }

        bool has_pid ( Pid pid ) const { return pid2pos.find(pid) != pid2pos.end(); }

   //     bool is_chan_var(const Var_dcl& vd) const { return (chanvar2pos.find(vd) != chanvar2pos.end()); }
    //    bool is_ev_var(const Var_dcl& vd) const { return (evvar2pos.find(vd) != evvar2pos.end()); }

        int size() const { return  pid2pos.size() + glbvar2pos.size() ; } //TODO-ASSMP, assume that this is fine

        bool empty() const { 
            return pid2pos.empty() && glbvar2pos.empty() ;
        }

        template<class T>
        void add_conc_val(Pid pid, const Var_dcl& v, const T& oracle, Conc_sd& csd) const {
                list<Const> cl;
                oracle.get_range_values(pid,v,cl);
                //Consts& cs = csd[v];
                //cs.insert(cl.begin(), cl.end());
                csd.add_vals(pid,v,cl);
        }

        template<class T>
        void get_conc_sd(const T& oracle, Conc_sd& csd) const {
            const_foreach(it, pid2pos) {
                Pid pid = it->first;
                const map<Var_dcl,int>& var2pos = it->second;
                const_foreach(it2, var2pos) { add_conc_val(pid, it2->first, oracle, csd); }
            }
            const_foreach(it, glbvar2pos) { add_conc_val(-1, it->first, oracle, csd); }
        }

        void set_state_vars ( const map<Pid, Var_dcls >& _pid2stvars ) {
            pid2stvars = _pid2stvars;
        }
        void add_state_vars ( Pid pid, const Var_dcls& vds ) {
            pid2stvars[pid].insert( vds.begin(), vds.end() );
        }


        bool is_vd_st_var ( Pid pid, const Var_dcl& vd ) const ;

        template<class DP, class T>
            void DP_declare_one_var(DP& dp, const T& oracle, 
                                    Pid pid, const Var_dcl& vd, int depth) const {
                if ( vd.to_str() == "pc" ) return;
                //DP::DP_type ty = oracle.get_DP_ty(dp,vd,pid);
                Ty ty = oracle.get_ty(vd,pid);
                string name = oracle.get_glb_name(vd,pid,depth);
                dp.mk_var_decl(name,ty);
                //Util::msg(2,"declaring var "+name);
            }

        template<class DP, class T>
            void DP_declare_vars(DP& dp, const T& oracle, int depth) const {
                const_foreach(it, pid2pos) {
                    Pid pid = it->first;
                    const map<Var_dcl,int>& var2pos = it->second;
                    const_foreach(it2, var2pos) { 
                        const Var_dcl& vd = it2->first;
                        DP_declare_one_var(dp,oracle,pid,vd,depth);
                    }
                }
                const_foreach(it, glbvar2pos) { DP_declare_one_var(dp,oracle, -1, it->first, depth); }
            }

        template<class DP, class T>
            typename DP::expr get_expr_for_vd ( DP& dp, const T& oracle, 
                                    Pid pid, const Var_dcl& vd, int depth) const {
                string name = oracle.get_glb_name(vd,pid,depth);
                return dp.get_var_expr(name);
            }

    
        DP::expr get_diff_constraints_var(DP& dp, DP::expr e0, DP::expr e1) const {
            DP::expr res =  dp.mk_diseq(e0,e1);
            return res;
        }



            //improve by checking only state vars and pc
        template<class T>
            DP::expr get_diff_constraints(DP& dp, const T& oracle, int d0, int d1) const {
                DP::expr res = dp.mk_false();
                DP::expr a[2];
                CHK ( !pid2stvars.empty() );

                const_foreach(it, pid2pos) {
                    Pid pid = it->first;
                    const map<Var_dcl,int>& var2pos = it->second;
                    const_foreach(it2, var2pos) { 
                        const Var_dcl& vd = it2->first;
                        if ( vd.to_str() == "pc" ) continue;
                        if ( ! is_vd_st_var(pid,vd) ) continue;

                        DP::expr e0 = get_expr_for_vd(dp,oracle,pid, vd, d0);
                        DP::expr e1 = get_expr_for_vd(dp,oracle,pid, vd, d1);

                        a[0] = res;
                        a[1] = get_diff_constraints_var(dp,e0,e1);
                        res = dp.mk_or(a, 2);

                        Ty ty = oracle.get_ty(vd,pid);
                        //res2 = dp.mk_and( res2, get_bv_constraints(dp,e0,e1,ty) );
                    }
                }
                const_foreach(it, glbvar2pos) { 
                    const Var_dcl& vd = it->first;
                    if ( ! is_vd_st_var(-1,vd) ) continue;
                    DP::expr e0 = get_expr_for_vd(dp,oracle,-1, vd, d0);
                    DP::expr e1 = get_expr_for_vd(dp,oracle,-1, vd, d1);
                    a[0] = res;
                    a[1] = get_diff_constraints_var(dp,e0,e1); 
                    res = dp.mk_or(a, 2);

                    Ty ty = oracle.get_ty(vd,-1);
                   // res2 = dp.mk_and( res2, get_bv_constraints(dp,e0,e1,ty) );
                }
                //res = dp.mk_and(res,res2);
                return res;
            }
        
        /*template<class DP>
        typename DP::expr get_bv_constraints ( DP& dp, typename DP::expr e0, typename DP::expr e1, Ty ty ) const {
            typename DP::expr res = dp.mk_true();
            if ( is_ty_bv(ty) ) {
                int ub = get_ty_size(ty);
                //Util::msg(2,"adding bv constraints..");

                res = dp.mk_and( res, dp.mk_ge(e0, dp.mk_num(0)) );
                res = dp.mk_and( res, dp.mk_gt( dp.mk_num(ub), e0) );

                res = dp.mk_and( res, dp.mk_ge(e1, dp.mk_num(0)) );
                res = dp.mk_and( res, dp.mk_gt( dp.mk_num(ub), e1) );
            }
            return res;
        }*/

        template<class DP, class T>
            typename DP::expr get_bv_constraints(DP& dp, const T& oracle, Pid pid, 
                                                const Var_dcl& vd, int depth) const {
            typename DP::expr res = dp.mk_true() ;
            Ty ty = oracle.get_ty(vd,pid);

            if ( is_ty_bv(ty) ) {
                int ub = get_ty_size(ty);
                typename DP::expr e = get_expr_for_vd(dp,oracle,pid, vd, depth);
                res = dp.mk_ge(e, dp.mk_num(0)) ;
                res = dp.mk_and(res, dp.mk_gt( dp.mk_num(ub), e) );
            }
            return res;
        }

        template<class DP, class T>
           typename DP::expr get_bv_constraints(DP& dp, const T& oracle, int depth) const {
               typename DP::expr res = dp.mk_true();

                const_foreach(it, pid2pos) {
                    Pid pid = it->first;
                    const map<Var_dcl,int>& var2pos = it->second;
                    const_foreach(it2, var2pos) { 
                        const Var_dcl& vd = it2->first;
                        if ( vd.to_str() == "pc" ) continue;
                        res = dp.mk_and( res, get_bv_constraints(dp,oracle,pid,vd,depth) );
                    }
                }

                const_foreach(it, glbvar2pos) { 
                    const Var_dcl& vd = it->first;
                    res = dp.mk_and( res, get_bv_constraints(dp,oracle,-1,vd,depth) );
                }
                return res;
            }

};

//-- Proc_states, Glb_state 
typedef vector<Local_val> Proc_state;

struct Proc_states_diff {
    map<int, map<int,Local_val> > vars_diff; //pid -> (pos -> val)

    Proc_states_diff() {}
    template<class Psd>
    Proc_states_diff(const Psd& psd, const Mem_map& mem) {
        const map<Pid, map<Var_dcl,Const> >& psd_vars_diff = psd.vars_diff;
        const_foreach(it, psd_vars_diff) {
            Pid pid = it->first;
            const map<Var_dcl,Const>& var_map = it->second;
            const_foreach(it2, var_map) {
                size_t pos = mem.get_loc_pos(pid,it2->first);
                set_val(pid, pos, eval_const(it2->second) );
            }
        }
    }
    Local_val get_val(int pid, int pos) const { 
        CHK( vars_diff.find(pid) != vars_diff.end() );
        CHK( vars_diff.find(pid)->second.find(pos) != vars_diff.find(pid)->second.end() );
        return vars_diff.find(pid)->second.find(pos)->second; 
    }

    void set_val(int pid, int pos, Local_val val) { vars_diff[pid][pos]= val; }
    bool empty() const { return vars_diff.empty(); }
    void clear() { vars_diff.clear(); }

    string to_str(const Mem_map& mem) const {
        string s;
        const_foreach(it, vars_diff) {
            Pid pid = it->first;
            map<int, map<int,Local_val> >::const_iterator it = vars_diff.find(pid);
            if( it == vars_diff.end()) continue;
            s+="["+int2str(pid)+"] ";

            const map<Var_dcl,int>& var_map = mem.get_loc_var_map(pid);
            const_foreach(itv, var_map) {
                map<int, Local_val>::const_iterator it2 = it->second.find(itv->second);
                if( it2 == it->second.end()) continue;
                s += itv->first.to_str() + ":" + it2->second.to_str() + " ";
            }
        }
        return s;
    }

                    //this is conjunctive add
    bool add( const Proc_states_diff& psd, Upd_mode U )  {
        for(map<int, map<int,Local_val> >::const_iterator it = psd.vars_diff.begin(); it != psd.vars_diff.end(); ++it) {
            int pid = it->first;
            const map<int,Local_val>& pos2val = it->second; 
            const_foreach(it2, pos2val) {
                int pos = it2->first;
                Local_val val = it2->second;
                if ( vars_diff[pid].find(pos) == vars_diff[pid].end() ) {
                     set_val(pid,pos,val);
                }
                else if ( U == OW ) {
                    if ( !val.is_x() ) set_val(pid,pos,val);
                    //irrespective of prev val
                }
                else if ( U == NO_OW) {
                    Val prev_val = get_val(pid,pos);
                    tbool tb = (prev_val != val);
                    if ( tb.is_t() ) return false;
                    //if is_f, then vals are same, do nothing
                    if ( tb.is_x() && prev_val.is_x() ) set_val(pid,pos,val); //irresp of new val
                }
                else assert(0);
            }
        }
        return true;
    }

    //not clear, what happens when pids are not the same for the two ps_diffs
    void project_to( Proc_states_diff& psd ) const {
        for(map<int, map<int,Local_val> >::const_iterator it = psd.vars_diff.begin(); it != psd.vars_diff.end(); ++it) {
            int pid = it->first;
            const map<int,Local_val>& pos2val = it->second; 
            const_foreach(it2, pos2val) {
                int pos = it2->first;
                psd.set_val( pid, pos, get_val(pid,pos) );
            }
        }
    }

};

struct Proc_states {
    map<int, Proc_state > loc_vals; //pid -> vector of loc_var values

    void init_pid_size(Pid pid, uint size) { 
        Proc_state& ls = loc_vals[pid];
        ls.resize(size);
        //for(uint i=0; i<size; i++)  ls[i]=0; //TODO-ASSMP
    }

    void init_size(const Mem_map& mem) {
        list<Pid> pid_list;
        mem.get_pids(pid_list);
        foreach(it, pid_list) {
            Pid pid = *it;
            init_pid_size(pid, mem.get_num_loc_vars(pid));
        }
    }


    //destructive in-place check/update
    /*bool apply_diff(const Proc_states_diff& psd, Upd_mode m=OW) {
        for(map<int, map<int,Local_val> >::const_iterator it = psd.vars_diff.begin(); 
                it != psd.vars_diff.end(); ++it) {

            int pid = it->first;
            const map<int,Local_val>& pos2val = it->second; 
            Util::msg(5,"pid is "+int2str(pid));
            CHK ( loc_vals.find(pid) != loc_vals.end()) ;

            Proc_state& ls = (loc_vals[pid]);

            const_foreach(it2, pos2val) {
                if( m == NO_OW ) {
                    if( ls[it2->first] != it2->second ) return false;
                }
                else if ( m == OW ) ls[it2->first] = it2->second;
                else assert(0); 
            }
        }
        return true;
    }*/

    void apply_diff(const Proc_states_diff& psd) {
        for(map<int, map<int,Local_val> >::const_iterator it = psd.vars_diff.begin(); 
                it != psd.vars_diff.end(); ++it) {

            int pid = it->first;
            const map<int,Local_val>& pos2val = it->second; 
            Util::msg(5,"pid is "+int2str(pid));
            CHK ( loc_vals.find(pid) != loc_vals.end()) ;

            Proc_state& ls = (loc_vals[pid]);
            const_foreach(it2, pos2val) {
                ls[it2->first] = it2->second;
            }
        }
    }

    /*void get_diff_eqv( const Proc_states& ps2, Proc_states_diff& pdiff, Proc_states_diff& peqv ) const {
        const map<int, Proc_state>& loc_vals2 = ps2.loc_vals;
        const_foreach(it, loc_vals) {
            int pid = it->first;
            CHK( loc_vals2.find(pid) != loc_vals2.end() );
            const Proc_state& ls = it->second;
            const Proc_state& ls2 = loc_vals2.find(pid)->second;

            //get diff and eqv of ls and ls2
            //ls.get_diff(ls2, pd);
            CHK(ls.size() == ls2.size());
            for(size_t pos=0; pos< ls.size(); ++pos) {
                if( ls[pos] != ls2[pos]) pdiff.set_val(pid,pos,ls2[pos]);
                else peqv.set_val(pid,pos,ls[pos]);
            }
        }
    }*/

    template<class Psd>
    void convert_to_diff( Psd& psd, const Mem_map& mem ) const {
        const_foreach(it, loc_vals) {
            Pid pid = it->first;
            const map<Var_dcl,int>& var_map = mem.get_loc_var_map(pid);
            const_foreach(it, var_map) {
                const Var_dcl& vd= it->first;
                size_t pos = it->second;
                psd.set_val(pid,vd,Const(get_loc_val(pid,pos).to_int()));
            }
        }
    }

    Local_val get_loc_val(int pid, int pos) const { return (loc_vals.find(pid)->second)[pos]; }

    void set_val(int pid, int pos, Local_val pc_val) {
        (loc_vals.find(pid)->second)[pos] = pc_val; 
    }

    const Proc_state& get_proc_state(Pid pid) const {
        CHK ( loc_vals.find(pid) != loc_vals.end() );
        return loc_vals.find(pid)->second;
    }
    void set_proc_state( Pid pid, const Proc_state& ps ) { loc_vals[pid] = ps; }

    void get_X_mem_map(const Mem_map& mem, Mem_map& imem) const {
        const_foreach(it, loc_vals) {
            Pid pid = it->first;
            const map<Var_dcl,int>& var_map = mem.get_loc_var_map(pid);
            const_foreach(it, var_map) {
                if ( get_loc_val(pid,it->second).is_x() ) {
         //       Util::msg(2,"getXmemmap: adding loc var " + it->first.to_str());
                    imem.add_loc_var(pid,it->first);
                }
          //      else Util::msg(2,"var "+it->first.to_str()+ " is not x");
            }
        }
    }

    /*void abstract_using_mem_map ( const Mem_map& mem, const Mem_map& imem, Proc_states& pss ) const {
        const_foreach(it, loc_vals) {
            Pid pid = it->first;
            if ( !imem.has_pid(pid) ) continue;
            const map<Var_dcl,int>& var_map = mem.get_loc_var_map(pid);
            const map<Var_dcl,int>& ivar_map = imem.get_loc_var_map(pid);
            CHK ( var_map.size() == ivar_map.size() );
            pss.set_proc_state(pid, get_proc_state(pid) );
        }

    }*/

    void clone_abs ( const Mem_map& my_mem, const Mem_map& o_mem, const Proc_states& pss ) {
        init_size(my_mem);
        const_foreach(it, loc_vals) {
            Pid pid = it->first;
            CHK ( o_mem.has_pid(pid) ) ;
            const map<Var_dcl,int>& my_var_map = my_mem.get_loc_var_map(pid);
            const map<Var_dcl,int>& o_var_map = o_mem.get_loc_var_map(pid);
            CHK ( my_var_map.size() == o_var_map.size() );
            set_proc_state(pid, pss.get_proc_state(pid) );
        }

    }
    string to_str(const Mem_map& mem) const {
        string s;
        const_foreach(it, loc_vals) {
            Pid pid = it->first;
            s+="["+int2str(pid)+"] ";
            const map<Var_dcl,int>& var_map = mem.get_loc_var_map(pid);
            const_foreach(it, var_map) {
                s += it->first.to_str() + ":" + get_loc_val(pid,it->second).to_str() + " ";
            }
        }
        return s;
    }
};

struct Glb_state_diff {
    map<int, Global_val> vars_diff; //pos to new val
    //map<int, Ev_val> ev_vars_diff; //pos to new val
    //map<int, Chan_val> qs_diff; //pos to new q val
    
    Glb_state_diff() {}
    template<class Gsd>
        Glb_state_diff(Gsd& gsd, const Mem_map& mem) {
            const map<Var_dcl,Const>& gsd_vars_diff = gsd.vars_diff;
            const_foreach(it, gsd_vars_diff) {
                size_t pos = mem.get_glb_pos(it->first);
                set_glb_val(pos, eval_const(it->second) );
            }
        }
    bool empty() const { return (vars_diff.empty()); }

    Global_val get_glb_val(int offset) const { return vars_diff.find(offset)->second; }
    //const Ev_val& get_ev_val(int offset) const { return ev_vars_diff.find(offset)->second; }
    //const Chan_val& get_chan_val(int offset) const { return qs_diff.find(offset)->second; }

    void set_glb_val(int pos, const Global_val& glb_val) { vars_diff[pos] = glb_val; }
    //void set_ev_val(int pos, const Ev_val& ev_val) { ev_vars_diff[pos] = ev_val; }
    //void set_chan_val(int pos, const Chan_val& qval) { qs_diff[pos] = qval; }

    string to_str(const Mem_map& mem) const {
        string s;
        if( vars_diff.empty() ) return s;

        const map<Var_dcl,int>& var_map = mem.get_glb_var_map();
        const_foreach(it, var_map) {
            if( vars_diff.find(it->second) == vars_diff.end()) continue;
            s += it->first.to_str() + ":" + get_glb_val(it->second).to_str() + " ";
        }

        return s;
    }

    bool add (const Glb_state_diff& gsd, Upd_mode U) {
        const map<int, Global_val>& pos2gv = gsd.vars_diff;
        const_foreach(it, pos2gv) {
            int pos = it->first;
            const Val& val = it->second;
            if ( vars_diff.find(it->first) == vars_diff.end() ) set_glb_val(pos,val);
            else if ( U == OW ) {
                if ( !val.is_x() ) set_glb_val(pos,val);
                //irrespective of prev val
            }
            else if ( U == NO_OW) {
                Val prev_val = get_glb_val(pos);
                tbool tb = (prev_val != val);
                if ( tb.is_t() ) return false;
                //if is_f, then vals are same, do nothing
                if ( tb.is_x() && prev_val.is_x() ) set_glb_val(pos,val); //irresp of new val
            }
            else assert(0);
        }

        return true;
    }

    void project_to( Glb_state_diff& gsd ) const {
        const map<int, Global_val>& pos2gv = gsd.vars_diff;
        const_foreach(it, pos2gv) {
            CHK ( vars_diff.find(it->first) != vars_diff.end());
            gsd.set_glb_val( it->first, get_glb_val(it->first) );
        }

    }

    void get_rwm (const Mem_map& mem, RW_map::RW_enum mode, RW_map& rwm) const {

        const map<Var_dcl,int>& var_map = mem.get_glb_var_map();
        const_foreach(it, var_map) {
            if( vars_diff.find(it->second) == vars_diff.end()) continue;
            rwm.add_val(mode, it->first, get_glb_val(it->second).to_int() );
        }
    }

};

struct Glb_state { //TODO  handle queues
    vector<Global_val> vars; //global vars
    //vector<Ev_val> ev_vars; // event vars
    //vector<Chan_val> qs; //queue vals


    void init_size( const Mem_map& mem ) { 
        int glb_val_num = mem.get_num_glb_vars();
        vars.resize(glb_val_num); 
    }
    
    void init_val() { foreach(it,vars) *it = 0; //todo-assmp 
    }

    Global_val get_glb_val(size_t offset) const { CHK( offset < vars.size()); return vars[offset]; }
    //const Ev_val& get_ev_val(size_t offset) const { CHK( offset < ev_vars.size()); return ev_vars[offset]; }
    //const Chan_val& get_chan_val(size_t offset) const { CHK( offset < qs.size()); return qs[offset]; }

    void set_val(int pos, Global_val glb_val) { vars[pos] = glb_val; }
    //void set_ev_val(int pos, Ev_val ev_val) { ev_vars[pos] = ev_val; }
    //void set_chan_val(int pos, const Chan_val& qval) { qs[pos] = qval;}

    /*bool apply_diff(const Glb_state_diff& gsd, Upd_mode m = OW) {
        const map<int, Global_val>& pos2gv = gsd.vars_diff;
        const_foreach(it, pos2gv) {
            if( m == OW) set_val(it->first,it->second);
            else if( m == NO_OW ) {
                if ( get_glb_val(it->first) != it->second ) return false;
            }
            else assert(0);
        }
        return true;
    }*/

    void apply_diff(const Glb_state_diff& gsd) {
        const map<int, Global_val>& pos2gv = gsd.vars_diff;
        const_foreach(it, pos2gv) {
            set_val(it->first,it->second);
        }
    }

    /*void get_diff_eqv( const Glb_state& gs2, Glb_state_diff& gdiff, Glb_state_diff& geqv ) const {
        const vector<Global_val>& vars2 = gs2.vars;
        CHK( vars.size() == vars2.size() );
        for(size_t pos = 0; pos<vars.size(); ++pos) {
            if( vars[pos] != vars2[pos] ) gdiff.set_glb_val(pos, vars2[pos]);
            else geqv.set_glb_val(pos, vars2[pos]);
        }
    }*/

    bool agrees_with( const Glb_state_diff& gsd) const {
        const map<int, Global_val>& pos2gv = gsd.vars_diff;
        const_foreach(it, pos2gv) {
            const Val& val1 = get_glb_val(it->first);
            const Val& val2 = it->second;
            tbool tb = (val1 != val2);
            if ( tb.is_t() ) return false;
            //otherwise, they actually match or may match
            
            //if( get_glb_val(it->first) != it->second ) return false;
        }
        return true;
    }

    template<class Gsd>
    void convert_to_diff(Gsd& gsd, const Mem_map& mem) const {
        const map<Var_dcl,int>& var_map = mem.get_glb_var_map();
        const_foreach(it, var_map) {
            const Var_dcl& vd= it->first;
            size_t pos = it->second;
            CHK( pos < vars.size() );
            gsd.set_val(vd, Const(vars[pos].to_int()));
        }
    }


    void clear_events(const Mem_map& mem) { 
        const set<int>& evvar_pos = mem.get_evvar_pos();
        const_foreach(it, evvar_pos) {  set_val(*it, 0); }
    }

    void get_X_mem_map(const Mem_map& mem, Mem_map& imem) const {
        const map<Var_dcl,int>& var_map = mem.get_glb_var_map();
        const_foreach(it, var_map) {
            if( get_glb_val(it->second).is_x() ) {
         //       Util::msg(2,"getXmemmap: adding glb var " + it->first.to_str());
                imem.add_glb_var(it->first);
            }
        }
    }

    //merge abstract using and clone_abs
    /*
    void abstract_using_mem_map ( const Mem_map& mem, const Mem_map& imem, Glb_state& gs ) const {
        const map<Var_dcl,int>& var_map = mem.get_glb_var_map();
        const map<Var_dcl,int>& ivar_map = imem.get_glb_var_map();
        const_foreach(it, var_map) {
            if ( ivar_map.find (it->first) != ivar_map.end() ) {
                Val val = get_glb_val(it->second);
                size_t pos = imem.get_glb_pos(it->first);
                gs.set_val( pos, val);
            }
        }
    }*/

    void clone_abs ( const Mem_map& my_mem, const Mem_map& o_mem, const Glb_state& gs ) {
        init_size(my_mem);
        const map<Var_dcl,int>& my_var_map = my_mem.get_glb_var_map();
        const map<Var_dcl,int>& o_var_map = o_mem.get_glb_var_map();
        const_foreach(it, my_var_map) {
            CHK ( o_var_map.find (it->first) != o_var_map.end() ) ;
            size_t o_pos = o_mem.get_glb_pos(it->first);
            Val val = gs.get_glb_val(o_pos);
            size_t my_pos = it->second;
            set_val( my_pos, val);
        }
    }


    string to_str(const Mem_map& mem) const {
        string s;
        const map<Var_dcl,int>& var_map = mem.get_glb_var_map();
        const_foreach(it, var_map) {
            s += it->first.to_str() + ":" + (get_glb_val(it->second)).to_str() + " ";
        }
        return s;
    }
};

struct CFA_states_diff {
    map<int, CFA_state> pos2st;
    bool empty() const { return pos2st.empty(); }
    void clear() { pos2st.clear(); }

    bool has_state(int pos) const {
        return ( pos2st.find(pos) != pos2st.end() ) ;
    }

    const CFA_state& get_state (int pos) const { 
        //Util::msg(2,"cfa_states_diff: getting state pos "+ int2str(pos));
        CHK ( has_state(pos) ); 
        //const CFA_state& st = pos2st.find(pos)->second; 
        //Util::msg(2,"cfa_states_diff: State is "+ st.to_str());
        return pos2st.find(pos)->second;
    }

    void set_state (int pos, const CFA_state& st) { 
        //Util::msg(2,"cfa_states_diff: setting state pos "+ int2str(pos) + "to "+ st.to_str());
        pos2st[pos] = st; 
    }

    string to_str(const Mem_map& mem) const {
        string s;
        const_foreach(it, pos2st) {
           // s += int2str(mem.get_fa_pid(it->first)) + ":" ;
            s += it->second.to_str() + " ";
        }
        return s;
    }

    bool add( const CFA_states_diff& csd, Upd_mode U) {
        const map<int, CFA_state>& _pos2st = csd.pos2st;
        const_foreach(it, _pos2st) {
            if ( (!has_state(it->first)) || U == OW)
                set_state(it->first,it->second);
            else {
                if ( ! (get_state(it->first) == it->second) ) 
                return false;
            }
        }
        return true;
    }
};

struct CFA_states  {
    map<Pid, CFA_state> pos2st; 
    typedef map<Pid, CFA_state> Super;
    public:
    CFA_states () {}
    //CFA_states ( Super &sup ) : pos2(sup) {}
    //void init_size(size_t size) { Super::resize(size); }
    bool has_pos(size_t pos) const { return pos2st.find(pos) != pos2st.end() ; }
    void set_state(size_t pos, const CFA_state& st) {  
        //Util::msg(2,"cfa_states: setting state pos "+ int2str(pos) + "to "+ st.to_str());
        pos2st[pos] = st; 
    }
    const CFA_state& get_state ( int pos ) const { 
        //Util::msg(2,"cfa_states: getting state pos "+ int2str(pos) );
        CHK( has_pos(pos)); 
        //const CFA_state& st = pos2st.find(pos)->second; 
        //Util::msg(2,"cfa_states: State is "+ st.to_str() );
        return pos2st.find(pos)->second; 
    }

    /*bool apply_diff(const CFA_states_diff& cfas_diff, Upd_mode m = OW) {
        const map<int, CFA_state>& _pos2st = cfas_diff.pos2st;
        const_foreach(it, _pos2st) {
            if ( m == OW ) {
                set_state(it->first,it->second);
            }
            else if (m == NO_OW) {
                if(!(get_state(it->first) == it->second)) return false;
            }
            else assert(0);
        }
        return true;
    }*/
    void apply_diff(const CFA_states_diff& cfas_diff) {
        const map<int, CFA_state>& _pos2st = cfas_diff.pos2st;
        const_foreach(it, _pos2st) {
            set_state(it->first,it->second);
        }
    }
    string to_str(const Mem_map& mem) const {
        string s = " ";
        const_foreach(it, pos2st) {
            Pid pid = it->first;
            s += "{" + int2str( pid ) + ": "+ it->second.to_str() + "} ";
        }
        return s;
    }
    bool empty() const { return pos2st.empty(); }

};

struct Prop_state_diff {
    Prop_state ps; //TODO: try if set<state> works
    public:
    Prop_state_diff() {}
    const Prop_state& get_state() const { CHK(!empty()); return ps; }
    void set_state(const Prop_state& _ps) { ps = _ps; }
    bool empty() const { return (ps == Prop_state()); }
    void clear() { ps = Prop_state(); }

    string to_str() const { string s; if( !empty() ) s = ps.to_str(); return s;}
    bool add( const Prop_state_diff& prd, Upd_mode U) {
        if ( prd.empty() ) return true;
        if ( empty() || U==OW) set_state(prd.get_state() );
        else {
            if ( ! (get_state() == prd.get_state()) ) return false;
        }
        return true;
    }

};

//----- Sys_state_diff

class Sys_state_diff {
    protected:
        const Mem_map* mem;
        Proc_states_diff ps_diff;
        Glb_state_diff gs_diff;
        CFA_states_diff cfas_diff;
        Prop_state_diff prs_diff;
    public:
        Sys_state_diff(const Mem_map* _mem) : mem(_mem) {}
        Sys_state_diff(const Mem_map& _mem) : mem(&_mem) {}
        template<class Diff>
        Sys_state_diff(const Mem_map& _mem, const Diff& d) 
            : mem(&_mem), ps_diff(d.get_psd(),*mem), gs_diff(d.get_gsd(), *mem) {
        }

        const Mem_map& get_mem() const { CHK(mem != NULL); return *mem;}
        void get_diff(Sys_state_diff& sd) const {  
            //Util::msg(2,"sys_state_diff::get_diff add_check");
            sd.add_check(*this) ;  }
        void set_mem(const Mem_map& _mem) { mem = &_mem; }


        void set_pcval(int pid, Local_val pc_val) {
            CHK(pid != -1);
            CHK(mem != NULL);
            int offset = mem->get_loc_pos(pid,Var_dcl(Symbol("pc")));
            ps_diff.set_val(pid, offset, pc_val);
        }

        void set_loc_val(int pid, const  Var_dcl& vd, Local_val val) {
            CHK(mem != NULL);
            int offset = mem->get_loc_pos(pid,vd);
            return ps_diff.set_val(pid,offset,val);
        }

        void set_glb_val(const Var_dcl& vd, Global_val val) {
            CHK(mem != NULL);
            int offset = mem->get_glb_pos(vd);
            return gs_diff.set_glb_val(offset,val);
        }

/*        void set_ev_val(const Var_dcl& vd, Ev_val val) {
            CHK(mem != NULL);
            int offset = mem->get_ev_pos(vd);
            return gs_diff.set_ev_val(offset,val);
        }

        void set_chan_val(const Var_dcl& vd, const Chan_val& qval) {
            CHK(mem != NULL);
            if( mem->is_chan_var(vd) ) {
                int offset = mem->get_chan_pos(vd);
                return gs_diff.set_chan_val(offset,qval);
            }
            else if ( mem->is_ev_var(vd) ) set_ev_val(vd,qval);
            else assert(0);
        }
*/
        void set_val(int pid, const Var_dcl& vd, const Val& val) {
            CHK(mem != NULL);
            if( mem->is_glb_var(vd) ) set_glb_val(vd,val);
            else if (pid != -1 && mem->is_loc_var(pid,vd) ) set_loc_val(pid,vd,val);
            else assert(0);
        }
        void set_val(Pid pid, const Var_dcl& vd, const Const& c) {
            set_val(pid,vd,Val(c.to_num()));
        }
        void set_val(const Var_dcl& vd, const Val& v) {
            set_val(-1,vd,v);
        }
        /*void set_val(int pid, const Var_dcl& vd, const Const& c) {
            set_val(pid, vd, Val(c.to_num()));
        }*/


        void set_cfa_state (int pid, const CFA_state& st) {
            CHK(mem != NULL);
            //int pos = mem->get_fa_pos(pid);
            cfas_diff.set_state(pid, st);
        }

        void set_prop_state(const Prop_state& ps) { prs_diff.set_state(ps); }
        const Prop_state& get_prop_state () const { return prs_diff.get_state(); }

        bool empty() const { 
            CHK(mem != NULL);
            return (ps_diff.empty() && gs_diff.empty() && cfas_diff.empty() && prs_diff.empty() );
        }

        bool add(const Sys_state_diff& s_diff, Upd_mode U) ;
        bool add_check( const Sys_state_diff& sd, Upd_mode U = NO_OW) { 
            //Util::msg(2,"****************** add_check: U is "+int2str(U)); 
            bool ret = add(sd,U); 
            //if( !ret ) fprintf(stderr,"*********************************** add_check: ret is 0\n");
            return ret;
        }
        void update_with( const Sys_state_diff& sd) { add_check(sd,OW); }
        
        void abstract_to_globals ( Sys_state_diff& sd ) const;

        void get_rwm (RW_map::RW_enum mode, RW_map& rw) const {
            CHK ( mem != NULL );
            gs_diff.get_rwm (*mem, mode, rw);
        }

        string to_str() const { 
            //if( mem == NULL) return "":
            CHK(mem != NULL);
            string s = "(" + ps_diff.to_str(*mem) + "| "+ gs_diff.to_str(*mem) + "| "+ cfas_diff.to_str(*mem)+ "| "+ prs_diff.to_str()+")"; 
            return s;
        }

        bool operator < (const Sys_state_diff& s_diff) const { return (to_str() < s_diff.to_str()); }
        friend class Sys_state;
};

typedef set<Sys_state_diff> Sys_state_diffs;


//----- Sys_state :
class Sys_state : public Node {
    protected:
        const Mem_map* mem;
        Proc_states proc_states; //local states of cfgs
        Glb_state gs; //global state of cfgs
        CFA_states  cfa_states; //!!FIX: set up diffs for this
        Prop_state ps;

        //clones and if necessary, abstracts 
        void clone_abs ( const Sys_state& s) {
            //CHK ( mem != NULL) ; 
            CHK ( s.mem != NULL );
            if( mem == NULL ) mem = s.mem; //can introduce problems
            if ( mem != s.mem ) {
                proc_states.clone_abs(*mem, *(s.mem), s.proc_states);
                gs.clone_abs(*mem, *(s.mem), s.gs);
            }
            else {
                proc_states = s.proc_states;
                gs = s.gs;
            }
            cfa_states = s.cfa_states;
            ps = s.ps;
        }

    public:
        friend class Sys_state_diff;
        typedef Sys_state_diff Diff;
        
        Sys_state() { mem = NULL;}
        Sys_state ( const Mem_map& _mem) {  mem = &_mem; }
        void set_mem(const Mem_map& _mem) { CHK ( mem == NULL);  mem = (&_mem); }

        Sys_state ( const Sys_state& s) {
            mem = s.mem; clone_abs(s);
        }

        const Sys_state& operator = (const Sys_state& s) {
            clone_abs(s); return *this;
        }

        //initialize sizes
        void initialize() {
            proc_states.init_size(*mem);

            //gs.init_size(mem->get_num_glb_vars(), mem->get_num_ev_vars(), mem->get_num_chan_vars());
            gs.init_size(*mem);
            //gs.init_val();

            //cfa_states.init_size( mem->get_num_fas() );
        }

        //use for initializing also		
        void set_cfa_state (int pid, const CFA_state& st) {
            //int pos = mem->get_fa_pos(pid);
            cfa_states.set_state(pid, st);
        }
        const CFA_state& get_cfa_state (int pid) const {
            //int pos = mem->get_fa_pos(pid);
            return cfa_states.get_state(pid);
        }

        void set_prop_state(const Prop_state& _ps) { ps = _ps;}
        const Prop_state& get_prop_state() const { return ps;}

        const Mem_map& get_mem() const { return *mem;}

        int get_pcval(int pid) const { 
            CHK(pid != -1);
            return eval(pid,Var_dcl(Symbol("pc"))).to_int();
        }


        void set_pcval(int pid, Local_val pc_val) {
            CHK(pid != -1);
            set_loc_val(pid, Var_dcl(Symbol("pc")), pc_val);
        }


        Val eval(int pid, const Var_dcl& vd) const {
            if(mem->is_glb_var(vd)) {//global
                int offset = mem->get_glb_pos(vd);
                return gs.get_glb_val(offset);
            }
            else if( mem->is_loc_var(pid,vd) ) {//local
                int offset = mem->get_loc_pos(pid,vd);
                return proc_states.get_loc_val(pid,offset);
            }
            else {
                Util::error(true,"Var %s is neither global nor local\n",vd.to_str().c_str());
                assert(0);
                return Val();
            }
        }
/*
        const Ev_val& eval_evvar(const Var_dcl& vd) const {
            CHK(mem->is_ev_var(vd)) ;
            int offset = mem->get_ev_pos(vd);
            return gs.get_ev_val(offset);
        }

        const Chan_val& eval_chanvar(const Var_dcl& vd) const {
            //hack for now: distinguish between different sends and recvs
            //on channels, events, globals etc
            if (mem->is_chan_var(vd)) {
            int offset = mem->get_chan_pos(vd);
            return gs.get_chan_val(offset);
            }
            else if( mem->is_ev_var(vd) ) { return eval_evvar(vd); }
            else assert(0);
        }
*/
        void set_val(int pid, const Var_dcl& vd, const Val& val) {
            if( mem->is_glb_var(vd) ) set_glb_val(vd,val);
            else if ( mem->is_loc_var(pid,vd) ) set_loc_val(pid,vd,val);
            else assert(0);
        }


        void set_val(Pid pid, const Var_dcl& vd, const Const& c) {
            Val v(c.to_num());
            set_val(pid,vd,v);
        }

        void set_val(const Var_dcl& vd, const Const& c) {
            if ( mem->is_glb_var(vd) ) set_val(-1, vd, c);
            //else set_chan_val(vd,c.to_num());
            else assert(0);
        }

        void set_loc_val(int pid, const  Var_dcl& vd, Val val) {
            int offset = mem->get_loc_pos(pid,vd);
            return proc_states.set_val(pid,offset,val);
        }

        void set_glb_val(const Var_dcl& vd, Global_val val) {
            int offset = mem->get_glb_pos(vd);
            return gs.set_val(offset,val);
        }

        // diff = s2 xor s1, eqv = s2 exnor s1
        /*void get_diff_eqv( const Sys_state& s2, Sys_state_diff& sdiff, Sys_state_diff& seqv) const {
            proc_states.get_diff_eqv(s2.proc_states, sdiff.ps_diff, seqv.ps_diff);
            gs.get_diff_eqv(s2.gs, sdiff.gs_diff, seqv.gs_diff);
            //no diff due to cfa_states or prop_state
        }*/

        template<class St_d>
        void convert_to_diff( St_d& sd) const {
            proc_states.convert_to_diff(sd.get_psd(), *mem);
            gs.convert_to_diff(sd.get_gsd(), *mem);
            //no diff due to cfa_states or prop_state
        }
/*
        bool apply_diff(const Sys_state_diff& s_diff, Upd_mode m = OW) {
            bool ret = proc_states.apply_diff(s_diff.ps_diff,m);
            if( !ret ) return false;
            ret = gs.apply_diff(s_diff.gs_diff,m);
            if( !ret ) return false;
            ret = cfa_states.apply_diff(s_diff.cfas_diff,m);
            if( !ret ) return false;
            if( !s_diff.prs_diff.empty()) {
                //if( m == OW ) ps = s_diff.prs_diff.get_state();
                //else if( m == NO_OW ) {
                //    if (! (s_diff.prs_diff.get_state() == ps) ) return false;
                //}
                //else assert(0);
                
                ps = s_diff.prs_diff.get_state();
            }
            return true;
        }
        bool apply_diff_check(const Sys_state_diff& sd, Upd_mode m = OW) {
            bool ret = apply_diff(sd,m);
            if( !ret ) { Util::msg(2,"sys_state:: apply_diff_check: ret is 0");}
            return ret;
        }
*/

        void apply_diff(const Sys_state_diff& s_diff) {
            proc_states.apply_diff(s_diff.ps_diff);
            gs.apply_diff(s_diff.gs_diff);
            cfa_states.apply_diff(s_diff.cfas_diff);
            if( !s_diff.prs_diff.empty()) {
                ps = s_diff.prs_diff.get_state();
            }
        }

        void apply_diff_check(const Sys_state_diff& sd) {
            apply_diff(sd);
        }

        bool agrees_with_global( const Sys_state_diff& sd) const {
            CHK( sd.ps_diff.empty() );
            CHK( sd.cfas_diff.empty() );
            //CHK( sd.prs_diff.empty() );
            return gs.agrees_with(sd.gs_diff);
        }

        void clear_events() { gs.clear_events(*mem); }

        void get_X_mem_map(Mem_map& imem) const {
            proc_states.get_X_mem_map(*mem, imem);
            gs.get_X_mem_map(*mem, imem);
        }

        /*void abstract_using_mem_map ( const Mem_map& imem, Sys_state& s) const {
            proc_states.abstract_using_mem_map(*mem, imem, s.proc_states);
            gs.abstract_using_mem_map(*mem, imem, s.gs);
            CHK ( cfa_states.empty() );
            //CHK ( ps.empty() );
        }*/

        string to_str() const { 
            string s = "(" + proc_states.to_str(*mem)  + "| " + gs.to_str(*mem);
            s += "| "+ cfa_states.to_str(*mem) + "| "+  ps.to_str() + ") "; 
            return s;
        }
        bool operator < (const Sys_state& s) const { return (to_str() < s.to_str()); }
        bool operator == (const Sys_state& s) const { return (to_str() == s.to_str()); }
        bool operator != (const Sys_state& s) const { return (to_str() != s.to_str()); }

};

typedef set<Sys_state> Sys_states;

//Val eval_expr(const Expr* e, const Sys_state& s, int pid) ;
//void eval_act(const Sys_state& s, const CFG_action& act, int pid, Sys_state_diff& s_diff);

#endif 
