#ifndef __TEACHER_H__
#define __TEACHER_H__

#include"fa.h"
#include"sys_expl.h"
#include"sys_expl_sym.h"
#include"stats.h"

template<class Sys, class Sys_CE, class _Prop, class Sys_expl>
class Teacher
{
    public:
        typedef Sys System;
        typedef Sys_expl Sys_explorer;
        typedef typename System::DFAt _DFA;
        typedef typename _DFA::St St;
        typedef typename _DFA::Sig Sig;
        typedef typename _DFA::Trace Trace;
        typedef Trace CE;

        //typedef DFA<St,Sig> _DFA;
        typedef set<Sig> Sigma;
        typedef _Prop Prop;



    protected:
        const System* sys;
        const Prop* spec;
        Sigma sigma;

        set<CE> ce_history;


        //cache for membership queries
        map<Trace,bool> queryCache;

        static int num_q;
        static int num_q_in_cache;


        /*********************************************************************/
        //check for prefixes in cache : only for prefix-closed U
        /*********************************************************************/
        bool do_prefix_closed_optimization(const Trace& t)
        {
            set<Trace> prefixes;
            t.get_prefixes(prefixes);
            //for(set<Trace>::const_iterator it = prefixes.begin();it != prefixes.end();++it) 
            typename_const_foreach(it, prefixes) {
                if((queryCache.find(*it) != queryCache.end()) && !queryCache[*it]) {
                    Util::message(3,"found false prefix %s in query cache ...\n",it->to_str().c_str());
                    queryCache[t] = false;
                    return true;
                }
            }
            return false;
        }


    public: 

        /*********************************************************************/
        //constructors and destructors
        /*********************************************************************/
        Teacher() : sys(NULL), spec(NULL) {  }
        Teacher(const System& _sys, const Prop& _spec, const Sigma& _sigma)
            : sys(&_sys), spec(&_spec), sigma(_sigma) {}

        const Sigma& get_sigma() const { return sigma; }
        Sigma get_basis_sigma() const { Sigma basis; basis.insert(Sig()); return basis; }
        Sig get_basis_sig() const { CHK( !sigma.empty() ); return *(sigma.begin()); }


        //bool check_M_t_satisfy_spec(Trace& t, Sys_CE& ce, bool gen_ce = true, bool lift=false, bool fixpt = true ) const { // M||t sat Spec -> ce
        bool check_M_t_satisfy_spec(Trace& t, Sys_CE& ce, bool gen_ce, bool lift, bool fixpt ) const { // M||t sat Spec -> ce
            _DFA fa(pair<Trace,Sigma>(t,sigma));
            CHK ( sys != NULL);
            Util::msg(2,"check_M_t_satisfy_spec: start build fa..");
            System sys_with_t( *sys, fa );
            Util::msg(2,"check_M_t_satisfy_spec: build fa..now calling check_prop");
            bool res;
            CHK( spec != NULL );
            res = Sys_explorer( sys_with_t ).check_prop( *spec, ce, gen_ce, lift, fixpt );
            Util::message(3,"check_M_t_satisfy_spec: for p=%s, res is %d.\n", t.to_str().c_str(), res);
            return res;
        }

        // rest_ce rest_ce_abs full_ce
        bool check_M_ce_satisfy_spec(const Sys_CE& t, Trace& t_abs, Sys_CE& ce) const { //usual mem q: M||p sat Spec
            //Trace t1_abs ; t.abstract_to_globals(*(sigma.begin()),t1_abs); 
            //t.abstract_to_globals(t_abs); 

            t.abstract_to_globals(*(sigma.begin()),t_abs); 
            
            bool res = check_M_t_satisfy_spec(t_abs,ce,true,false,false);
            //t_abs = t1_abs;
            return res;
        }

        bool ask_mem_q(Trace &t) {
            Statistics::num_mem_q++;
            //first check the cache
            Timer timer; timer.start();
            Util::message(2,"\n\n*********** asking mem query for %s \n",t.to_str().c_str());
            if(queryCache.find(t) != queryCache.end()) {
                Util::message(3,"found %s in cache ...\n",t.to_str().c_str());
                Statistics::num_mem_q_in_cache++;
            } 
            //not found in cache - do it the hard way
            else {
                Util::message(3,"not found %s in cache ...\n",t.to_str().c_str());
                Util::message(3,"checking path %s for containment in U \n",t.to_str().c_str());
                Sys_CE dummy;
                queryCache[t] = check_M_t_satisfy_spec(t,dummy,false,false,false);
            }
            //all done
            timer.stop();
            Util::message(2,"************  end mem query; ret = %d, time taken = %f\n\n\n",queryCache[t],timer.read());
            Statistics::mem_q_time += timer.read();
            return queryCache[t];
        }

        bool ask_cand_q(const _DFA&dfa, CE &ce) {
            Statistics::num_cand_q++;

            Util::message(2,"\nasking candidate query  ...\n");
            Timer timer; timer.start();
            CHK ( sys != NULL);
            System sys_with_dfa( *sys, dfa );
            Sys_CE full_ce;
            CHK(spec != NULL);
            bool res = Sys_explorer( sys_with_dfa ).check_prop( *spec , full_ce , true, false, true);  

            if(!res) {
                /*
                CE ce_abs;
                full_ce.abstract_to_globals(*(sigma.begin()), ce_abs);
                full_ce = Sys_CE();
                //TO GENERALIZE
                //res = check_M_t_satisfy_spec(ce_abs,full_ce,true,true,false);
                res = check_M_t_satisfy_spec(ce_abs,full_ce,true,false,false);
                CHK ( !res );
                */
                Util::msg(2,"sigma-begin is "+(sigma.begin())->to_str_const());
                full_ce.abstract_to_globals(*(sigma.begin()), ce);
                CHK( !(Key_In_Bag(ce,ce_history)) );
                ce_history.insert(ce);
                Util::message(2,"candidate query failed ...");
                //Util::message(3,"CE is : %s ...\n",full_ce.to_str().c_str());
                //full_ce.project_last(sigma,ce);
            } else Util::message(2,"candidate query passed   ...");
            timer.stop();
            Util::message(2,"time taken = %f\n\n\n",timer.read());
            Statistics::cand_q_time += timer.read();
            
            return res;
        }

        bool enlarge_split(const Trace& U, const Trace& V, const Sig& A, Sig& refiner_sig) {
            return Sys_explorer( *sys ).generalize_split( *spec , U , V, A, refiner_sig );  

        }
};


#endif //__TEACHER_H__
