#ifndef UPD_MAP_H
#define UPD_MAP_H

#include"sys_state.h"
#include"id_diff.h"
//--- Update map 

//AND diffs with sd
template<class Upd_diff>
void upd_diffs_by_diff( const set<Upd_diff>& diffs, const Upd_diff& sd, set<Upd_diff>& new_diffs)  {

    if( diffs.empty() ) {  } //earlier diffs intersected to nothing
    else {
        typename_const_foreach(it, diffs) {
            Upd_diff _sd = *it;
            //Util::msg(2,"upd_map: add_check");
            bool ret = _sd.add_check(sd);
            if (ret) new_diffs.insert(_sd);
        }
    }
}

/*bool check_upd_diffs_by_diff( const Sys_state_diffs& diffs, const Sys_state_diff& sd, 
            Sys_state_diffs& new_diffs, Sys_state_diffs& err_diffs) ;
*/


template<class Upd_diff>
class Upd_map {
    protected:
        typedef set<Upd_diff> Upd_diffs;
        typedef map<Sys_state, Upd_diffs > St2Diff;
        St2Diff st2diff;

        void record_update( const Sys_state& s) {

        }

        void add_intersect( const Upd_diff& sd, const Sys_state& s, 
                        Upd_diffs& new_diffs) {
            CHK( !sd.empty() );
            CHK( st2diff.find(s) != st2diff.end());
            /*if ( st2diff.find(s) == st2diff.end() ) {
                new_diffs.insert(sd);
            }
            else */
            upd_diffs_by_diff(st2diff[s], sd, new_diffs);
        }


    public:
        Upd_map() {}

        string to_str() const {
            string s;
            typename_const_foreach( it, st2diff ) {
                const Sys_state& st = it->first;
                const Upd_diffs& upd_diffs = it->second;
                s += st.to_str() + " ::: \n";
                typename_const_foreach(it2, upd_diffs) {
                    s += it2->to_str() + "\n";
                }
            }
            return s;
        }

        bool empty() const { return st2diff.empty(); }
        void get_diffs(const Sys_state& s, Upd_diffs& diffs) const { 
            typename St2Diff::const_iterator it = st2diff.find(s);
            if( it != st2diff.end() ) diffs = it->second;
        }

        void add_union( const Upd_diff& sd, const Sys_state& s) {
            CHK ( !sd.empty() );
            record_update(s);
            //Util::msg(2,"Upd_map::add_union:  sd = " + sd.to_str() );
            st2diff[s].insert(sd);
        }

        void add_intersect( const Upd_map& um ) {
            const St2Diff& st2diff2 = um.st2diff;
            typename_const_foreach( it, st2diff2 ) {
                const Sys_state& s = it->first;
                const Upd_diffs& sds = it->second;
                Upd_diffs new_diffs;
                //Util::msg(2,"{add_intersect for state "+ s.to_str());

                if ( st2diff.find(s) == st2diff.end() ) {
                    new_diffs = sds;
                  //  Util::msg(2,"st2diff empty, new_diffs = "+ set_to_str(new_diffs));
                }
                else {

                    typename_const_foreach( itd, sds ) {
                        add_intersect(*itd, s, new_diffs);
                    }
                   // Util::msg(2,"new_diffs = "+ set_to_str(new_diffs));
                }
                record_update(s);
                st2diff[s] = new_diffs; //new_diffs may be empty
                //Util::msg(2,"}done add_intersect");
            }
        }

        void get_one_image( const Sys_state& old_s, const Upd_diffs& diffs, Sys_states& upd_states) {
            CHK ( !diffs.empty() );
            typename_const_foreach(it2,diffs) {
                Sys_state new_s = old_s;
                const Upd_diff& u_diff = *it2;
                Sys_state::Diff s_diff(new_s.get_mem());
                u_diff.get_diff(s_diff);
                new_s.clear_events();
                new_s.apply_diff_check(s_diff);

                upd_states.insert(new_s);
            }
        }

        void get_image_states(Sys_states& upd_states) { //initially upd_sts has prev states
            typename_const_foreach(it, st2diff) {
                const Sys_state& old_s = it->first;
                CHK( upd_states.find(old_s) != upd_states.end() );
                const Upd_diffs& diffs = it->second;

                if( !diffs.empty() ) { //not already in upd_states
                    upd_states.erase(old_s);
                    get_one_image(old_s,diffs,upd_states); 
                }
            }
        }

        //st2pred : map of reachable states to their predecessors
        //new_states=image\st2pred : frontier for next step
        void get_new_states(set<Sys_state>& new_states, map<Sys_state,Sys_state>& st2pred) {
            typename_const_foreach(it, st2diff) {
                const Sys_state& old_s = it->first;
                const Upd_diffs& diffs = it->second;
                typename_const_foreach(it2,diffs) {
                    Sys_state new_s = old_s;
                    const Upd_diff& u_diff = *it2;
                    new_s.clear_events();
                    Sys_state::Diff s_diff (new_s.get_mem());
                    u_diff.get_diff(s_diff);
                    new_s.apply_diff_check( s_diff );

                    //insert in st2pred
                    if( st2pred.find(new_s) == st2pred.end() ) {
                        st2pred[new_s].set_mem(old_s.get_mem());
                        st2pred[new_s] = old_s;

                        Util::msg(2,"adding new state to new_states " + new_s.to_str() +"\n");
                        Util::msg(2,"s_diff was "+s_diff.to_str());
                        new_states.insert(new_s);
                    }
                }
            }
        }
        
        //get new prop matching states
        void get_new_prop_states ( const Sys_state& new_s, const Sys_state_diffs& prop_diffs,
                                   Sys_states& new_prop_states) {
            const_foreach(itp, prop_diffs) {
                Sys_state new_prop_st = (new_s);
                //if( new_prop_st.apply_diff_check(*itp,NO_OW) ) //if apply_diff_check succeeds
                 if ( new_prop_st.agrees_with_global(*itp) ) {
                   // Util::msg(2,"new_prop_st = "+new_prop_st.to_str() + "\nagrees with " + itp->to_str());
                    new_prop_st.apply_diff_check(*itp);
                    new_prop_states.insert(new_prop_st);
                 }
            }
        }

        void get_new_states_matching_prop(const map<Sys_state, Sys_state_diffs>& prop_diffs_map, 
                    set<Sys_state>& new_states, map<Sys_state,Sys_state>& st2pred) {
            typename_const_foreach(it, st2diff) {
                const Sys_state& old_s = it->first;
                const Upd_diffs& diffs = it->second;

                CHK( prop_diffs_map.find(old_s) != prop_diffs_map.end() );
                //if not, then continue; //here either propfa is not complete 
                const Sys_state_diffs& prop_diffs = prop_diffs_map.find(old_s)->second;

                typename_const_foreach(it2,diffs) {
                    Sys_state new_s = old_s;
                    const Upd_diff& u_diff = *it2;
                    new_s.clear_events();
                    Sys_state::Diff s_diff (new_s.get_mem());
                    u_diff.get_diff(s_diff);
                    new_s.apply_diff_check( s_diff );

                    Sys_states new_prop_states;
                    get_new_prop_states(new_s, prop_diffs, new_prop_states);

                    const_foreach( itp, new_prop_states)  {
                        //insert in st2pred
                        if( st2pred.find(*itp) == st2pred.end() ) {
                            st2pred[*itp].set_mem(old_s.get_mem());
                            st2pred[*itp] = old_s;

                            Util::msg(2,"adding new state to new_states " + itp->to_str() );
                            Util::msg(2,"s_diff was "+s_diff.to_str());
                            new_states.insert(*itp);
                        }
                    }
                }
            }
        }


};

typedef Upd_map<ID_diff> ID_upd_map;
#endif
