#include<cstdarg>
#include"util.h"



string int2str(int i) {
  char s[100];
  sprintf(s,"%d",i);
  return string(s);
}

namespace Util {

tag curr_tmp_tag = 0;
tag curr_tmp_etag = 0;

//public
int VERBOSITY_LEVEL = 1;
tag get_new_tag() { return curr_tmp_tag++; }
tag get_new_etag() { return curr_tmp_etag++; }

Pid tmp_cfg_pid=0, tmp_cfa_pid = 0;
Pid get_new_cfg_comp_pid() { return tmp_cfg_pid++; }
Pid get_new_cfa_comp_pid() { return tmp_cfa_pid++; }

string to_str(tag i) {
  char s[100];
  sprintf(s,"%d",i);
  return string(s);
}

string num_to_bv_string(int n) {
    string s;
    while (1) {
        if (n%2 == 1) s="1"+s; 
        else s="0"+s;
        n = n/2;
        if ( n== 0) break;
    }
    return s;
}

/* Message and Error utils (C) Sagar Chaki*/
/*********************************************************************/
//prints a message and exits.
/*********************************************************************/
//__attribute__ ((noreturn)) void error(bool exit_val, char *format,...)
void error(bool exit_val, char *format,...)
{
  va_list argp;
  va_start(argp,format);
  vprintf(format,argp);
  va_end(argp);
  if( exit_val )
  exit(1);
}

void err(bool exit_val, string message)
{
  fprintf(stderr,"Error: %s\n",message.c_str());
  if( exit_val )
  exit(1);
}

/*********************************************************************/
//prints a message. the first argument is the verbosity level of the
//message.
/*********************************************************************/
void message(int vl,char *format,...) {

  if(vl <= VERBOSITY_LEVEL) {
    va_list argp;
    va_start(argp,format);
    vprintf(format,argp);
    va_end(argp);
    printf("\n");
    fflush(stdout);
  }

}

void msg(int vl,char *format,...)
{
  if(vl <= VERBOSITY_LEVEL) {
    va_list argp;
    va_start(argp,format);
    vprintf(format,argp);
    va_end(argp);
    printf("\n");
    fflush(stdout);
  }
}

void msg(int vl,string mes)
{
  if(vl <= Util::VERBOSITY_LEVEL) {
    fprintf(stderr,"%s\n",mes.c_str());
    fflush(stderr);
  }
}

/*
 TIMER DEFINITION: (C) Sagar Chaki, SEI, CMU.
*/
/*********************************************************************/
//constructors and destructors
/*********************************************************************/
Timer::Timer(const Timer &rhs) { *this = rhs; }

/*********************************************************************/
//operators
/*********************************************************************/
const Timer &Timer::operator = (const Timer &rhs)
{
  start_t = rhs.start_t;
  run_t = rhs.run_t;
  return *this;
}

/*********************************************************************/
//return true if the timer is running and false otherwise
/*********************************************************************/
bool Timer::running() const { return (start_t != -1); }

/*********************************************************************/
//read the current time. the timer must be stopped.
/*********************************************************************/
double Timer::read() const
{
  if(start_t != -1) {
    Util::error(true,"ERROR: trying to read a running timer ...\n");
  }
  return run_t;
}

/*********************************************************************/
//forward a timer by the supplied amount of time
/*********************************************************************/
void Timer::forward(const Timer &arg)
{
  if(arg.start_t != -1) {
    Util::error(true,"ERROR: trying to forward by a running timer ...\n");
  }
  if(start_t != -1) {
    Util::error(true,"ERROR: trying to forward a running timer ...\n");
  }
  run_t += arg.run_t;
}

/*********************************************************************/
//start the timer
/*********************************************************************/
void Timer::start()
{
  if(start_t != -1) {
    Util::error(true,"ERROR: trying to start a running timer ...\n");
  }
#ifdef __MAGIC_USE_CL__
  start_t = static_cast<double>(::time(NULL));
#else
  struct timeval tv;
  if(gettimeofday(&tv,NULL) == -1) {
    Util::error(true,"ERROR: could not get current time for starting timer ...\n");
  }
  start_t = tv.tv_sec + (tv.tv_usec / 1000000.0);
#endif //__MAGIC_USE_CL__
}

/*********************************************************************/
//stop the timer
/*********************************************************************/
void Timer::stop()
{  
  if(start_t == -1) {
    assert(false);
    Util::error(true,"ERROR: trying to stop an unstarted timer ...\n");
  }
#ifdef __MAGIC_USE_CL__
  double currTime = static_cast<double>(::time(NULL));
#else
  struct timeval tv;
  if(gettimeofday(&tv,NULL) == -1) {
    Util::error(true,"ERROR: could not get current time for stopping timer ...\n");
  }
  double currTime = tv.tv_sec + (tv.tv_usec / 1000000.0);
#endif //__MAGIC_USE_CL__
  run_t += (currTime - start_t);
  start_t = -1;
}


}
