#ifndef UTIL_H
#define UTIL_H


#include <cstdio> 
#include <cassert> 

#ifdef __USE_CL__ 
#include <time.h> 
#else 
#include <sys/time.h> 
#include <sys/resource.h> 
#include <unistd.h> 
#endif //__USE_CL__ 

#include"common.h"


string int2str(int i);

template<typename T> 
string vec_to_str(const vector<T>& v) {
	string s = "( ";
   typename_const_foreach(it, v) { s += it->to_str() + " ";}
	s  += ")";
	return s;
}


template<typename T> 
string set_to_str(set<T>& v) {
    string s = "( ";
    //typename_foreach(it, v) { 
    typename set<T>::iterator it = v.begin();
    for( ; it != v.end(); ++it ) {
        string st = it->to_str();
        s += st + " ";
        if( st.size() > 20 ) s+= "\n";
    }
    s  += ")";
    return s;
}

template<typename T> 
string set_to_str_const(const set<T>& v) {
    string s = "( ";
    typename_const_foreach(it, v) { 
        string st = it->to_str_const();
        s += st;
        if( st.size() > 10 ) s+= "\n";
    }
    s  += ")";
    return s;
}

template<typename T> 
string list_to_str(const list<T>& v) {
    string s = "( ";
    typename_const_foreach(it, v) { 
        string st = (it)->to_str();
        s += st + " ";
        //if( st.size() > 10 ) s+= "\n";
    }
    s  += ")";
    return s;
}
template<typename T> 
string list_ptr_to_str(const list<T>& v) {
    string s = "( ";
    typename_const_foreach(it, v) { 
        string st = (*it)->to_str();
        s += st + " ";
        //if( st.size() > 10 ) s+= "\n";
    }
    s  += ")";
    return s;
}
namespace Util {

extern int VERBOSITY_LEVEL;

typedef int tag;
tag get_new_tag();
tag get_new_etag();
string to_str(tag i) ;

string num_to_bv_string(int n);

Pid get_new_cfg_comp_pid();
Pid get_new_cfa_comp_pid();
//print an error message and quit
void error(bool exit_val, char *format,...);
void err(bool exit_val, string message);

//print an error message
void message(int vl,char *format,...);
void msg(int vl,char *format,...);
void msg(int vl,string mes);

class Timer
{
 private:
  double start_t;
  double run_t;

 public:
  Timer() { start_t = -1; run_t = 0; }
  Timer(const Timer &rhs);  
  const Timer &operator = (const Timer &rhs);

  bool running() const;
  void start();
  void stop();
  void forward(const Timer &arg);
  double read() const;
};


}
#endif
